@@ -1,4 +1,9 @@
-#ifdef __BEOS__
+/* beos.c
+ * (c) 2002 Mikulas Patocka
+ * This file is a part of the Links program, released under GPL
+ */
+
+#if defined(__BEOS__) || defined(__HAIKU__)
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -6,6 +11,7 @@
 #include <fcntl.h>
 #include <sys/types.h>
 #include <sys/socket.h>
+#include <sys/time.h>
 #include <netinet/in.h>
 #include <be/kernel/OS.h>
 
@@ -15,6 +21,10 @@
 #define MAXINT 0x7fffffff
 #endif
 
+#ifdef __HAIKU__
+int closesocket(int);
+#endif
+
 int be_read(int s, void *ptr, int len)
 {
 	if (s >= SHS) return recv(s - SHS, ptr, len, 0);
@@ -40,7 +50,7 @@ int be_socket(int af, int sock, int prot)
 	return h + SHS;
 }
 
-int  be_connect(int s, struct sockaddr *sa, int sal)
+int be_connect(int s, struct sockaddr *sa, int sal)
 {
 	return connect(s - SHS, sa, sal);
 }
@@ -91,11 +101,11 @@ int be_pipe(int *fd)
 	struct sockaddr_in sa1, sa2;
 	int retry_count = 0;
 	again:
-	if ((s1 = be_socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
+	if ((s1 = be_socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
 		/*perror("socket1");*/
 		goto fatal_retry;
 	}
-	if ((s2 = be_socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
+	if ((s2 = be_socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
 		/*perror("socket2");*/
 		be_close(s1);
 		goto fatal_retry;
@@ -148,7 +158,7 @@ int be_select(int n, struct fd_set *rd, struct fd_set *wr, struct fd_set *exc, s
 	if (!rd) rd = &d;
 	if (!wr) wr = &d;
 	if (!exc) exc = &d;
-	if (n >= FDSETSIZE) n = FDSETSIZE;
+	if (n >= FD_SETSIZE) n = FD_SETSIZE;
 	FD_ZERO(exc);
 	for (i = 0; i < n; i++) if ((i < SHS && FD_ISSET(i, rd)) || FD_ISSET(i, wr)) {
 		for (i = SHS; i < n; i++) FD_CLR(i, rd);
@@ -203,7 +213,8 @@ int get_input_handle()
 	if ((inth = start_thr(input_handle_th, NULL, "input_thread")) < 0) {
 		closesocket(ihpipe[0]);
 		closesocket(ihpipe[1]);
-		return -1;
+		fprintf(stderr, "Can't spawn input thread");
+		exit(4);
 	}
 	return h = ihpipe[0];
 }
@@ -240,9 +251,14 @@ int get_output_handle()
 	if (start_thr(output_handle_th, NULL, "output_thread") < 0) {
 		closesocket(ohpipe[0]);
 		closesocket(ohpipe[1]);
-		return -1;
+		fprintf(stderr, "Can't spawn output thread");
+		exit(4);
 	}
 	return h = ohpipe[1];
 }*/
 
+#else
+
+typedef int beos_c_no_empty_unit;
+
 #endif
