@@ -7,13 +7,19 @@ struct memory_list *getml(void *p, ...)
 	int n = 0;
 	void *q = p;
 	va_start(ap, p);
-	while (q) n++, q = va_arg(ap, void *);
-	if (!(ml = mem_alloc(sizeof(struct memory_list) + n * sizeof(void *)))) return NULL;
+	while (q) {
+		if (n == MAXINT) overalloc();
+		n++, q = va_arg(ap, void *);
+	}
+	if ((unsigned)n > (MAXINT - sizeof(struct memory_list)) / sizeof(void *)) overalloc();
+	ml = mem_alloc(sizeof(struct memory_list) + n * sizeof(void *));
 	ml->n = n;
 	n = 0;
 	q = p;
+	va_end(ap);
 	va_start(ap, p);
 	while (q) ml->p[n++] = q, q = va_arg(ap, void *);
+	va_end(ap);
 	return ml;
 }
 
@@ -24,16 +30,21 @@ void add_to_ml(struct memory_list **ml, ...)
 	int n = 0;
 	void *q;
 	if (!*ml) {
-		if (!(*ml = mem_alloc(sizeof(struct memory_list)))) return;
+		*ml = mem_alloc(sizeof(struct memory_list));
 		(*ml)->n = 0;
 	}
 	va_start(ap, ml);
-	while ((q = va_arg(ap, void *))) n++;
-	if (!(nml = mem_realloc(*ml, sizeof(struct memory_list) + (n + (*ml)->n) * sizeof(void *))))
-		return;
+	while ((q = va_arg(ap, void *))) {
+		if (n == MAXINT) overalloc();
+		n++;
+	}
+	if ((unsigned)n + (unsigned)((*ml)->n) > (MAXINT - sizeof(struct memory_list)) / sizeof(void *)) overalloc();
+	nml = mem_realloc(*ml, sizeof(struct memory_list) + (n + (*ml)->n) * sizeof(void *));
+	va_end(ap);
 	va_start(ap, ml);
 	while ((q = va_arg(ap, void *))) nml->p[nml->n++] = q;
 	*ml = nml;
+	va_end(ap);
 }
 
 void freeml(struct memory_list *ml)
@@ -53,20 +64,12 @@ void dialog_func(struct window *, struct event *, int);
 void do_menu_selected(struct terminal *term, struct menu_item *items, void *data, int selected)
 {
 	struct menu *menu;
-	if ((menu = mem_alloc(sizeof(struct menu)))) {
-		menu->selected = selected;
-		menu->view = 0;
-		menu->items = items;
-		menu->data = data;
-		add_window(term, menu_func, menu);
-	} else if (items->free_i) {
-		int i;
-		for (i = 0; items[i].text; i++) {
-			if (items[i].free_i & 2) mem_free(items[i].text);
-			if (items[i].free_i & 4) mem_free(items[i].rtext);
-		}
-		mem_free(items);
-	}
+	menu = mem_alloc(sizeof(struct menu));
+	menu->selected = selected;
+	menu->view = 0;
+	menu->items = items;
+	menu->data = data;
+	add_window(term, menu_func, menu);
 }
 
 void do_menu(struct terminal *term, struct menu_item *items, void *data)
@@ -179,7 +182,7 @@ void menu_func(struct window *win, struct event *ev, int fwd)
 	struct menu *menu = win->data;
 	struct window *w1;
 	menu->win = win;
-	switch (ev->ev) {
+	switch ((int)ev->ev) {
 		case EV_INIT:
 		case EV_RESIZE:
 		case EV_REDRAW:
@@ -210,7 +213,7 @@ void menu_func(struct window *win, struct event *ev, int fwd)
 						menu->selected = s;
 						scroll_menu(menu, 0);
 						display_menu(win->term, menu);
-						if ((ev->b & BM_ACT) == B_UP || menu->items[s].in_m) select_menu(win->term, menu);
+						if ((ev->b & BM_ACT) == B_UP/* || menu->items[s].in_m*/) select_menu(win->term, menu);
 					}
 				}
 			}
@@ -285,16 +288,15 @@ void menu_func(struct window *win, struct event *ev, int fwd)
 void do_mainmenu(struct terminal *term, struct menu_item *items, void *data, int sel)
 {
 	struct mainmenu *menu;
-	if ((menu = mem_alloc(sizeof(struct mainmenu)))) {
-		menu->selected = sel == -1 ? 0 : sel;
-		menu->items = items;
-		menu->data = data;
-		add_window(term, mainmenu_func, menu);
-		if (sel != -1) {
-			struct event ev = {EV_KBD, KBD_ENTER, 0, 0};
-			struct window *win = term->windows.next;
-			win->handler(win, &ev, 0);
-		}
+	menu = mem_alloc(sizeof(struct mainmenu));
+	menu->selected = sel == -1 ? 0 : sel;
+	menu->items = items;
+	menu->data = data;
+	add_window(term, mainmenu_func, menu);
+	if (sel != -1) {
+		struct event ev = {EV_KBD, KBD_ENTER, 0, 0};
+		struct window *win = term->windows.next;
+		win->handler(win, &ev, 0);
 	}
 }
 
@@ -330,8 +332,8 @@ void select_mainmenu(struct terminal *term, struct mainmenu *menu)
 	struct menu_item *it = &menu->items[menu->selected];
 	if (menu->selected < 0 || menu->selected >= menu->ni || it->hotkey == M_BAR) return;
 	if (!it->in_m) {
-		struct window *win;
-		for (win = term->windows.next; (void *)win != &term->windows && (win->handler == menu_func || win->handler == mainmenu_func); delete_window(win)) ;
+		struct window *win, *win1;
+		for (win = term->windows.next; (void *)win != &term->windows && (win->handler == menu_func || win->handler == mainmenu_func); win1 = win->next, delete_window(win), win = win1) ;
 	}
 	it->func(term, it->data, menu->data);
 }
@@ -341,7 +343,7 @@ void mainmenu_func(struct window *win, struct event *ev, int fwd)
 	int s = 0;
 	struct mainmenu *menu = win->data;
 	menu->win = win;
-	switch(ev->ev) {
+	switch ((int)ev->ev) {
 		case EV_INIT:
 		case EV_RESIZE:
 		case EV_REDRAW:
@@ -378,6 +380,14 @@ void mainmenu_func(struct window *win, struct event *ev, int fwd)
 				if (++menu->selected >= menu->ni) menu->selected = 0;
 				s = 1;
 			}
+			if (ev->x == KBD_HOME) {
+				menu->selected = 0;
+				s = 1;
+			}
+			if (ev->x == KBD_END) {
+				menu->selected = menu->ni - 1;
+				s = 1;
+			}
 			if ((ev->x == KBD_LEFT || ev->x == KBD_RIGHT) && fwd) {
 				display_mainmenu(win->term, menu);
 				select_mainmenu(win->term, menu);
@@ -392,7 +402,7 @@ void mainmenu_func(struct window *win, struct event *ev, int fwd)
 						s = 2;
 					}
 			} else if (!s) {
-				delete_window_ev(win, ev->x != KBD_ESC ? ev : NULL);
+				delete_window_ev(win, (ev->x >= KBD_F1 && ev->x <= KBD_F12) || ev->y & KBD_ALT ? ev : NULL);
 				break;
 			}
 			display_mainmenu(win->term, menu);
@@ -406,7 +416,7 @@ void mainmenu_func(struct window *win, struct event *ev, int fwd)
 struct menu_item *new_menu(int free_i)
 {
 	struct menu_item *mi;
-	if (!(mi = mem_alloc(sizeof(struct menu_item)))) return NULL;
+	mi = mem_alloc(sizeof(struct menu_item));
 	memset(mi, 0, sizeof(struct menu_item));
 	mi->free_i = free_i;
 	return mi;
@@ -416,8 +426,9 @@ void add_to_menu(struct menu_item **mi, unsigned char *text, unsigned char *rtex
 {
 	struct menu_item *mii;
 	int n;
-	for (n = 0; (*mi)[n].text; n++) ;
-	if (!(mii = mem_realloc(*mi, (n + 2) * sizeof(struct menu_item)))) return;
+	for (n = 0; (*mi)[n].text; n++) if (n == MAXINT) overalloc();
+	if (((unsigned)n + 2) > MAXINT / sizeof(struct menu_item)) overalloc();
+	mii = mem_realloc(*mi, (n + 2) * sizeof(struct menu_item));
 	*mi = mii;
 	memcpy(mii + n + 1, mii + n, sizeof(struct menu_item));
 	mii[n].text = text;
@@ -433,8 +444,12 @@ void do_dialog(struct terminal *term, struct dialog *dlg, struct memory_list *ml
 	struct dialog_data *dd;
 	struct dialog_item *d;
 	int n = 0;
-	for (d = dlg->items; d->type != D_END; d++) n++;
-	if (!(dd = mem_alloc(sizeof(struct dialog_data) + sizeof(struct dialog_item_data) * n))) return;
+	for (d = dlg->items; d->type != D_END; d++) {
+		if (n == MAXINT) overalloc();
+		n++;
+	}
+	if ((unsigned)n > (MAXINT - sizeof(struct dialog_data)) / sizeof(struct dialog_item_data)) overalloc();
+	dd = mem_alloc(sizeof(struct dialog_data) + sizeof(struct dialog_item_data) * n);
 	dd->dlg = dlg;
 	dd->n = n;
 	dd->ml = ml;
@@ -462,9 +477,10 @@ void display_dlg_item(struct dialog_data *dlg, struct dialog_item_data *di, int
 			if (di->vpos < 0) di->vpos = 0;
 			fill_area(term, di->x, di->y, di->l, 1, COLOR_DIALOG_FIELD);
 			if (di->item->type == D_FIELD) {
-				print_text(term, di->x, di->y, strlen(di->cdata + di->vpos) <= di->l ? strlen(di->cdata + di->vpos) : di->l, di->cdata + di->vpos, COLOR_DIALOG_FIELD_TEXT);
+				print_text(term, di->x, di->y, strlen(di->cdata + di->vpos) <= (size_t)di->l ? (int)strlen(di->cdata + di->vpos) : di->l, di->cdata + di->vpos, COLOR_DIALOG_FIELD_TEXT);
+			} else {
+				fill_area(term, di->x, di->y, strlen(di->cdata + di->vpos) <= (size_t)di->l ? (int)strlen(di->cdata + di->vpos) : di->l, 1, COLOR_DIALOG_FIELD_TEXT | '*');
 			}
-			else fill_area(term, di->x, di->y, strlen(di->cdata + di->vpos) <= di->l ? strlen(di->cdata + di->vpos) : di->l, 1, COLOR_DIALOG_FIELD_TEXT | '*');
 			if (sel) {
 				set_cursor(term, di->x + di->cpos - di->vpos, di->y, di->x + di->cpos - di->vpos, di->y);
 				set_window_ptr(dlg->win, di->x, di->y);
@@ -486,7 +502,7 @@ void display_dlg_item(struct dialog_data *dlg, struct dialog_item_data *di, int
 			show_dlg_item_box(dlg, di);
 			break;
 		default:
-			debug("Tried to draw unknown ");
+			internal("Tried to draw unknown");
 	}
 }
 
@@ -519,13 +535,14 @@ void dlg_set_history(struct dialog_item_data *di)
 	memcpy(di->cdata, s, l);
 	di->cdata[l] = 0;
 	di->cpos = l;
+	di->vpos = 0;
 }
 
 int dlg_mouse(struct dialog_data *dlg, struct dialog_item_data *di, struct event *ev)
 {
 	switch (di->item->type) {
 		case D_BUTTON:
-			if (ev->y != di->y || ev->x < di->x || ev->x >= di->x + strlen(_(di->item->text, dlg->win->term)) + 4) return 0;
+			if (ev->y != di->y || ev->x < di->x || (size_t)ev->x >= di->x + strlen(_(di->item->text, dlg->win->term)) + 4) return 0;
 			display_dlg_item(dlg, &dlg->items[dlg->selected], 0);
 			dlg->selected = di - dlg->items;
 			display_dlg_item(dlg, di, 1);
@@ -534,7 +551,7 @@ int dlg_mouse(struct dialog_data *dlg, struct dialog_item_data *di, struct event
 		case D_FIELD_PASS:
 		case D_FIELD:
 			if (ev->y != di->y || ev->x < di->x || ev->x >= di->x + di->l) return 0;
-			if ((di->cpos = di->vpos + ev->x - di->x) > strlen(di->cdata)) di->cpos = strlen(di->cdata);
+			if ((size_t)(di->cpos = di->vpos + ev->x - di->x) > strlen(di->cdata)) di->cpos = strlen(di->cdata);
 			display_dlg_item(dlg, &dlg->items[dlg->selected], 0);
 			dlg->selected = di - dlg->items;
 			display_dlg_item(dlg, di, 1);
@@ -596,14 +613,11 @@ void do_tab_compl(struct terminal *term, struct list_head *history, struct windo
 	unsigned char *cdata = ((struct dialog_data*)win->data)->items[((struct dialog_data*)win->data)->selected].cdata;
 	int l = strlen(cdata), n = 0;
 	struct history_item *hi;
-	struct menu_item *items = DUMMY, *i;
+	struct menu_item *items = DUMMY;
 	foreach(hi, *history) if (!strncmp(cdata, hi->d, l)) {
 		if (!(n & (ALLOC_GR - 1))) {
-			if (!(i = mem_realloc(items, (n + ALLOC_GR + 1) * sizeof(struct menu_item)))) {
-				mem_free(items);
-				return;
-			}
-			items = i;
+			if ((unsigned)n > MAXINT / sizeof(struct menu_item) - ALLOC_GR - 1) overalloc();
+			items = mem_realloc(items, (n + ALLOC_GR + 1) * sizeof(struct menu_item));
 		}
 		items[n].text = hi->d;
 		items[n].rtext = "";
@@ -613,6 +627,7 @@ void do_tab_compl(struct terminal *term, struct list_head *history, struct windo
 		items[n].data = hi->d;
 		items[n].in_m = 0;
 		items[n].free_i = 1;
+		if (n == MAXINT) overalloc();
 		n++;
 	}
 	if (n == 1) {
@@ -622,7 +637,7 @@ void do_tab_compl(struct terminal *term, struct list_head *history, struct windo
 	}
 	if (n) {
 		memset(&items[n], 0, sizeof(struct menu_item));
-		do_menu(term, items, win);
+		do_menu_selected(term, items, win, n - 1);
 	}
 }
 
@@ -640,14 +655,15 @@ void dialog_func(struct window *win, struct event *ev, int fwd)
 		return;
 	}
 	
-	switch (ev->ev) {
+	switch ((int)ev->ev) {
 		case EV_INIT:
 			for (i = 0; i < dlg->n; i++) {
-				struct dialog_item_data *di = &dlg->items[i];
+				/* volatile because of a compiler bug */
+				struct dialog_item_data * volatile di = &dlg->items[i];
 				memset(di, 0, sizeof(struct dialog_item_data));
 				di->item = &dlg->dlg->items[i];
-				if ((di->cdata = mem_alloc(di->item->dlen)))
-					memcpy(di->cdata, di->item->data, di->item->dlen);
+				di->cdata = mem_alloc(di->item->dlen);
+				memcpy(di->cdata, di->item->data, di->item->dlen);
 				if (di->item->type == D_CHECKBOX) {
 					if (di->item->gid) {
 						if (*(int *)di->cdata == di->item->gnum) di->checked = 1;
@@ -670,7 +686,7 @@ void dialog_func(struct window *win, struct event *ev, int fwd)
 						/*int l = di->item->dlen;*/
 						foreach(j, di->item->history->items) {
 							struct history_item *hi;
-							if (!(hi = mem_alloc(sizeof(struct history_item) + strlen(j->d) + 1))) continue;
+							hi = mem_alloc(sizeof(struct history_item) + strlen(j->d) + 1);
 							strcpy(hi->d, j->d);
 							add_to_list(di->history, hi);
 						}
@@ -706,7 +722,7 @@ void dialog_func(struct window *win, struct event *ev, int fwd)
 						}
 						break;
 					case ACT_RIGHT:
-						if (di->cpos < strlen(di->cdata)) di->cpos++;
+						if ((size_t)di->cpos < strlen(di->cdata)) di->cpos++;
 						goto dsp_f;
 					case ACT_LEFT:
 						if (di->cpos > 0) di->cpos--;
@@ -724,7 +740,7 @@ void dialog_func(struct window *win, struct event *ev, int fwd)
 						}
 						goto dsp_f;
 					case ACT_DELETE:
-						if (di->cpos < strlen(di->cdata))
+						if ((size_t)di->cpos < strlen(di->cdata))
 							memmove(di->cdata + di->cpos, di->cdata + di->cpos + 1, strlen(di->cdata) - di->cpos + 1);
 						goto dsp_f;
 					case ACT_KILL_TO_BOL:
@@ -747,10 +763,11 @@ void dialog_func(struct window *win, struct event *ev, int fwd)
 					case ACT_PASTE_CLIPBOARD: {
 						/* Paste from clipboard */
 						unsigned char * clipboard = get_clipboard_text();
-						strncpy(di->cdata , clipboard, di->item->dlen);
-						di->cdata[di->item->dlen - 1] = 0;
-						di->cpos = strlen(di->cdata);
-						mem_free(clipboard);
+						if (clipboard) {
+							safe_strncpy(di->cdata, clipboard, di->item->dlen);
+							di->cpos = strlen(di->cdata);
+							mem_free(clipboard);
+						}
 						goto dsp_f;
 					}
 					case ACT_AUTO_COMPLETE: 
@@ -758,7 +775,7 @@ void dialog_func(struct window *win, struct event *ev, int fwd)
 						goto dsp_f;
 					default:
 						if (ev->x >= ' ' && ev->x < 0x100 && !ev->y) {
-							if (strlen(di->cdata) < di->item->dlen - 1) {
+							if (strlen(di->cdata) + 1 < (size_t)di->item->dlen) {
 								memmove(di->cdata + di->cpos + 1, di->cdata + di->cpos, strlen(di->cdata) - di->cpos + 1);
 								di->cdata[di->cpos++] = ev->x;
 							}
@@ -823,13 +840,13 @@ void dialog_func(struct window *win, struct event *ev, int fwd)
 int check_number(struct dialog_data *dlg, struct dialog_item_data *di)
 {
 	unsigned char *end;
-	long l = strtol(di->cdata, (char **)&end, 10);
+	long l = strtol(di->cdata, (char **)(void *)&end, 10);
 	if (!*di->cdata || *end) {
-		msg_box(dlg->win->term, NULL, TEXT(T_BAD_NUMBER), AL_CENTER, TEXT(T_NUMBER_EXPECTED), NULL, 1, TEXT(T_CANCEL), NULL, B_ENTER | B_ESC);
+		msg_box(dlg->win->term, NULL, TEXT_(T_BAD_NUMBER), AL_CENTER, TEXT_(T_NUMBER_EXPECTED), NULL, 1, TEXT_(T_CANCEL), NULL, B_ENTER | B_ESC);
 		return 1;
 	}
 	if (l < di->item->gid || l > di->item->gnum) {
-		msg_box(dlg->win->term, NULL, TEXT(T_BAD_NUMBER), AL_CENTER, TEXT(T_NUMBER_OUT_OF_RANGE), NULL, 1, TEXT(T_CANCEL), NULL, B_ENTER | B_ESC);
+		msg_box(dlg->win->term, NULL, TEXT_(T_BAD_NUMBER), AL_CENTER, TEXT_(T_NUMBER_OUT_OF_RANGE), NULL, 1, TEXT_(T_CANCEL), NULL, B_ENTER | B_ESC);
 		return 1;
 	}
 	return 0;
@@ -839,7 +856,7 @@ int check_nonempty(struct dialog_data *dlg, struct dialog_item_data *di)
 {
 	unsigned char *p;
 	for (p = di->cdata; *p; p++) if (*p > ' ') return 0;
-	msg_box(dlg->win->term, NULL, TEXT(T_BAD_STRING), AL_CENTER, TEXT(T_EMPTY_STRING_NOT_ALLOWED), NULL, 1, TEXT(T_CANCEL), NULL, B_ENTER | B_ESC);
+	msg_box(dlg->win->term, NULL, TEXT_(T_BAD_STRING), AL_CENTER, TEXT_(T_EMPTY_STRING_NOT_ALLOWED), NULL, 1, TEXT_(T_CANCEL), NULL, B_ENTER | B_ESC);
 	return 1;
 }
 
@@ -864,17 +881,26 @@ int check_dialog(struct dialog_data *dlg)
 	return 0;
 }
 
-int ok_dialog(struct dialog_data *dlg, struct dialog_item_data *di)
+void get_dialog_data(struct dialog_data *dlg)
 {
 	int i;
+	for (i = 0; i < dlg->n; i++) {
+		/* volatile because of a compiler bug */
+		void * volatile p1 = dlg->dlg->items[i].data;
+		void * volatile p2 = dlg->items[i].cdata;
+		volatile int l = dlg->dlg->items[i].dlen;
+		memcpy(p1, p2, l);
+	}
+}
+
+int ok_dialog(struct dialog_data *dlg, struct dialog_item_data *di)
+{
 	void (*fn)(void *) = dlg->dlg->refresh;
 	void *data = dlg->dlg->refresh_data;
 	if (check_dialog(dlg)) return 1;
-	for (i = 0; i < dlg->n; i++)
-		memcpy(dlg->dlg->items[i].data, dlg->items[i].cdata, dlg->dlg->items[i].dlen);
+	get_dialog_data(dlg);
 	if (fn) fn(data);
-	i = cancel_dialog(dlg, di);
-	return i;
+	return cancel_dialog(dlg, di);
 }
 
 void center_dlg(struct dialog_data *dlg)
@@ -1040,7 +1066,7 @@ void max_group_width(struct terminal *term, unsigned char **texts, struct dialog
 {
 	int ww = 0;
 	while (n--) {
-		int wx = item->item->type == D_CHECKBOX ? 4 : item->item->type == D_BUTTON ? strlen(_(item->item->text, term)) + 5 : item->item->dlen + 1;
+		int wx = item->item->type == D_CHECKBOX ? 4 : item->item->type == D_BUTTON ? strlen(_(item->item->text, term)) + 5 : (size_t)item->item->dlen + 1;
 		wx += strlen(_(texts[0], term));
 		if (n) wx++;
 		ww += wx;
@@ -1053,7 +1079,7 @@ void max_group_width(struct terminal *term, unsigned char **texts, struct dialog
 void min_group_width(struct terminal *term, unsigned char **texts, struct dialog_item_data *item, int n, int *w)
 {
 	while (n--) {
-		int wx = item->item->type == D_CHECKBOX ? 4 : item->item->type == D_BUTTON ? strlen(_(item->item->text, term)) + 5 : item->item->dlen + 1;
+		int wx = item->item->type == D_CHECKBOX ? 4 : item->item->type == D_BUTTON ? strlen(_(item->item->text, term)) + 5 : (size_t)item->item->dlen + 1;
 		wx += strlen(_(texts[0], term));
 		if (wx > *w) *w = wx;
 		texts++;
@@ -1066,7 +1092,7 @@ void dlg_format_group(struct terminal *term, struct terminal *t2, unsigned char
 	int nx = 0;
 	while (n--) {
 		int sl;
-		int wx = item->item->type == D_CHECKBOX ? 4 : item->item->type == D_BUTTON ? strlen(_(item->item->text, t2)) + 5 : item->item->dlen + 1;
+		int wx = item->item->type == D_CHECKBOX ? 4 : item->item->type == D_BUTTON ? strlen(_(item->item->text, t2)) + 5 : (size_t)item->item->dlen + 1;
 		if (_(texts[0], t2)[0]) sl = strlen(_(texts[0], t2));
 		else sl = -1;
 		wx += sl;
@@ -1207,14 +1233,11 @@ void msg_box(struct terminal *term, struct memory_list *ml, unsigned char *title
 	udata = DUMMY;
 	udatan = 0;
 	do {
-		unsigned char **udata_;
 		text = va_arg(ap, unsigned char *);
 		na_kovarne__to_je_narez:
-		if (!(udata_ = mem_realloc(udata, ++udatan * sizeof(unsigned char *)))) {
-			mem_free(udata);
-			return;
-		}
-		udata = udata_;
+		udatan++;
+		if ((unsigned)udatan > MAXINT / sizeof(unsigned char *)) overalloc();
+		udata = mem_realloc(udata, udatan * sizeof(unsigned char *));
 		udata[udatan - 1] = text;
 		if (text && !(align & AL_EXTD_TEXT)) {
 			text = NULL;
@@ -1223,10 +1246,8 @@ void msg_box(struct terminal *term, struct memory_list *ml, unsigned char *title
 	} while (text);
 	udata2 = va_arg(ap, void *);
 	n = va_arg(ap, int);
-	if (!(dlg = mem_alloc(sizeof(struct dialog) + (n + 1) * sizeof(struct dialog_item)))) {
-		mem_free(udata);
-		return;
-	}
+	if ((unsigned)n > (MAXINT - sizeof(struct dialog)) / sizeof(struct dialog_item) - 1) overalloc();
+	dlg = mem_alloc(sizeof(struct dialog) + (n + 1) * sizeof(struct dialog_item));
 	memset(dlg, 0, sizeof(struct dialog) + (n + 1) * sizeof(struct dialog_item));
 	dlg->title = title;
 	dlg->fn = msg_box_fn;
@@ -1251,20 +1272,22 @@ void msg_box(struct terminal *term, struct memory_list *ml, unsigned char *title
 		dlg->items[i].text = m;
 		dlg->items[i].udata = fn;
 	}
+	va_end(ap);
 	dlg->items[i].type = D_END;
 	add_to_ml(&ml, dlg, udata, NULL);
 	do_dialog(term, dlg, ml);
 }
 
-void add_to_history(struct history *h, unsigned char *t)
+void add_to_history(struct history *h, unsigned char *t, int check_duplicates)
 {
 	struct history_item *hi, *hs;
-	int l;
+	size_t l;
 	if (!h || !t || !*t) return;
-	l = strlen(t) + 1;
-	if (!(hi = mem_alloc(sizeof(struct history_item) + l))) return;
-	memcpy(hi->d, t, l);
-	foreach(hs, h->items) if (!strcmp(hs->d, t)) {
+	l = strlen(t);
+	if (l > MAXINT - sizeof(struct history_item)) overalloc();
+	hi = mem_alloc(sizeof(struct history_item) + l);
+	memcpy(hi->d, t, l + 1);
+	if (check_duplicates) foreach(hs, h->items) if (!strcmp(hs->d, t)) {
 		struct history_item *hd = hs;
 		hs = hs->prev;
 		del_from_list(hd);
@@ -1301,7 +1324,7 @@ int input_field_ok(struct dialog_data *dlg, struct dialog_item_data *di)
 	void *data = dlg->dlg->udata2;
 	unsigned char *text = dlg->items->cdata;
 	if (check_dialog(dlg)) return 1;
-	add_to_history(dlg->dlg->items->history, text);
+	add_to_history(dlg->dlg->items->history, text, 1);
 	if (fn) fn(data, text);
 	ok_dialog(dlg, di);
 	return 0;
@@ -1342,12 +1365,14 @@ void input_field(struct terminal *term, struct memory_list *ml, unsigned char *t
 {
 	struct dialog *dlg;
 	unsigned char *field;
-	if (!(dlg = mem_alloc(sizeof(struct dialog) + 4 * sizeof(struct dialog_item) + l)))
-		return;
+	if ((unsigned)l > MAXINT - sizeof(struct dialog) + 4 * sizeof(struct dialog_item)) overalloc();
+	dlg = mem_alloc(sizeof(struct dialog) + 4 * sizeof(struct dialog_item) + l);
 	memset(dlg, 0, sizeof(struct dialog) + 4 * sizeof(struct dialog_item) + l);
 	*(field = (unsigned char *)dlg + sizeof(struct dialog) + 4 * sizeof(struct dialog_item)) = 0;
-	if (def) if (strlen(def) + 1 > l) memcpy(field, def, l - 1);
-		 else strcpy(field, def);
+	if (def) {
+		if (strlen(def) + 1 > (size_t)l) memcpy(field, def, l - 1);
+		else strcpy(field, def);
+	}
 	dlg->title = title;
 	dlg->fn = input_field_fn;
 	dlg->udata = text;
@@ -1437,7 +1462,6 @@ void show_dlg_item_box(struct dialog_data *dlg, struct dialog_item_data *box_ite
 	struct terminal *term = dlg->win->term;
 	struct dlg_data_item_data_box *box;
 	struct box_item *citem;	/* Item currently being shown */
-	struct box_item *items;	/* List of items to be shown */
 	int n;	/* Index of item currently being displayed */
 
 	box = (struct dlg_data_item_data_box *)(box_item_data->item->data);
