@@ -43,6 +43,7 @@ int find_in_cache(unsigned char *url, struct cache_entry **f)
 	struct cache_entry *e;
 	url = extract_proxy(url);
 	foreach(e, cache) if (!strcmp(e->url, url)) {
+		e->refcount++;
 		del_from_list(e);
 		add_to_list(cache, e);
 		*f = e;
@@ -57,45 +58,38 @@ int get_cache_entry(unsigned char *url, struct cache_entry **f)
 	if (!find_in_cache(url, f)) return 0;
 	shrink_memory(0);
 	url = extract_proxy(url);
-	if (!(e = mem_alloc(sizeof(struct cache_entry)))) return -1;
+	e = mem_alloc(sizeof(struct cache_entry));
 	memset(e, 0, sizeof(struct cache_entry));
-	if (!(e->url = mem_alloc(strlen(url)+1))) {
-		mem_free(e);
-		return -1;
-	}
+	e->url = mem_alloc(strlen(url) + 1);
 	strcpy(e->url, url);
 	e->length = 0;
 	e->incomplete = 1;
 	e->data_size = 0;
 	init_list(e->frag);
 	e->count = cache_count++;
-	e->refcount = 0;
+	e->refcount = 1;
 	add_to_list(cache, e);
 	*f = e;
 	return 0;
 }
 
-int get_cache_data(struct cache_entry *e, unsigned char **d, int *l)
-{
-	struct fragment *frag;
-	*d = NULL; *l = 0;
-	if ((void *)(frag = e->frag.next) == &e->frag) return -1;
-	*d = frag->data;
-	*l = frag->length;
-	return 0;
-}
+#define sf(x) e->data_size += (x), cache_size += (x)
 
-#define sf(x) e->data_size += x, cache_size += x
+int page_size = 4096;
 
-#define C_ALIGN(x) (((x) | 0x3fff) + 1)
+#define C_ALIGN(x) ((((x) + sizeof(struct fragment) + 64) | (page_size - 1)) - sizeof(struct fragment) - 64)
 
-int add_fragment(struct cache_entry *e, int offset, unsigned char *data, int length)
+int add_fragment(struct cache_entry *e, off_t offset, unsigned char *data, off_t length)
 {
 	struct fragment *f;
 	struct fragment *nf;
 	int a = 0;
 	int trunc = 0;
+	volatile off_t ca;
 	if (!length) return 0;
+	e->incomplete = 1;
+	if (offset + length < 0 || offset + length < offset) overalloc();
+	if (offset + (off_t)C_ALIGN(length) < 0 || offset + (off_t)C_ALIGN(length) < offset) overalloc();
 	if (e->length < offset + length) e->length = offset + length;
 	e->count = cache_count++;
 	foreach(f, e->frag) {
@@ -109,6 +103,7 @@ int add_fragment(struct cache_entry *e, int offset, unsigned char *data, int len
 					f->length = offset-f->offset+length;
 				}
 				else {
+					sf(-(f->offset+f->length-offset));
 					f->length = offset-f->offset;
 					f = f->next;
 					break;
@@ -120,7 +115,11 @@ int add_fragment(struct cache_entry *e, int offset, unsigned char *data, int len
 			goto ch_o;
 		}
 	}
-	if (!(nf = mem_alloc(sizeof(struct fragment)+C_ALIGN(length)))) return -1;
+/* Intel C 9 has a bug and miscompiles this statement (< 0 test is true) */
+	/*if (C_ALIGN(length) > MAXINT - sizeof(struct fragment) || C_ALIGN(length) < 0) overalloc();*/
+	ca = C_ALIGN(length);
+	if (ca > MAXINT - (int)sizeof(struct fragment) || ca < 0) overalloc();
+	nf = mem_alloc(sizeof(struct fragment) + ca);
 	a = 1;
 	sf(length);
 	nf->offset = offset;
@@ -132,7 +131,7 @@ int add_fragment(struct cache_entry *e, int offset, unsigned char *data, int len
 	ch_o:
 	while ((void *)f->next != &e->frag && f->offset+f->length > f->next->offset) {
 		if (f->offset+f->length < f->next->offset+f->next->length) {
-			if (!(nf = mem_realloc(f, sizeof(struct fragment)+f->next->offset-f->offset+f->next->length))) goto ff;
+			nf = mem_realloc(f, sizeof(struct fragment)+f->next->offset-f->offset+f->next->length);
 			nf->prev->next = nf;
 			nf->next->prev = nf;
 			f = nf;
@@ -140,7 +139,6 @@ int add_fragment(struct cache_entry *e, int offset, unsigned char *data, int len
 			memcpy(f->data+f->length, f->next->data+f->offset+f->length-f->next->offset, f->next->offset+f->next->length-f->offset-f->length);
 			sf(f->next->offset+f->next->length-f->offset-f->length);
 			f->length = f->real_length = f->next->offset+f->next->length-f->offset;
-			ff:;
 		} else {
 			if (memcmp(f->data+f->next->offset-f->offset, f->next->data, f->next->length)) trunc = 1;
 		}
@@ -160,7 +158,7 @@ int add_fragment(struct cache_entry *e, int offset, unsigned char *data, int len
 void defrag_entry(struct cache_entry *e)
 {
 	struct fragment *f, *g, *h, *n, *x;
-	int l;
+	off_t l;
 	if (list_empty(e->frag)) return;
 	f = e->frag.next;
 	if (f->offset) return;
@@ -168,9 +166,10 @@ void defrag_entry(struct cache_entry *e)
 		internal("fragments overlay");
 		return;
 	}
-	if (g == f->next) return;
+	if (g == f->next && f->length == f->real_length) return;
 	for (l = 0, h = f; h != g; h = h->next) l += h->length;
-	if (!(n = mem_alloc(sizeof(struct fragment) + l))) return;
+	if (l > MAXINT - (int)sizeof(struct fragment) || l < 0) overalloc();
+	n = mem_alloc(sizeof(struct fragment) + l);
 	n->offset = 0;
 	n->length = l;
 	n->real_length = l;
@@ -194,26 +193,30 @@ void defrag_entry(struct cache_entry *e)
 	}*/
 }
 
-void truncate_entry(struct cache_entry *e, int off, int final)
+void truncate_entry(struct cache_entry *e, off_t off, int final)
 {
+	int modified = 0;
 	struct fragment *f, *g;
 	if (e->length > off) e->length = off, e->incomplete = 1;
 	foreach(f, e->frag) {
 		if (f->offset >= off) {
 			del:
 			while ((void *)f != &e->frag) {
+				modified = 1;
 				sf(-f->length);
 				g = f->next;
 				del_from_list(f);
 				mem_free(f);
 				f = g;
 			}
-			return;
+			goto ret;
 		}
 		if (f->offset + f->length > off) {
-			f->length = off - f->offset;
+			modified = 1;
 			sf(-(f->offset + f->length - off));
-			if (final && (g = mem_realloc(f, sizeof(struct fragment) + f->length))) {
+			f->length = off - f->offset;
+			if (final) {
+				g = mem_realloc(f, sizeof(struct fragment) + f->length);
 				g->next->prev = g;
 				g->prev->next = g;
 				f = g;
@@ -223,11 +226,16 @@ void truncate_entry(struct cache_entry *e, int off, int final)
 			goto del;
 		}
 	}
+	ret:
+	if (modified) {
+		e->count = cache_count++;
+	}
 }
 
-void free_entry_to(struct cache_entry *e, int off)
+void free_entry_to(struct cache_entry *e, off_t off)
 {
 	struct fragment *f, *g;
+	e->incomplete = 1;
 	foreach(f, e->frag) {
 		if (f->offset + f->length <= off) {
 			sf(-f->length);
@@ -237,7 +245,7 @@ void free_entry_to(struct cache_entry *e, int off)
 			mem_free(g);
 		} else if (f->offset < off) {
 			sf(f->offset - off);
-			memmove(f->data, f->data + off - f->offset, f->length -= off - f->offset);
+			memmove(f->data, f->data + (off - f->offset), f->length -= off - f->offset);
 			f->offset = off;
 		} else break;
 	}
@@ -283,7 +291,6 @@ void garbage_collection(int u)
 	struct cache_entry *e, *f;
 	long ncs = cache_size;
 	long ccs = 0;
-	int no = 0;
 	if (!u && cache_size <= memory_cache_size) return;
 	foreach(e, cache) {
 		if (e->refcount || is_entry_used(e)) if ((ncs -= e->data_size) < 0) {
@@ -295,9 +302,8 @@ void garbage_collection(int u)
 	if (ccs != cache_size) internal("cache size badly computed: %ld != %ld", cache_size, ccs), cache_size = ccs;
 	if (!u && ncs <= memory_cache_size) return;
 	for (e = cache.prev; (void *)e != &cache; e = e->prev) {
-		if (!u && ncs <= memory_cache_size * MEMORY_CACHE_GC_PERCENT) goto g;
+		if (!u && ncs <= (longlong)memory_cache_size * MEMORY_CACHE_GC_PERCENT) goto g;
 		if (e->refcount || is_entry_used(e)) {
-			no = 1;
 			e->tgc = 0;
 			continue;
 		}
@@ -311,7 +317,7 @@ void garbage_collection(int u)
 	g:
 	if ((void *)(e = e->next) == &cache) return;
 	if (!u) for (f = e; (void *)f != &cache; f = f->next) {
-		if (ncs + f->data_size <= memory_cache_size * MEMORY_CACHE_GC_PERCENT) {
+		if (ncs + f->data_size <= (longlong)memory_cache_size * MEMORY_CACHE_GC_PERCENT) {
 			ncs += f->data_size;
 			f->tgc = 0;
 		}
@@ -320,7 +326,16 @@ void garbage_collection(int u)
 		f = f->next;
 		if (f->prev->tgc) delete_cache_entry(f->prev);
 	}
-	/*if (!no && cache_size > memory_cache_size * MEMORY_CACHE_GC_PERCENT) {
+	/*if (!no && cache_size > (longlong)memory_cache_size * MEMORY_CACHE_GC_PERCENT) {
 		internal("garbage collection doesn't work, cache size %ld", cache_size);
 	}*/
 }
+
+void init_cache(void)
+{
+#ifdef HAVE_GETPAGESIZE
+	int getpg = getpagesize();
+	if (getpg > 0 && getpg < 0x10000 && !(getpg & (getpg - 1))) page_size = getpg;
+#endif
+}
+
