@@ -14,8 +14,9 @@ struct codepage_desc {
 #include "codepage.inc"
 #include "uni_7b.inc"
 #include "entity.inc"
+#include "upcase.inc"
 
-char *strings[256] = {
+char strings[256][2] = {
 	"\000", "\001", "\002", "\003", "\004", "\005", "\006", "\007",
 	"\010", "\011", "\012", "\013", "\014", "\015", "\016", "\017",
 	"\020", "\021", "\022", "\023", "\024", "\025", "\026", "\033",
@@ -62,25 +63,24 @@ unsigned char *no_str = NULL;
 void new_translation_table(struct conv_table *p)
 {
 	int i;
-	if (!no_str) no_str = stracpy("*");
 	for (i = 0; i < 256; i++) if (p[i].t) free_translation_table(p[i].u.tbl);
 	for (i = 0; i < 128; i++) p[i].t = 0, p[i].u.str = strings[i];
 	for (; i < 256; i++) p[i].t = 0, p[i].u.str = no_str;
 }
 
-#define BIN_SEARCH(table, entry, entries, key, result)					\
-{											\
-	int _s = 0, _e = (entries) - 1;							\
-	while (_s <= _e || !((result) = -1)) {						\
-		int _m = (_s + _e) / 2;							\
-		if ((table)[_m].entry == (key)) {					\
-			(result) = _m;							\
-			break;								\
-		}									\
-		if ((table)[_m].entry > (key)) _e = _m - 1;				\
-		if ((table)[_m].entry < (key)) _s = _m + 1;				\
-	}										\
-}											\
+#define BIN_SEARCH(table, entry, entries, key, result)			\
+{									\
+	int _s = 0, _e = (entries) - 1;					\
+	while (_s <= _e || !((result) = -1)) {				\
+		int _m = (_s + _e) / 2;					\
+		if ((table)[_m].entry == (key)) {			\
+			(result) = _m;					\
+			break;						\
+		}							\
+		if ((table)[_m].entry > (key)) _e = _m - 1;		\
+		if ((table)[_m].entry < (key)) _s = _m + 1;		\
+	}								\
+}									\
 
 int strange_chars[32] = {
 0x20ac, 0x0000, 0x002a, 0x0000, 0x201e, 0x2026, 0x2020, 0x2021,
@@ -89,24 +89,35 @@ int strange_chars[32] = {
 0x007e, 0x2122, 0x0161, 0x003e, 0x0153, 0x0000, 0x0000, 0x0000,
 };
 
-static inline unsigned char *u2cp(int u, int to)
+static inline unsigned char *u2cp(int u, int to, int fallback)
 {
 	int j, s;
+	again:
 	if (u < 128) return strings[u];
 	if (u == 0xa0) return "\001";
 	if (u == 0xad) return "";
 	if (u < 0xa0) {
-		if (!strange_chars[u - 0x80]) return NULL;
-		return u2cp(strange_chars[u - 0x80], to);
+		u = strange_chars[u - 0x80];
+		if (!u) return NULL;
+		goto again;
 	}
 	for (j = 0; codepages[to].table[j].c; j++)
 		if (codepages[to].table[j].u == u)
 			return strings[codepages[to].table[j].c];
+	if (!fallback) return NULL;
 	BIN_SEARCH(unicode_7b, x, N_UNICODE_7B, u, s);
 	if (s != -1) return unicode_7b[s].s;
 	return NULL;
 }
 
+int cp2u(unsigned char ch, int from)
+{
+	struct table_entry *e;
+	if (from < 0 || ch < 0x80) return ch;
+	for (e = codepages[from].table; e->c; e++) if (e->c == ch) return e->u;
+	return -1;
+}
+
 unsigned char utf_buffer[7];
 
 unsigned char *encode_utf_8(int u)
@@ -151,7 +162,7 @@ void add_utf_8(struct conv_table *ct, int u, unsigned char *str)
 				internal("bad utf encoding #1");
 				return;
 			}
-			if (!(nct = mem_alloc(sizeof(struct conv_table) * 256))) return;
+			nct = mem_alloc(sizeof(struct conv_table) * 256);
 			memset(nct, 0, sizeof(struct conv_table) * 256);
 			new_translation_table(nct);
 			ct[*p].t = 1;
@@ -208,7 +219,7 @@ void free_conv_table()
 	if (!utf_table_init) free_utf_table();
 	if (first) memset(table, 0, sizeof(struct conv_table) * 256), first = 0;
 	new_translation_table(table);
-	mem_free(no_str), no_str = NULL;
+	if (no_str) mem_free(no_str), no_str = NULL;
 }
 
 struct conv_table *get_translation_table(int from, int to)
@@ -216,6 +227,7 @@ struct conv_table *get_translation_table(int from, int to)
 	int i;
 	static int lfr = -1;
 	static int lto = -1;
+	if (!no_str) no_str = stracpy("*");
 	if (first) memset(table, 0, sizeof(struct conv_table) * 256), first = 0;
 	if (/*from == to ||*/ from == -1 || to == -1) return NULL;
 	if (codepages[to].table == table_utf_8) return get_translation_table_to_utf_8(from);
@@ -224,7 +236,7 @@ struct conv_table *get_translation_table(int from, int to)
 	new_translation_table(table);
 	if (codepages[from].table == table_utf_8) {
 		int j;
-		for (j = 0; codepages[to].table[j].c; j++) add_utf_8(table, codepages[to].table[j].u, strings[codepages[to].table[j].c]);
+		for (j = 0; codepages[to].table[j].c; j++) add_utf_8(table, codepages[to].table[j].u, codepages[to].table[j].u == 0xa0 ? "\001" : codepages[to].table[j].u == 0xad ? "" : strings[codepages[to].table[j].c]);
 		for (i = 0; unicode_7b[i].x != -1; i++) if (unicode_7b[i].x >= 0x80) add_utf_8(table, unicode_7b[i].x, unicode_7b[i].s);
 	} else for (i = 128; i < 256; i++) {
 		int j;
@@ -234,7 +246,7 @@ struct conv_table *get_translation_table(int from, int to)
 		}
 		continue;
 		f:
-		u = u2cp(codepages[from].table[j].u, to);
+		u = u2cp(codepages[from].table[j].u, to, 1);
 		if (u) table[i].u.str = u;
 	}
 	return table;
@@ -250,34 +262,39 @@ static inline int xxstrcmp(unsigned char *s1, unsigned char *s2, int l2)
 	return !!*s1;
 }
 
+int get_entity_number(unsigned char *st, int l)
+{
+	int n = 0;
+	if (upcase(st[0]) == 'X') {
+		st++, l--;
+		if (!l) return -1;
+		do {
+			char c = upcase(*(st++));
+			if (c >= '0' && c <= '9') n = n * 16 + c - '0';
+			else if (c >= 'A' && c <= 'F') n = n * 16 + c - 'A' + 10;
+			else return -1;
+			if (n >= 0x10000) return -1;
+		} while (--l);
+	} else {
+		if (!l) return -1;
+		do {
+			char c = *(st++);
+			if (c >= '0' && c <= '9') n = n * 10 + c - '0';
+			else return -1;
+			if (n >= 0x10000) return -1;
+		} while (--l);
+	}
+	return n;
+}
+
 unsigned char *get_entity_string(unsigned char *st, int l, int encoding)
 {
 	int n;
 	if (l <= 0) return NULL;
 	if (st[0] == '#') {
 		if (l == 1) return NULL;
-		if (st[1] == 'x') {
-			n = 0;
-			if (l == 2 || l > 6) return NULL;
-			st += 2, l -= 2;
-			do {
-				char c = upcase(*(st++));
-				if (c >= '0' && c <= '9') n = n * 16 + c - '0';
-				else if (c >= 'A' && c <= 'F') n = n * 16 + c - 'A' + 10;
-				else return NULL;
-				if (n >= 0x10000) return NULL;
-			} while (--l);
-		} else {
-			n = 0;
-			if (l > 6) return NULL;
-			st++, l--;
-			do {
-				char c = *(st++);
-				if (c >= '0' && c <= '9') n = n * 10 + c - '0';
-				else return NULL;
-				if (n >= 0x10000) return NULL;
-			} while (--l);
-		}
+		if ((n = get_entity_number(st + 1, l - 1)) == -1) return NULL;
+		if (n < 32 && get_attr_val_nl != 2) n = 32;
 	} else {
 		int s = 0, e = N_ENTITIES - 1;
 		while (s <= e) {
@@ -295,13 +312,12 @@ unsigned char *get_entity_string(unsigned char *st, int l, int encoding)
 		f:;
 	}
 
-	return u2cp(n, encoding);
+	return u2cp(n, encoding, 1);
 }
 
 unsigned char *convert_string(struct conv_table *ct, unsigned char *c, int l)
 {
 	unsigned char *buffer;
-	unsigned char *b;
 	int bp = 0;
 	int pp = 0;
 	if (!ct) {
@@ -310,19 +326,22 @@ unsigned char *convert_string(struct conv_table *ct, unsigned char *c, int l)
 		return memacpy(c, l);
 		xx:;
 	}
-	if (!(buffer = mem_alloc(ALLOC_GR))) return NULL;
+	buffer = mem_alloc(ALLOC_GR);
 	while (pp < l) {
 		unsigned char *e;
 		if (c[pp] < 128 && c[pp] != '&') {
-			putc:
+			put_c:
 			buffer[bp++] = c[pp++];
-			if (!(bp & (ALLOC_GR - 1))) if ((b = mem_realloc(buffer, bp + ALLOC_GR))) buffer = b; else bp--;
+			if (!(bp & (ALLOC_GR - 1))) {
+				if ((unsigned)bp > MAXINT - ALLOC_GR) overalloc();
+				buffer = mem_realloc(buffer, bp + ALLOC_GR);
+			}
 			continue;
 		}
 		if (c[pp] != '&') {
 			struct conv_table *t;
 			int i;
-			if (!ct) goto putc;
+			if (!ct) goto put_c;
 			t = ct;
 			i = pp;
 			decode:
@@ -330,26 +349,32 @@ unsigned char *convert_string(struct conv_table *ct, unsigned char *c, int l)
 				e = t[c[i]].u.str;
 			} else {
 				t = t[c[i++]].u.tbl;
-				if (i >= l) goto putc;
+				if (i >= l) goto put_c;
 				goto decode;
 			}
 			pp = i + 1;
 		} else {
 			int i = pp + 1;
-			if (d_opt->plain) goto putc;
+			if (d_opt->plain) goto put_c;
 			while (i < l && c[i] != ';' && c[i] != '&' && c[i] > ' ') i++;
-			if (!(e = get_entity_string(&c[pp + 1], i - pp - 1, d_opt->cp))) goto putc;
+			if (!(e = get_entity_string(&c[pp + 1], i - pp - 1, d_opt->cp))) goto put_c;
 			pp = i + (i < l && c[i] == ';');
 		}
 		if (!e[0]) continue;
 		if (!e[1]) {
 			buffer[bp++] = e[0];
-			if (!(bp & (ALLOC_GR - 1))) if ((b = mem_realloc(buffer, bp + ALLOC_GR))) buffer = b; else bp--;
+			if (!(bp & (ALLOC_GR - 1))) {
+				if ((unsigned)bp > MAXINT - ALLOC_GR) overalloc();
+				buffer = mem_realloc(buffer, bp + ALLOC_GR);
+			}
 			continue;
 		}
 		while (*e) {
 			buffer[bp++] = *(e++);
-			if (!(bp & (ALLOC_GR - 1))) if ((b = mem_realloc(buffer, bp + ALLOC_GR))) buffer = b; else bp--;
+			if (!(bp & (ALLOC_GR - 1))) {
+				if ((unsigned)bp > MAXINT - ALLOC_GR) overalloc();
+				buffer = mem_realloc(buffer, bp + ALLOC_GR);
+			}
 		}
 	}
 	buffer[bp] = 0;
@@ -367,7 +392,7 @@ int get_cp_index(unsigned char *n)
 					for (q = 1; codepages[i].aliases[a][q]; q++) {
 						if (upcase(n[p+q]) != upcase(codepages[i].aliases[a][q])) goto fail;
 					}
-					if (strlen(codepages[i].aliases[a]) > ll) {
+					if (strlen(codepages[i].aliases[a]) > (size_t)ll) {
 						ll = strlen(codepages[i].aliases[a]);
 						ii = i;
 					}
@@ -396,3 +421,24 @@ int is_cp_special(int index)
 {
 	return codepages[index].table == table_utf_8;
 }
+
+unsigned char charset_upcase(unsigned char ch, int cp)
+{
+	int u, res;
+	unsigned char *str;
+	if (ch < 0x80) return upcase(ch);
+	u = cp2u(ch, cp);
+	BIN_SEARCH(unicode_upcase, lo, sizeof(unicode_upcase) / sizeof(*unicode_upcase), u, res);
+	if (res == -1) return ch;
+	str = u2cp(unicode_upcase[res].up, cp, 0);
+	if (!str || !str[0] || str[1]) return ch;
+	return str[0];
+}
+
+void charset_upcase_string(unsigned char **chp, int cp)
+{
+	unsigned char *ch = *chp;
+	int i;
+	for (i = 0; ch[i]; i++) ch[i] = charset_upcase(ch[i], cp);
+}
+
