@@ -55,16 +55,13 @@ void make_connection(struct connection *c, int port, int *sock, void (*func)(str
 		abort_connection(c);
 		return;
 	}
-	if (!(b = mem_alloc(sizeof(struct conn_info)))) {
-		mem_free(host);
-		setcstate(c, S_OUT_OF_MEM);
-		retry_connection(c);
-		return;
-	}
+	if (c->newconn)
+		internal("already making a connection");
+	b = mem_alloc(sizeof(struct conn_info));
 	b->func = func;
 	b->sock = sock;
 	b->port = port;
-	c->buffer = b;
+	c->newconn = b;
 	log_data("\nCONNECTION: ", 13);
 	log_data(host, strlen(host));
 	log_data("\n", 1);
@@ -79,16 +76,16 @@ int get_pasv_socket(struct connection *c, int cc, int *sock, unsigned char *port
 	int s;
 	struct sockaddr_in sa;
 	struct sockaddr_in sb;
-	int len = sizeof(sa);
+	socklen_t len = sizeof(sa);
 	memset(&sa, 0, sizeof sa);
 	memset(&sb, 0, sizeof sb);
 	if (getsockname(cc, (struct sockaddr *)&sa, &len)) {
 		e:
-		setcstate(c, -errno);
+		setcstate(c, get_error_from_errno(errno));
 		retry_connection(c);
-		return -2;
+		return -1;
 	}
-	if ((s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) goto e;
+	if ((s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) goto e;
 	*sock = s;
 	fcntl(s, F_SETFL, O_NONBLOCK);
 	memcpy(&sb, &sa, sizeof(struct sockaddr_in));
@@ -99,32 +96,34 @@ int get_pasv_socket(struct connection *c, int cc, int *sock, unsigned char *port
 	if (listen(s, 1)) goto e;
 	memcpy(port, &sa.sin_addr.s_addr, 4);
 	memcpy(port + 4, &sa.sin_port, 2);
-#if defined(IP_TOS) && defined(IPTOS_THROUGHPUT)
-	{
-		int on = IPTOS_THROUGHPUT;
-		setsockopt(s, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int));
-	}
-#endif
 	return 0;
 }
 
 #ifdef HAVE_SSL
 void ssl_want_read(struct connection *c)
 {
-	struct conn_info *b = c->buffer;
+	struct conn_info *b = c->newconn;
+
+	set_timeout(c);
 
 	if (c->no_tsl) c->ssl->options |= SSL_OP_NO_TLSv1;
 	switch (SSL_get_error(c->ssl, SSL_connect(c->ssl))) {
 		case SSL_ERROR_NONE:
-			c->buffer = NULL;
+			c->newconn = NULL;
 			b->func(c);
 			mem_free(b);
+			break;
 		case SSL_ERROR_WANT_READ:
+			set_handlers(*b->sock, (void(*)(void *))ssl_want_read, NULL, (void(*)(void *))exception, c);
+			break;
+		case SSL_ERROR_WANT_WRITE:
+			set_handlers(*b->sock, NULL, (void(*)(void *))ssl_want_read, (void(*)(void *))exception, c);
 			break;
 		default:
 			c->no_tsl++;
 			setcstate(c, S_SSL_ERROR);
 			retry_connection(c);
+			break;
 	}
 }
 #endif
@@ -132,104 +131,84 @@ void ssl_want_read(struct connection *c)
 void dns_found(struct connection *c, int state)
 {
 	int s;
-	struct conn_info *b = c->buffer;
+	struct conn_info *b = c->newconn;
 	if (state) {
 		setcstate(c, S_NO_DNS);
-		retry_connection(c);
+		abort_connection(c);
 		return;
 	}
-	if ((s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
-		setcstate(c, -errno);
+	if ((s = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
+		setcstate(c, get_error_from_errno(errno));
 		retry_connection(c);
 		return;
 	}
 	*b->sock = s;
 	fcntl(s, F_SETFL, O_NONBLOCK);
 	memset(&b->sa, 0, sizeof(struct sockaddr_in));
-	b->sa.sin_family = PF_INET;
+	b->sa.sin_family = AF_INET;
 	b->sa.sin_addr.s_addr = b->addr;
 	b->sa.sin_port = htons(b->port);
 	if (connect(s, (struct sockaddr *)&b->sa, sizeof b->sa)) {
 		if (errno != EALREADY && errno != EINPROGRESS) {
-			setcstate(c, -errno);
+#ifdef BEOS
+			if (errno == EWOULDBLOCK) errno = ETIMEDOUT;
+#endif
+			setcstate(c, get_error_from_errno(errno));
 			retry_connection(c);
 			return;
 		}
 		set_handlers(s, NULL, (void(*)(void *))connected, (void(*)(void *))exception, c);
 		setcstate(c, S_CONN);
 	} else {
-#ifdef HAVE_SSL
-		if(c->ssl) {
-			c->ssl = getSSL();
-			SSL_set_fd(c->ssl,s);
-			if (c->no_tsl) c->ssl->options |= SSL_OP_NO_TLSv1;
-			switch (SSL_get_error(c->ssl, SSL_connect(c->ssl))) {
-				case SSL_ERROR_WANT_READ:
-					setcstate(c, S_SSL_NEG);
-					set_handlers(s, (void(*)(void *))ssl_want_read, NULL, (void(*)(void *))exception, c);
-					return;
-				case SSL_ERROR_NONE:
-					break;
-				default:
-					c->no_tsl++;
-					setcstate(c, S_SSL_ERROR);
-					retry_connection(c);
-					return;
-			}
-		}
-#endif
-		c->buffer = NULL;
-		b->func(c);
-		mem_free(b);
+		connected(c);
 	}
 }
 
 void connected(struct connection *c)
 {
-	struct conn_info *b = c->buffer;
-	void (*func)(struct connection *) = b->func;
+	struct conn_info *b = c->newconn;
 	int err = 0;
-	int len = sizeof(int);
-	/*if (!connect(*b->sock, (struct sockaddr *)&b->sa, sizeof b->sa) || errno == EISCONN) {
-		mem_free(b);
-		func(c);
-		return;
-	}
-	mem_free(b);
-	setcstate(c, -errno);*/
+	socklen_t len = sizeof(int);
 	if (getsockopt(*b->sock, SOL_SOCKET, SO_ERROR, (void *)&err, &len))
 		if (!(err = errno)) {
 			err = -(S_STATE);
 			goto bla;
 		}
 	if (err >= 10000) err -= 10000;	/* Why does EMX return so large values? */
-	bla:
-	if (err > 0) setcstate(c, -err), retry_connection(c);
-	else {
-#ifdef HAVE_SSL
-		if(c->ssl) {
-			c->ssl = getSSL();
-			SSL_set_fd(c->ssl, *b->sock);
-			if (c->no_tsl) c->ssl->options |= SSL_OP_NO_TLSv1;
-			switch (SSL_get_error(c->ssl, SSL_connect(c->ssl))) {
-				case SSL_ERROR_WANT_READ:
-					setcstate(c, S_SSL_NEG);
-					set_handlers(*b->sock, (void(*)(void *))ssl_want_read, NULL, (void(*)(void *))exception, c);
-					return;
-				case SSL_ERROR_NONE:
-					break;
-				default:
-					c->no_tsl++;
-					setcstate(c, S_SSL_ERROR);
-					retry_connection(c);
-					return;
-			}
-		}
-#endif
-		c->buffer = NULL;
-		mem_free(b);
-		func(c);
+	if (err > 0) {
+		bla:
+		setcstate(c, get_error_from_errno(err));
+		retry_connection(c);
+		return;
 	}
+	set_timeout(c);
+#ifdef HAVE_SSL
+	if (c->ssl) {
+		c->ssl = getSSL();
+		SSL_set_fd(c->ssl, *b->sock);
+		if (c->no_tsl) c->ssl->options |= SSL_OP_NO_TLSv1;
+		switch (SSL_get_error(c->ssl, SSL_connect(c->ssl))) {
+			case SSL_ERROR_WANT_READ:
+				setcstate(c, S_SSL_NEG);
+				set_handlers(*b->sock, (void(*)(void *))ssl_want_read, NULL, (void(*)(void *))exception, c);
+				return;
+			case SSL_ERROR_WANT_WRITE:
+				setcstate(c, S_SSL_NEG);
+				set_handlers(*b->sock, NULL, (void(*)(void *))ssl_want_read, (void(*)(void *))exception, c);
+				return;
+			case SSL_ERROR_NONE:
+				break;
+			default:
+				c->no_tsl++;
+				setcstate(c, S_SSL_ERROR);
+				retry_connection(c);
+				return;
+		}
+	}
+#endif
+	c->newconn = NULL;
+	b->func(c);
+	mem_free(b);
 }
 
 struct write_buffer {
@@ -250,6 +229,7 @@ void write_select(struct connection *c)
 		abort_connection(c);
 		return;
 	}
+	set_timeout(c);
 	/*printf("ws: %d\n",wb->len-wb->pos);
 	for (wr = wb->pos; wr < wb->len; wr++) printf("%c", wb->data[wr]);
 	printf("-\n");*/
@@ -259,7 +239,7 @@ void write_select(struct connection *c)
 		if ((wr = SSL_write(c->ssl, wb->data + wb->pos, wb->len - wb->pos)) <= 0) {
 			int err;
 			if ((err = SSL_get_error(c->ssl, wr)) != SSL_ERROR_WANT_WRITE) {
-				setcstate(c, wr ? (err == SSL_ERROR_SYSCALL ? -errno : S_SSL_ERROR) : S_CANT_WRITE);
+				setcstate(c, wr ? (err == SSL_ERROR_SYSCALL ? get_error_from_errno(errno) : S_SSL_ERROR) : S_CANT_WRITE);
 				if (!wr || err == SSL_ERROR_SYSCALL) retry_connection(c);
 				else abort_connection(c);
 				return;
@@ -269,7 +249,13 @@ void write_select(struct connection *c)
 	} else
 #endif
 		if ((wr = write(wb->sock, wb->data + wb->pos, wb->len - wb->pos)) <= 0) {
-			setcstate(c, wr ? -errno : S_CANT_WRITE);
+#ifdef ATHEOS
+	/* Workaround for a bug in Syllable */
+			if (wr && errno == EAGAIN) {
+				return;
+			}
+#endif
+			setcstate(c, wr ? get_error_from_errno(errno) : S_CANT_WRITE);
 			retry_connection(c);
 			return;
 		}
@@ -288,11 +274,8 @@ void write_to_socket(struct connection *c, int s, unsigned char *data, int len,
 {
 	struct write_buffer *wb;
 	log_data(data, len);
-	if (!(wb = mem_alloc(sizeof(struct write_buffer) + len))) {
-		setcstate(c, S_OUT_OF_MEM);
-		abort_connection(c);
-		return;
-	}
+	if ((unsigned)len > MAXINT - sizeof(struct write_buffer)) overalloc();
+	wb = mem_alloc(sizeof(struct write_buffer) + len);
 	wb->sock = s;
 	wb->len = len;
 	wb->pos = 0;
@@ -300,10 +283,10 @@ void write_to_socket(struct connection *c, int s, unsigned char *data, int len,
 	memcpy(wb->data, data, len);
 	if (c->buffer) mem_free(c->buffer);
 	c->buffer = wb;
-	set_handlers(s, NULL, (void (*)())write_select, (void (*)())exception, c);
+	set_handlers(s, NULL, (void (*)(void *))write_select, (void (*)(void *))exception, c);
 }
 
-#define READ_SIZE 16384
+#define READ_SIZE	64240
 
 void read_select(struct connection *c)
 {
@@ -316,11 +299,8 @@ void read_select(struct connection *c)
 		return;
 	}
 	set_handlers(rb->sock, NULL, NULL, NULL, NULL);
-	if (!(rb = mem_realloc(rb, sizeof(struct read_buffer) + rb->len + READ_SIZE))) {
-		setcstate(c, S_OUT_OF_MEM);
-		abort_connection(c);
-		return;
-	}
+	if ((unsigned)rb->len > MAXINT - sizeof(struct read_buffer) - READ_SIZE) overalloc();
+	rb = mem_realloc(rb, sizeof(struct read_buffer) + rb->len + READ_SIZE);
 	c->buffer = rb;
 
 #ifdef HAVE_SSL
@@ -336,7 +316,7 @@ void read_select(struct connection *c)
 				rb->done(c, rb);
 				return;
 			}
-			setcstate(c, rd ? (err == SSL_ERROR_SYSCALL ? -errno : S_SSL_ERROR) : S_CANT_READ);
+			setcstate(c, rd ? (err == SSL_ERROR_SYSCALL ? get_error_from_errno(errno) : S_SSL_ERROR) : S_CANT_READ);
 			/*mem_free(rb);*/
 			if (!rd || err == SSL_ERROR_SYSCALL) retry_connection(c);
 			else abort_connection(c);
@@ -350,7 +330,7 @@ void read_select(struct connection *c)
 				rb->done(c, rb);
 				return;
 			}
-			setcstate(c, rd ? -errno : S_CANT_READ);
+			setcstate(c, rd ? get_error_from_errno(errno) : S_CANT_READ);
 			/*mem_free(rb);*/
 			retry_connection(c);
 			return;
@@ -363,11 +343,7 @@ void read_select(struct connection *c)
 struct read_buffer *alloc_read_buffer(struct connection *c)
 {
 	struct read_buffer *rb;
-	if (!(rb = mem_alloc(sizeof(struct read_buffer) + READ_SIZE))) {
-		setcstate(c, S_OUT_OF_MEM);
-		abort_connection(c);
-		return NULL;
-	}
+	rb = mem_alloc(sizeof(struct read_buffer) + READ_SIZE);
 	memset(rb, 0, sizeof(struct read_buffer));
 	return rb;
 }
@@ -378,7 +354,7 @@ void read_from_socket(struct connection *c, int s, struct read_buffer *buf, void
 	buf->sock = s;
 	if (c->buffer && buf != c->buffer) mem_free(c->buffer);
 	c->buffer = buf;
-	set_handlers(s, (void (*)())read_select, NULL, (void (*)())exception, c);
+	set_handlers(s, (void (*)(void *))read_select, NULL, (void (*)(void *))exception, c);
 }
 
 void kill_buffer_data(struct read_buffer *rb, int n)
