@@ -44,7 +44,7 @@ void delete_cookie(struct cookie *);
 void free_cookie(struct cookie *c)
 {
 	mem_free(c->name);
-	mem_free(c->value);
+	if (c->value) mem_free(c->value);
 	if (c->server) mem_free(c->server);
 	if (c->path) mem_free(c->path);
 	if (c->domain) mem_free(c->domain);
@@ -52,7 +52,8 @@ void free_cookie(struct cookie *c)
 
 int check_domain_security(unsigned char *server, unsigned char *domain)
 {
-	int i, j, dl, nd;
+	size_t i, j, dl;
+	int nd;
 	if (domain[0] == '.') domain++;
 	dl = strlen(domain);
 	if (dl > strlen(server)) return 1;
@@ -73,18 +74,21 @@ int check_domain_security(unsigned char *server, unsigned char *domain)
 
 int set_cookie(struct terminal *term, unsigned char *url, unsigned char *str)
 {
+	int noval = 0;
 	struct cookie *cookie;
 	struct c_server *cs;
-	unsigned char *p, *q, *s, *server, *date, *document;
+	unsigned char *p, *q, *s, *server, *date;
 	if (accept_cookies == ACCEPT_NONE) return 0;
 	for (p = str; *p != ';' && *p; p++) /*if (WHITECHAR(*p)) return 0*/;
-	for (q = str; *q != '='; q++) if (!*q || q >= p) return 0;
+	for (q = str; *q != '='; q++) if (!*q || q >= p) {
+		noval = 1;
+		break;
+	}
 	if (str == q || q + 1 == p) return 0;
-	if (!(cookie = mem_alloc(sizeof(struct cookie)))) return 0;
-	document = get_url_data(url);
+	cookie = mem_alloc(sizeof(struct cookie));
 	server = get_host_name(url);
 	cookie->name = memacpy(str, q - str);
-	cookie->value = memacpy(q + 1, p - q - 1);
+	cookie->value = !noval ? memacpy(q + 1, p - q - 1) : NULL;
 	cookie->server = stracpy(server);
 	date = parse_header_param(str, "expires");
 	if (date) {
@@ -95,8 +99,9 @@ int set_cookie(struct terminal *term, unsigned char *url, unsigned char *str)
 	} else
 		cookie->expires = 0;
 	if (!(cookie->path = parse_header_param(str, "path"))) {
-		unsigned char *w;
+		/*unsigned char *w;*/
 		cookie->path = stracpy("/");
+		/*
 		add_to_strn(&cookie->path, document);
 		for (w = cookie->path; *w; w++) if (end_of_dir(*w)) {
 			*w = 0;
@@ -107,9 +112,8 @@ int set_cookie(struct terminal *term, unsigned char *url, unsigned char *str)
 				w[1] = 0;
 				break;
 			}
+		*/
 	} else {
-		if (!cookie->path[0] || cookie->path[strlen(cookie->path) - 1] != '/')
-			add_to_strn(&cookie->path, "/");
 		if (cookie->path[0] != '/') {
 			add_to_strn(&cookie->path, "x");
 			memmove(cookie->path + 1, cookie->path, strlen(cookie->path) - 1);
@@ -159,9 +163,14 @@ void accept_cookie(struct cookie *c)
 		free_cookie(e);
 		mem_free(e);
 	}
+	if (c->value && !strcasecmp(c->value, "deleted")) {
+		free_cookie(c);
+		mem_free(c);
+		return;
+	}
 	add_to_list(cookies, c);
 	foreach(cd, c_domains) if (!strcasecmp(cd->domain, c->domain)) return;
-	if (!(cd = mem_alloc(sizeof(struct c_domain) + strlen(c->domain) + 1))) return;
+	cd = mem_alloc(sizeof(struct c_domain) + strlen(c->domain) + 1);
 	strcpy(cd->domain, c->domain);
 	add_to_list(c_domains, cd);
 }
@@ -203,12 +212,11 @@ void cookie_default(void *idp, int a)
 	struct c_server *s;
 	if (!(c = find_cookie_id(idp))) return;
 	foreach(s, c_servers) if (!strcasecmp(s->server, c->server)) goto found;
-	if ((s = mem_alloc(sizeof(struct c_server) + strlen(c->server) + 1))) {
-		strcpy(s->server, c->server);
-		add_to_list(c_servers, s);
-		found:
-		s->accept = a;
-	}
+	s = mem_alloc(sizeof(struct c_server) + strlen(c->server) + 1);
+	strcpy(s->server, c->server);
+	add_to_list(c_servers, s);
+	found:
+	s->accept = a;
 }
 
 void accept_cookie_always(void *idp)
@@ -236,8 +244,10 @@ int is_path_prefix(unsigned char *d, unsigned char *s)
 {
 	int dl = strlen(d);
 	int sl = strlen(s);
+	if (!dl) return 1;
 	if (dl > sl) return 0;
-	return !memcmp(d, s, dl);
+	if (memcmp(d, s, dl)) return 0;
+	return d[dl - 1] == '/' || !s[dl] || s[dl] == '/' || s[dl] == POST_CHAR || s[dl] == '?' || s[dl] == '&';
 }
 
 int cookie_expired(struct cookie *c)	/* parse_http_date is broken */
@@ -266,12 +276,14 @@ void send_cookies(unsigned char **s, int *l, unsigned char *url)
 			mem_free(d);
 			continue;
 		}
-		if (c->secure) continue;
+		if (c->secure && casecmp(url, "https://", 8)) continue;
 		if (!nc) add_to_str(s, l, "Cookie: "), nc = 1;
 		else add_to_str(s, l, "; ");
 		add_to_str(s, l, c->name);
-		add_to_str(s, l, "=");
-		add_to_str(s, l, c->value);
+		if (c->value) {
+			add_to_str(s, l, "=");
+			add_to_str(s, l, c->value);
+		}
 	}
 	if (nc) add_to_str(s, l, "\r\n");
 	mem_free(server);
