@@ -2,8 +2,6 @@
 
 void get_system_name()
 {
-	FILE *f;
-	unsigned char *p;
 #if defined(HAVE_SYS_UTSNAME_H) && defined(HAVE_UNAME)
 	struct utsname name;
 	memset(&name, 0, sizeof name);
@@ -22,18 +20,22 @@ void get_system_name()
 	}
 #endif
 #ifdef HAVE_POPEN
-	memset(system_name, 0, MAX_STR_LEN);
-	if (!(f = popen("uname -srm", "r"))) goto fail;
-	if (fread(system_name, 1, MAX_STR_LEN - 1, f) <= 0) {
+	{
+		FILE *f;
+		unsigned char *p;
+		memset(system_name, 0, MAX_STR_LEN);
+		if (!(f = popen("uname -srm", "r"))) goto fail;
+		if (fread(system_name, 1, MAX_STR_LEN - 1, f) <= 0) {
+			pclose(f);
+			goto fail;
+		}
 		pclose(f);
-		goto fail;
+		for (p = system_name; *p; p++) if (*p < ' ') {
+			*p = 0;
+			break;
+		}
+		if (system_name[0]) return;
 	}
-	pclose(f);
-	for (p = system_name; *p; p++) if (*p < ' ') {
-		*p = 0;
-		break;
-	}
-	if (system_name[0]) return;
 	fail:
 #endif
 	strcpy(system_name, SYSTEM_NAME);
@@ -44,7 +46,7 @@ extern struct option html_options[];
 
 struct option *all_options[] = { links_options, html_options, NULL, };
 
-unsigned char *_parse_options(int argc, unsigned char *argv[], struct option **opt)
+unsigned char *p_arse_options(int argc, unsigned char *argv[], struct option **opt)
 {
 	unsigned char *e, *u = NULL;
 	while (argc) {
@@ -75,7 +77,7 @@ unsigned char *_parse_options(int argc, unsigned char *argv[], struct option **o
 
 unsigned char *parse_options(int argc, unsigned char *argv[])
 {
-	return _parse_options(argc, argv, all_options);
+	return p_arse_options(argc, argv, all_options);
 }
 
 unsigned char *get_token(unsigned char **line)
@@ -140,7 +142,7 @@ void parse_config_file(unsigned char *name, unsigned char *file, struct option *
 		if (!(tok = get_token(&n))) goto f;
 		nl = strlen(tok);
 		for (op = opt; (options = *op); op++) 
-		    	for (i = 0; options[i].p; i++) if (options[i].cfg_name && nl == strlen(options[i].cfg_name) && !casecmp(tok, options[i].cfg_name, nl)) {
+		    	for (i = 0; options[i].p; i++) if (options[i].cfg_name && (size_t)nl == strlen(options[i].cfg_name) && !casecmp(tok, options[i].cfg_name, nl)) {
 				unsigned char *o = memacpy(p, pl);
 				if ((e = options[i].rd_cfg(&options[i], o))) {
 					if (e[0]) fprintf(stderr, "Error parsing config file %s, line %d: %s\n", name, line, e), err = 1;
@@ -192,30 +194,92 @@ unsigned char *read_config_file(unsigned char *name)
 
 int write_to_config_file(unsigned char *name, unsigned char *c)
 {
-	int rr = strlen(c);
-	int r = rr;
+	int rr;
+	int r;
 	int h, w;
-	if ((h = open(name, O_WRONLY | O_NOCTTY | O_CREAT | O_TRUNC, 0666)) == -1) return -1;
+	unsigned char *tmp_name = stracpy(name);
+	unsigned char *ds, *dt, *dot;
+	for (dt = ds = tmp_name; *dt; dt++) if (dir_sep(*dt)) ds = dt;
+	if ((dot = strchr(ds, '.'))) *dot = 0;
+	add_to_strn(&tmp_name, ".tmp");
+	if ((h = open(tmp_name, O_WRONLY | O_NOCTTY | O_CREAT | O_TRUNC, 0666)) == -1) {
+		mem_free(tmp_name);
+		return get_error_from_errno(errno);
+	}
 	set_bin(h);
+	rr = strlen(c);
+	r = rr;
 	while (r > 0) {
 		if ((w = write(h, c + rr - r, r)) <= 0) {
+			int err = !w ? ENOSPC : errno;
 			close(h);
-			return -1;
+			unlink(tmp_name);
+			mem_free(tmp_name);
+			return get_error_from_errno(err);
 		}
 		r -= w;
 	}
 	close(h);
+#ifndef RENAME_OVER_EXISTING_FILES
+	unlink(name);
+#endif
+	if (rename(tmp_name, name)) {
+		int err = errno;
+		unlink(tmp_name);
+		mem_free(tmp_name);
+		return get_error_from_errno(err);
+	}
+	mem_free(tmp_name);
 	return 0;
 }
 
 unsigned char *get_home(int *n)
 {
 	struct stat st;
-	unsigned char *home = stracpy(getenv("HOME"));
+	unsigned char *home = NULL;
 	unsigned char *home_links;
 	unsigned char *config_dir = stracpy(getenv("CONFIG_DIR"));
 
 	if (n) *n = 1;
+#ifdef WIN32
+	if (!home) {
+		home = stracpy(getenv("APPDATA"));
+#ifdef HAVE_CYGWIN_CONV_PATH
+		/*
+		 * Newer Cygwin complains about windows-style path, so
+		 * we have to convert it.
+		 */
+		if (home) {
+			unsigned char *new_path;
+			ssize_t sz = cygwin_conv_path(CCP_WIN_A_TO_POSIX | CCP_ABSOLUTE, home, NULL, 0);
+			if (sz < 0)
+				goto skip_path_conv;
+			new_path = mem_alloc(sz);
+			sz = cygwin_conv_path(CCP_WIN_A_TO_POSIX | CCP_ABSOLUTE, home, new_path, sz);
+			if (sz < 0) {
+				mem_free(new_path);
+				goto skip_path_conv;
+			}
+			mem_free(home);
+			home = new_path;
+skip_path_conv:;
+		}
+#endif
+		if (home && (stat(home, &st) == -1 || !S_ISDIR(st.st_mode))) {
+			mem_free(home);
+			home = NULL;
+		}
+	}
+#endif
+	if (!home) home = stracpy(getenv("HOME"));
+#ifdef WIN32
+/* When we run in Cygwin without Cygwin environment, it reports home "/".
+   Unfortunatelly, it can't write anything to that directory */
+	if (home && !strcmp(home, "/")) {
+		mem_free(home);
+		home = NULL;
+	}
+#endif
 	if (!home) {
 		int i;
 		home = stracpy(path_to_exe);
@@ -249,20 +313,33 @@ unsigned char *get_home(int *n)
 		mem_free(config_dir);
 	} else add_to_strn(&home_links, ".links");
 	if (stat(home_links, &st)) {
+#ifdef HAVE_MKDIR
 		if (!mkdir(home_links, 0777)) goto home_creat;
+#endif
 		if (config_dir) goto failed;
 		goto first_failed;
 	}
 	if (S_ISDIR(st.st_mode)) goto home_ok;
+	/* This is a Cygwin hack! Cygwin reports stat for "links" if no
+	   "links" exists and only "links.exe" does. So try to create directory
+	   anyway. */
+	if (!mkdir(home_links, 0777)) goto home_creat;
 	first_failed:
 	mem_free(home_links);
 	home_links = stracpy(home);
 	add_to_strn(&home_links, "links");
 	if (stat(home_links, &st)) {
+#ifdef HAVE_MKDIR
 		if (!mkdir(home_links, 0777)) goto home_creat;
+#else
+		mem_free(home_links);
+		home_links = stracpy(home);
+		goto home_ok;
+#endif
 		goto failed;
 	}
 	if (S_ISDIR(st.st_mode)) goto home_ok;
+	if (!mkdir(home_links, 0777)) goto home_creat;
 	failed:
 	mem_free(home_links);
 	mem_free(home);
@@ -313,14 +390,17 @@ void load_config_file(unsigned char *prefix, unsigned char *name)
 
 void load_config()
 {
-	load_config_file("/etc/", "links.cfg");
+#ifdef SHARED_CONFIG_DIR
+	load_config_file(SHARED_CONFIG_DIR, "links.cfg");
+#endif
 	load_config_file(links_home, "links.cfg");
 	load_config_file(links_home, "html.cfg");
 	load_config_file(links_home, "user.cfg");
 }
 
-int write_config_file(unsigned char *prefix, unsigned char *name, struct option *o, struct terminal *term)
+int write_config_data(unsigned char *prefix, unsigned char *name, struct option *o, struct terminal *term)
 {
+	int err;
 	unsigned char *c, *config_file;
 	if (!(c = create_config_string(o))) return -1;
 	config_file = stracpy(prefix);
@@ -329,8 +409,8 @@ int write_config_file(unsigned char *prefix, unsigned char *name, struct option
 		return -1;
 	}
 	add_to_strn(&config_file, name);
-	if (write_to_config_file(config_file, c)) {
-		if (term) msg_box(term, NULL, TEXT(T_CONFIG_ERROR), AL_CENTER, TEXT(T_UNABLE_TO_WRITE_TO_CONFIG_FILE), NULL, 1, TEXT(T_CANCEL), NULL, B_ENTER | B_ESC);
+	if ((err = write_to_config_file(config_file, c))) {
+		if (term) msg_box(term, NULL, TEXT_(T_CONFIG_ERROR), AL_CENTER | AL_EXTD_TEXT, TEXT_(T_UNABLE_TO_WRITE_TO_CONFIG_FILE), ": ", get_err_msg(err), NULL, NULL, 1, TEXT_(T_CANCEL), NULL, B_ENTER | B_ESC);
 		mem_free(c);
 		mem_free(config_file);
 		return -1;
@@ -342,12 +422,12 @@ int write_config_file(unsigned char *prefix, unsigned char *name, struct option
 
 void write_config(struct terminal *term)
 {
-	write_config_file(links_home, "links.cfg", links_options, term);
+	write_config_data(links_home, "links.cfg", links_options, term);
 }
 
 void write_html_config(struct terminal *term)
 {
-	write_config_file(links_home, "html.cfg", html_options, term);
+	write_config_data(links_home, "html.cfg", html_options, term);
 }
 
 void add_nm(struct option *o, unsigned char **s, int *l)
@@ -399,7 +479,7 @@ unsigned char *str_rd(struct option *o, unsigned char *c)
 	unsigned char *tok = get_token(&c);
 	unsigned char *e = NULL;
 	if (!tok) return NULL;
-	if (strlen(tok) + 1 > o->max) e = "String too long";
+	if (strlen(tok) + 1 > (size_t)o->max) e = "String too long";
 	else strcpy(o->ptr, tok);
 	mem_free(tok);
 	return e;
@@ -408,7 +488,7 @@ unsigned char *str_rd(struct option *o, unsigned char *c)
 void str_wr(struct option *o, unsigned char **s, int *l)
 {
 	add_nm(o, s, l);
-	if (strlen(o->ptr) > o->max - 1) {
+	if (strlen(o->ptr) + 1 > (size_t)o->max) {
 		unsigned char *s1 = init_str();
 		int l1 = 0;
 		add_bytes_to_str(&s1, &l1, o->ptr, o->max - 1);
@@ -424,8 +504,8 @@ unsigned char *cp_rd(struct option *o, unsigned char *c)
 	unsigned char *e = NULL;
 	int i;
 	if (!tok) return "Missing argument";
-	/*if (!strcasecmp(c, "none")) i = -1;
-	else */if ((i = get_cp_index(tok)) == -1) e = "Unknown codepage";
+	if ((i = get_cp_index(tok)) == -1) e = "Unknown codepage";
+	else if (o->min == 1 && is_cp_special(i)) e = "UTF-8 can't be here";
 	else *(int *)o->ptr = i;
 	mem_free(tok);
 	return e;
@@ -462,7 +542,7 @@ void lang_wr(struct option *o, unsigned char **s, int *l)
 int getnum(unsigned char *s, int *n, int r1, int r2)
 {
 	unsigned char *e;
-	long l = strtol(s, (char **)&e, 10);
+	long l = strtol(s, (char **)(void *)&e, 10);
 	if (*e || !*s) return -1;
 	if (l < r1 || l >= r2) return -1;
 	*n = (int)l;
@@ -601,7 +681,7 @@ unsigned char *term_rd(struct option *o, unsigned char *c)
 	ts->block_cursor = !!((w[0] - '0') & 4);
 	mem_free(w);
 	if (!(w = get_token(&c))) goto err;
-	if ((i = get_cp_index(w)) == -1) goto err_f;
+	if ((i = get_cp_index(w)) == -1 || is_cp_special(i)) goto err_f;
 	ts->charset = i;
 	mem_free(w);
 	end:
@@ -640,7 +720,7 @@ unsigned char *term2_rd(struct option *o, unsigned char *c)
 	ts->col = w[0] - '0';
 	mem_free(w);
 	if (!(w = get_token(&c))) goto err;
-	if ((i = get_cp_index(w)) == -1) goto err_f;
+	if ((i = get_cp_index(w)) == -1 || is_cp_special(i)) goto err_f;
 	ts->charset = i;
 	mem_free(w);
 	end:
@@ -670,12 +750,18 @@ void term_wr(struct option *o, unsigned char **s, int *l)
 
 unsigned char *gen_cmd(struct option *o, unsigned char ***argv, int *argc)
 {
+	unsigned char *e;
+	int l;
 	unsigned char *r;
 	if (!*argc) return "Parameter expected";
+	e = init_str();
+	l = 0;
+	add_quoted_to_str(&e, &l, **argv);
+	r = o->rd_cfg(o, e);
+	mem_free(e);
+	if (r) return r;
 	(*argv)++; (*argc)--;
-	if (!(r = o->rd_cfg(o, *(*argv - 1)))) return NULL;
-	(*argv)--; (*argc)++;
-	return r;
+	return NULL;
 }
 
 unsigned char *lookup_cmd(struct option *o, unsigned char ***argv, int *argc)
@@ -686,7 +772,7 @@ unsigned char *lookup_cmd(struct option *o, unsigned char ***argv, int *argc)
 	if (*argc >= 2) return "Too many parameters";
 	(*argv)++; (*argc)--;
 	if (do_real_lookup(*(*argv - 1), &addr)) {
-#ifdef HAVE_HERROR
+#if defined(HAVE_GETHOSTBYNAME) && defined(HAVE_HERROR)
 		herror("error");
 #else
 		fprintf(stderr, "error: host not found\n");
@@ -716,11 +802,13 @@ unsigned char *anonymous_cmd(struct option *o, unsigned char ***argv, int *argc)
 	anonymous = 1;
 	return NULL;
 }
+
 unsigned char *force_html_cmd(struct option *o, unsigned char ***argv, int *argc)
 {
 	force_html = 1;
 	return NULL;
 }
+
 unsigned char *dump_cmd(struct option *o, unsigned char ***argv, int *argc)
 {
 	if (dmp != o->min && dmp) return "Can't use both -dump and -source";
@@ -741,7 +829,10 @@ fprintf(stdout, "%s%s%s%s%s%s\n",
 Options are:\n\
 \n\
  -async-dns <0>/<1>\n\
-  Asynchronous DNS resolver on(1)/off(0). \n\
+  Asynchronous DNS resolver on(1)/off(0).\n\
+\n\
+ -download-utime <0>/<1>\n\
+  Set time of downloaded files to time from server.\n\
 \n\
  -max-connections <max>\n\
   Maximum number of concurrent connections.\n\
@@ -782,10 +873,59 @@ Options are:\n\
  -download-dir <path>\n\
   Default download directory.\n\
   (default: actual dir)\n\
+\n\
+ -http-bugs.http10 <0>/<1>\n\
+  \"1\" forces using only HTTP/1.0 protocol.\n\
+\n\
+ -http-bugs.allow-blacklist <0>/<1>\n\
+  \"1\" defaults to using list of servers that have broken HTTP/1.1 support.\n\
+  When links finds such server, it will retry the request with HTTP/1.0.\n\
+\n\
+ -http-bugs.bug-302-redirect <0>/<1>\n\
+  Process 302 redirect in a way that is incompatible with RFC1945 and RFC2068,\n\
+  but the same as Netscape and MSIE. Many pages depend on it.\n\
+\n\
+ -http-bugs.bug-post-no-keepalive <0>/<1>\n\
+  No keepalive connection after post requests. For some buggy servers.\n\
+\n\
+ -http-bugs.bug-no-accept-charset <0>/<1>\n\
+  Do not send Accept-Charset field of HTTP header.\n\
+\n\
+ -ftp.anonymous-password <string>\n\
+  Use ftp PASV command to bypass firewalls.\n\
+\n\
+ -ftp.fast <0>/<1>\n\
+  Send more ftp commands simultaneously. Faster response when\n\
+  browsing ftp directories, but it is incompatible with RFC\n\
+  and some servers don't like it.\n\
+\n\
+ -ftp.set-iptos <0>/<1>\n\
+  Set IP Type-of-service to high throughput on ftp connections.\n\
 \n"),
-(" -assume-codepage <codepage>\n\
+(" -html-assume-codepage <codepage>\n\
   Use the given codepage when the webpage did not specify\n\
   its codepage. (default: ISO 8859-1)\n\
+\n\
+ -html-tables <0>/<1>\n\
+  Render tables.\n\
+\n\
+ -html-frames <0>/<1>\n\
+  Render frames.\n\
+\n\
+ -html-images <0>/<1>\n\
+  Display links to images.\n\
+\n\
+ -html-numbered-links <0>/<1>\n\
+  Link numbering.\n\
+\n\
+ -html-table-order <0>/<1>\n\
+  Walk table by rows (0) or columns (1).\n\
+\n\
+ -html-margin <margin>\n\
+  Text margin.\n\
+\n\
+ -language <language>\n\
+  User interface language.\n\
 \n\
  -anonymous\n\
   Restrict links so that it can run on an anonymous account.\n\
@@ -795,20 +935,26 @@ Options are:\n\
 \n\
  -force-html\n\
   Treat file as if it had an .html extension.\n\
+\n\
+ -source\n\
+  Write the given HTML document in source form to stdout.\n\
 \n\
  -dump\n\
   Write a plain-text version of the given HTML document to\n\
   stdout.\n\
 \n\
  -width <size>\n\
-  Size of screen in characters, used in combination with -dump\n\
+  Size of screen in characters, used in combination with -dump.\n\
+\n\
+ -codepage <codepage>\n\
+  Character set of output of -dump.\n\
 \n\
  -no-connect\n\
   Runs links as a separate instance - instead of connecting to\n\
   existing instance.\n\
 \n\
- -source\n\
-  Write the given HTML document in source form to stdout.\n\
+ -lookup <host>\n\
+  Do lookup like \"host\" command.\n\
 \n\
  -version\n\
   Prints the links version number and exit.\n\
@@ -827,13 +973,17 @@ Options are:\n\
 	<-	 go back\n\
 	g	 go to url\n\
 	G	 go to url based on current url\n\
+	^R	 reload\n\
 	/	 search\n\
 	?	 search back\n\
 	n	 find next\n\
 	N	 find previous\n\
 	=	 document info\n\
 	\\	 document source\n\
+	|	 HTTP header\n\
+	*	 toggle displaying of image links\n\
 	d	 download\n\
+	s	 bookmarks\n\
 	q	 quit\n"));
 
 	fflush(stdout);
@@ -867,9 +1017,10 @@ int receive_timeout = 120;
 int unrestartable_receive_timeout = 600;
 
 int screen_width = 80;
+int dump_codepage = -1;
 
 int max_format_cache_entries = 5;
-long memory_cache_size = 1048576;
+int memory_cache_size = 1048576;
 
 int enable_html_tables = 1;
 int enable_html_frames = 1;
@@ -877,10 +1028,10 @@ int display_images = 1;
 
 struct document_setup dds = { 0, 0, 1, 1, 0, 3, 0, 0 };
 
-struct rgb default_fg = { 191, 191, 191 };
-struct rgb default_bg = { 0, 0, 0 };
-struct rgb default_link = { 255, 255, 255 };
-struct rgb default_vlink = { 255, 255, 0 };
+struct rgb default_fg = { 191, 191, 191, 0 };
+struct rgb default_bg = { 0, 0, 0, 0 };
+struct rgb default_link = { 255, 255, 255, 0 };
+struct rgb default_vlink = { 255, 255, 0, 0 };
 
 int default_left_margin = HTML_LEFT_MARGIN;
 
@@ -889,7 +1040,7 @@ unsigned char ftp_proxy[MAX_STR_LEN] = "";
 
 unsigned char download_dir[MAX_STR_LEN] = "";
 
-unsigned char default_anon_pass[MAX_STR_LEN] = "somebody@host.domain";
+struct ftp_options ftp_options = { "somebody@host.domain", 0, 0, 1 };
 
 /* These are workarounds for some CGI script bugs */
 struct http_bugs http_bugs = { 0, 1, 1, 0, 0 };
@@ -901,125 +1052,120 @@ struct http_bugs http_bugs = { 0, 1, 1, 0, 0 };
 	   results if GET wants to retreive data POSTed in the same connection */
 
 struct option links_options[] = {
-	1, printhelp_cmd, NULL, NULL, 0, 0, NULL, NULL, "?",
-	1, printhelp_cmd, NULL, NULL, 0, 0, NULL, NULL, "h",
-	1, printhelp_cmd, NULL, NULL, 0, 0, NULL, NULL, "help",
-	1, printhelp_cmd, NULL, NULL, 0, 0, NULL, NULL, "-help",
-	1, lookup_cmd, NULL, NULL, 0, 0, NULL, NULL, "lookup",
-	1, version_cmd, NULL, NULL, 0, 0, NULL, NULL, "version",
-	1, no_connect_cmd, NULL, NULL, 0, 0, NULL, NULL, "no-connect",
-	1, anonymous_cmd, NULL, NULL, 0, 0, NULL, NULL, "anonymous",
-	1, gen_cmd, num_rd, NULL, 0, MAXINT, &base_session, NULL, "base-session",
-	1, force_html_cmd, NULL, NULL, 0, 0, NULL, NULL, "force-html",
-	1, dump_cmd, NULL, NULL, D_DUMP, 0, NULL, NULL, "dump",
-	1, dump_cmd, NULL, NULL, D_SOURCE, 0, NULL, NULL, "source",
-	1, gen_cmd, num_rd, num_wr, 0, 1, &async_lookup, "async_dns", "async-dns",
-	1, gen_cmd, num_rd, num_wr, 0, 1, &download_utime, "download_utime", "download-utime",
-	1, gen_cmd, num_rd, num_wr, 1, 99, &max_connections, "max_connections", "max-connections",
-	1, gen_cmd, num_rd, num_wr, 1, 99, &max_connections_to_host, "max_connections_to_host", "max-connections-to-host",
-	1, gen_cmd, num_rd, num_wr, 0, 16, &max_tries, "retries", "retries",
-	1, gen_cmd, num_rd, num_wr, 1, 1800, &receive_timeout, "receive_timeout", "receive-timeout",
-	1, gen_cmd, num_rd, num_wr, 1, 1800, &unrestartable_receive_timeout, "unrestartable_receive_timeout", "unrestartable-receive-timeout",
-	1, gen_cmd, num_rd, NULL, 10, 512, &screen_width, NULL, "width",
-	1, gen_cmd, num_rd, num_wr, 0, 256, &max_format_cache_entries, "format_cache_size", "format-cache-size",
-	1, gen_cmd, num_rd, num_wr, 0, MAXINT, &memory_cache_size, "memory_cache_size", "memory-cache-size",
-	1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, http_proxy, "http_proxy", "http-proxy",
-	1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, ftp_proxy, "ftp_proxy", "ftp-proxy",
-	1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, download_dir, "download_dir", "download-dir",
-	1, gen_cmd, lang_rd, lang_wr, 0, 0, &current_language, "language", "language",
-	1, gen_cmd, num_rd, num_wr, 0, 1, &http_bugs.http10, "http_bugs.http10", "http-bugs.http10",
-	1, gen_cmd, num_rd, num_wr, 0, 1, &http_bugs.allow_blacklist, "http_bugs.allow_blacklist", "http-bugs.allow-blacklist",
-	1, gen_cmd, num_rd, num_wr, 0, 1, &http_bugs.bug_302_redirect, "http_bugs.bug_302_redirect", "http-bugs.bug-302-redirect",
-	1, gen_cmd, num_rd, num_wr, 0, 1, &http_bugs.bug_post_no_keepalive, "http_bugs.bug_post_no_keepalive", "http-bugs.bug-post-no-keepalive",
-	1, gen_cmd, num_rd, num_wr, 0, 1, &http_bugs.no_accept_charset, "http_bugs.no_accept_charset", "http-bugs.bug-no-accept-charset",
-	1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, default_anon_pass, "ftp.anonymous_password", "ftp.anonymous-password",
-	{1, gen_cmd, num_rd, num_wr, 0, 1, &passive_ftp, "ftp.use_passive", "ftp.use-passive"},
-	{1, gen_cmd, num_rd, num_wr, 0, 1, &fast_ftp, "ftp.fast", "ftp.fast"},
-	1, gen_cmd, cp_rd, NULL, 0, 0, &dds.assume_cp, "assume_codepage", "assume-codepage",
-	1, NULL, term_rd, term_wr, 0, 0, NULL, "terminal", NULL,
-	1, NULL, term2_rd, NULL, 0, 0, NULL, "terminal2", NULL,
-	1, NULL, type_rd, type_wr, 0, 0, NULL, "association", NULL,
-	1, NULL, ext_rd, ext_wr, 0, 0, NULL, "extension", NULL,
-	1, NULL, prog_rd, prog_wr, 0, 0, &mailto_prog, "mailto", NULL,
-	1, NULL, prog_rd, prog_wr, 0, 0, &telnet_prog, "telnet", NULL,
-	1, NULL, prog_rd, prog_wr, 0, 0, &tn3270_prog, "tn3270", NULL,
-	1, NULL, bind_rd, NULL, 0, 0, NULL, "bind", NULL,
-	1, NULL, unbind_rd, NULL, 0, 0, NULL, "unbind", NULL,
-	0, NULL, NULL, NULL, 0, 0, NULL, NULL, NULL,
+	{ 1, printhelp_cmd, NULL, NULL, 0, 0, NULL, NULL, "?" },
+	{ 1, printhelp_cmd, NULL, NULL, 0, 0, NULL, NULL, "h" },
+	{ 1, printhelp_cmd, NULL, NULL, 0, 0, NULL, NULL, "help" },
+	{ 1, printhelp_cmd, NULL, NULL, 0, 0, NULL, NULL, "-help" },
+	{ 1, lookup_cmd, NULL, NULL, 0, 0, NULL, NULL, "lookup" },
+	{ 1, version_cmd, NULL, NULL, 0, 0, NULL, NULL, "version" },
+	{ 1, no_connect_cmd, NULL, NULL, 0, 0, NULL, NULL, "no-connect" },
+	{ 1, anonymous_cmd, NULL, NULL, 0, 0, NULL, NULL, "anonymous" },
+	{ 1, gen_cmd, num_rd, NULL, 0, MAXINT, &base_session, NULL, "base-session" },
+	{ 1, force_html_cmd, NULL, NULL, 0, 0, NULL, NULL, "force-html" },
+	{ 1, dump_cmd, NULL, NULL, D_SOURCE, 0, NULL, NULL, "source" },
+	{ 1, dump_cmd, NULL, NULL, D_DUMP, 0, NULL, NULL, "dump" },
+	{ 1, gen_cmd, num_rd, NULL, 10, 512, &screen_width, "dump_width", "width" },
+	{ 1, gen_cmd, cp_rd, NULL, 1, 0, &dump_codepage, "dump_codepage", "codepage" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &async_lookup, "async_dns", "async-dns" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &download_utime, "download_utime", "download-utime" },
+	{ 1, gen_cmd, num_rd, num_wr, 1, 99, &max_connections, "max_connections", "max-connections" },
+	{ 1, gen_cmd, num_rd, num_wr, 1, 99, &max_connections_to_host, "max_connections_to_host", "max-connections-to-host" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 16, &max_tries, "retries", "retries" },
+	{ 1, gen_cmd, num_rd, num_wr, 1, 1800, &receive_timeout, "receive_timeout", "receive-timeout" },
+	{ 1, gen_cmd, num_rd, num_wr, 1, 1800, &unrestartable_receive_timeout, "unrestartable_receive_timeout", "unrestartable-receive-timeout" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 256, &max_format_cache_entries, "format_cache_size", "format-cache-size" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, MAXINT, &memory_cache_size, "memory_cache_size", "memory-cache-size" },
+	{ 1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, http_proxy, "http_proxy", "http-proxy" },
+	{ 1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, ftp_proxy, "ftp_proxy", "ftp-proxy" },
+	{ 1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, download_dir, "download_dir", "download-dir" },
+	{ 1, gen_cmd, lang_rd, lang_wr, 0, 0, &current_language, "language", "language" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &http_bugs.http10, "http_bugs.http10", "http-bugs.http10" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &http_bugs.allow_blacklist, "http_bugs.allow_blacklist", "http-bugs.allow-blacklist" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &http_bugs.bug_302_redirect, "http_bugs.bug_302_redirect", "http-bugs.bug-302-redirect" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &http_bugs.bug_post_no_keepalive, "http_bugs.bug_post_no_keepalive", "http-bugs.bug-post-no-keepalive" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &http_bugs.no_accept_charset, "http_bugs.no_accept_charset", "http-bugs.bug-no-accept-charset" },
+	{ 1, gen_cmd, str_rd, str_wr, 0, MAX_STR_LEN, ftp_options.anon_pass, "ftp.anonymous_password", "ftp.anonymous-password" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &ftp_options.passive_ftp, "ftp.use_passive", "ftp.use-passive" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &ftp_options.fast_ftp, "ftp.fast", "ftp.fast" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &ftp_options.set_tos, "ftp.set_iptos", "ftp.set-iptos" },
+	{ 1, gen_cmd, cp_rd, NULL, 0, 0, &dds.assume_cp, "assume_codepage", "assume-codepage" },
+	{ 1, NULL, term_rd, term_wr, 0, 0, NULL, "terminal", NULL },
+	{ 1, NULL, term2_rd, NULL, 0, 0, NULL, "terminal2", NULL },
+	{ 1, NULL, type_rd, type_wr, 0, 0, NULL, "association", NULL },
+	{ 1, NULL, ext_rd, ext_wr, 0, 0, NULL, "extension", NULL },
+	{ 1, NULL, prog_rd, prog_wr, 0, 0, &mailto_prog, "mailto", NULL },
+	{ 1, NULL, prog_rd, prog_wr, 0, 0, &telnet_prog, "telnet", NULL },
+	{ 1, NULL, prog_rd, prog_wr, 0, 0, &tn3270_prog, "tn3270", NULL },
+	{ 1, NULL, prog_rd, prog_wr, 0, 0, &mms_prog, "mms", NULL },
+	{ 1, NULL, bind_rd, NULL, 0, 0, NULL, "bind", NULL },
+	{ 1, NULL, unbind_rd, NULL, 0, 0, NULL, "unbind", NULL },
+	{ 0, NULL, NULL, NULL, 0, 0, NULL, NULL, NULL },
 };
 
 struct option html_options[] = {
-	1, gen_cmd, num_rd, num_wr, 0, 1, &dds.hard_assume, "html_hard_assume", "html-hard-assume",
-	1, gen_cmd, cp_rd, cp_wr, 0, 0, &dds.assume_cp, "html_assume_codepage", "html-assume-codepage",
-	1, gen_cmd, num_rd, num_wr, 0, 1, &dds.tables, "html_tables", "html-tables",
-	1, gen_cmd, num_rd, num_wr, 0, 1, &dds.frames, "html_frames", "html-frames",
-	1, gen_cmd, num_rd, num_wr, 0, 1, &dds.images, "html_images", "html-images",
-	1, gen_cmd, num_rd, num_wr, 0, 1, &dds.num_links, "html_numbered_links", "html-numbered-links",
-	1, gen_cmd, num_rd, num_wr, 0, 1, &dds.table_order, "html_table_order", "html-table-order",
-	1, gen_cmd, num_rd, num_wr, 0, 9, &dds.margin, "html_margin", "html-margin",
-	0, NULL, NULL, NULL, 0, 0, NULL, NULL, NULL,
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &dds.hard_assume, "html_hard_assume", "html-hard-assume" },
+	{ 1, gen_cmd, cp_rd, cp_wr, 0, 0, &dds.assume_cp, "html_assume_codepage", "html-assume-codepage" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &dds.tables, "html_tables", "html-tables" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &dds.frames, "html_frames", "html-frames" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &dds.images, "html_images", "html-images" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &dds.num_links, "html_numbered_links", "html-numbered-links" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 1, &dds.table_order, "html_table_order", "html-table-order" },
+	{ 1, gen_cmd, num_rd, num_wr, 0, 9, &dds.margin, "html_margin", "html-margin" },
+	{ 0, NULL, NULL, NULL, 0, 0, NULL, NULL, NULL },
 };
 
-extern struct history goto_url_history ;
-
-int load_url_history()
+void load_url_history(void)
 {
-	unsigned char *history_file ;
-	FILE *fp ;
-	char	url[MAX_INPUT_URL_LEN];
+	unsigned char *history_file, *hs;
+	unsigned char *hsp;
 
-	if (anonymous) return 0;
+	if (anonymous) return;
 	/* Must have been called after init_home */
-	if (!links_home)
-		return 0;
-	history_file = stracpy (links_home) ;
-	if (!history_file)
-		return 0;
+	if (!links_home) return;
+	history_file = stracpy(links_home);
 	add_to_strn(&history_file, "links.his");
-	if (!(fp = fopen(history_file, "r")))
-	{
-		mem_free (history_file) ;
-		return 0;
+	hs = read_config_file(history_file);
+	mem_free(history_file);
+	if (!hs) return;
+	for (hsp = hs; *hsp; ) {
+		unsigned char *hsl, *hsc;
+		for (hsl = hsp; *hsl && *hsl != 10 && *hsl != 13; hsl++) ;
+		hsc = memacpy(hsp, hsl - hsp);
+		add_to_history(&goto_url_history, hsc, 0);
+		mem_free(hsc);
+		hsp = hsl;
+		while (*hsp == 10 || *hsp == 13) hsp++;
 	}
-	while (fgets (url, MAX_INPUT_URL_LEN, fp))
-	{
-		url[MAX_INPUT_URL_LEN - 1] = 0;
-		if (*url) url[strlen(url)-1] = 0 ;
-		add_to_history(&goto_url_history, url) ;
-	}
-	fclose (fp) ;
-	mem_free (history_file) ;
-	return 0 ;
+	mem_free(hs);
 }
 
-int save_url_history()
+void save_url_history(void)
 {
-	struct history_item* hi ;
-	unsigned char *history_file ;
-	int  i = 0 ;
-	FILE *fp ;
-	if (anonymous) return 0;
+	struct history_item *hi;
+	unsigned char *history_file;
+	unsigned char *hs;
+	int hsl = 0;
+	int i = 0;
+	if (anonymous) return;
 
 	/* Must have been called after init_home */
-	if (!links_home)
-		return 0;
-	history_file = stracpy (links_home) ;
-	if (!history_file)
-		return 0;
+	if (!links_home) return;
+	history_file = stracpy(links_home);
 	add_to_strn(&history_file, "links.his");
-	if (!(fp = fopen(history_file, "w"))) {
-		mem_free(history_file);
-		return 0;
-	}
-	foreachback(hi,goto_url_history.items)
-	{
+	hs = init_str();
+	hsl = 0;
+	foreachback(hi, goto_url_history.items) {
+		if (!*hi->d || strchr(hi->d, 10) || strchr(hi->d, 13)) continue;
 		if (i++ > MAX_HISTORY_ITEMS)
-			break ;
-		else
-			fputs (hi->d, fp), fputc ('\n', fp) ;
+			break;
+		else {
+			add_to_str(&hs, &hsl, hi->d);
+			add_to_str(&hs, &hsl, NEWLINE);
+		}
 	}
-	fclose (fp) ;
-	mem_free (history_file) ;
-	return 0 ;
+	write_to_config_file(history_file, hs);
+	mem_free(history_file);
+	mem_free(hs);
+	return;
 }
 
