@@ -1,15 +1,17 @@
 #include "links.h"
 
+#if defined(HAVE_GETHOSTBYNAME_BUG) || !defined(HAVE_GETHOSTBYNAME)
+#define EXTERNAL_LOOKUP
+#endif
+
 struct dnsentry {
 	struct dnsentry *next;
 	struct dnsentry *prev;
 	ttime get_time;
 	ip addr;
-	char name[1];
+	unsigned char name[1];
 };
 
-#define DNS_PIPE_BUFFER	32
-
 #ifndef THREAD_SAFE_LOOKUP
 struct dnsquery *dns_queue = NULL;
 #endif
@@ -24,21 +26,112 @@ struct dnsquery {
 	int h;
 	struct dnsquery **s;
 	ip *addr;
-	char name[1];
+	unsigned char name[1];
 };
 
 struct list_head dns_cache = {&dns_cache, &dns_cache};
 
+static int get_addr_byte(unsigned char **ptr, unsigned char *res, unsigned char stp)
+{
+	unsigned u = 0;
+	if (!(**ptr >= '0' && **ptr <= '9')) return -1;
+	while (**ptr >= '0' && **ptr <= '9') {
+		u = u * 10 + **ptr - '0';
+		if (u >= 256) return -1;
+		(*ptr)++;
+	}
+	if (stp != 255 && **ptr != stp) return -1;
+	(*ptr)++;
+	*res = u;
+	return 0;
+}
+
+#ifdef EXTERNAL_LOOKUP
+
+static int do_external_lookup(unsigned char *name, ip *host)
+{
+	unsigned char buffer[1024];
+	unsigned char sink[16];
+	int rd;
+	int pi[2];
+	pid_t f;
+	unsigned char *n;
+	if (pipe(pi) == -1)
+		return -1;
+	f = fork();
+	if (f == -1) {
+		close(pi[0]);
+		close(pi[1]);
+		return -1;
+	}
+	if (!f) {
+		close(pi[0]);
+		if (dup2(pi[1], 1) == -1) _exit(1);
+		if (dup2(pi[1], 2) == -1) _exit(1);
+		close(pi[1]);
+		execlp("host", "host", name, NULL);
+		execl("/usr/sbin/host", "host", name, NULL);
+		_exit(1);
+	}
+	close(pi[1]);
+	rd = hard_read(pi[0], buffer, sizeof buffer - 1);
+	if (rd >= 0) buffer[rd] = 0;
+	if (rd > 0) {
+		while (hard_read(pi[0], sink, sizeof sink) > 0);
+	}
+	close(pi[0]);
+	/* Don't wait for the process, we already have sigchld handler that
+	 * does cleanup.
+	 * waitpid(f, NULL, 0); */
+	if (rd < 0) return -1;
+	/*fprintf(stderr, "query: '%s', result: %s\n", name, buffer);*/
+	while ((n = strstr(buffer, name))) {
+		memset(n, '-', strlen(name));
+	}
+	for (n = buffer; n < buffer + rd; n++) {
+		if (*n >= '0' && *n <= '9') {
+			if (get_addr_byte(&n, ((unsigned char *)host + 0), '.')) goto skip_addr;
+			if (get_addr_byte(&n, ((unsigned char *)host + 1), '.')) goto skip_addr;
+			if (get_addr_byte(&n, ((unsigned char *)host + 2), '.')) goto skip_addr;
+			if (get_addr_byte(&n, ((unsigned char *)host + 3), 255)) goto skip_addr;
+			return 0;
+skip_addr:
+			if (n >= buffer + rd) break;
+		}
+	}
+	return -1;
+}
+
+#endif
+
 int do_real_lookup(unsigned char *name, ip *host)
 {
-	char *n;
+	unsigned char *n;
 	struct hostent *hst;
+	if (!*name) return -1;
 	for (n = name; *n; n++) if (*n != '.' && (*n < '0' || *n > '9')) goto nogethostbyaddr;
+	n = name;
+	if (get_addr_byte(&n, ((unsigned char *)host + 0), '.')) goto skip_addr;
+	if (get_addr_byte(&n, ((unsigned char *)host + 1), '.')) goto skip_addr;
+	if (get_addr_byte(&n, ((unsigned char *)host + 2), '.')) goto skip_addr;
+	if (get_addr_byte(&n, ((unsigned char *)host + 3), 0)) goto skip_addr;
+	return 0;
+	skip_addr:
 #ifdef HAVE_GETHOSTBYADDR
-	if (!(hst = gethostbyaddr (name, strlen(name), AF_INET)))
+	if (!(hst = gethostbyaddr(name, strlen(name), AF_INET)))
+#endif
+	{
+		nogethostbyaddr:
+#ifdef HAVE_GETHOSTBYNAME
+		if (!(hst = gethostbyname(name)))
+#endif
+		{
+#ifdef EXTERNAL_LOOKUP
+			return do_external_lookup(name, host);
 #endif
-		nogethostbyaddr: if (!(hst = gethostbyname(name)))
 			return -1;
+		}
+	}
 	memcpy(host, hst->h_addr_list[0], sizeof(ip));
 	return 0;
 }
@@ -76,7 +169,9 @@ int do_lookup(struct dnsquery *q, int force_async)
 	if (!async_lookup && !force_async) {
 #endif
 		int r;
+#ifndef NO_ASYNC_LOOKUP
 		sync_lookup:
+#endif
 		r = do_real_lookup(q->name, q->addr);
 		q->xfn(q, r);
 		return 0;
@@ -109,7 +204,7 @@ int do_queued_lookup(struct dnsquery *q)
 #endif
 }
 
-int find_in_dns_cache(char *name, struct dnsentry **dnsentry)
+int find_in_dns_cache(unsigned char *name, struct dnsentry **dnsentry)
 {
 	struct dnsentry *e;
 	foreach(e, dns_cache)
@@ -148,12 +243,11 @@ void end_dns_lookup(struct dnsquery *q, int a)
 		mem_free(dnsentry);
 	}
 	if (a) goto e;
-	if ((dnsentry = mem_alloc(sizeof(struct dnsentry) + strlen(q->name) + 1))) {
-		strcpy(dnsentry->name, q->name);
-		memcpy(&dnsentry->addr, q->addr, sizeof(ip));
-		dnsentry->get_time = get_time();
-		add_to_list(dns_cache, dnsentry);
-	}
+	dnsentry = mem_alloc(sizeof(struct dnsentry) + strlen(q->name) + 1);
+	strcpy(dnsentry->name, q->name);
+	memcpy(&dnsentry->addr, q->addr, sizeof(ip));
+	dnsentry->get_time = get_time();
+	add_to_list(dns_cache, dnsentry);
 	e:
 	if (q->s) *q->s = NULL;
 	fn = q->fn;
@@ -165,7 +259,7 @@ void end_dns_lookup(struct dnsquery *q, int a)
 int find_host_no_cache(unsigned char *name, ip *addr, void **qp, void (*fn)(void *, int), void *data)
 {
 	struct dnsquery *q;
-	if (!(q = malloc(sizeof(struct dnsquery) + strlen(name) + 1))) {
+	if (!(q = (struct dnsquery *)malloc(sizeof(struct dnsquery) + strlen(name) + 1))) {
 		fn(data, -1);
 		return 0;
 	}
@@ -184,7 +278,7 @@ int find_host(unsigned char *name, ip *addr, void **qp, void (*fn)(void *, int),
 	struct dnsentry *dnsentry;
 	if (qp) *qp = NULL;
 	if (!find_in_dns_cache(name, &dnsentry)) {
-		if (dnsentry->get_time + DNS_TIMEOUT < get_time()) goto timeout;
+		if ((uttime)get_time() - (uttime)dnsentry->get_time > DNS_TIMEOUT) goto timeout;
 		memcpy(addr, &dnsentry->addr, sizeof(ip));
 		fn(data, 0);
 		return 0;
@@ -207,7 +301,7 @@ void kill_dns_request(void **qp)
 void shrink_dns_cache(int u)
 {
 	struct dnsentry *d, *e;
-	foreach(d, dns_cache) if (u || d->get_time + DNS_TIMEOUT < get_time()) {
+	foreach(d, dns_cache) if (u || (uttime)get_time() - (uttime)d->get_time > DNS_TIMEOUT) {
 		e = d;
 		d = d->prev;
 		del_from_list(e);
