@@ -1,8 +1,12 @@
 #include "links.h"
 
+char dummy_val;
+volatile char *dummy_ptr = &dummy_val;
+
 void do_not_optimize_here(void *p)
 {
-	/* stop GCC optimization - avoid bugs in it */
+	*dummy_ptr = 0;
+	/* break ANSI aliasing */
 }
 
 #ifdef LEAK_DEBUG
@@ -16,7 +20,9 @@ struct list_head memory_list = { &memory_list, &memory_list };
 static inline void force_dump()
 {
 	fprintf(stderr, "\n\033[1m%s\033[0m\n", "Forcing core dump");
+	fflush(stdout);
 	fflush(stderr);
+	fatal_tty_exit();
 	raise(SIGSEGV);
 }
 
@@ -31,7 +37,7 @@ void check_memory_leaks()
 			int r = 0;
 			struct alloc_header *ah;
 			foreach (ah, memory_list) {
-				fprintf(stderr, "%s%p:%d @ %s:%d", r ? ", ": "", (char *)ah + L_D_S, ah->size, ah->file, ah->line), r = 1;
+				fprintf(stderr, "%s%p:%lu @ %s:%d", r ? ", ": "", (char *)ah + L_D_S, (unsigned long)ah->size, ah->file, ah->line), r = 1;
 				if (ah->comment) fprintf(stderr, ":\"%s\"", ah->comment);
 			}
 			fprintf(stderr, "\n");
@@ -45,7 +51,11 @@ void check_memory_leaks()
 void er(int b, unsigned char *m, va_list l)
 {
 	if (b) fprintf(stderr, "%c", (char)7);
+#ifdef HAVE_VPRINTF
 	vfprintf(stderr, m, l);
+#else
+	fprintf(stderr, "%s", m);
+#endif
 	fprintf(stderr, "\n");
 	sleep(1);
 }
@@ -93,12 +103,15 @@ void *debug_mem_alloc(unsigned char *file, int line, size_t size)
 	struct alloc_header *ah;
 #endif
 	if (!size) return DUMMY;
+	if (size > MAXINT) overalloc();
 #ifdef LEAK_DEBUG
 	mem_amount += size;
 	size += L_D_S;
 #endif
 	if (!(p = xmalloc(size))) {
-		error("ERROR: out of memory (malloc returned NULL)\n");
+		error("ERROR: out of memory (malloc returned NULL)");
+		fatal_tty_exit();
+		exit(RET_FATAL);
 		return NULL;
 	}
 #ifdef LEAK_DEBUG
@@ -151,8 +164,11 @@ void *debug_mem_realloc(unsigned char *file, int line, void *p, size_t size)
 		debug_mem_free(file, line, p);
 		return DUMMY;
 	}
+	if (size > MAXINT) overalloc();
 	if (!(p = xrealloc((char *)p - L_D_S, size + L_D_S))) {
-		error("ERROR: out of memory (realloc returned NULL)\n");
+		error("ERROR: out of memory (realloc returned NULL)");
+		fatal_tty_exit();
+		exit(RET_FATAL);
 		return NULL;
 	}
 #ifdef LEAK_DEBUG
