@@ -1,12 +1,17 @@
-#include "links.h"
+/* ftp.c
+ * ftp:// processing
+ * (c) 2002 Mikulas Patocka
+ * This file is a part of the Links program, released under GPL.
+ */
 
-int fast_ftp = 0;
+#include "links.h"
 
 #define FTP_BUF	16384
 
 struct ftp_connection_info {
 	int pending_commands;
 	int opc;
+	int pasv;
 	int dir;
 	int rest_sent;
 	int conn_st;
@@ -17,31 +22,36 @@ struct ftp_connection_info {
 	unsigned char cmdbuf[1];
 };
 
-void ftp_login(struct connection *);
-void ftp_logged(struct connection *);
-void ftp_sent_passwd(struct connection *);
-void ftp_got_reply(struct connection *, struct read_buffer *);
-void ftp_got_info(struct connection *, struct read_buffer *);
-void ftp_got_user_info(struct connection *, struct read_buffer *);
-void ftp_dummy_info(struct connection *, struct read_buffer *);
-void ftp_pass_info(struct connection *, struct read_buffer *);
-void ftp_send_retr_req(struct connection *, int);
-struct ftp_connection_info *add_file_cmd_to_str(struct connection *);
-void ftp_retr_1(struct connection *);
-void ftp_retr_file(struct connection *, struct read_buffer *);
-void ftp_got_final_response(struct connection *, struct read_buffer *);
-void got_something_from_data_connection(struct connection *);
-void ftp_end_request(struct connection *);
+static void ftp_get_banner(struct connection *);
+static void ftp_got_banner(struct connection *, struct read_buffer *);
+static void ftp_login(struct connection *);
+static void ftp_logged(struct connection *);
+static void ftp_sent_passwd(struct connection *);
+static void ftp_got_info(struct connection *, struct read_buffer *);
+static void ftp_got_user_info(struct connection *, struct read_buffer *);
+static void ftp_dummy_info(struct connection *, struct read_buffer *);
+static void ftp_pass_info(struct connection *, struct read_buffer *);
+static void ftp_send_retr_req(struct connection *, int);
+static struct ftp_connection_info *add_file_cmd_to_str(struct connection *);
+static void ftp_retr_1(struct connection *);
+static void ftp_retr_file(struct connection *, struct read_buffer *);
+static void ftp_got_final_response(struct connection *, struct read_buffer *);
+static void created_data_connection(struct connection *);
+static void got_something_from_data_connection(struct connection *);
+static void ftp_end_request(struct connection *);
+static int get_ftp_response(struct connection *, struct read_buffer *, int);
+static int ftp_process_dirlist(struct cache_entry *, off_t *, int *, unsigned char *, int, int, int *);
 
-int get_ftp_response(struct connection *c, struct read_buffer *rb, int part)
+
+static int get_ftp_response(struct connection *c, struct read_buffer *rb, int part)
 {
 	int l;
 	set_timeout(c);
 	again:
 	for (l = 0; l < rb->len; l++) if (rb->data[l] == 10) {
 		unsigned char *e;
-		int k = strtoul(rb->data, (char **)&e, 10);
-		if (e != rb->data + 3 || k < 100) return -1;
+		long k = strtoul(rb->data, (char **)(void *)&e, 10);
+		if (e != rb->data + 3 || k < 100 || k >= 1000) return -1;
 		if (*e == '-') {
 			int i;
 			for (i = 0; i < rb->len - 5; i++) {
@@ -68,7 +78,7 @@ int get_ftp_response(struct connection *c, struct read_buffer *rb, int part)
 void ftp_func(struct connection *c)
 {
 	/*setcstate(c, S_CONN);*/
-	set_timeout(c);
+	/*set_timeout(c);*/
 	if (get_keepalive_socket(c)) {
 		int p;
 		if ((p = get_port(c->url)) == -1) {
@@ -76,29 +86,44 @@ void ftp_func(struct connection *c)
 			abort_connection(c);
 			return;
 		}
-		make_connection(c, p, &c->sock1, ftp_login);
+		make_connection(c, p, &c->sock1, ftp_options.fast_ftp ? ftp_login : ftp_get_banner);
 	} else ftp_send_retr_req(c, S_SENT);
 }
 
-void ftp_login(struct connection *c)
+static void ftp_get_banner(struct connection *c)
+{
+	struct read_buffer *rb;
+	set_timeout(c);
+	setcstate(c, S_SENT);
+	if (!(rb = alloc_read_buffer(c))) return;
+	read_from_socket(c, c->sock1, rb, ftp_got_banner);
+}
+
+static void ftp_got_banner(struct connection *c, struct read_buffer *rb)
+{
+	int g = get_ftp_response(c, rb, 0);
+	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
+	if (!g) { read_from_socket(c, c->sock1, rb, ftp_got_banner); return; }
+	if (g >= 400) { setcstate(c, S_FTP_UNAVAIL); retry_connection(c); return; }
+	ftp_login(c);
+}
+
+static void ftp_login(struct connection *c)
 {
 	unsigned char *login;
 	unsigned char *u;
 	int logl = 0;
-	if (!(login = init_str())) {
-		setcstate(c, S_OUT_OF_MEM);
-		abort_connection(c);
-		return;
-	}
+	set_timeout(c);
+	login = init_str();
 	add_to_str(&login, &logl, "USER ");
 	if ((u = get_user_name(c->url)) && *u) add_to_str(&login, &logl, u);
 	else add_to_str(&login, &logl, "anonymous");
 	if (u) mem_free(u);
-	if (fast_ftp) {
+	if (ftp_options.fast_ftp) {
 		struct ftp_connection_info *fi;
 		add_to_str(&login, &logl, "\r\nPASS ");
 		if ((u = get_pass(c->url)) && *u) add_to_str(&login, &logl, u);
-		else add_to_str(&login, &logl, default_anon_pass);
+		else add_to_str(&login, &logl, ftp_options.anon_pass);
 		if (u) mem_free(u);
 		add_to_str(&login, &logl, "\r\n");
 		if (!(fi = add_file_cmd_to_str(c))) {
@@ -112,14 +137,18 @@ void ftp_login(struct connection *c)
 	setcstate(c, S_SENT);
 }
 
-void ftp_logged(struct connection *c)
+static void ftp_logged(struct connection *c)
 {
 	struct read_buffer *rb;
 	if (!(rb = alloc_read_buffer(c))) return;
+	if (!ftp_options.fast_ftp) {
+		ftp_got_user_info(c, rb);
+		return;
+	}
 	read_from_socket(c, c->sock1, rb, ftp_got_info);
 }
 
-void ftp_got_info(struct connection *c, struct read_buffer *rb)
+static void ftp_got_info(struct connection *c, struct read_buffer *rb)
 {
 	int g = get_ftp_response(c, rb, 0);
 	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
@@ -128,7 +157,7 @@ void ftp_got_info(struct connection *c, struct read_buffer *rb)
 	ftp_got_user_info(c, rb);
 }
 
-void ftp_got_user_info(struct connection *c, struct read_buffer *rb)
+static void ftp_got_user_info(struct connection *c, struct read_buffer *rb)
 {
 	int g = get_ftp_response(c, rb, 0);
 	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
@@ -136,22 +165,18 @@ void ftp_got_user_info(struct connection *c, struct read_buffer *rb)
 	if (g >= 530 && g < 540) { setcstate(c, S_FTP_LOGIN); retry_connection(c); return; }
 	if (g >= 400) { setcstate(c, S_FTP_UNAVAIL); retry_connection(c); return; }
 	if (g >= 200 && g < 300) {
-		if (fast_ftp) ftp_dummy_info(c, rb);
+		if (ftp_options.fast_ftp) ftp_dummy_info(c, rb);
 		else ftp_send_retr_req(c, S_GETH);
 	} else {
-		if (fast_ftp) ftp_pass_info(c, rb);
+		if (ftp_options.fast_ftp) ftp_pass_info(c, rb);
 		else {
 			unsigned char *login;
 			unsigned char *u;
 			int logl = 0;
-			if (!(login = init_str())) {
-				setcstate(c, S_OUT_OF_MEM);
-				abort_connection(c);
-				return;
-			}
+			login = init_str();
 			add_to_str(&login, &logl, "PASS ");
 			if ((u = get_pass(c->url)) && *u) add_to_str(&login, &logl, u);
-			else add_to_str(&login, &logl, default_anon_pass);
+			else add_to_str(&login, &logl, ftp_options.anon_pass);
 			if (u) mem_free(u);
 			add_to_str(&login, &logl, "\r\n");
 			write_to_socket(c, c->sock1, login, strlen(login), ftp_sent_passwd);
@@ -161,7 +186,7 @@ void ftp_got_user_info(struct connection *c, struct read_buffer *rb)
 	}
 }
 
-void ftp_dummy_info(struct connection *c, struct read_buffer *rb)
+static void ftp_dummy_info(struct connection *c, struct read_buffer *rb)
 {
 	int g = get_ftp_response(c, rb, 0);
 	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
@@ -169,122 +194,140 @@ void ftp_dummy_info(struct connection *c, struct read_buffer *rb)
 	ftp_retr_file(c, rb);
 }
 
-void ftp_sent_passwd(struct connection *c)
+static void ftp_sent_passwd(struct connection *c)
 {
 	struct read_buffer *rb;
 	if (!(rb = alloc_read_buffer(c))) return;
 	read_from_socket(c, c->sock1, rb, ftp_pass_info);
 }
 
-void ftp_pass_info(struct connection *c, struct read_buffer *rb)
+static void ftp_pass_info(struct connection *c, struct read_buffer *rb)
 {
 	int g = get_ftp_response(c, rb, 0);
 	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
 	if (!g) { read_from_socket(c, c->sock1, rb, ftp_pass_info); setcstate(c, S_LOGIN); return; }
 	if (g >= 530 && g < 540) { setcstate(c, S_FTP_LOGIN); abort_connection(c); return; }
 	if (g >= 400) { setcstate(c, S_FTP_UNAVAIL); abort_connection(c); return; }
-	if (fast_ftp) ftp_retr_file(c, rb);
+	if (ftp_options.fast_ftp) ftp_retr_file(c, rb);
 	else ftp_send_retr_req(c, S_GETH);
 }
 
-struct ftp_connection_info *add_file_cmd_to_str(struct connection *c)
+static struct ftp_connection_info *add_file_cmd_to_str(struct connection *c)
 {
 	unsigned char *d = get_url_data(c->url);
 	unsigned char *de;
+	int del;
 	unsigned char pc[6];
 	int ps;
 	struct ftp_connection_info *inf, *inf2;
 	unsigned char *s;
 	int l;
-	if (!(inf = mem_alloc(sizeof(struct ftp_connection_info)))) {
-		oom:
-		setcstate(c, S_OUT_OF_MEM);
-		abort_connection(c);
-		return NULL;
-	}
-	memset(inf, 0, sizeof(struct ftp_connection_info));
-	l = 0;
-	if (!(s = init_str())) {
-		mem_free(inf);
-		goto oom;
-	}
-	c->info = inf;
-	if ((ps = get_pasv_socket(c, c->sock1, &c->sock2, pc))) return NULL;
 	if (!d) {
 		internal("get_url_data failed");
 		setcstate(c, S_INTERNAL);
 		abort_connection(c);
 		return NULL;
 	}
+	de = init_str(), del = 0;
+	add_conv_str(&de, &del, d, strlen(d), -2);
+	d = de;
+	inf = mem_alloc(sizeof(struct ftp_connection_info));
+	memset(inf, 0, sizeof(struct ftp_connection_info));
+	l = 0;
+	s = init_str();
+	inf->pasv = ftp_options.passive_ftp;
+	c->info = inf;
+	if (!inf->pasv) if ((ps = get_pasv_socket(c, c->sock1, &c->sock2, pc))) {
+		mem_free(d);
+		return NULL;
+	}
+#ifdef HAVE_IPTOS
+	if (ftp_options.set_tos) {
+		int on = IPTOS_THROUGHPUT;
+		setsockopt(c->sock2, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int));
+	}
+#endif
 	if (!(de = strchr(d, POST_CHAR))) de = d + strlen(d);
 	if (d == de || de[-1] == '/') {
 		inf->dir = 1;
 		inf->pending_commands = 4;
-		add_to_str(&s, &l, "TYPE A\r\nPORT ");
-		add_num_to_str(&s, &l, pc[0]);
-		add_chr_to_str(&s, &l, ',');
-		add_num_to_str(&s, &l, pc[1]);
-		add_chr_to_str(&s, &l, ',');
-		add_num_to_str(&s, &l, pc[2]);
-		add_chr_to_str(&s, &l, ',');
-		add_num_to_str(&s, &l, pc[3]);
-		add_chr_to_str(&s, &l, ',');
-		add_num_to_str(&s, &l, pc[4]);
-		add_chr_to_str(&s, &l, ',');
-		add_num_to_str(&s, &l, pc[5]);
-		add_to_str(&s, &l, "\r\nCWD /");
+		add_to_str(&s, &l, "TYPE A\r\n");
+		if (!inf->pasv) {
+			add_to_str(&s, &l, "PORT ");
+			add_num_to_str(&s, &l, pc[0]);
+			add_chr_to_str(&s, &l, ',');
+			add_num_to_str(&s, &l, pc[1]);
+			add_chr_to_str(&s, &l, ',');
+			add_num_to_str(&s, &l, pc[2]);
+			add_chr_to_str(&s, &l, ',');
+			add_num_to_str(&s, &l, pc[3]);
+			add_chr_to_str(&s, &l, ',');
+			add_num_to_str(&s, &l, pc[4]);
+			add_chr_to_str(&s, &l, ',');
+			add_num_to_str(&s, &l, pc[5]);
+			add_to_str(&s, &l, "\r\n");
+		} else {
+			add_to_str(&s, &l, "PASV\r\n");
+		}
+		add_to_str(&s, &l, "CWD /");
 		add_bytes_to_str(&s, &l, d, de - d);
 		add_to_str(&s, &l, "\r\nLIST\r\n");
 		c->from = 0;
 	} else {
 		inf->dir = 0;
 		inf->pending_commands = 3;
-		add_to_str(&s, &l, "TYPE I\r\nPORT ");
-		add_num_to_str(&s, &l, pc[0]);
-		add_chr_to_str(&s, &l, ',');
-		add_num_to_str(&s, &l, pc[1]);
-		add_chr_to_str(&s, &l, ',');
-		add_num_to_str(&s, &l, pc[2]);
-		add_chr_to_str(&s, &l, ',');
-		add_num_to_str(&s, &l, pc[3]);
-		add_chr_to_str(&s, &l, ',');
-		add_num_to_str(&s, &l, pc[4]);
-		add_chr_to_str(&s, &l, ',');
-		add_num_to_str(&s, &l, pc[5]);
-		if (c->from) {
-			add_to_str(&s, &l, "\r\nREST ");
+		add_to_str(&s, &l, "TYPE I\r\n");
+		if (!inf->pasv) {
+			add_to_str(&s, &l, "PORT ");
+			add_num_to_str(&s, &l, pc[0]);
+			add_chr_to_str(&s, &l, ',');
+			add_num_to_str(&s, &l, pc[1]);
+			add_chr_to_str(&s, &l, ',');
+			add_num_to_str(&s, &l, pc[2]);
+			add_chr_to_str(&s, &l, ',');
+			add_num_to_str(&s, &l, pc[3]);
+			add_chr_to_str(&s, &l, ',');
+			add_num_to_str(&s, &l, pc[4]);
+			add_chr_to_str(&s, &l, ',');
+			add_num_to_str(&s, &l, pc[5]);
+			add_to_str(&s, &l, "\r\n");
+		} else {
+			add_to_str(&s, &l, "PASV\r\n");
+		}
+		if (c->from && c->no_cache < NC_IF_MOD) {
+			add_to_str(&s, &l, "REST ");
 			add_num_to_str(&s, &l, c->from);
+			add_to_str(&s, &l, "\r\n");
 			inf->rest_sent = 1;
 			inf->pending_commands++;
-		}
-		add_to_str(&s, &l, "\r\nRETR /");
+		} else c->from = 0;
+		add_to_str(&s, &l, "RETR /");
 		add_bytes_to_str(&s, &l, d, de - d);
 		add_to_str(&s, &l, "\r\n");
 	}
 	inf->opc = inf->pending_commands;
-	if ((inf2 = mem_realloc(inf, sizeof(struct ftp_connection_info) + l + 1)))
-		strcpy((inf = inf2)->cmdbuf, s);
+	if ((unsigned)l > MAXINT - sizeof(struct ftp_connection_info) - 1) overalloc();
+	inf2 = mem_realloc(inf, sizeof(struct ftp_connection_info) + l + 1);
+	strcpy((inf = inf2)->cmdbuf, s);
 	mem_free(s);
 	c->info = inf;
+	mem_free(d);
 	return inf;
 }
 
 
-void ftp_send_retr_req(struct connection *c, int state)
+static void ftp_send_retr_req(struct connection *c, int state)
 {
 	struct ftp_connection_info *fi;
 	unsigned char *login;
 	int logl = 0;
-	if (!(login = init_str())) {
-		setcstate(c, S_OUT_OF_MEM);
-		abort_connection(c);
-		return;
-	}
+	set_timeout(c);
+	login = init_str();
 	if (!c->info && !(fi = add_file_cmd_to_str(c))) {
 		mem_free(login);
 		return;
 	} else fi = c->info;
-	if (fast_ftp) a:add_to_str(&login, &logl, fi->cmdbuf);
+	if (ftp_options.fast_ftp) a:add_to_str(&login, &logl, fi->cmdbuf);
 	else {
 		unsigned char *nl = strchr(fi->cmdbuf, '\n');
 		if (!nl) goto a;
@@ -297,25 +340,59 @@ void ftp_send_retr_req(struct connection *c, int state)
 	setcstate(c, state);
 }
 
-void ftp_retr_1(struct connection *c)
+static void ftp_retr_1(struct connection *c)
 {
 	struct read_buffer *rb;
 	if (!(rb = alloc_read_buffer(c))) return;
 	read_from_socket(c, c->sock1, rb, ftp_retr_file);
 }
 
-void ftp_retr_file(struct connection *c, struct read_buffer *rb)
+static void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 {
 	int g;
 	struct ftp_connection_info *inf = c->info;
 	if (0) {
 		rep:
-		if (!fast_ftp) {
+		if (!ftp_options.fast_ftp) {
 			ftp_send_retr_req(c, S_GETH);
 			return;
 		}
 	}
 	if (inf->pending_commands > 1) {
+		unsigned char pc[6];
+		if (inf->pasv && inf->opc - (inf->pending_commands - 1) == 2) {
+			int i = 3, j;
+			while (i < rb->len) {
+				if (rb->data[i] >= '0' && rb->data[i] <= '9') {
+					for (j = 0; j < 6; j++) {
+						int n = 0;
+						while (rb->data[i] >= '0' && rb->data[i] <= '9') {
+							n = n * 10 + rb->data[i] - '0';
+							if (n >= 256) goto no_pasv;
+							if (++i >= rb->len) goto no_pasv;
+						}
+						pc[j] = n;
+						if (j != 5) {
+							if (rb->data[i] != ',') goto xa;
+							if (++i >= rb->len) goto xa;
+							if (rb->data[i] < '0' || rb->data[i] > '9') {
+								xa:
+								if (j != 1) goto no_pasv;
+								pc[4] = pc[0];
+								pc[5] = pc[1];
+								pc[0] = pc[1] = pc[2] = pc[3] = 0;
+								goto pasv_ok;
+							}
+						}
+					}
+					goto pasv_ok;
+				}
+				i++;
+			}
+			no_pasv:
+			memset(pc, 0, sizeof pc);
+			pasv_ok:;
+		}
 		g = get_ftp_response(c, rb, 0);
 		if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
 		if (!g) { read_from_socket(c, c->sock1, rb, ftp_retr_file); setcstate(c, S_GETH); return; }
@@ -325,6 +402,14 @@ void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 				goto rep;
 			case 2:		/* PORT */
 				if (g >= 400) { setcstate(c, S_FTP_PORT); abort_connection(c); return; }
+				if (inf->pasv) {
+					if (!pc[4] && !pc[5]) {
+						setcstate(c, S_FTP_ERROR);
+						retry_connection(c);
+						return;
+					}
+					make_connection(c, (pc[4] << 8) + pc[5], &c->sock2, created_data_connection);
+				}
 				goto rep;
 			case 3:		/* REST / CWD */
 				if (g >= 400) {
@@ -338,7 +423,6 @@ void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 	g = get_ftp_response(c, rb, 2);
 	if (!g) { read_from_socket(c, c->sock1, rb, ftp_retr_file); setcstate(c, S_GETH); return; }
 	if (g >= 100 && g < 200) {
-		int est;
 		unsigned char *d = rb->data;
 		int i, p = 0;
 		for (i = 0; i < rb->len && d[i] != 10; i++) if (d[i] == '(') p = i;
@@ -351,26 +435,41 @@ void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 		for (; i < rb->len; i++) if (d[i] != ' ') break;
 		if (i + 4 > rb->len) goto nol;
 		if (casecmp(&d[i], "byte", 4)) goto nol;
-		est = strtol(&d[p], NULL, 10);
-		if (est && !c->from) c->est_length = est; /* !!! FIXME: when I add appedning to downloaded file */
+		{
+#if defined(HAVE_STRTOLL)
+			long long est = strtoll(&d[p], NULL, 10);
+#elif defined(HAVE_STRTOQ)
+			longlong est = strtoq(&d[p], NULL, 10);
+#else
+			long est = strtol(&d[p], NULL, 10);
+			if (est == MAXLONG) est = -1;
+#endif
+			if (est < 0 || (off_t)est < 0 || (off_t)est != est) est = 0;
+			if (est && !c->from) c->est_length = est; /* !!! FIXME: when I add appedning to downloaded file */
+		}
 		nol:;
 	}
-	set_handlers(c->sock2, (void (*)(void *))got_something_from_data_connection, NULL, NULL, c);
+	if (!inf->pasv)
+		set_handlers(c->sock2, (void (*)(void *))got_something_from_data_connection, NULL, NULL, c);
 	/*read_from_socket(c, c->sock1, rb, ftp_got_final_response);*/
 	ftp_got_final_response(c, rb);
 }
 
-void ftp_got_final_response(struct connection *c, struct read_buffer *rb)
+static void ftp_got_final_response(struct connection *c, struct read_buffer *rb)
 {
 	struct ftp_connection_info *inf = c->info;
 	int g = get_ftp_response(c, rb, 0);
 	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
 	if (!g) { read_from_socket(c, c->sock1, rb, ftp_got_final_response); if (c->state != S_TRANS) setcstate(c, S_GETH); return; }
-	if (g == 550) {
-		if (!c->cache && get_cache_entry(c->url, &c->cache)) {
-			setcstate(c, S_OUT_OF_MEM);
-			abort_connection(c);
-			return;
+	if (g == 425 || g == 450 || g == 500 || g == 501 || g == 550) {
+		if (c->url[strlen(c->url) - 1] == '/') goto skip_redir;
+		if (!c->cache) {
+			if (get_cache_entry(c->url, &c->cache)) {
+				setcstate(c, S_OUT_OF_MEM);
+				abort_connection(c);
+				return;
+			}
+			c->cache->refcount--;
 		}
 		if (c->cache->redirect) mem_free(c->cache->redirect);
 		c->cache->redirect = stracpy(c->url);
@@ -382,6 +481,7 @@ void ftp_got_final_response(struct connection *c, struct read_buffer *rb)
 		abort_connection(c);
 		return;
 	}
+	skip_redir:
 	if (g >= 400) { setcstate(c, S_FTP_FILE_ERROR); abort_connection(c); return; }
 	if (inf->conn_st == 2) {
 		setcstate(c, S_OK);
@@ -392,27 +492,27 @@ void ftp_got_final_response(struct connection *c, struct read_buffer *rb)
 	}
 }
 
-#define accept_char(x) ((upcase((x)) >= 'A' && upcase((x)) <= 'Z') || ((x) >= '0' && (x) <= '9') || (x) == ' ' || (x) == '-' || (x) == '_' || (x) == '.' || (x) == ':' || (x) == ';')
-
-void add_conv_str(unsigned char **s, int *l, unsigned char *b, int ll)
+static int is_date(char *data)	/* can touch at most data[-4] --- "n 12 "<--if fed with this --- if you change it, fix the caller */
 {
-	for (; ll; ll--, b++) {
-		if (accept_char(*b)) add_chr_to_str(s, l, *b);
-		else {
-			add_to_str(s, l, "&#");
-			add_num_to_str(s, l, (int)*b);
-			add_chr_to_str(s, l, ';');
-		}
-	}
+	/* fix for ftp://ftp.su.se/ */
+	if (*data == ' ') data--;
+	if (data[0] >= '0' && data[0] <= '9' && data[-1] >= '0' && data[-1] <= '9') data -= 2;
+	else if (data[0] >= '1' && data[0] <= '9' && data[-1] == ' ') data -= 1 + (data[-2] == ' ');
+	else return 0;
+	if (data[0] == ':') return 1;
+	if (data[0] != ' ') return 0;
+	if ((data[-1] < 'a' || data[-1] > 'z') && (data[-1] < 'A' || data[-1] > 'Z')) return 0;
+	return 1;
 }
 
-int ftp_process_dirlist(struct cache_entry *ce, int *pos, int *d, unsigned char *bf, int ln, int fin, int *tr)
+static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsigned char *bf, int ln, int fin, int *tr)
 {
 	unsigned char *str, *buf;
 	int sl;
 	int ret = 0;
 	int p;
 	int len;
+	int f;
 	again:
 	buf = bf + ret;
 	len = ln - ret;
@@ -429,31 +529,105 @@ int ftp_process_dirlist(struct cache_entry *ce, int *pos, int *d, unsigned char
 	str = init_str();
 	sl = 0;
 	/*add_to_str(&str, &sl, "   ");*/
+	f = *d;
 	if (*d && *d < p && WHITECHAR(buf[*d - 1])) {
-		int ee;
+		int ee, dir;
 		ppp:
 		for (ee = *d; ee <= p - 4; ee++)
 			if (!memcmp(buf + ee, " -> ", 4)) goto syml;
 		ee = p;
 		syml:
-		add_conv_str(&str, &sl, buf, *d);
-		add_to_str(&str, &sl, "<a href=\"");
-		add_conv_str(&str, &sl, buf + *d, ee - *d);
-		if (buf[0] == 'd') add_chr_to_str(&str, &sl, '/');
+		if (!f) {
+			if ((ee - *d != 1 || buf[*d] != '.') &&
+			    (ee - *d != 2 || buf[*d] != '.' || buf[*d + 1] != '.')) {
+				int i;
+				for (i = 0; i < *d; i++) add_chr_to_str(&str, &sl, ' ');
+				add_to_str(&str, &sl, "<a href=\"../\">..</a>\n");
+			}
+		}
+		dir = buf[0] == 'd';
+		if (!dir) {
+			unsigned char *p = memacpy(buf, *d);
+			if (strstr(p, "<DIR>")) dir = 1;
+			mem_free(p);
+		};
+		add_conv_str(&str, &sl, buf, *d, 0);
+		add_to_str(&str, &sl, "<a href=\"./");
+		add_conv_str(&str, &sl, buf + *d, ee - *d, 1);
+		if (dir) add_chr_to_str(&str, &sl, '/');
 		add_to_str(&str, &sl, "\">");
-		add_conv_str(&str, &sl, buf + *d, ee - *d);
+		add_conv_str(&str, &sl, buf + *d, ee - *d, 0);
 		add_to_str(&str, &sl, "</a>");
-		add_conv_str(&str, &sl, buf + ee, p - ee);
+		add_conv_str(&str, &sl, buf + ee, p - ee, 0);
 	} else {
-		int pp;
+		int pp, ppos;
+		int bp, bn;
 		if (p > 5 && !casecmp(buf, "total", 5)) goto raw;
+		for (pp = p - 1; pp >= 0; pp--) if (!WHITECHAR(buf[pp])) break;
+		if (pp < 0) goto raw;
+		if (pp < p - 1) pp++;
+		ppos = -1;
+		for (; pp >= 10; pp--) if (WHITECHAR(buf[pp])) {
+			if (is_date(&buf[pp - 6]) &&
+			    buf[pp - 5] == ' ' &&
+			    ((buf[pp - 4] == '2' && buf[pp - 3] == '0') ||
+			     (buf[pp - 4] == '1' && buf[pp - 3] == '9')) &&
+			    buf[pp - 2] >= '0' && buf[pp - 2] <= '9' &&
+			    buf[pp - 1] >= '0' && buf[pp - 1] <= '9') {
+				if (pp < p - 2 && buf[pp + 1] == ' ' && buf[pp + 2] != ' ') ppos = pp + 1;
+				else ppos = pp;
+			}
+			if (buf[pp - 6] == ' ' &&
+			    ((buf[pp - 5] >= '0' && buf[pp - 5] <= '2') || buf[pp - 5] == ' ') &&
+			    buf[pp - 4] >= '0' && buf[pp - 4] <= '9' &&
+			    buf[pp - 3] == ':' &&
+			    buf[pp - 2] >= '0' && buf[pp - 2] <= '5' &&
+			    buf[pp - 1] >= '0' && buf[pp - 1] <= '9') ppos = pp;
+		}
+		if (ppos != -1) {
+			pp = ppos;
+			goto done;
+		}
+
+		for (pp = 0; pp + 5 <= p; pp++)
+			if (!casecmp(&buf[pp], "<DIR>", 5)) {
+				pp += 4;
+				while (pp + 1 < p && WHITECHAR(buf[pp + 1])) pp++;
+				if (pp + 1 < p) goto done;
+			}
+		
+		bn = -1;
+		bp = 0;		/* warning, go away */
+		for (pp = 0; pp < p; ) {
+			if (buf[pp] >= '0' && buf[pp] <= '9') {
+				int i;
+				for (i = pp; i < p; i++)
+					if (buf[i] < '0' || buf[i] > '9') break;
+				if (i < p && WHITECHAR(buf[i])) {
+					if (i - pp > bn) {
+						bn = i - pp;
+						bp = pp;
+					}
+				}
+				pp = i;
+			}
+			while (pp < p && !WHITECHAR(buf[pp])) pp++;
+			while (pp < p && WHITECHAR(buf[pp])) pp++;
+		}
+		if (bn >= 0) {
+			pp = bp + bn;
+			while (pp + 1 < p && WHITECHAR(buf[pp + 1])) pp++;
+			if (pp + 1 < p) goto done;
+		}
+
 		for (pp = p - 1; pp >= 0; pp--) if (!WHITECHAR(buf[pp])) break;
 		if (pp < 0) goto raw;
 		for (; pp >= 0; pp--) if (WHITECHAR(buf[pp]) && (pp < 3 || memcmp(buf + pp - 3, " -> ", 4)) && (pp > p - 4 || memcmp(buf + pp, " -> ", 4))) break;
+		done:
 		*d = pp + 1;
 		goto ppp;
 		raw:
-		add_conv_str(&str, &sl, buf, p);
+		add_conv_str(&str, &sl, buf, p, 0);
 	}
 	add_chr_to_str(&str, &sl, '\n');
 	if (add_fragment(ce, *pos, str, sl)) *tr = 0;
@@ -462,7 +636,20 @@ int ftp_process_dirlist(struct cache_entry *ce, int *pos, int *d, unsigned char
 	goto again;
 }
 
-void got_something_from_data_connection(struct connection *c)
+static void created_data_connection(struct connection *c)
+{
+	struct ftp_connection_info *inf = c->info;
+#ifdef HAVE_IPTOS
+	if (ftp_options.set_tos) {
+		int on = IPTOS_THROUGHPUT;
+		setsockopt(c->sock2, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int));
+	}
+#endif
+	inf->d = 1;
+	set_handlers(c->sock2, (void (*)(void *))got_something_from_data_connection, NULL, NULL, c);
+}
+
+static void got_something_from_data_connection(struct connection *c)
 {
 	struct ftp_connection_info *inf = c->info;
 	int l;
@@ -477,13 +664,18 @@ void got_something_from_data_connection(struct connection *c)
 		set_handlers(ns, (void (*)(void *))got_something_from_data_connection, NULL, NULL, c);
 		return;
 	}
-	if (!c->cache && get_cache_entry(c->url, &c->cache)) {
-		setcstate(c, S_OUT_OF_MEM);
-		abort_connection(c);
-		return;
+	if (!c->cache) {
+		if (get_cache_entry(c->url, &c->cache)) {
+			setcstate(c, S_OUT_OF_MEM);
+			abort_connection(c);
+			return;
+		}
+		c->cache->refcount--;
 	}
 	if (inf->dir && !c->from) {
 		unsigned char *ud;
+		unsigned char *s0;
+		int s0l;
 		static unsigned char ftp_head[] = "<html><head><title>/";
 		static unsigned char ftp_head2[] = "</title></head><body><h2>Directory /";
 		static unsigned char ftp_head3[] = "</h2><pre>";
@@ -491,23 +683,38 @@ void got_something_from_data_connection(struct connection *c)
 		A(ftp_head);
 		ud = stracpy(get_url_data(c->url));
 		if (strchr(ud, POST_CHAR)) *strchr(ud, POST_CHAR) = 0;
-		if (ud) A(ud);
+		s0 = init_str();
+		s0l = 0;
+		add_conv_str(&s0, &s0l, ud, strlen(ud), -1);
+		mem_free(ud);
+		A(s0);
 		A(ftp_head2);
-		if (ud) A(ud);
+		A(s0);
 		A(ftp_head3);
-		if (ud) mem_free(ud);
+		mem_free(s0);
 		if (!c->cache->head) c->cache->head = stracpy("\r\n");
 		add_to_strn(&c->cache->head, "Content-Type: text/html\r\n");
 #undef A
 	}
 	if ((l = read(c->sock2, inf->ftp_buffer + inf->buf_pos, FTP_BUF - inf->buf_pos)) == -1) {
 		e:
-		setcstate(c, -errno);
+		if (inf->conn_st != 1 && !inf->dir && !c->from) {
+			set_handlers(c->sock2, NULL, NULL, NULL, NULL);
+			close_socket(&c->sock2);
+			inf->conn_st = 2;
+			return;
+		}
+		setcstate(c, get_error_from_errno(errno));
 		retry_connection(c);
 		return;
 	}
 	if (l > 0) {
 		if (!inf->dir) {
+			if ((off_t)(0UL + c->from + l) < 0) {
+				setcstate(c, S_LARGE_FILE);
+				abort_connection(c);
+				return;
+			}
 			c->received += l;
 			if (add_fragment(c->cache, c->from, inf->ftp_buffer, l) == 1) c->tries = 0;
 			c->from += l;
@@ -529,11 +736,10 @@ void got_something_from_data_connection(struct connection *c)
 		ftp_end_request(c);
 	} else {
 		inf->conn_st = 2;
-		setcstate(c, S_TRANS);
 	}
 }
 
-void ftp_end_request(struct connection *c)
+static void ftp_end_request(struct connection *c)
 {
 	if (c->state == S_OK) {
 		if (c->cache) {
