@@ -1,10 +1,12 @@
 #include "links.h"
 
+#define format format_
+
 struct list_head html_stack = {&html_stack, &html_stack};
 
 static inline int atchr(unsigned char c)
 {
-	return isA(c) || (c > ' ' && c != '=' && c != '<' && c != '>');
+	return /*isA(c) ||*/ (c > ' ' && c != '=' && c != '<' && c != '>');
 }
 
 int parse_element(unsigned char *e, unsigned char *eof, unsigned char **name, int *namelen, unsigned char **attr, unsigned char **end)
@@ -19,7 +21,7 @@ int parse_element(unsigned char *e, unsigned char *eof, unsigned char **name, in
 		e++;
 		if (e >= eof) return -1;
 	}
-	if ((!WHITECHAR(*e) && *e != '>' && *e != '<' && *e != '/' && *e != ':')) return -1;
+	/*if ((!WHITECHAR(*e) && *e != '>' && *e != '<' && *e != '/' && *e != ':')) return -1;*/
 	if (name && namelen) *namelen = e - *name;
 	while ((WHITECHAR(*e) || *e == '/' || *e == ':')) {
 		e++;
@@ -51,7 +53,7 @@ int parse_element(unsigned char *e, unsigned char *eof, unsigned char **name, in
 	}
 	if (U(*e)) {
 		unsigned char uu = *e;
-		u:
+		/*u:*/
 		goto x3;
 		while (e < eof && *e != uu && *e /*(WHITECHAR(*e) || *e > ' ')*/) {
 			x3:
@@ -61,7 +63,7 @@ int parse_element(unsigned char *e, unsigned char *eof, unsigned char **name, in
 		if (*e < ' ') return -1;
 		e++;
 		if (e >= eof /*|| (!WHITECHAR(*e) && *e != uu && *e != '>' && *e != '<')*/) return -1;
-		if (*e == uu) goto u;
+		/*if (*e == uu) goto u;*/
 	} else {
 		while (!WHITECHAR(*e) && *e != '>' && *e != '<') {
 			e++;
@@ -79,14 +81,21 @@ int parse_element(unsigned char *e, unsigned char *eof, unsigned char **name, in
 	return 0;
 }
 
-unsigned char *_xx_;
-
-#define add_chr(s, l, c)	do {\
-	if ((_xx_ = s, l % ALLOC_GR) || (_xx_ = mem_realloc(s, l + ALLOC_GR))) s = _xx_, s[l++] = c;	\
-	} while (0)
+#define add_chr(s, l, c)						\
+do {									\
+	if (!((l) & (32 - 1))) {					\
+		if ((unsigned)(l) > MAXINT - 32) overalloc();		\
+		(s) = mem_realloc((s), (l) + 32);			\
+	}								\
+	(s)[(l)++] = (c);						\
+} while (0)
 
 int get_attr_val_nl = 0;
 
+/* parses html element attributes */
+/* e is attr pointer previously get from parse_element, DON'T PASS HERE ANY OTHER VALUE!!! */
+/* name is searched attribute */
+/* returns allocated string containing the attribute, or NULL on unsuccess */
 unsigned char *get_attr_val(unsigned char *e, unsigned char *name)
 {
 	unsigned char *n;
@@ -111,24 +120,27 @@ unsigned char *get_attr_val(unsigned char *e, unsigned char *name)
 		}
 	} else {
 		char uu = *e;
-		a:
+		/*a:*/
 		e++;
 		while (*e != uu) {
 			if (!*e) {
 				mem_free(a);
 				return NULL;
 			}
-			if (!f && *e != 13) {
-				if (*e != 9 && *e != 10) add_chr(a, l, *e);
-				else if (!get_attr_val_nl) add_chr(a, l, ' ');
+			if (!f) {
+				if (get_attr_val_nl == 2) goto exact;
+				if (*e != 13) {
+					if (*e != 9 && *e != 10) exact:add_chr(a, l, *e);
+					else if (!get_attr_val_nl) add_chr(a, l, ' ');
+				}
 			}
 			e++;
 		}
 		e++;
-		if (*e == uu) {
+		/*if (*e == uu) {
 			if (!f) add_chr(a, l, *e);
 			goto a;
-		}
+		}*/
 	}
 	ea:
 	if (!f) {
@@ -136,12 +148,18 @@ unsigned char *get_attr_val(unsigned char *e, unsigned char *name)
 		add_chr(a, l, 0);
 		if (strchr(a, '&')) {
 			unsigned char *aa = a;
+			int c = d_opt->cp;
+			d_opt->cp = d_opt->real_cp;
 			a = convert_string(NULL, aa, strlen(aa));
+			d_opt->cp = c;
 			mem_free(aa);
 		}
-		for (b = a; *b == ' '; b++);
-		if (b != a) memmove(a, b, strlen(b) + 1);
-		for (b = a + strlen(a) - 1; b >= a && *b == ' '; b--) *b = 0;
+		while ((b = strchr(a, 1))) *b = ' ';
+		if (get_attr_val_nl != 2) {
+			for (b = a; *b == ' '; b++);
+			if (b != a) memmove(a, b, strlen(b) + 1);
+			for (b = a + strlen(a) - 1; b >= a && *b == ' '; b--) *b = 0;
+		}
 		set_mem_comment(a, name, strlen(name));
 		return a;
 	}
@@ -182,37 +200,57 @@ unsigned char *get_url_val(unsigned char *e, unsigned char *name)
 	return a;
 }
 
+unsigned char *get_exact_attr_val(unsigned char *e, unsigned char *name)
+{
+	unsigned char *a;
+	get_attr_val_nl = 2;
+	a = get_attr_val(e, name);
+	get_attr_val_nl = 0;
+	if (a) {
+		unsigned char *x1, *x2;
+		for (x1 = x2 = a; *x1; x1++, x2++) {
+			if (x1[0] == '\r') {
+				*x2 = '\n';
+				if (x1[1] == '\n') x1++;
+			} else {
+				*x2 = *x1;
+			}
+		}
+		*x2 = 0;
+	}
+	return a;
+}
 
 struct {
-	int n;
+	unsigned short int n;
 	char *s;
 } roman_tbl[] = {
-	1000,	"m",
-	999,	"im",
-/*	995,	"vm",*/
-	990,	"xm",
-/*	950,	"lm",*/
-	900,	"cm",
-	500,	"d",
-	499,	"id",
-/*	495,	"vd",*/
-	490,	"xd",
-/*	450,	"ld",*/
-	400,	"cd",
-	100,	"c",
-	99,	"ic",
-/*	95,	"vc",*/
-	90,	"xc",
-	50,	"l",
-	49,	"il",
-/*	45,	"vl",*/
-	40,	"xl",
-	10,	"x",
-	9,	"ix",
-	5,	"v",
-	4,	"iv",
-	1,	"i",
-	0,	NULL
+	{ 1000,	"m" },
+	{ 999,	"im" },
+/*	{ 995,	"vm" },*/
+	{ 990,	"xm" },
+/*	{ 950,	"lm" },*/
+	{ 900,	"cm" },
+	{ 500,	"d" },
+	{ 499,	"id" },
+/*	{ 495,	"vd" },*/
+	{ 490,	"xd" },
+/*	{ 450,	"ld" },*/
+	{ 400,	"cd" },
+	{ 100,	"c" },
+	{ 99,	"ic" },
+/*	{ 95,	"vc" },*/
+	{ 90,	"xc" },
+	{ 50,	"l" },
+	{ 49,	"il" },
+/*	{ 45,	"vl" },*/
+	{ 40,	"xl" },
+	{ 10,	"x" },
+	{ 9,	"ix" },
+	{ 5,	"v" },
+	{ 4,	"iv" },
+	{ 1,	"i" },
+	{ 0,	NULL },
 };
 
 void roman(char *p, unsigned n)
@@ -392,7 +430,7 @@ struct color_spec color_specs[] = {
 
 int decode_color(unsigned char *str, struct rgb *col)
 {
-	int ch;
+	unsigned long ch;
 	if (*str != '#') {
 		struct color_spec *cs;
 		for (cs = color_specs; cs < endof(color_specs); cs++)
@@ -406,7 +444,7 @@ int decode_color(unsigned char *str, struct rgb *col)
 	if (strlen(str) == 6) {
 		char *end;
 		ch = strtoul(str, &end, 16);
-		if (!*end) {
+		if (!*end && ch < 0x1000000) {
 found:
 			col->r = ch / 0x10000;
 			col->g = ch / 0x100 % 0x100;
@@ -471,7 +509,7 @@ void kill_html_stack_item(struct html_element *e)
 
 static inline void kill_elem(char *e)
 {
-	if (html_top.namelen == strlen(e) && !casecmp(html_top.name, e, html_top.namelen))
+	if ((size_t)html_top.namelen == strlen(e) && !casecmp(html_top.name, e, html_top.namelen))
 		kill_html_stack_item(&html_top);
 }
 
@@ -500,14 +538,14 @@ void html_stack_dup()
 	   	internal("html stack empty");
 		return;
 	}
-	if (!(e = mem_alloc(sizeof(struct html_element)))) return;
+	e = mem_alloc(sizeof(struct html_element));
 	memcpy(e, ep, sizeof(struct html_element));
-	copy_string(&e->attr.link, ep->attr.link);
-	copy_string(&e->attr.target, ep->attr.target);
-	copy_string(&e->attr.image, ep->attr.image);
-	copy_string(&e->attr.href_base, ep->attr.href_base);
-	copy_string(&e->attr.target_base, ep->attr.target_base);
-	copy_string(&e->attr.select, ep->attr.select);
+	e->attr.link = stracpy(ep->attr.link);
+	e->attr.target = stracpy(ep->attr.target);
+	e->attr.image = stracpy(ep->attr.image);
+	e->attr.href_base = stracpy(ep->attr.href_base);
+	e->attr.target_base = stracpy(ep->attr.target_base);
+	e->attr.select = stracpy(ep->attr.select);
 	/*if (e->name) {
 		if (e->attr.link) set_mem_comment(e->attr.link, e->name, e->namelen);
 		if (e->attr.target) set_mem_comment(e->attr.target, e->name, e->namelen);
@@ -523,9 +561,8 @@ void html_stack_dup()
 }
 
 void *ff;
-void (*put_chars_f)(void *, unsigned char *, int);
+int (*put_chars_f)(void *, unsigned char *, int);
 void (*line_break_f)(void *);
-void (*init_f)(void *);
 void *(*special_f)(void *, int, ...);
 
 unsigned char *eoff;
@@ -545,14 +582,14 @@ void ln_break(int n, void (*line_break)(void *), void *f)
 	if (!n || html_top.invisible) return;
 	while (n > line_breax) line_breax++, line_break(f);
 	pos = 0;
-	putsp = -1; /* ??? */
+	putsp = -1;
 }
 
-void put_chrs(unsigned char *start, int len, void (*put_chars)(void *, unsigned char *, int), void *f)
+void put_chrs(unsigned char *start, int len, int (*put_chars)(void *, unsigned char *, int), void *f)
 {
 	if (par_format.align == AL_NO) putsp = 0;
 	if (!len || html_top.invisible) return;
-	if (putsp == 1) put_chars(f, " ", 1), pos++, putsp = -1;
+	if (putsp == 1) pos += put_chars(f, " ", 1), putsp = -1;
 	if (putsp == -1) {
 		if (start[0] == ' ') start++, len--;
 		putsp = 0;
@@ -565,8 +602,7 @@ void put_chrs(unsigned char *start, int len, void (*put_chars)(void *, unsigned
 	if (start[len - 1] == ' ') putsp = -1;
 	if (par_format.align == AL_NO) putsp = 0;
 	was_br = 0;
-	put_chars(f, start, len);
-	pos += len;
+	pos += put_chars(f, start, len);
 	line_breax = 0;
 }
 
@@ -583,16 +619,16 @@ void kill_until(int ls, ...)
 			char *s = va_arg(arg, char *);
 			if (!s) break;
 			if (!*s) sk++;
-			else if (e->namelen == strlen(s) && !casecmp(e->name, s, strlen(s)))
+			else if ((size_t)e->namelen == strlen(s) && !casecmp(e->name, s, strlen(s))) {
 				if (!sk) {
 					if (e->dontkill) break;
 					va_end(arg);
 					goto killll;
-				}
-				else if (sk == 1) {
+				} else if (sk == 1) {
 					va_end(arg);
 					goto killl;
 				} else break;
+			}
 		}
 		va_end(arg);
 		if (e->dontkill || (e->namelen == 5 && !casecmp(e->name, "TABLE", 5))) break;
@@ -618,8 +654,8 @@ int get_num(unsigned char *a, unsigned char *n)
 	char *al;
 	if ((al = get_attr_val(a, n))) {
 		char *end;
-		int s = strtoul(al, &end, 10);
-		if (!*al || *end || s < 0) s = -1;
+		unsigned long s = strtoul(al, &end, 10);
+		if (!*al || *end || s > 10000) s = -1;
 		mem_free(al);
 		return s;
 	}
@@ -630,7 +666,7 @@ int parse_width(unsigned char *w, int trunc)
 {
 	unsigned char *end;
 	int p = 0;
-	int s;
+	long s;
 	int l;
 	while (WHITECHAR(*w)) w++;
 	for (l = 0; w[l] && w[l] != ','; l++) ;
@@ -639,8 +675,8 @@ int parse_width(unsigned char *w, int trunc)
 	if (w[l - 1] == '%') l--, p = 1;
 	while (l && WHITECHAR(w[l - 1])) l--;
 	if (!l) return -1;
-	s = strtoul((char *)w, (char **)&end, 10);
-	if (end - w < l) return -1;
+	s = strtoul((char *)w, (char **)(void *)&end, 10);
+	if (end - w < l || s < 0 || s > 10000) return -1;
 	if (p) {
 		if (trunc) s = s * (par_format.width - par_format.leftmargin - par_format.rightmargin) / 100;
 		else return -1;
@@ -723,7 +759,7 @@ void html_font(unsigned char *a)
 	char *al;
 	if ((al = get_attr_val(a, "size"))) {
 		int p = 0;
-		unsigned s;
+		unsigned long s;
 		char *nn = al;
 		char *end;
 		if (*al == '+') p = 1, nn++;
@@ -753,10 +789,9 @@ void html_img(unsigned char *a)
 		if (format.link) mem_free(format.link);
 		if (format.form) format.form = NULL;
 		u = join_urls(format.href_base, al);
-		if ((format.link = mem_alloc(strlen(u) + 5))) {
-			strcpy(format.link, "MAP@");
-			strcat(format.link, u);
-		}
+		format.link = mem_alloc(strlen(u) + 5);
+		strcpy(format.link, "MAP@");
+		strcat(format.link, u);
 		format.attr |= AT_BOLD;
 		mem_free(u);
 		mem_free(al);
@@ -897,7 +932,7 @@ void html_pre(unsigned char *a)
 void html_hr(unsigned char *a)
 {
 	int i/* = par_format.width - 10*/;
-	char r = 205;
+	unsigned char r = 205;
 	int q = get_num(a, "size");
 	if (q >= 0 && q < 2) r = 196;
 	html_stack_dup();
@@ -973,8 +1008,8 @@ void html_ul(unsigned char *a)
 		if (!strcasecmp(al, "square")) par_format.flags = P_PLUS;
 		mem_free(al);
 	}
-	if ((par_format.leftmargin += 2 + (par_format.list_level > 1)) > par_format.width / 2 && !table_level)
-		par_format.leftmargin = par_format.width / 2;
+	if ((par_format.leftmargin += 2 + (par_format.list_level > 1)) > par_format.width * 2 / 3 && !table_level)
+		par_format.leftmargin = par_format.width * 2 / 3;
 	par_format.align = AL_LEFT;
 	html_top.dontkill = 1;
 }
@@ -998,8 +1033,8 @@ void html_ol(unsigned char *a)
 		if (!strcmp(al, "I")) par_format.flags = P_ROMAN;
 		mem_free(al);
 	}
-	if ((par_format.leftmargin += (par_format.list_level > 1)) > par_format.width / 2 && !table_level)
-		par_format.leftmargin = par_format.width / 2;
+	if ((par_format.leftmargin += (par_format.list_level > 1)) > par_format.width * 2 / 3 && !table_level)
+		par_format.leftmargin = par_format.width * 2 / 3;
 	par_format.align = AL_LEFT;
 	html_top.dontkill = 1;
 }
@@ -1036,8 +1071,8 @@ void html_li(unsigned char *a)
 		put_chrs(".&nbsp;", 7, put_chars_f, ff);
 		par_format.leftmargin += strlen(n) + c + 2;
 		par_format.align = AL_LEFT;
+		html_top.next->parattr.list_number = par_format.list_number + 1;
 		par_format.list_number = 0;
-		html_top.next->parattr.list_number++;
 		putsp = -1;
 	}
 	line_breax = 2;
@@ -1071,8 +1106,8 @@ void html_dt(unsigned char *a)
 void html_dd(unsigned char *a)
 {
 	kill_until(0, "", "DL", NULL);
-	if ((par_format.leftmargin = par_format.dd_margin + (table_level ? 3 : 8)) > par_format.width / 2 && !table_level)
-		par_format.leftmargin = par_format.width / 2;
+	if ((par_format.leftmargin = par_format.dd_margin + (table_level ? 3 : 8)) > par_format.width * 2 / 3 && !table_level)
+		par_format.leftmargin = par_format.width * 2 / 3;
 	par_format.align = AL_LEFT;
 }
 
@@ -1111,21 +1146,27 @@ void get_html_form(unsigned char *a, struct form *form)
 	form->num = a - startf;
 }
 
-void find_form_for_input(unsigned char *i)
+void find_form_for_input(unsigned char *i, int when_unused)
 {
 	unsigned char *s, *ss, *name, *attr, *lf, *la;
 	int namelen;
 	if (form.action) mem_free(form.action);
 	if (form.target) mem_free(form.target);
 	memset(&form, 0, sizeof(form));
-	if (!special_f(ff, SP_USED, NULL)) return;
-	if (last_input_tag && i <= last_input_tag && i > last_form_tag) {
+	if (!when_unused && !special_f(ff, SP_USED, NULL)) return;
+	if (last_form_tag && last_input_tag && i <= last_input_tag && i > last_form_tag) {
 		get_html_form(last_form_attr, &form);
 		return;
 	}
-	if (last_input_tag && i > last_input_tag) s = last_form_tag;
-	else s = startf;
-	lf = NULL, la = NULL;
+	if (last_form_tag && last_input_tag && i > last_input_tag) {
+		if (parse_element(last_form_tag, i, &name, &namelen, &la, &s))
+			internal("couldn't parse already parsed tag");
+		lf = last_form_tag;
+		s = last_input_tag;
+	} else {
+		lf = NULL, la = NULL;
+		s = startf;
+	}
 	se:
 	while (s < i && *s != '<') sp:s++;
 	if (s >= i) goto end_parse;
@@ -1155,8 +1196,8 @@ void html_button(unsigned char *a)
 {
 	char *al;
 	struct form_control *fc;
-	find_form_for_input(a);
-	if (!(fc = mem_alloc(sizeof(struct form_control)))) return;
+	find_form_for_input(a, 0);
+	fc = mem_alloc(sizeof(struct form_control));
 	memset(fc, 0, sizeof(struct form_control));
 	if (!(al = get_attr_val(a, "type"))) {
 		fc->type = FC_SUBMIT;
@@ -1187,7 +1228,7 @@ void html_button(unsigned char *a)
 	fc->method = form.method;
 	fc->action = stracpy(form.action);
 	fc->name = get_attr_val(a, "name");
-	fc->default_value = get_attr_val(a, "value");
+	fc->default_value = get_exact_attr_val(a, "value");
 	fc->ro = has_attr(a, "disabled") ? 2 : has_attr(a, "readonly") ? 1 : 0;
 	if (fc->type == FC_IMAGE) fc->alt = get_attr_val(a, "alt");
 	if (fc->type == FC_SUBMIT && !fc->default_value) fc->default_value = stracpy("Submit");
@@ -1207,8 +1248,8 @@ void html_input(unsigned char *a)
 	int i;
 	unsigned char *al;
 	struct form_control *fc;
-	find_form_for_input(a);
-	if (!(fc = mem_alloc(sizeof(struct form_control)))) return;
+	find_form_for_input(a, 0);
+	fc = mem_alloc(sizeof(struct form_control));
 	memset(fc, 0, sizeof(struct form_control));
 	if (!(al = get_attr_val(a, "type"))) {
 		fc->type = FC_TEXT;
@@ -1243,15 +1284,20 @@ void html_input(unsigned char *a)
 	fc->action = stracpy(form.action);
 	fc->target = stracpy(form.target);
 	fc->name = get_attr_val(a, "name");
-	if (fc->type != FC_FILE) fc->default_value = get_attr_val(a, "value");
+	if (fc->type == FC_TEXT || fc->type == FC_PASSWORD) fc->default_value = get_attr_val(a, "value");
+	else if (fc->type != FC_FILE) fc->default_value = get_exact_attr_val(a, "value");
 	if (fc->type == FC_CHECKBOX && !fc->default_value) fc->default_value = stracpy("on");
 	if ((fc->size = get_num(a, "size")) == -1) fc->size = HTML_DEFAULT_INPUT_SIZE;
 	fc->size++;
 	if (fc->size > d_opt->xw) fc->size = d_opt->xw;
-	if ((fc->maxlength = get_num(a, "maxlength")) == -1) fc->maxlength = MAXINT;
+	if ((fc->maxlength = get_num(a, "maxlength")) == -1) fc->maxlength = MAXINT / 4;
 	if (fc->type == FC_CHECKBOX || fc->type == FC_RADIO) fc->default_state = has_attr(a, "checked");
 	fc->ro = has_attr(a, "disabled") ? 2 : has_attr(a, "readonly") ? 1 : 0;
-	if (fc->type == FC_IMAGE) fc->alt = get_attr_val(a, "alt");
+	if (fc->type == FC_IMAGE) {
+		fc->alt = get_attr_val(a, "alt");
+		if (!fc->alt) fc->alt = get_attr_val(a, "title");
+		if (!fc->alt) fc->alt = get_attr_val(a, "name");
+	}
 	if (fc->type == FC_SUBMIT && !fc->default_value) fc->default_value = stracpy("Submit");
 	if (fc->type == FC_RESET && !fc->default_value) fc->default_value = stracpy("Reset");
 	if (!fc->default_value) fc->default_value = stracpy("");
@@ -1305,6 +1351,7 @@ void html_select(unsigned char *a)
 	char *al;
 	if (!(al = get_attr_val(a, "name"))) return;
 	html_top.dontkill = 1;
+	if (format.select) mem_free(format.select);
 	format.select = al;
 	format.select_disabled = 2 * has_attr(a, "disabled");
 }
@@ -1313,17 +1360,17 @@ void html_option(unsigned char *a)
 {
 	struct form_control *fc;
 	unsigned char *val;
-	find_form_for_input(a);
+	find_form_for_input(a, 0);
 	if (!format.select) return;
-	if (!(fc = mem_alloc(sizeof(struct form_control)))) return;
+	fc = mem_alloc(sizeof(struct form_control));
 	memset(fc, 0, sizeof(struct form_control));
-	if (!(val = get_attr_val(a, "value"))) {
+	if (!(val = get_exact_attr_val(a, "value"))) {
 		unsigned char *p, *r;
 		unsigned char *name;
 		int namelen;
 		int l = 0;
 		for (p = a - 1; *p != '<'; p--) ;
-		if (!(val = init_str())) goto x;
+		val = init_str();
 		if (parse_element(p, eoff, NULL, NULL, NULL, &p)) {
 			internal("parse element failed");
 			goto x;
@@ -1373,12 +1420,17 @@ void html_option(unsigned char *a)
 
 void clr_spaces(unsigned char *name)
 {
-	unsigned char *nm;
-	for (nm = name; *nm; nm++)
-		if (WHITECHAR(*nm) || *nm == 1) *nm = ' ';
-	for (nm = name; *nm; nm++)
+	unsigned char *n1, *n2;
+	for (n1 = name; *n1; n1++)
+		if (WHITECHAR(*n1) || *n1 == 1) *n1 = ' ';
+	/*for (nm = name; *nm; nm++)
 		while (nm[0] == ' ' && (nm == name || nm[1] == ' ' || !nm[1]))
-			memmove(nm, nm + 1, strlen(nm));
+			memmove(nm, nm + 1, strlen(nm));*/
+	if (!strchr(name, ' ')) return;
+	for (n1 = name, n2 = name; *n1; n1++)
+		if (!(n1[0] == ' ' && (n2 == name || n1[1] == ' ' || !n1[1])))
+			*n2++ = *n1;
+	*n2 = 0;
 }
 
 int menu_stack_size;
@@ -1394,21 +1446,15 @@ void new_menu_item(unsigned char *name, int data, int fullname)
 		if (name[0] == 1) name[0] = ' ';
 	}
 	if (name && data == -1) {
-		if (!(nmenu = mem_alloc(sizeof(struct menu_item)))) {
-			mem_free(name);
-			return;
-		}
+		nmenu = mem_alloc(sizeof(struct menu_item));
 		memset(nmenu, 0, sizeof(struct menu_item));
 		/*nmenu->text = "";*/
 	}
 	if (menu_stack_size && name) {
 		top = item = menu_stack[menu_stack_size - 1];
 		while (item->text) item++;
-		if (!(top = mem_realloc(top, (char *)(item + 2) - (char *)top))) {
-			if (data == -1) mem_free(nmenu);
-			mem_free(name);
-			return;
-		}
+		if ((size_t)((char *)(item + 2) - (char *)top) > MAXINT) overalloc();
+		top = mem_realloc(top, (char *)(item + 2) - (char *)top);
 		item = item - menu_stack[menu_stack_size - 1] + top;
 		menu_stack[menu_stack_size - 1] = top;
 		if (menu_stack_size >= 2) {
@@ -1428,9 +1474,8 @@ void new_menu_item(unsigned char *name, int data, int fullname)
 		/*item->text = "";*/
 	} else if (name) mem_free(name);
 	if (name && data == -1) {
-		struct menu_item **ms;
-		if (!(ms = mem_realloc(menu_stack, (menu_stack_size + 1) * sizeof(struct menu_item *)))) return;
-		menu_stack = ms;
+		if ((unsigned)menu_stack_size > MAXINT / sizeof(struct menu_item *) - 1) overalloc();
+		menu_stack = mem_realloc(menu_stack, (menu_stack_size + 1) * sizeof(struct menu_item *));
 		menu_stack[menu_stack_size++] = nmenu;
 	}
 	if (!name) menu_stack_size--;
@@ -1511,14 +1556,19 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	int t_namelen;
 	unsigned char *lbl;
 	int lbl_l;
+	unsigned char *vlbl;
+	int vlbl_l;
 	int nnmi = 0;
 	struct conv_table *ct = special_f(f, SP_TABLE, NULL);
 	unsigned char **val, **lbls;
 	int order, preselect, group;
 	int i, mw;
 	if (has_attr(attr, "multiple")) return 1;
-	find_form_for_input(attr);
+	find_form_for_input(attr, 0);
 	lbl = NULL;
+	lbl_l = 0;
+	vlbl = NULL;
+	vlbl_l = 0;
 	val = DUMMY;
 	order = 0, group = 0, preselect = -1;
 	init_menu();
@@ -1532,6 +1582,7 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 		abort:
 		*end = html;
 		if (lbl) mem_free(lbl);
+		if (vlbl) mem_free(vlbl);
 		for (i = 0; i < order; i++) if (val[i]) mem_free(val[i]);
 		mem_free(val);
 		destroy_menu();
@@ -1545,6 +1596,7 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 		while (l && WHITECHAR(s[l-1])) l--;
 		q = convert_string(ct, s, l);
 		if (q) add_to_str(&lbl, &lbl_l, q), mem_free(q);
+		add_bytes_to_str(&vlbl, &vlbl_l, s, l);
 	}
 	if (html + 2 <= eof && (html[1] == '!' || html[1] == '?')) {
 		html = skip_comment(html, eof);
@@ -1556,55 +1608,64 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	}
 	if (t_namelen == 7 && !casecmp(t_name, "/SELECT", 7)) {
 		if (lbl) {
-			if (!val[order - 1]) val[order - 1] = stracpy(lbl);
+			if (!val[order - 1]) val[order - 1] = stracpy(vlbl);
 			if (!nnmi) new_menu_item(lbl, order - 1, 1), lbl = NULL;
 			else mem_free(lbl), lbl = NULL;
+			mem_free(vlbl), vlbl = NULL;
 		}
 		goto end_parse;
 	}
 	if (t_namelen == 7 && !casecmp(t_name, "/OPTION", 7)) {
 		if (lbl) {
-			if (!val[order - 1]) val[order - 1] = stracpy(lbl);
+			if (!val[order - 1]) val[order - 1] = stracpy(vlbl);
 			if (!nnmi) new_menu_item(lbl, order - 1, 1), lbl = NULL;
 			else mem_free(lbl), lbl = NULL;
+			mem_free(vlbl), vlbl = NULL;
 		}
 		goto see;
 	}
 	if (t_namelen == 6 && !casecmp(t_name, "OPTION", 6)) {
 		unsigned char *v, *vx;
 		if (lbl) {
-			if (!val[order - 1]) val[order - 1] = stracpy(lbl);
+			if (!val[order - 1]) val[order - 1] = stracpy(vlbl);
 			if (!nnmi) new_menu_item(lbl, order - 1, 1), lbl = NULL;
 			else mem_free(lbl), lbl = NULL;
+			mem_free(vlbl), vlbl = NULL;
 		}
 		if (has_attr(t_attr, "disabled")) goto see;
 		if (preselect == -1 && has_attr(t_attr, "selected")) preselect = order;
-		v = get_attr_val(t_attr, "value");
+		v = get_exact_attr_val(t_attr, "value");
 		if (!(order & (ALLOC_GR - 1))) {
-			unsigned char **vv;
-			if (!(vv = mem_realloc(val, (order + ALLOC_GR) * sizeof(unsigned char *)))) goto abort;
-			val = vv;
+			if ((unsigned)order > MAXINT / sizeof(unsigned char *) - ALLOC_GR) overalloc();
+			if ((unsigned)order > MAXINT / sizeof(char *) - ALLOC_GR) overalloc();
+			val = mem_realloc(val, (order + ALLOC_GR) * sizeof(unsigned char *));
 		}
 		val[order++] = v;
-		if ((vx = get_attr_val(t_attr, "label"))) new_menu_item(vx, order - 1, 0);
+		if ((vx = get_attr_val(t_attr, "label"))) {
+			new_menu_item(convert_string(ct, vx, strlen(vx)), order - 1, 0);
+			mem_free(vx);
+		}
 		if (!v || !vx) {
 			lbl = init_str(), lbl_l = 0;
+			vlbl = init_str(), vlbl_l = 0;
 			nnmi = !!vx;
 		}
 		goto see;
 	}
 	if ((t_namelen == 8 && !casecmp(t_name, "OPTGROUP", 8)) || (t_namelen == 9 && !casecmp(t_name, "/OPTGROUP", 9))) {
 		if (lbl) {
-			if (!val[order - 1]) val[order - 1] = stracpy(lbl);
+			if (!val[order - 1]) val[order - 1] = stracpy(vlbl);
 			if (!nnmi) new_menu_item(lbl, order - 1, 1), lbl = NULL;
 			else mem_free(lbl), lbl = NULL;
+			mem_free(vlbl), vlbl = NULL;
 		}
 		if (group) new_menu_item(NULL, -1, 0), group = 0;
 	}
 	if (t_namelen == 8 && !casecmp(t_name, "OPTGROUP", 8)) {
 		char *la;
-		if (!(la = get_attr_val(t_attr, "label"))) if (!(la = stracpy(""))) goto see;
-		new_menu_item(la, -1, 0);
+		if (!(la = get_attr_val(t_attr, "label"))) la = stracpy("");
+		new_menu_item(convert_string(ct, la, strlen(la)), -1, 0);
+		mem_free(la);
 		group = 1;
 	}
 	goto see;
@@ -1612,12 +1673,9 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	end_parse:
 	*end = en;
 	if (!order) goto abort;
-	if (!(fc = mem_alloc(sizeof(struct form_control)))) goto abort;
+	fc = mem_alloc(sizeof(struct form_control));
 	memset(fc, 0, sizeof(struct form_control));
-	if (!(lbls = mem_alloc(order * sizeof(char *)))) {
-		mem_free(fc);
-		goto abort;
-	}
+	lbls = mem_alloc(order * sizeof(char *));
 	memset(lbls, 0, order * sizeof(char *));
 	fc->form_num = last_form_tag - startf;
 	fc->ctrl_num = attr - last_form_tag;
@@ -1639,7 +1697,7 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	format.form = fc;
 	format.attr |= AT_BOLD;
 	mw = 0;
-	for (i = 0; i < order; i++) if (lbls[i] && strlen(lbls[i]) > mw) mw = strlen(lbls[i]);
+	for (i = 0; i < order; i++) if (lbls[i] && strlen(lbls[i]) > (size_t)mw) mw = strlen(lbls[i]);
 	for (i = 0; i < mw; i++) put_chrs("_", 1, put_chars_f, f);
 	kill_html_stack_item(&html_top);
 	put_chrs("]", 1, put_chars_f, f);
@@ -1655,11 +1713,11 @@ void html_textarea(unsigned char *a)
 void do_html_textarea(unsigned char *attr, unsigned char *html, unsigned char *eof, unsigned char **end, void *f)
 {
 	struct form_control *fc;
-	unsigned char *p, *t_name;
+	unsigned char *p, *t_name, *w;
 	int t_namelen;
 	int cols, rows;
 	int i;
-	find_form_for_input(attr);
+	find_form_for_input(attr, 1);
 	while (html < eof && (*html == '\n' || *html == '\r')) html++;
 	p = html;
 	while (p < eof && *p != '<') {
@@ -1672,7 +1730,8 @@ void do_html_textarea(unsigned char *attr, unsigned char *html, unsigned char *e
 	}
 	if (parse_element(p, eof, &t_name, &t_namelen, NULL, end)) goto pp;
 	if (t_namelen != 9 || casecmp(t_name, "/TEXTAREA", 9)) goto pp;
-	if (!(fc = mem_alloc(sizeof(struct form_control)))) return;
+	if (!last_form_tag) return;
+	fc = mem_alloc(sizeof(struct form_control));
 	memset(fc, 0, sizeof(struct form_control));
 	fc->form_num = last_form_tag - startf;
 	fc->ctrl_num = attr - last_form_tag;
@@ -1683,18 +1742,20 @@ void do_html_textarea(unsigned char *attr, unsigned char *html, unsigned char *e
 	fc->type = FC_TEXTAREA;;
 	fc->ro = has_attr(attr, "disabled") ? 2 : has_attr(attr, "readonly") ? 1 : 0;
 	fc->default_value = memacpy(html, p - html);
-	for (p = fc->default_value; p && p[0]; p++) if (p[0] == '\r')
-		if (p[1] == '\n' || (p > fc->default_value && p[-1] == '\n')) memcpy(p, p + 1, strlen(p)), p--;
-		else p[0] = '\n';
-	if ((cols = get_num(attr, "cols")) <= 0) cols = HTML_DEFAULT_INPUT_SIZE;
+	if ((cols = get_num(attr, "cols")) <= 5) cols = HTML_DEFAULT_TEXTAREA_WIDTH;
 	cols++;
-	if ((rows = get_num(attr, "rows")) <= 0) rows = 1;
+	if ((rows = get_num(attr, "rows")) <= 0) rows = HTML_DEFAULT_TEXTAREA_HEIGHT;
 	if (cols > d_opt->xw) cols = d_opt->xw;
 	if (rows > d_opt->yw) rows = d_opt->yw;
 	fc->cols = cols;
 	fc->rows = rows;
-	fc->wrap = has_attr(attr, "wrap");
-	if ((fc->maxlength = get_num(attr, "maxlength")) == -1) fc->maxlength = MAXINT;
+	fc->wrap = 1;
+	if ((w = get_attr_val(attr, "wrap"))) {
+		if (!strcasecmp(w, "hard") || !strcasecmp(w, "physical")) fc->wrap = 2;
+		else if (!strcasecmp(w, "off")) fc->wrap = 0;
+		mem_free(w);
+	}
+	if ((fc->maxlength = get_num(attr, "maxlength")) == -1) fc->maxlength = MAXINT / 4;
 	if (rows > 1) ln_break(1, line_break_f, f);
 	else put_chrs(" ", 1, put_chars_f, f);
 	html_stack_dup();
@@ -1767,12 +1828,15 @@ void parse_frame_widths(unsigned char *a, int ww, int www, int **op, int *olp)
 	o = DUMMY;
 	new_ch:
 	while (WHITECHAR(*a)) a++;
-	n = strtoul(a, (char **)&a, 10);
+	n = strtoul(a, (char **)(void *)&a, 10);
+	if (n > 10000) n = 10000;
 	q = n;
 	if (*a == '%') q = q * ww / 100;
-	else if (*a != '*') q = (q + (www - 1) / 2) / www;
+	else if (*a != '*') q = (q + (www - 1) / 2) / (www ? www : 1);
 	else if (!(q = -q)) q = -1;
-	if ((oo = mem_realloc(o, (ol + 1) * sizeof(int)))) (o = oo)[ol++] = q;
+	if ((unsigned)ol > MAXINT / sizeof(int) - 1) overalloc();
+	o = mem_realloc(o, (ol + 1) * sizeof(int));
+	o[ol++] = q;
 	if ((aa = strchr(a, ','))) {
 		a = aa + 1;
 		goto new_ch;
@@ -1789,10 +1853,10 @@ void parse_frame_widths(unsigned char *a, int ww, int www, int **op, int *olp)
 		for (i = 0; i < ol; i++) d += o[i];
 		qq = q;
 		for (i = 0; i < ol; i++) {
-			q -= o[i] - o[i] * (d - qq) / d;
+			q -= o[i] - o[i] * (d - qq) / (d ? d : 1);
 			do_not_optimize_here(&d);
 				/* SIGH! gcc 2.7.2.* has an optimizer bug! */
-			o[i] = o[i] * (d - qq) / d;
+			o[i] = o[i] * (d - qq) / (d ? d : 1);
 		}
 		while (q) {
 			nn = 0;
@@ -1807,10 +1871,8 @@ void parse_frame_widths(unsigned char *a, int ww, int www, int **op, int *olp)
 		int nn = 0;
 		for (i = 0; i < ol; i++) if (o[i] < 0) nn = 1;
 		if (!nn) goto distribute;
-		if (!(oo = mem_alloc(ol * sizeof(int)))) {
-			*olp = 0;
-			return;
-		}
+		if ((unsigned)ol > MAXINT / sizeof(int)) overalloc();
+		oo = mem_alloc(ol * sizeof(int));
 		memcpy(oo, o, ol * sizeof(int));
 		for (i = 0; i < ol; i++) if (o[i] < 1) o[i] = 1;
 		q = ww - q;
@@ -1818,14 +1880,20 @@ void parse_frame_widths(unsigned char *a, int ww, int www, int **op, int *olp)
 		for (i = 0; i < ol; i++) if (oo[i] < 0) d += -oo[i];
 		qq = q;
 		for (i = 0; i < ol; i++) if (oo[i] < 0) {
-			o[i] += (-oo[i] * qq / d);
-			q -= (-oo[i] * qq / d);
+			o[i] += (-oo[i] * qq / (d ? d : 1));
+			q -= (-oo[i] * qq / (d ? d : 1));
+		}
+		if (q < 0) {
+			q = 0;
+			/*internal("parse_frame_widths: q < 0"); may happen when page contains too big values */
 		}
-		if (q < 0) internal("parse_frame_widths: q < 0");
 		for (i = 0; i < ol; i++) if (oo[i] < 0) {
 			if (q) o[i]++, q--;
 		}
-		if (q > 0) internal("parse_frame_widths: q > 0");
+		if (q > 0) {
+			q = 0;
+			/*internal("parse_frame_widths: q > 0"); may happen when page contains too big values */
+		}
 		mem_free(oo);
 	}
 	for (i = 0; i < ol; i++) if (!o[i]) {
@@ -1843,11 +1911,8 @@ void html_frameset(unsigned char *a)
 	struct frameset_param fp;
 	unsigned char *c, *d;
 	if (!d_opt->frames || !special_f(ff, SP_USED, NULL)) return;
-	if (!(c = get_attr_val(a, "cols"))) if (!(c = stracpy("100%"))) return;
-	if (!(d = get_attr_val(a, "rows"))) if (!(d = stracpy("100%"))) {
-		mem_free(c);
-		return;
-	}
+	if (!(c = get_attr_val(a, "cols"))) c = stracpy("100%");
+	if (!(d = get_attr_val(a, "rows"))) d = stracpy("100%");
 	if (!html_top.frameset) {
 		x = d_opt->xw;
 		y = d_opt->yw;
@@ -1879,7 +1944,7 @@ void html_frameset(unsigned char *a)
 		horiz = 1;
 		if (!(c = get_attr_val(a, "rows"))) return;
 	}
-	if (!(fp = mem_alloc(sizeof(struct frameset_param)))) goto f;
+	fp = mem_alloc(sizeof(struct frameset_param));
 	fp->n = 0;
 	fp->horiz = horiz;
 	par_format.leftmargin = par_format.rightmargin = 0;
@@ -1892,11 +1957,9 @@ void html_frameset(unsigned char *a)
 			continue;
 		}
 		if ((w = parse_width(d, 1)) != -1) {
-			struct frameset_param *fpp;
-			if ((fpp = mem_realloc(fp, sizeof(struct frameset_param) + (fp->n + 1) * sizeof(int)))) {
-				fp = fpp;
-				fp->width[fp->n++] = w;
-			}
+			if ((unsigned)fp->n > (MAXINT - sizeof(struct frameset_param)) / sizeof(int) - 1) overalloc();
+			fp = mem_realloc(fp, sizeof(struct frameset_param) + (fp->n + 1) * sizeof(int));
+			fp->width[fp->n++] = w;
 		}
 		if (!(d = strchr(d, ','))) break;
 		d++;
@@ -1910,17 +1973,36 @@ void html_frameset(unsigned char *a)
 
 void html_link(unsigned char *a)
 {
-	unsigned char *name, *url;
+	unsigned char *name, *url, *title;
 	if ((name = get_attr_val(a, "type"))) {
-		if (casecmp(a, "text/css", 8)) {
+		if (strcasecmp(name, "text/html")) {
 			mem_free(name);
 			return;
 		}
 		mem_free(name);
 	}
 	if (!(url = get_url_val(a, "href"))) return;
-	if (!(name = get_attr_val(a, "rel"))) if (!(name = get_attr_val(a, "rev"))) name = stracpy(url);
-	if (strcasecmp(name, "STYLESHEET") && strcasecmp(name, "made") && strcasecmp(name, "SHORTCUT ICON")) put_link_line("Link: ", name, url, format.target_base);
+	if (!(name = get_attr_val(a, "rel")))
+		if (!(name = get_attr_val(a, "rev")))
+			if (!(name = get_attr_val(a, "ref")))
+				name = stracpy(url);
+	if (!strcasecmp(name, "stylesheet") ||
+	    !strcasecmp(name, "alternate stylesheet") ||
+	    !strcasecmp(name, "made") ||
+	    !strcasecmp(name, "icon") ||
+	    !strcasecmp(name, "shortcut icon") ||
+	    !strcasecmp(name, "apple-touch-icon") ||
+	    !strcasecmp(name, "meta") ||
+	    !strcasecmp(name, "pingback") ||
+	    !strcasecmp(name, "File-List") ||
+	    !casecmp(name, "schema", 6)) goto skip;
+	if ((title = get_attr_val(a, "title"))) {
+		add_to_strn(&name, ": ");
+		add_to_strn(&name, title);
+		mem_free(title);
+	}
+	put_link_line("Link: ", name, url, format.target_base);
+	skip:
 	mem_free(name);
 	mem_free(url);
 }
@@ -1939,6 +2021,7 @@ struct element_info elements[] = {
 	{"DFN",		html_bold,	0, 0},
 	{"I",		html_italic,	0, 0},
 	{"Q",		html_italic,	0, 0},
+	{"CITE",	html_italic,	0, 0},
 	{"EM",		html_italic,	0, 0},
 	{"ABBR",	html_italic,	0, 0},
 	{"U",		html_underline,	0, 0},
@@ -2030,7 +2113,7 @@ void process_head(unsigned char *head)
 {
 	unsigned char *r, *p;
 	if ((r = parse_http_header(head, "Refresh", NULL))) {
-		if ((p = parse_header_param(r, "URL"))) {
+		if ((p = parse_header_param(r, "URL")) || (p = parse_header_param(r, ""))) {
 			put_link_line("Refresh: ", p, p, d_opt->framename);
 			mem_free(p);
 		}
@@ -2038,7 +2121,29 @@ void process_head(unsigned char *head)
 	}
 }
 
-void parse_html(unsigned char *html, unsigned char *eof, void (*put_chars)(void *, unsigned char *, int), void (*line_break)(void *), void (*init)(void *), void *(*special)(void *, int, ...), void *f, unsigned char *head)
+int qd(unsigned char *html, unsigned char *eof, int *len)
+{
+	int l;
+	*len = 1;
+	if (html >= eof) {
+		internal("qd: out of data, html == %p, eof == %p", html, eof);
+		return -1;
+	}
+	if (html[0] != '&' || d_opt->plain) return html[0];
+	if (html + 1 >= eof) return -1;
+	if (html[1] != '#') return -1;
+	for (l = 2; l < 10 && html + l < eof; l++) if (html[l] == ';') {
+		int n = get_entity_number(html + 2, l - 2);
+		if (n >= 0) {
+			*len = l + 1;
+			return n;
+		}
+		break;
+	}
+	return -1;
+}
+
+void parse_html(unsigned char *html, unsigned char *eof, int (*put_chars)(void *, unsigned char *, int), void (*line_break)(void *), void *(*special)(void *, int, ...), void *f, unsigned char *head)
 {
 	/*unsigned char *start = html;*/
 	unsigned char *lt;
@@ -2048,7 +2153,6 @@ void parse_html(unsigned char *html, unsigned char *eof, void (*put_chars)(void
 	was_br = 0;
 	put_chars_f = put_chars;
 	line_break_f = line_break;
-	init_f = init;
 	special_f = special;
 	ff = f;
 	eoff = eof;
@@ -2056,7 +2160,6 @@ void parse_html(unsigned char *html, unsigned char *eof, void (*put_chars)(void
 	set_lt:
 	put_chars_f = put_chars;
 	line_break_f = line_break;
-	init_f = init;
 	special_f = special;
 	ff = f;
 	eoff = eof;
@@ -2078,7 +2181,7 @@ void parse_html(unsigned char *html, unsigned char *eof, void (*put_chars)(void
 				if (!parse_element(h, eof, &name, &namelen, &attr, &end)) {
 					put_chrs(lt, html - lt, put_chars, f);
 					lt = html = h;
-					putsp = 1;
+					if (!html_top.invisible) putsp = 1;
 					goto element;
 				}
 			}
@@ -2099,20 +2202,24 @@ void parse_html(unsigned char *html, unsigned char *eof, void (*put_chars)(void
 			put_chrs(" ", 1, put_chars, f);
 			/*putsp = -1;*/
 		}
-		if (par_format.align == AL_NO) {
+		if (par_format.align == AL_NO && (*html < 32 || *html == '&')) {
+			int l;
+			int q = qd(html, eof, &l);
 			putsp = 0;
-			if (*html == 9) {
+			if (q == 9) {
 				put_chrs(lt, html - lt, put_chars, f);
-				put_chrs("        ", 8 - pos%8, put_chars, f);
-				html++;
+				put_chrs("        ", 8 - pos % 8, put_chars, f);
+				html += l;
 				goto set_lt;
-			} else if (*html == 13 || *html == 10) {
+			} else if (q == 13 || q == 10) {
 				put_chrs(lt, html - lt, put_chars, f);
 				next_break:
-				if (*html == 13 && html < eof-1 && html[1] == 10) html++;
+				html += l;
+				if (q == 13 && html < eof - 1 && qd(html, eof, &l) == 10) html += l;
 				ln_break(1, line_break, f);
-				html++;
-				if (*html == 13 || *html == 10) {
+				if (html >= eof) goto set_lt;
+				q = qd(html, eof, &l);
+				if (q == 13 || q == 10) {
 					line_breax = 0;
 					goto next_break;
 				}
@@ -2158,7 +2265,7 @@ void parse_html(unsigned char *html, unsigned char *eof, void (*put_chars)(void
 		html = end;
 		for (ei = elements; ei->name; ei++) {
 			if (ei->name &&
-			   (strlen(ei->name) != namelen || casecmp(ei->name, name, namelen)))
+			   (strlen(ei->name) != (size_t)namelen || casecmp(ei->name, name, namelen)))
 				continue;
 			if (!inv) {
 				char *a;
@@ -2220,9 +2327,10 @@ void parse_html(unsigned char *html, unsigned char *eof, void (*put_chars)(void
 				/*debug_stack();*/
 				foreach(e, html_stack) {
 					if (e->linebreak && !ei->linebreak && ei->name) xxx = 1;
-					if (e->namelen != namelen || casecmp(e->name, name, e->namelen))
+					if (e->namelen != namelen || casecmp(e->name, name, e->namelen)) {
 						if (e->dontkill) break;
 						else continue;
+					}
 					if (xxx) {
 						kill_html_stack_item(e);
 						break;
@@ -2263,7 +2371,7 @@ int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, uns
 	scan_http_equiv(s, eof, &hd, &hdl, NULL);
 	ct = get_convert_table(hd, to, def, NULL, NULL, hdef);
 	mem_free(hd);
-	if (!(*menu = mem_alloc(sizeof(struct menu_item)))) return -1;
+	*menu = mem_alloc(sizeof(struct menu_item));
 	memset(*menu, 0, sizeof(struct menu_item));
 	/*(*menu)->text = NULL;*/
 	se:
@@ -2346,12 +2454,7 @@ int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, uns
 	}
 	if (!(target = get_target(attr)) && !(target = stracpy(target_base)))
 		target = stracpy("");
-	if (!(ld = mem_alloc(sizeof(struct link_def)))) {
-		if (label) mem_free(label);
-		mem_free(href);
-		mem_free(target);
-		goto se2;
-	}
+	ld = mem_alloc(sizeof(struct link_def));
 	if (!(ld->link = join_urls(href_base, href))) {
 		mem_free(href);
 		mem_free(target);
@@ -2372,21 +2475,17 @@ int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, uns
 	}
 	if (label) clr_spaces(label);
 	if (label && !*label) mem_free(label), label = NULL;
-	if (!label) if (!(label = stracpy(ld->link))) {
-		mem_free(href);
-		mem_free(target);
-		goto se2;
-	}
-	if ((nm = mem_realloc(*menu, (nmenu + 2) * sizeof(struct menu_item)))) {
-		*menu = nm;
-		memset(&nm[nmenu], 0, 2 * sizeof(struct menu_item));
-		nm[nmenu].text = label;
-		nm[nmenu].rtext = "";
-		nm[nmenu].hotkey = "";
-		nm[nmenu].func = MENU_FUNC map_selected;
-		nm[nmenu].data = ld;
-		nm[++nmenu].text = NULL;
-	}
+	if (!label) label = stracpy(ld->link);
+	if ((unsigned)nmenu > MAXINT / sizeof(struct menu_item) - 2) overalloc();
+	nm = mem_realloc(*menu, (nmenu + 2) * sizeof(struct menu_item));
+	*menu = nm;
+	memset(&nm[nmenu], 0, 2 * sizeof(struct menu_item));
+	nm[nmenu].text = label;
+	nm[nmenu].rtext = "";
+	nm[nmenu].hotkey = "";
+	nm[nmenu].func = MENU_FUNC map_selected;
+	nm[nmenu].data = ld;
+	nm[++nmenu].text = NULL;
 	add_to_ml(ml, ld, ld->link, ld->target, label, NULL);
 	goto se2;
 	done:
@@ -2422,7 +2521,10 @@ void scan_http_equiv(unsigned char *s, unsigned char *eof, unsigned char **head,
 			s = skip_comment(s, eof);
 			goto xse;
 		}
-		if (parse_element(s, eof, &name, &namelen, &attr, &s)) goto xsp;
+		if (parse_element(s, eof, &name, &namelen, &attr, &s)) {
+			s1 = s;
+			goto xsp;
+		}
 		clr_spaces(*title);
 		goto ps;
 	}
