@@ -1,5 +1,7 @@
 #include "links.h"
 
+#define format format_
+
 static inline int color_distance(struct rgb *c1, struct rgb *c2)
 {
 	return
@@ -9,55 +11,55 @@ static inline int color_distance(struct rgb *c1, struct rgb *c2)
 }
 
 struct rgb palette[] = {
-/*	{0x00, 0x00, 0x00},
-	{0x80, 0x00, 0x00},
-	{0x00, 0x80, 0x00},
-	{0x80, 0x80, 0x00},
-	{0x00, 0x00, 0x80},
-	{0x80, 0x00, 0x80},
-	{0x00, 0x80, 0x80},
-	{0xC0, 0xC0, 0xC0},
-	{0x80, 0x80, 0x80},
-	{0xff, 0x00, 0x00},
-	{0x00, 0xff, 0x00},
-	{0xff, 0xff, 0x00},
-	{0x00, 0x00, 0xff},
-	{0xff, 0x00, 0xff},
-	{0x00, 0xff, 0xff},
-	{0xff, 0xff, 0xff},*/
-	/*{0x00, 0x00, 0x00},
-	{0xaa, 0x00, 0x00},
-	{0x00, 0xaa, 0x00},
-	{0xaa, 0x55, 0x00},
-	{0x00, 0x00, 0xaa},
-	{0xaa, 0x00, 0xaa},
-	{0x00, 0xaa, 0xaa},
-	{0xaa, 0xaa, 0xaa},
-	{0x55, 0x55, 0x55},
-	{0xff, 0x55, 0x55},
-	{0x55, 0xff, 0x55},
-	{0xff, 0xff, 0x55},
-	{0x55, 0x55, 0xff},
-	{0xff, 0x55, 0xff},
-	{0x55, 0xff, 0xff},
-	{0xff, 0xff, 0xff},*/
-	{0x00, 0x00, 0x00},
-	{0x80, 0x00, 0x00},
-	{0x00, 0x80, 0x00},
-	{0xaa, 0x55, 0x00},
-	{0x00, 0x00, 0x80},
-	{0x80, 0x00, 0x80},
-	{0x00, 0x80, 0x80},
-	{0xaa, 0xaa, 0xaa},
-	{0x55, 0x55, 0x55},
-	{0xff, 0x55, 0x55},
-	{0x55, 0xff, 0x55},
-	{0xff, 0xff, 0x55},
-	{0x55, 0x55, 0xff},
-	{0xff, 0x55, 0xff},
-	{0x55, 0xff, 0xff},
-	{0xff, 0xff, 0xff},
-	{-1, -1, -1}
+/*	{0x00, 0x00, 0x00, 0},
+	{0x80, 0x00, 0x00, 0},
+	{0x00, 0x80, 0x00, 0},
+	{0x80, 0x80, 0x00, 0},
+	{0x00, 0x00, 0x80, 0},
+	{0x80, 0x00, 0x80, 0},
+	{0x00, 0x80, 0x80, 0},
+	{0xC0, 0xC0, 0xC0, 0},
+	{0x80, 0x80, 0x80, 0},
+	{0xff, 0x00, 0x00, 0},
+	{0x00, 0xff, 0x00, 0},
+	{0xff, 0xff, 0x00, 0},
+	{0x00, 0x00, 0xff, 0},
+	{0xff, 0x00, 0xff, 0},
+	{0x00, 0xff, 0xff, 0},
+	{0xff, 0xff, 0xff, 0},*/
+	/*{0x00, 0x00, 0x00, 0},
+	{0xaa, 0x00, 0x00, 0},
+	{0x00, 0xaa, 0x00, 0},
+	{0xaa, 0x55, 0x00, 0},
+	{0x00, 0x00, 0xaa, 0},
+	{0xaa, 0x00, 0xaa, 0},
+	{0x00, 0xaa, 0xaa, 0},
+	{0xaa, 0xaa, 0xaa, 0},
+	{0x55, 0x55, 0x55, 0},
+	{0xff, 0x55, 0x55, 0},
+	{0x55, 0xff, 0x55, 0},
+	{0xff, 0xff, 0x55, 0},
+	{0x55, 0x55, 0xff, 0},
+	{0xff, 0x55, 0xff, 0},
+	{0x55, 0xff, 0xff, 0},
+	{0xff, 0xff, 0xff, 0},*/
+	{0x00, 0x00, 0x00, 0},
+	{0x80, 0x00, 0x00, 0},
+	{0x00, 0x80, 0x00, 0},
+	{0xaa, 0x55, 0x00, 0},
+	{0x00, 0x00, 0x80, 0},
+	{0x80, 0x00, 0x80, 0},
+	{0x00, 0x80, 0x80, 0},
+	{0xaa, 0xaa, 0xaa, 0},
+	{0x55, 0x55, 0x55, 0},
+	{0xff, 0x55, 0x55, 0},
+	{0x55, 0xff, 0x55, 0},
+	{0xff, 0xff, 0x55, 0},
+	{0x55, 0x55, 0xff, 0},
+	{0xff, 0x55, 0xff, 0},
+	{0x55, 0xff, 0xff, 0},
+	{0xff, 0xff, 0xff, 0},
+	{-1, -1, -1, 0},
 };
 
 /*struct rgb rgbcache = {0, 0, 0};
@@ -123,26 +125,21 @@ static inline int fg_color(int fg, int bg)
 	return fg;
 }
 
-#ifdef ALIGN
-#undef ALIGN
-#endif
-
-#define ALIGN(x) (((x)+0x7f)&~0x7f)
+#define XALIGN(x) (((x)+0x7f)&~0x7f)
 
 int nowrap = 0;
 
-static inline int xpand_lines(struct part *p, int y)
+static inline void xpand_lines(struct part *p, int y)
 {
 	/*if (y >= p->y) p->y = y + 1;*/
-	if (!p->data) return 0;
+	if (!p->data) return;
+	if (XALIGN((unsigned)y + (unsigned)p->yp) > MAXINT) overalloc();
 	y += p->yp;
 	if (y >= p->data->y) {			/* !!! FIXME: out of inline */
 		int i;
-		if (ALIGN(y + 1) >= ALIGN(p->data->y)) {
-			struct line *l;
-			if (!(l = mem_realloc(p->data->data, ALIGN(y+1)*sizeof(struct line))))
-				return -1;
-			p->data->data = l;
+		if (XALIGN(y + 1) > XALIGN(p->data->y)) {
+			if (XALIGN((unsigned)y + 1) > MAXINT / sizeof(struct line)) overalloc();
+			p->data->data = mem_realloc(p->data->data, XALIGN(y+1)*sizeof(struct line));
 		}
 		for (i = p->data->y; i <= y; i++) {
 			p->data->data[i].l = 0;
@@ -151,50 +148,46 @@ static inline int xpand_lines(struct part *p, int y)
 		}
 		p->data->y = i;
 	}
-	return 0;
 }
 
-static inline int xpand_line(struct part *p, int y, int x)
+static inline void xpand_line(struct part *p, int y, int x)
 {
-	if (!p->data) return 0; /* !!! FIXME: p->x (?) */
+	if (!p->data) return; /* !!! FIXME: p->x (?) */
+	if (XALIGN((unsigned)x + (unsigned)p->xp) > MAXINT) overalloc();
 	x += p->xp;
 	y += p->yp;
 #ifdef DEBUG
 	if (y >= p->data->y) {
 		internal("line does not exist");
-		return -1;
+		return;
 	}
 #endif
 	if (x >= p->data->data[y].l) {		/* !!! FIXME: out of inline */
 		int i;
-		if (ALIGN(x+1) >= ALIGN(p->data->data[y].l)) {
-			chr *l;
-			if (!(l = mem_realloc(p->data->data[y].d, ALIGN(x+1)*sizeof(chr))))
-				return -1;
-			p->data->data[y].d = l;
+		if (XALIGN(x+1) > XALIGN(p->data->data[y].l)) {
+			if (XALIGN((unsigned)x + 1) > MAXINT / sizeof(chr)) overalloc();
+			p->data->data[y].d = mem_realloc(p->data->data[y].d, XALIGN(x+1)*sizeof(chr));
 		}
 		for (i = p->data->data[y].l; i <= x; i++)
 			p->data->data[y].d[i] = (p->data->data[y].c << 11) | ' ';
 		p->data->data[y].c = p->bgcolor;
 		p->data->data[y].l = i;
 	}
-	return 0;
 }
 
-int r_xpand_spaces(struct part *p, int l)
+void r_xpand_spaces(struct part *p, int l)
 {
 	unsigned char *c;
-	if (!(c = mem_realloc(p->spaces, l + 1))) return -1;
+	if ((unsigned)l >= MAXINT) overalloc();
+	c = mem_realloc(p->spaces, l + 1);
 	memset(c + p->spl, 0, l - p->spl + 1);
 	p->spl = l + 1;
 	p->spaces = c;
-	return 0;
 }
 
-static inline int xpand_spaces(struct part *p, int l)
+static inline void xpand_spaces(struct part *p, int l)
 {
-	if (l >= p->spl) return r_xpand_spaces(p, l);
-	return 0;
+	if ((unsigned)l >= (unsigned)p->spl) r_xpand_spaces(p, l);
 }
 
 #define POS(x, y) (p->data->data[p->yp + (y)].d[p->xp + (x)])
@@ -205,15 +198,15 @@ static inline int xpand_spaces(struct part *p, int l)
 
 static inline void set_hchar(struct part *p, int x, int y, unsigned c)
 {
-	if (xpand_lines(p, y)) return;
-	if (xpand_line(p, y, x)) return;
+	xpand_lines(p, y);
+	xpand_line(p, y, x);
 	POS(x, y) = c;
 }
 
 static inline void set_hchars(struct part *p, int x, int y, int xl, unsigned c)
 {
-	if (xpand_lines(p, y)) return;
-	if (xpand_line(p, y, x+xl-1)) return;
+	xpand_lines(p, y);
+	xpand_line(p, y, x+xl-1);
 	for (; xl; xl--, x++) POS(x, y) = c;
 }
 
@@ -227,21 +220,21 @@ void xset_hchars(struct part *p, int x, int y, int xl, unsigned c)
 	set_hchars(p, x, y, xl, c);
 }
 
-int xxpand_lines(struct part *p, int y)
+void xxpand_lines(struct part *p, int y)
 {
-	return xpand_lines(p, y);
+	xpand_lines(p, y);
 }
 
-int xxpand_line(struct part *p, int y, int x)
+void xxpand_line(struct part *p, int y, int x)
 {
-	return xpand_line(p, y, x);
+	xpand_line(p, y, x);
 }
 
 static inline void set_hline(struct part *p, int x, int y, int xl, unsigned char *d, unsigned c, int spc)
 {
-	if (xpand_lines(p, y)) return;
-	if (xpand_line(p, y, x+xl-1)) return;
-	if (spc && xpand_spaces(p, x+xl-1)) return;
+	xpand_lines(p, y);
+	xpand_line(p, y, x+xl-1);
+	if (spc) xpand_spaces(p, x+xl-1);
 	for (; xl; xl--, x++, d++) {
 		if (spc) p->spaces[x] = *d == ' ';
 		if (p->data) POS(x, y) = *d | c;
@@ -263,11 +256,13 @@ static inline void move_links(struct part *p, int xf, int yf, int xt, int yt)
 		int i;
 		struct link *link = &p->data->links[n];
 			/*printf("ml: %d %d %d %d",link->pos[0].x,link->pos[0].y,X(xf),Y(yf));fflush(stdout);sleep(1);*/
-		for (i = 0; i < link->n; i++) if (link->pos[i].y == Y(yf)) {
+		/*for (i = 0; i < link->n; i++) fprintf(stderr, "%d.%d -> %d.%d: %d.%d : %d %d\n", X(xf), Y(yf), X(xt), yt != -1 ? Y(yt) : -1, n, i, link->pos[i].x, link->pos[i].y);*/
+		for (i = 0; i < link->n; i++) if (link->pos[i].y >= Y(yf)) {
 			w = 1;
-			if (link->pos[i].x >= X(xf))
+			if (link->pos[i].y == Y(yf) && link->pos[i].x >= X(xf)) {
 				if (yt >= 0) link->pos[i].y = Y(yt), link->pos[i].x += -xf + xt;
 				else memmove(&link->pos[i], &link->pos[i+1], (link->n-i-1) * sizeof(struct point)), link->n--, i--;
+			}
 		}
 		/*if (!link->n) {
 			if (link->where) mem_free(link->where);
@@ -295,8 +290,8 @@ static inline void move_links(struct part *p, int xf, int yf, int xt, int yt)
 static inline void copy_chars(struct part *p, int x, int y, int xl, chr *d)
 {
 	if (xl <= 0) return;
-	if (xpand_lines(p, y)) return;
-	if (xpand_line(p, y, x+xl-1)) return;
+	xpand_lines(p, y);
+	xpand_line(p, y, x+xl-1);
 	for (; xl; xl--, x++, d++) POS(x, y) = *d;
 }
 
@@ -312,7 +307,8 @@ static inline void shift_chars(struct part *p, int y, int s)
 {
 	chr *a;
 	int l = LEN(y);
-	if (!(a = mem_alloc(l * sizeof(chr)))) return;
+	if ((unsigned)l > MAXINT / sizeof(chr)) overalloc();
+	a = mem_alloc(l * sizeof(chr));
 	memcpy(a, &POS(0, y), l * sizeof(chr));
 	set_hchars(p, 0, y, s, (p->data->data[y].c << 11) | ' ');
 	copy_chars(p, s, y, l, a);
@@ -384,9 +380,8 @@ struct link *new_link(struct f_data *f)
 {
 	if (!f) return NULL;
 	if (!(f->nlinks & (ALLOC_GR - 1))) {
-		struct link *l;
-		if (!(l = mem_realloc(f->links, (f->nlinks + ALLOC_GR) * sizeof(struct link)))) return NULL;
-		f->links = l;
+		if ((unsigned)f->nlinks > MAXINT / sizeof(struct link) - ALLOC_GR) overalloc();
+		f->links = mem_realloc(f->links, (f->nlinks + ALLOC_GR) * sizeof(struct link));
 	}
 	memset(&f->links[f->nlinks], 0, sizeof(struct link));
 	return &f->links[f->nlinks++];
@@ -395,14 +390,19 @@ struct link *new_link(struct f_data *f)
 void html_tag(struct f_data *f, unsigned char *t, int x, int y)
 {
 	struct tag *tag;
+	unsigned char *tt;
+	int ll;
 	if (!f) return;
-	if ((tag = mem_alloc(sizeof(struct tag) + strlen(t) + 1))) {
-		tag->x = x;
-		tag->y = y;
-		strcpy(tag->name, t);
-		add_to_list(f->tags, tag);
-		if ((void *)last_tag_for_newline == &f->tags) last_tag_for_newline = tag;
-	}
+	tt = init_str();
+	ll = 0;
+	add_conv_str(&tt, &ll, t, strlen(t), -2);
+	tag = mem_alloc(sizeof(struct tag) + strlen(tt) + 1);
+	tag->x = x;
+	tag->y = y;
+	strcpy(tag->name, tt);
+	add_to_list(f->tags, tag);
+	if ((void *)last_tag_for_newline == &f->tags) last_tag_for_newline = tag;
+	mem_free(tt);
 }
 
 unsigned char *last_link;
@@ -418,20 +418,21 @@ void put_chars(struct part *, unsigned char *, int);
 
 #define CH_BUF	256
 
-void put_chars_conv(struct part *p, unsigned char *c, int l)
+int put_chars_conv(struct part *p, unsigned char *c, int l)
 {
 	static char buffer[CH_BUF];
 	int bp = 0;
 	int pp = 0;
+	int total = 0;
 	if (format.attr & AT_GRAPHICS) {
 		put_chars(p, c, l);
-		return;
+		return l;
 	}
 	if (!l) put_chars(p, NULL, 0);
 	while (pp < l) {
 		unsigned char *e;
 		if (c[pp] < 128 && c[pp] != '&') {
-			putc:
+			put_c:
 			buffer[bp++] = c[pp++];
 			if (bp < CH_BUF) continue;
 			goto flush;
@@ -439,7 +440,7 @@ void put_chars_conv(struct part *p, unsigned char *c, int l)
 		if (c[pp] != '&') {
 			struct conv_table *t;
 			int i;
-			if (!convert_table) goto putc;
+			if (!convert_table) goto put_c;
 			t = convert_table;
 			i = pp;
 			decode:
@@ -447,15 +448,15 @@ void put_chars_conv(struct part *p, unsigned char *c, int l)
 				e = t[c[i]].u.str;
 			} else {
 				t = t[c[i++]].u.tbl;
-				if (i >= l) goto putc;
+				if (i >= l) goto put_c;
 				goto decode;
 			}
 			pp = i + 1;
 		} else {
 			int i = pp + 1;
-			if (d_opt->plain) goto putc;
+			if (d_opt->plain) goto put_c;
 			while (i < l && c[i] != ';' && c[i] != '&' && c[i] > ' ') i++;
-			if (!(e = get_entity_string(&c[pp + 1], i - pp - 1, d_opt->cp))) goto putc;
+			if (!(e = get_entity_string(&c[pp + 1], i - pp - 1, d_opt->cp))) goto put_c;
 			pp = i + (i < l && c[i] == ';');
 		}
 		if (!e[0]) continue;
@@ -471,15 +472,18 @@ void put_chars_conv(struct part *p, unsigned char *c, int l)
 			if (bp < CH_BUF) continue;
 			flush1:
 			put_chars(p, buffer, bp);
+			total += bp;
 			bp = 0;
 		}
 	}
 	if (bp) put_chars(p, buffer, bp);
+	total += bp;
+	return total;
 }
 
 void put_chars(struct part *p, unsigned char *c, int l)
 {
-	static struct text_attrib_beginning ta_cache = { -1, { 0, 0, 0 }, { 0, 0, 0 } };
+	static struct text_attrib_beginning ta_cache = { -1, { 0, 0, 0, 0 }, { 0, 0, 0, 0 } };
 	static int bg_cache;
 	static int fg_cache;
 
@@ -487,13 +491,14 @@ void put_chars(struct part *p, unsigned char *c, int l)
 	int i;
 	struct link *link;
 	struct point *pt;
+	if (l < 0) overalloc();
 	/*printf("%d-", p->cx);for (i=0; i<l; i++) printf("%c", c[i]); printf("-\n");sleep(1);*/
 	while (par_format.align != AL_NO && p->cx == -1 && l && *c == ' ') c++, l--;
 	if (!l) return;
 	if (c[0] != ' ' || (c[1] && c[1] != ' ')) {
 		last_tag_for_newline = (void *)&p->data->tags;
 	}
-	if (p->cx == -1) p->cx = par_format.leftmargin;
+	if (p->cx < par_format.leftmargin) p->cx = par_format.leftmargin;
 	if (last_link || last_image || last_form || format.link || format.image || format.form) goto process_link;
 	no_l:
 	/*printf("%d %d\n",p->cx, p->cy);*/
@@ -585,12 +590,12 @@ void put_chars(struct part *p, unsigned char *c, int l)
 		link->sel_color = ((fg & 8) << 3) | (fg & 7) | (bg << 3);
 		link->n = 0;
 		set_link:
-		if ((pt = mem_realloc(link->pos, (link->n + l) * sizeof(struct point)))) {
-			link->pos = pt;
-			for (i = 0; i < l; i++) pt[link->n + i].x = X(p->cx) + i,
-						pt[link->n + i].y = Y(p->cy);
-			link->n += i;
-		}
+		if ((unsigned)link->n + (unsigned)l > MAXINT / sizeof(struct point)) overalloc();
+		pt = mem_realloc(link->pos, (link->n + l) * sizeof(struct point));
+		link->pos = pt;
+		for (i = 0; i < l; i++) pt[link->n + i].x = X(p->cx) + i,
+					pt[link->n + i].y = Y(p->cy);
+		link->n += l;
 	}
 	goto no_l;
 
@@ -635,11 +640,6 @@ void line_break(struct part *p)
 	memset(p->spaces, 0, p->spl);
 }
 
-void html_init(struct part *p)
-{
-	/* !!! FIXME: background */
-}
-
 int g_ctrl_num;
 
 void html_form_control(struct part *p, struct form_control *fc)
@@ -664,6 +664,13 @@ void html_form_control(struct part *p, struct form_control *fc)
 		}
 		*/
 	}
+	if (fc->type == FC_TEXTAREA) {
+		unsigned char *p;
+		for (p = fc->default_value; p[0]; p++) if (p[0] == '\r') {
+			if (p[1] == '\n') memmove(p, p + 1, strlen(p)), p--;
+			else p[0] = '\n';
+		}
+	}
 	add_to_list(p->data->forms, fc);
 }
 
@@ -684,7 +691,9 @@ struct frameset_desc *create_frameset(struct f_data *fda, struct frameset_param
 		internal("zero size of frameset");
 		return NULL;
 	}
-	if (!(fd = mem_alloc(sizeof(struct frameset_desc) + fp->x * fp->y * sizeof(struct frame_desc)))) return NULL;
+	if (fp->x && (unsigned)fp->x * (unsigned)fp->y / (unsigned)fp->x != (unsigned)fp->y) overalloc();
+	if ((unsigned)fp->x * (unsigned)fp->y > (MAXINT - sizeof(struct frameset_desc)) / sizeof(struct frame_desc)) overalloc();
+	fd = mem_alloc(sizeof(struct frameset_desc) + fp->x * fp->y * sizeof(struct frame_desc));
 	memset(fd, 0, sizeof(struct frameset_desc) + fp->x * fp->y * sizeof(struct frame_desc));
 	fd->n = fp->x * fp->y;
 	fd->x = fp->x;
@@ -751,7 +760,7 @@ void *html_special(struct part *p, int c, ...)
 
 void do_format(char *start, char *end, struct part *part, unsigned char *head)
 {
-	parse_html(start, end, (void (*)(void *, unsigned char *, int)) put_chars_conv, (void (*)(void *)) line_break, (void (*)(void *)) html_init, (void *(*)(void *, int, ...)) html_special, part, head);
+	parse_html(start, end, (int (*)(void *, unsigned char *, int)) put_chars_conv, (void (*)(void *)) line_break, (void *(*)(void *, int, ...)) html_special, part, head);
 	/*if ((part->y -= line_breax) < 0) part->y = 0;*/
 }
 
@@ -760,6 +769,7 @@ int margin;
 struct table_cache_entry {
 	struct table_cache_entry *next;
 	struct table_cache_entry *prev;
+	struct table_cache_entry *hash_next;
 	unsigned char *start;
 	unsigned char *end;
 	int align;
@@ -772,8 +782,17 @@ struct table_cache_entry {
 
 struct list_head table_cache = { &table_cache, &table_cache };
 
+#define TC_HASH_SIZE	4096
+
+struct table_cache_entry *table_cache_hash[TC_HASH_SIZE];
+
 void free_table_cache()
 {
+	struct table_cache_entry *tce;
+	foreach(tce, table_cache) {
+		int hash = ((int)(unsigned long)tce->start + tce->xs) & (TC_HASH_SIZE - 1);
+		table_cache_hash[hash] = NULL;
+	}
 	free_list(table_cache);
 }
 
@@ -788,12 +807,15 @@ struct part *format_html_part(unsigned char *start, unsigned char *end, int alig
 	int ef = empty_format;
 	struct form_control *fc;
 	struct table_cache_entry *tce;
-	if (!data) foreach(tce, table_cache) {
-		if (tce->start == start && tce->end == end && tce->align == align && tce->m == m && tce->width == width && tce->xs == xs && tce->link_num == link_num) {
-			if ((p = mem_alloc(sizeof(struct part)))) {
+	if (!data) {
+		tce = table_cache_hash[((int)(unsigned long)start + xs) & (TC_HASH_SIZE - 1)];
+		while (tce) {
+			if (tce->start == start && tce->end == end && tce->align == align && tce->m == m && tce->width == width && tce->xs == xs && tce->link_num == link_num) {
+				p = mem_alloc(sizeof(struct part));
 				memcpy(p, &tce->p, sizeof(struct part));
 				return p;
 			}
+			tce = tce->hash_next;
 		}
 	}
 	if (ys < 0) {
@@ -802,12 +824,11 @@ struct part *format_html_part(unsigned char *start, unsigned char *end, int alig
 	}
 	if (data) {
 		struct node *n;
-		if ((n = mem_alloc(sizeof(struct node)))) {
-			n->x = xs;
-			n->y = ys;
-			n->xw = !table_level ? MAXINT : width;
-			add_to_list(data->nodes, n);
-		}
+		n = mem_alloc(sizeof(struct node));
+		n->x = xs;
+		n->y = ys;
+		n->xw = !table_level ? MAXINT : width;
+		add_to_list(data->nodes, n);
 		/*sdbg(data);*/
 	}
 	last_link_to_move = data ? data->nlinks : 0;
@@ -821,7 +842,7 @@ struct part *format_html_part(unsigned char *start, unsigned char *end, int alig
 	last_link = last_image = last_target = NULL;
 	last_form = NULL;
 	nobreak = 1;
-	if (!(p = mem_alloc(sizeof(struct part)))) goto ret;
+	p = mem_alloc(sizeof(struct part));
 	p->x = p->y = 0;
 	p->data = data;
 	p->xp = xs; p->yp = ys;
@@ -867,7 +888,6 @@ struct part *format_html_part(unsigned char *start, unsigned char *end, int alig
 	}
 	foreach(fc, p->uf) destroy_fc(fc);
 	free_list(p->uf);
-	ret:
 	last_link_to_move = llm;
 	last_tag_to_move = ltm;
 	/*last_tag_for_newline = ltn;*/
@@ -875,7 +895,9 @@ struct part *format_html_part(unsigned char *start, unsigned char *end, int alig
 	empty_format = ef;
 	last_link = last_image = last_target = NULL;
 	last_form = NULL;
-	if (table_level > 1 && !data && ((tce = mem_alloc(sizeof(struct table_cache_entry))))) {
+	if (table_level > 1 && !data) {
+		int hash;
+		tce = mem_alloc(sizeof(struct table_cache_entry));
 		tce->start = start;
 		tce->end = end;
 		tce->align = align;
@@ -885,6 +907,9 @@ struct part *format_html_part(unsigned char *start, unsigned char *end, int alig
 		tce->link_num = link_num;
 		memcpy(&tce->p, p, sizeof(struct part));
 		add_to_list(table_cache, tce);
+		hash = ((int)(unsigned long)start + xs) & (TC_HASH_SIZE - 1);
+		tce->hash_next = table_cache_hash[hash];
+		table_cache_hash[hash] = tce;
 	}
 	return p;
 }
@@ -896,7 +921,7 @@ void push_base_format(unsigned char *url, struct document_options *opt)
 		internal("something on html stack");
 		init_list(html_stack);
 	}
-	if (!(e = mem_alloc(sizeof(struct html_element)))) return;
+	e = mem_alloc(sizeof(struct html_element));
 	memset(e, 0, sizeof(struct html_element));
 	add_to_list(html_stack, e);
 	format.attr = 0;
@@ -963,6 +988,7 @@ void format_html(struct cache_entry *ce, struct f_data *screen)
 	unsigned char *head, *t;
 	int hdl;
 	int i;
+	memset(table_cache_hash, 0, sizeof(table_cache_hash));
 	d_opt = &screen->opt;
 	screen->use_tag = ce->count;
 	defrag_entry(ce);
@@ -975,6 +1001,7 @@ void format_html(struct cache_entry *ce, struct f_data *screen)
 	if (ce->head) add_to_str(&head, &hdl, ce->head);
 	scan_http_equiv(start, end, &head, &hdl, &t);
 	convert_table = get_convert_table(head, screen->opt.cp, screen->opt.assume_cp, &screen->cp, &screen->ass, screen->opt.hard_assume);
+	screen->opt.real_cp = screen->cp;
 	i = d_opt->plain; d_opt->plain = 0;
 	screen->title = convert_string(convert_table, t, strlen(t));
 	d_opt->plain = i;
@@ -1081,9 +1108,11 @@ void delete_unused_format_cache_entries()
 {
 	struct f_data *ce;
 	foreach(ce, format_cache) {
-		struct cache_entry *cee = NULL;
-		if (!ce->refcount && (find_in_cache(ce->url, &cee) || !cee || cee->count != ce->use_tag)) {
-			if (!cee) internal("file %s disappeared from cache", ce->url);
+		struct cache_entry *cee;
+		if (find_in_cache(ce->url, &cee))
+			internal("file %s disappeared from cache", ce->url);
+		cee->refcount--;
+		if (!ce->refcount && cee->count != ce->use_tag) {
 			ce = ce->prev;
 			destroy_formatted(ce->next);
 			format_cache_entries--;
@@ -1108,14 +1137,15 @@ void cached_format_html(struct view_state *vs, struct f_data_c *screen, struct d
 	screen->name = n;
 	screen->link_bg = NULL;
 	screen->link_bg_n = 0;
-	if (vs) vs->f = screen;
 	screen->vs = vs;
 	screen->xl = screen->yl = -1;
 	screen->f_data = NULL;
 	foreach(ce, format_cache) if (!strcmp(ce->url, vs->url) && !compare_opt(&ce->opt, opt)) {
 		cee = NULL;
-		if (find_in_cache(vs->url, &cee) || !cee || cee->count != ce->use_tag) {
-			if (!cee) internal("file %s disappeared from cache", ce->url);
+		if (find_in_cache(ce->url, &cee))
+			internal("file %s disappeared from cache", ce->url);
+		cee->refcount--;
+		if (cee->count != ce->use_tag) {
 			if (!ce->refcount) {
 				ce = ce->prev;
 				destroy_formatted(ce->next);
@@ -1128,24 +1158,16 @@ void cached_format_html(struct view_state *vs, struct f_data_c *screen, struct d
 		screen->f_data = ce;
 		goto sx;
 	}
-	if (find_in_cache(vs->url, &cee) || !cee) {
+	if (find_in_cache(vs->url, &cee)) {
 		internal("document to format not found");
 		return;
 	}
-	cee->refcount++;
+	/*cee->refcount++;*/
 	shrink_memory(0);
-	if (!(ce = mem_alloc(SIZEOF_F_DATA))) {
-		cee->refcount--;
-		return;
-	}
+	ce = mem_alloc(SIZEOF_F_DATA);
 	init_formatted(ce);
 	ce->refcount = 1;
-	if (!(ce->url = mem_alloc(strlen(vs->url) + 1))) {
-		mem_free(ce);
-		cee->refcount--;
-		return;
-	}
-	strcpy(ce->url, vs->url);
+	ce->url = stracpy(vs->url);
 	copy_opt(&ce->opt, opt);
 	add_to_list(format_cache, ce);
 	screen->f_data = ce;
@@ -1178,14 +1200,15 @@ long formatted_info(int type)
 
 void add_frame_to_list(struct session *ses, struct f_data_c *fd)
 {
-	struct f_data_c *f;
+	struct f_data_c *f, *fp;
 	foreach(f, ses->scrn_frames) {
 		if (f->yp > fd->yp || (f->yp == fd->yp && f->xp > fd->xp)) {
 			add_at_pos(f->prev, fd);
 			return;
 		}
 	}
-	add_at_pos((struct f_data_c *)ses->scrn_frames.prev, fd);
+	fp = (struct f_data_c *)ses->scrn_frames.prev;
+	add_at_pos(fp, fd);
 }
 
 struct f_data_c *find_fd(struct session *ses, unsigned char *name, int depth, int x, int y)
@@ -1196,7 +1219,7 @@ struct f_data_c *find_fd(struct session *ses, unsigned char *name, int depth, in
 		fd->depth = depth;
 		return fd;
 	}
-	if (!(fd = mem_alloc(sizeof(struct f_data_c)))) return NULL;
+	fd = mem_alloc(sizeof(struct f_data_c));
 	memset(fd, 0, sizeof(struct f_data_c));
 	fd->used = 1;
 	fd->name = stracpy(name);
@@ -1217,7 +1240,8 @@ struct f_data_c *format_frame(struct session *ses, unsigned char *name, unsigned
 	repeat:
 	if (!(fr = ses_find_frame(ses, name))) return NULL;
 	vs = &fr->vs;
-	if (find_in_cache(vs->url, &ce) || !ce) return NULL;
+	if (find_in_cache(vs->url, &ce)) return NULL;
+	ce->refcount--;
 	if (ce->redirect && fr->redirect_cnt < MAX_REDIRECTS) {
 		unsigned char *u;
 		if ((u = join_urls(vs->url, ce->redirect))) {
@@ -1275,8 +1299,9 @@ void html_interpret(struct session *ses)
 	struct document_options o;
 	struct f_data_c *fd, *cf = NULL;
 	/*debug("I");*/
+	memset(&o, 0, sizeof(struct document_options));
 	if (!ses->screen) {
-		if (!(ses->screen = mem_alloc(sizeof(struct f_data_c)))) return;
+		ses->screen = mem_alloc(sizeof(struct f_data_c));
 		memset(ses->screen, 0, sizeof(struct f_data_c));
 		ses->screen->search_word = &ses->search_word;
 	}
@@ -1285,7 +1310,8 @@ void html_interpret(struct session *ses)
 	o.xp = 0;
 	o.yp = 1;
 	o.xw = ses->term->x;
-	o.yw = ses->term->y - 2;
+	if (ses->term->y < 2) o.yw = 0;
+	else o.yw = ses->term->y - 2;
 	o.col = ses->term->spec->col;
 	o.cp = ses->term->spec->charset;
 	ds2do(&ses->ds, &o);
@@ -1322,8 +1348,6 @@ void html_interpret(struct session *ses)
 	}
 }
 
-#define SRCH_ALLOC_GR	0x10000
-
 void add_srch_chr(struct f_data *f, unsigned char c, int x, int y, int nn)
 {
 	int n = f->nsearch;
@@ -1350,22 +1374,12 @@ void sort_srch(struct f_data *f)
 {
 	int i;
 	int *min, *max;
-	if (!(f->slines1 = mem_alloc(f->y * sizeof(struct search *)))) return;
-	if (!(f->slines2 = mem_alloc(f->y * sizeof(struct search *)))) {
-		mem_free(f->slines1);
-		return;
-	}
-	if (!(min = mem_alloc(f->y * sizeof(int)))) {
-		mem_free(f->slines1);
-		mem_free(f->slines2);
-		return;
-	}
-	if (!(max = mem_alloc(f->y * sizeof(int)))) {
-		mem_free(f->slines1);
-		mem_free(f->slines2);
-		mem_free(min);
-		return;
-	}
+	if ((unsigned)f->y > MAXINT / sizeof(struct search *)) overalloc();
+	if ((unsigned)f->y > MAXINT / sizeof(int)) overalloc();
+	f->slines1 = mem_alloc(f->y * sizeof(struct search *));
+	f->slines2 = mem_alloc(f->y * sizeof(struct search *));
+	min = mem_alloc(f->y * sizeof(int));
+	max = mem_alloc(f->y * sizeof(int));
 	memset(f->slines1, 0, f->y * sizeof(struct search *));
 	memset(f->slines2, 0, f->y * sizeof(struct search *));
 	for (i = 0; i < f->y; i++) min[i] = MAXINT, max[i] = 0;
@@ -1378,6 +1392,11 @@ void sort_srch(struct f_data *f)
 	mem_free(max);
 }
 
+static inline int is_spc(chr c)
+{
+	return (unsigned char)c <= ' ' || c & ATTR_FRAME;
+}
+
 int get_srch(struct f_data *f)
 {
 	struct node *n;
@@ -1392,8 +1411,9 @@ int get_srch(struct f_data *f)
 			int ns = 1;
 			for (x = n->x; x < xm && x < f->data[y].l; x++) {
 				unsigned char c = f->data[y].d[x];
-				if (c < ' ') c = ' ' ;
+				if (is_spc(f->data[y].d[x])) c = ' ';
 				if (c == ' ' && ns) continue;
+				c = charset_upcase(c, f->opt.cp);
 				if (ns) {
 					if (!cc) add_srch_chr(f, c, x, y, 1);
 					else cnt++;
@@ -1404,9 +1424,10 @@ int get_srch(struct f_data *f)
 					       else cnt++;
 				else {
 					int xx;
-					for (xx = x + 1; xx < xm && xx < f->data[y].l; xx++) if ((unsigned char) f->data[y].d[xx] >= ' ') goto ja_uz_z_toho_programovani_asi_zcvoknu;
+					for (xx = x + 1; xx < xm && xx < f->data[y].l; xx++) if (!is_spc(f->data[y].d[xx])) goto ja_uz_z_toho_programovani_asi_zcvoknu;
 					xx = x;
 					ja_uz_z_toho_programovani_asi_zcvoknu:
+				/* uz jsem zcvoknul, trpim poruchou osobnosti */
 					if (!cc) add_srch_chr(f, ' ', x, y, xx - x);
 					else cnt++;
 					if (xx == x) break;
@@ -1427,7 +1448,8 @@ void get_search_data(struct f_data *f)
 	if (f->search) return;
 	n = get_srch(f);
 	f->nsearch = 0;
-	if (!(f->search = mem_alloc(n * sizeof(struct search)))) return;
+	if ((unsigned)n > MAXINT / sizeof(struct search)) overalloc();
+	f->search = mem_alloc(n * sizeof(struct search));
 	get_srch(f);
 	while (f->nsearch && f->search[f->nsearch - 1].c == ' ') f->nsearch--;
 	/*debug("!");*/
