@@ -1,5 +1,7 @@
 #include "links.h"
 
+#define format format_
+
 #define AL_TR		-1
 
 #define VAL_TR		-1
@@ -56,10 +58,10 @@ void get_c_width(char *attr, int *w, int sh)
 	if ((al = get_attr_val(attr, "width"))) {
 		if (*al && al[strlen(al) - 1] == '*') {
 			char *en;
-			int n;
+			unsigned long n;
 			al[strlen(al) - 1] = 0;
 			n = strtoul(al, &en, 10);
-			if (n >= 0 && !*en) *w = W_REL - n;
+			if (n < 10000 && !*en) *w = W_REL - n;
 		} else {
 			int p = get_width(attr, "width", sh);
 			if (p >= 0) *w = p;
@@ -68,8 +70,8 @@ void get_c_width(char *attr, int *w, int sh)
 	}
 }
 
-#define INIT_X		2
-#define INIT_Y		2
+#define INIT_X		8
+#define INIT_Y		8
 
 struct table_cell {
 	int used;
@@ -141,24 +143,17 @@ unsigned char vline_table[3] = { 0x20, 0xb3, 0xba };
 struct table *new_table()
 {
 	struct table *t;
-	if (!(t = mem_alloc(sizeof(struct table)))) return NULL;
+	t = mem_alloc(sizeof(struct table));
 	memset(t, 0, sizeof(struct table));
 	t->p = NULL;
 	t->x = t->y = 0;
 	t->rx = INIT_X;
 	t->ry = INIT_Y;
-	if (!(t->cells = mem_alloc(INIT_X * INIT_Y * sizeof(struct table_cell)))) {
-		mem_free(t);
-		return NULL;
-	}
+	t->cells = mem_alloc(INIT_X * INIT_Y * sizeof(struct table_cell));
 	memset(t->cells, 0, INIT_X * INIT_Y * sizeof(struct table_cell));
 	t->c = 0;
 	t->rc = INIT_X;
-	if (!(t->cols = mem_alloc(INIT_X * sizeof(struct table_column)))) {
-		mem_free(t->cells);
-		mem_free(t);
-		return NULL;
-	}
+	t->cols = mem_alloc(INIT_X * sizeof(struct table_column));
 	memset(t->cols, 0, INIT_X * sizeof(struct table_column));
 	t->xcols = DUMMY;
 	t->xc = 0;
@@ -225,9 +220,17 @@ struct table_cell *new_cell(struct table *t, int x, int y)
 	}
 	nt.rx = t->rx;
 	nt.ry = t->ry;
-	while (x >= nt.rx) if (!(nt.rx *= 2)) return NULL;
-	while (y >= nt.ry) if (!(nt.ry *= 2)) return NULL;
-	if (!(nt.cells = mem_alloc(nt.rx * nt.ry * sizeof(struct table_cell)))) return NULL;
+	while (x >= nt.rx) {
+		if ((unsigned)nt.rx > MAXINT / 2) overalloc();
+		nt.rx *= 2;
+	}
+	while (y >= nt.ry) {
+		if ((unsigned)nt.ry > MAXINT / 2) overalloc();
+		nt.ry *= 2;
+	}
+	if ((unsigned)nt.rx * (unsigned)nt.ry / (unsigned)nt.rx != (unsigned)nt.ry) overalloc();
+	if ((unsigned)nt.rx * (unsigned)nt.ry > MAXINT / sizeof(struct table_cell)) overalloc();
+	nt.cells = mem_alloc(nt.rx * nt.ry * sizeof(struct table_cell));
 	memset(nt.cells, 0, nt.rx * nt.ry * sizeof(struct table_cell));
 	for (i = 0; i < t->x; i++)
 		for (j = 0; j < t->y; j++)
@@ -244,11 +247,16 @@ struct table_cell *new_cell(struct table *t, int x, int y)
 
 void new_columns(struct table *t, int span, int width, int align, int valign, int group)
 {
+	if ((unsigned)t->c + (unsigned)span > MAXINT) overalloc();
 	if (t->c + span > t->rc) {
 		int n = t->rc;
 		struct table_column *nc;
-		while (t->c + span > n) if (!(n *= 2)) return;
-		if (!(nc = mem_realloc(t->cols, n * sizeof(struct table_column)))) return;
+		while (t->c + span > n) {
+			if ((unsigned)n > MAXINT / 2) overalloc();
+			n *= 2;
+		}
+		if ((unsigned)n > MAXINT / sizeof(struct table_column)) overalloc();
+		nc = mem_realloc(t->cols, n * sizeof(struct table_column));
 		t->rc = n;
 		t->cols = nc;
 	}
@@ -264,13 +272,15 @@ void new_columns(struct table *t, int span, int width, int align, int valign, in
 void set_td_width(struct table *t, int x, int width, int f)
 {
 	if (x >= t->xc) {
-		int n = t->xc;
+		int n = t->xc ? t->xc : 1;
 		int i;
 		int *nc;
-		while (x >= n) if (!(n *= 2)) break;
-		if (!n && t->xc) return;
-		if (!n) n = x + 1;
-		if (!(nc = mem_realloc(t->xcols, n * sizeof(int)))) return;
+		while (x >= n) {
+			if ((unsigned)n > MAXINT / 2) overalloc();
+			n *= 2;
+		}
+		if ((unsigned)n > MAXINT / sizeof(int)) overalloc();
+		nc = mem_realloc(t->xcols, n * sizeof(int));
 		for (i = t->xc; i < n; i++) nc[i] = W_AUTO;
 		t->xc = n;
 		t->xcols = nc;
@@ -333,13 +343,11 @@ struct table *parse_table(unsigned char *html, unsigned char *eof, unsigned char
 	html = en;
 	if (bad_html && !p && !lbhp) {
 		if (!(*bhp & (ALLOC_GR-1))) {
-			struct s_e *s_e;
-			if (!(s_e = mem_realloc(*bad_html, (*bhp + ALLOC_GR) * sizeof(struct s_e)))) goto qwe;
-			*bad_html = s_e;
+			if ((unsigned)*bhp > MAXINT / sizeof(struct s_e) - ALLOC_GR) overalloc();
+			*bad_html = mem_realloc(*bad_html, (*bhp + ALLOC_GR) * sizeof(struct s_e));
 		}
 		lbhp = (*bad_html)[(*bhp)++].s = html;
 	}
-	qwe:
 	while (html < eof && *html != '<') html++;
 	if (html >= eof) {
 		if (p) CELL(t, x, y)->end = html;
@@ -429,7 +437,7 @@ struct table *parse_table(unsigned char *html, unsigned char *eof, unsigned char
 	if (p) CELL(t, x, y)->end = html, p = 0;
 	if (y == -1) y = 0, x = 0;
 	nc:
-	if (!(cell = new_cell(t, x, y))) goto see;
+	cell = new_cell(t, x, y);
 	if (cell->used) {
 		if (cell->colspan == -1) goto see;
 		x++;
@@ -456,6 +464,10 @@ struct table *parse_table(unsigned char *html, unsigned char *eof, unsigned char
 	if (!csp) csp = -1;
 	if ((rsp = get_num(t_attr, "rowspan")) == -1) rsp = 1;
 	if (!rsp) rsp = -1;
+	if (csp >= 0 && rsp >= 0 && csp * rsp > 100000) {
+		if (csp > 10) csp = -1;
+		if (rsp > 10) rsp = -1;
+	}
 	cell->colspan = csp;
 	cell->rowspan = rsp;
 	if (csp == 1) {
@@ -464,9 +476,9 @@ struct table *parse_table(unsigned char *html, unsigned char *eof, unsigned char
 		if (w != W_AUTO) set_td_width(t, x, w, 0);
 	}
 	qqq = t->x;
-	for (i = 1; csp != -1 ? i < csp : i < qqq; i++) {
-		struct table_cell *sc;
-		if (!(sc = new_cell(t, x + i, y)) || sc->used) {
+	for (i = 1; csp != -1 ? i < csp : x + i < qqq; i++) {
+		struct table_cell *sc = new_cell(t, x + i, y);
+		if (sc->used) {
 			csp = i;
 			for (k = 0; k < i; k++) CELL(t, x + k, y)->colspan = csp;
 			break;
@@ -478,10 +490,10 @@ struct table *parse_table(unsigned char *html, unsigned char *eof, unsigned char
 		sc->my = y;
 	}
 	qqq = t->y;
-	for (j = 1; rsp != -1 ? j < rsp : j < qqq; j++) {
+	for (j = 1; rsp != -1 ? j < rsp : y + j < qqq; j++) {
 		for (k = 0; k < i; k++) {
-			struct table_cell *sc;
-			if (!(sc = new_cell(t, x + k, y + j)) || sc->used) {
+			struct table_cell *sc = new_cell(t, x + k, y + j);
+			if (sc->used) {
 				int l, m;
 				if (sc->mx == x && sc->my == y) continue;
 				/*internal("boo");*/
@@ -511,10 +523,8 @@ struct table *parse_table(unsigned char *html, unsigned char *eof, unsigned char
 		}
 	}
 
-	if (!(t->r_heights = mem_alloc(t->y * sizeof(int)))) {
-		free_table(t);
-		return NULL;
-	}
+	if ((unsigned)t->y > MAXINT / sizeof(int)) overalloc();
+	t->r_heights = mem_alloc(t->y * sizeof(int));
 	memset(t->r_heights, 0, t->y * sizeof(int));
 
 	for (x = 0; x < t->c; x++) if (t->cols[x].width != W_AUTO) set_td_width(t, x, t->cols[x].width, 1);
@@ -533,7 +543,7 @@ void get_cell_width(char *start, char *end, int cellpd, int w, int a, int *min,
 	if (min) *min = p->x;
 	if (max) *max = p->xmax;
 	if (n_links) *n_links = p->link_num;
-	if (min && max && *min > *max) internal("get_cell_width: %d > %d", *min, *max);
+	/*if (min && max && *min > *max) internal("get_cell_width: %d > %d", *min, *max);*/
 	mem_free(p);
 }
 
@@ -541,11 +551,11 @@ static inline void check_cell_widths(struct table *t)
 {
 	int i, j;
 	for (j = 0; j < t->y; j++) for (i = 0; i < t->x; i++) {
-		int min, max, nl;
+		int min, max;
 		struct table_cell *c = CELL(t, i, j);
 		if (!c->start) continue;
 		get_cell_width(c->start, c->end, t->cellpd, 0, 0, &min, &max, c->link_num, NULL);
-		if (min != c->min_width || max < c->max_width) internal("check_cell_widths failed");
+		/*if (min != c->min_width || max < c->max_width) internal("check_cell_widths failed");*/
 	}
 }
 
@@ -573,6 +583,7 @@ void dst_width(int *p, int n, int w, int *lim)
 	int i, s = 0, d, r;
 	for (i = 0; i < n; i++) s += p[i];
 	if (s >= w) return;
+	if (!n) return;
 	again:
 	d = (w - s) / n;
 	r = (w - s) % n;
@@ -582,7 +593,7 @@ void dst_width(int *p, int n, int w, int *lim)
 		if (lim && p[i] > lim[i]) w += p[i] - lim[i], p[i] = lim[i];
 	}
 	if (w) {
-		if (!lim) internal("bug in dst_width");
+		/*if (!lim) internal("bug in dst_width");*/
 		lim = NULL;
 		s = 0;
 		goto again;
@@ -620,19 +631,10 @@ int get_hline_width(struct table *t, int row)
 int get_column_widths(struct table *t)
 {
 	int i, j, s, ns;
-	if (!t->min_c && !(t->min_c = mem_alloc(t->x * sizeof(int)))) return -1;
-	if (!t->max_c && !(t->max_c = mem_alloc(t->x * sizeof(int)))) {
-		mem_free(t->min_c);
-		t->min_c = NULL;
-		return -1;
-	}
-	if (!t->w_c && !(t->w_c = mem_alloc(t->x * sizeof(int)))) {
-		mem_free(t->min_c);
-		t->min_c = NULL;
-		mem_free(t->max_c);
-		t->max_c = NULL;
-		return -1;
-	}
+	if ((unsigned)t->x > MAXINT / sizeof(int)) overalloc();
+	if (!t->min_c) t->min_c = mem_alloc(t->x * sizeof(int));
+	if (!t->max_c) t->max_c = mem_alloc(t->x * sizeof(int));
+	if (!t->w_c) t->w_c = mem_alloc(t->x * sizeof(int));
 	memset(t->min_c, 0, t->x * sizeof(int));
 	memset(t->max_c, 0, t->x * sizeof(int));
 	s = 1;
@@ -642,8 +644,9 @@ int get_column_widths(struct table *t)
 			struct table_cell *c = CELL(t, i, j);
 			if (c->spanned || !c->used) continue;
 			if (c->colspan + i > t->x) {
-				internal("colspan out of table");
-				return -1;
+				/*internal("colspan out of table");
+				return -1;*/
+				continue;
 			}
 			if (c->colspan == s) {
 				int k, p = 0;
@@ -688,7 +691,7 @@ void get_table_width(struct table *t)
 	min += vl, max += vl;
 	t->min_t = min;
 	t->max_t = max;
-	if (min > max) internal("min(%d) > max(%d)", min, max);
+	/*if (min > max) internal("min(%d) > max(%d)", min, max);*/
 }
 
 void distribute_widths(struct table *t, int width)
@@ -701,14 +704,16 @@ void distribute_widths(struct table *t, int width)
 	int mmax_c = 0;
 	if (!t->x) return;
 	if (d < 0) {
-		internal("too small width %d, required %d", width, t->min_t);
+		/*internal("too small width %d, required %d", width, t->min_t);*/
+		return;
 	}
 	for (i = 0; i < t->x; i++) if (t->max_c[i] > mmax_c) mmax_c = t->max_c[i];
 	memcpy(t->w_c, t->min_c, t->x * sizeof(int));
 	t->rw = width;
+	if ((unsigned)t->x > MAXINT / sizeof(int)) overalloc();
 	u = mem_alloc(t->x);
-	if (!(w = mem_alloc(t->x * sizeof(int)))) goto end;
-	if (!(mx = mem_alloc(t->x * sizeof(int)))) goto end1;
+	w = mem_alloc(t->x * sizeof(int));
+	mx = mem_alloc(t->x * sizeof(int));
 	while (d) {
 		int mss, mii;
 		int p = 0;
@@ -752,7 +757,7 @@ void distribute_widths(struct table *t, int width)
 					w[i] = 1, mx[i] = MAXINT;
 					break;
 				default:
-					internal("could not expand table");
+					/*internal("could not expand table");*/
 					goto end2;
 			}
 			p += w[i];
@@ -779,16 +784,17 @@ void distribute_widths(struct table *t, int width)
 			t->w_c[mii] += mss;
 			d -= t->w_c[mii] - q;
 			while (d < 0) t->w_c[mii]--, d++;
-			if (t->w_c[mii] < q) internal("shrinking cell");
+			if (t->w_c[mii] < q) {
+				/*internal("shrinking cell");*/
+				t->w_c[mii] = q;
+			}
 			wq = 1;
 			if (d) goto a;
 		} else if (!wq) om++;
 	}
 	end2:
 	mem_free(mx);
-	end1:
 	mem_free(w);
-	end:
 	if (u) mem_free(u);
 }
 
@@ -799,7 +805,8 @@ void check_table_widths(struct table *t)
 	int i, j;
 	int s, ns;
 	int m, mi = 0; /* go away, warning! */
-	if (!(w = mem_alloc(t->x * sizeof(int)))) return;
+	if ((unsigned)t->x > MAXINT / sizeof(int)) overalloc();
+	w = mem_alloc(t->x * sizeof(int));
 	memset(w, 0, t->x * sizeof(int));
 	for (j = 0; j < t->y; j++) for (i = 0; i < t->x; i++) {
 		struct table_cell *c = CELL(t, i, j);
@@ -823,7 +830,7 @@ void check_table_widths(struct table *t)
 			struct table_cell *c = CELL(t, i, j);
 			if (!c->start) continue;
 			if (c->colspan + i > t->x) {
-				internal("colspan out of table");
+				/*internal("colspan out of table");*/
 				mem_free(w);
 				return;
 			}
@@ -991,12 +998,13 @@ void display_table_frames(struct table *t, int x, int y)
 	signed char *fh, *fv;
 	int i, j;
 	int cx, cy;
-	if (!(fh = mem_alloc((t->x + 2) * (t->y + 1)))) return;
+	if ((unsigned)t->x > MAXINT) overalloc();
+	if ((unsigned)t->y > MAXINT) overalloc();
+	if (((unsigned)t->x + 2) * ((unsigned)t->y + 2) / ((unsigned)t->x + 2) != ((unsigned)t->y + 2)) overalloc();
+	if (((unsigned)t->x + 2) * ((unsigned)t->y + 2) > MAXINT) overalloc();
+	fh = mem_alloc((t->x + 2) * (t->y + 1));
 	memset(fh, -1, (t->x + 2) * (t->y + 1));
-	if (!(fv = mem_alloc((t->x + 1) * (t->y + 2)))) {
-		mem_free(fh);
-		return;
-	}
+	fv = mem_alloc((t->x + 1) * (t->y + 2));
 	memset(fv, -1, (t->x + 1) * (t->y + 2));
 #ifndef DEBUG
 #define H_LINE_X(xx, yy) fh[(xx) + 1 + (t->x + 2) * (yy)]
@@ -1158,7 +1166,7 @@ void format_table(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	for (i = 0; i < bad_html_n; i++) {
 		while (bad_html[i].s < bad_html[i].e && WHITECHAR(*bad_html[i].s)) bad_html[i].s++;
 		while (bad_html[i].s < bad_html[i].e && WHITECHAR(bad_html[i].e[-1])) bad_html[i].e--;
-		if (bad_html[i].s < bad_html[i].e) parse_html(bad_html[i].s, bad_html[i].e, put_chars_f, line_break_f, init_f, special_f, p, NULL);
+		if (bad_html[i].s < bad_html[i].e) parse_html(bad_html[i].s, bad_html[i].e, put_chars_f, line_break_f, special_f, p, NULL);
 	}
 	mem_free(bad_html);
 	html_stack_dup();
@@ -1229,18 +1237,17 @@ void format_table(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	n->yw = p->yp - n->y + p->cy;
 	display_complicated_table(t, x, p->cy, &cye);
 	display_table_frames(t, x, p->cy);
-	if ((nn = mem_alloc(sizeof(struct node)))) {
-		nn->x = n->x;
-		nn->y = p->yp + cye;
-		nn->xw = n->xw;
-		add_to_list(p->data->nodes, nn);
-	}
+	nn = mem_alloc(sizeof(struct node));
+	nn->x = n->x;
+	nn->y = p->yp + cye;
+	nn->xw = n->xw;
+	add_to_list(p->data->nodes, nn);
 	/*sdbg(p->data);*/
 	/*for (y = p->cy; y < cye; y++) {
 		last_link_to_move = llm;
 		align_line(p, y);
 	}*/
-	if (p->cy + t->rh != cye) internal("size does not match; 1:%d, 2:%d", p->cy + t->rh, cye);
+	/*if (p->cy + t->rh != cye) internal("size does not match; 1:%d, 2:%d", p->cy + t->rh, cye);*/
 	p->cy = cye;
 	p->cx = -1;
 
