@@ -4,7 +4,7 @@ struct http_connection_info {
 	int bl_flags;
 	int http10;
 	int close;
-	int length;
+	off_t length;
 	int version;
 	int chunk_remaining;
 };
@@ -25,15 +25,14 @@ unsigned char *parse_http_header(unsigned char *head, unsigned char *item, unsig
 			for (g = ++f; *g >= ' '; g++) ;
 			while (g > f && g[-1] == ' ') g--;
 			if (h) mem_free(h);
-			if ((h = mem_alloc(g - f + 1))) {
-				memcpy(h, f, g - f);
-				h[g - f] = 0;
-				if (ptr) {
-					*ptr = f;
-					break;
-				}
-				return h;
+			h = mem_alloc(g - f + 1);
+			memcpy(h, f, g - f);
+			h[g - f] = 0;
+			if (ptr) {
+				*ptr = f;
+				break;
 			}
+			return h;
 		}
 		cont:;
 		f--;
@@ -43,7 +42,8 @@ unsigned char *parse_http_header(unsigned char *head, unsigned char *item, unsig
 
 unsigned char *parse_header_param(unsigned char *x, unsigned char *e)
 {
-	int le = strlen(e);
+	unsigned char u;
+	size_t le = strlen(e);
 	int lp;
 	unsigned char *y = x;
 	a:
@@ -53,9 +53,13 @@ unsigned char *parse_header_param(unsigned char *x, unsigned char *e)
 	if (casecmp(y, e, le)) goto a;
 	y += le;
 	while (*y && (*y <= ' ' || *y == '=')) y++;
-	if (!*y) return stracpy("");
+	u = ';';
+	if (*y == '\'' || *y == '"') u = *y++;
 	lp = 0;
-	while (y[lp] >= ' ' && y[lp] != ';') lp++;
+	while (y[lp] >= ' ' && y[lp] != u) {
+		lp++;
+		if (lp == MAXINT) overalloc();
+	}
 	return memacpy(y, lp);
 }
 
@@ -92,7 +96,7 @@ again:
 	while (start < e && *(e - 1) == ' ') e--;
 	if (start == e) return NULL;
 	
-	if (!(n = mem_alloc(e - start + 1))) return NULL;
+	n = mem_alloc(e - start + 1);
 	while (start < e) {
 		if (*start < ' ') n[i] = '.';
 		else n[i] = *start;
@@ -119,22 +123,33 @@ static int get_http_code(unsigned char *head, int *code, int *version)
 	return 0;
 }
 
-unsigned char *buggy_servers[] = { "mod_czech/3.1.0", "Purveyor", "Netscape-Enterprise", NULL };
+struct {
+	unsigned char *name;
+	int bugs;
+} buggy_servers[] = {
+	{ "mod_czech/3.1.0", BL_HTTP10 },
+	{ "Purveyor", BL_HTTP10 },
+	{ "Netscape-Enterprise", BL_HTTP10 | BL_NO_ACCEPT_LANGUAGE },
+	{ "Apache Coyote", BL_HTTP10 },
+	{ "lighttpd", BL_HTTP10 },
+	{ "FORPSI", BL_NO_RANGE },
+	{ "Sausalito", BL_HTTP10 },
+	{ NULL, 0 }
+};
 
 int check_http_server_bugs(unsigned char *url, struct http_connection_info *info, unsigned char *head)
 {
-	unsigned char *server, **s;
+	unsigned char *server;
+	int i, bugs;
 	if (!http_bugs.allow_blacklist || info->http10) return 0;
 	if (!(server = parse_http_header(head, "Server", NULL))) return 0;
-	for (s = buggy_servers; *s; s++) if (strstr(server, *s)) goto bug;
+	bugs = 0;
+	for (i = 0; buggy_servers[i].name; i++) if (strstr(server, buggy_servers[i].name)) bugs |= buggy_servers[i].bugs;
 	mem_free(server);
-	return 0;
-	bug:
-	mem_free(server);
-	if ((server = get_host_name(url))) {
-		add_blacklist_entry(server, BL_HTTP10);
+	if (bugs && (server = get_host_name(url))) {
+		add_blacklist_entry(server, bugs);
 		mem_free(server);
-		return 1;
+		return bugs & ~BL_NO_RANGE;
 	}
 	return 0;	
 }
@@ -189,14 +204,11 @@ void http_send_header(struct connection *c)
 	unsigned char *hdr;
 	unsigned char *h, *u, *uu, *sp;
 	int l = 0;
+	int la;
 	unsigned char *post;
 	unsigned char *host = upcase(c->url[0]) != 'P' ? c->url : get_url_data(c->url);
 	set_timeout(c);
-	if (!(info = mem_alloc(sizeof(struct http_connection_info)))) {
-		setcstate(c, S_OUT_OF_MEM);
-		abort_connection(c);
-		return;
-	}
+	info = mem_alloc(sizeof(struct http_connection_info));
 	memset(info, 0, sizeof(struct http_connection_info));
 	c->info = info;
 	if ((h = get_host_name(host))) {
@@ -207,11 +219,7 @@ void http_send_header(struct connection *c)
 	info->http10 = http10;
 	post = strchr(c->url, POST_CHAR);
 	if (post) post++;
-	if (!(hdr = init_str())) {
-		setcstate(c, S_OUT_OF_MEM);
-		http_end_request(c, 0);
-		return;
-	}
+	hdr = init_str();
 	if (!post) add_to_str(&hdr, &l, "GET ");
 	else {
 		add_to_str(&hdr, &l, "POST ");
@@ -235,14 +243,13 @@ void http_send_header(struct connection *c)
 	a:
 	for (sp = uu; *sp; sp++) if (*sp <= ' ') {
 		unsigned char *nu = mem_alloc(strlen(uu) + 3);
-		if (!nu) break;
 		memcpy(nu, uu, sp - uu);
 		sprintf(nu + (sp - uu), "%%%02X", (int)*sp);
 		strcat(nu, sp + 1);
 		mem_free(uu);
 		uu = nu;
 		goto a;
-	} else if (*sp == '\\') *sp = '/';
+	}
 	add_to_str(&hdr, &l, uu);
 	mem_free(uu);
 	if (!http10) add_to_str(&hdr, &l, " HTTP/1.1\r\n");
@@ -260,12 +267,11 @@ void http_send_header(struct connection *c)
 	}
 	add_to_str(&hdr, &l, "User-Agent: Links (" VERSION_STRING "; ");
 	add_to_str(&hdr, &l, system_name);
+	add_to_str(&hdr, &l, "; ");
 	if (!list_empty(terminals)) {
-		struct terminal *t = terminals.prev;
-		add_to_str(&hdr, &l, "; ");
-		add_num_to_str(&hdr, &l, t->x);
-		add_to_str(&hdr, &l, "x");
-		add_num_to_str(&hdr, &l, t->y);
+		add_to_str(&hdr, &l, "text");
+	} else {
+		add_to_str(&hdr, &l, "dump");
 	}
 	add_to_str(&hdr, &l, ")\r\n");
 	add_to_str(&hdr, &l, "Accept: */*\r\n");
@@ -285,6 +291,14 @@ void http_send_header(struct connection *c)
 		mem_free(ac);
 	}
 	if (!(info->bl_flags & BL_NO_CHARSET) && !http_bugs.no_accept_charset) add_to_str(&hdr, &l, accept_charset);
+	if (!(info->bl_flags & BL_NO_ACCEPT_LANGUAGE)) {
+		add_to_str(&hdr, &l, "Accept-Language: ");
+		la = l;
+		add_to_str(&hdr, &l, _(TEXT_(T__ACCEPT_LANGUAGE), NULL));
+		add_to_str(&hdr, &l, ", ");
+		if (!strstr(hdr + la, "en,") && !strstr(hdr + la, "en;")) add_to_str(&hdr, &l, "en;q=0.2, ");
+		add_to_str(&hdr, &l, "*;q=0.1\r\n");
+	}
 	if (!http10) {
 		if (upcase(c->url[0]) != 'P') add_to_str(&hdr, &l, "Connection: ");
 		else add_to_str(&hdr, &l, "Proxy-Connection: ");
@@ -292,6 +306,8 @@ void http_send_header(struct connection *c)
 		else add_to_str(&hdr, &l, "close\r\n");
 	}
 	if ((e = c->cache)) {
+		int code, vers;
+		if (get_http_code(e->head, &code, &vers) || code >= 400) goto skip_ifmod_and_range;
 		if (!e->incomplete && e->head && c->no_cache <= NC_IF_MOD &&
 		    e->last_modified) {
 			add_to_str(&hdr, &l, "If-Modified-Since: ");
@@ -299,18 +315,13 @@ void http_send_header(struct connection *c)
 			add_to_str(&hdr, &l, "\r\n");
 		}
 	}
-	if (c->no_cache >= NC_PR_NO_CACHE) add_to_str(&hdr, &l, "Pragma: no-cache\r\nCache-Control: no-cache\r\n");
-	if (c->from) {
+	if (c->from && (c->est_length == -1 || c->from < c->est_length) && c->no_cache < NC_IF_MOD && !(info->bl_flags & BL_NO_RANGE)) {
 		add_to_str(&hdr, &l, "Range: bytes=");
 		add_num_to_str(&hdr, &l, c->from);
 		add_to_str(&hdr, &l, "-\r\n");
 	}
-	if ((h = find_auth(host))) {
-		add_to_str(&hdr, &l, "Authorization: Basic ");
-		add_to_str(&hdr, &l, h);
-		add_to_str(&hdr, &l, "\r\n");
-		mem_free(h);
-	}
+	skip_ifmod_and_range:
+	if (c->no_cache >= NC_PR_NO_CACHE) add_to_str(&hdr, &l, "Pragma: no-cache\r\nCache-Control: no-cache\r\n");
 	if (post) {
 		unsigned char *pd = strchr(post, '\n');
 		if (pd) {
@@ -365,6 +376,11 @@ void read_http_data(struct connection *c, struct read_buffer *rb)
 	if (info->length != -2) {
 		int l = rb->len;
 		if (info->length >= 0 && info->length < l) l = info->length;
+		if ((off_t)(0UL + c->from + l) < 0) {
+			setcstate(c, S_LARGE_FILE);
+			abort_connection(c);
+			return;
+		}
 		c->received += l;
 		if (add_fragment(c->cache, c->from, rb->data, l) == 1) c->tries = 0;
 		if (info->length >= 0) info->length -= l;
@@ -397,9 +413,9 @@ void read_http_data(struct connection *c, struct read_buffer *rb)
 			int l;
 			if ((l = is_line_in_buffer(rb))) {
 				unsigned char *de;
-				int n;
-				if (l != -1) n = strtol(rb->data, (char **)&de, 16);
-				if (l == -1 || de == rb->data) {
+				long n = 0;	/* warning, go away */
+				if (l != -1) n = strtol(rb->data, (char **)(void *)&de, 16);
+				if (l == -1 || n < 0 || n >= MAXINT || de == rb->data) {
 					setcstate(c, S_HTTP_ERROR);
 					abort_connection(c);
 					return;
@@ -411,6 +427,11 @@ void read_http_data(struct connection *c, struct read_buffer *rb)
 		} else {
 			int l = info->chunk_remaining;
 			if (l > rb->len) l = rb->len;
+			if ((off_t)(0UL + c->from + l) < 0) {
+				setcstate(c, S_LARGE_FILE);
+				abort_connection(c);
+				return;
+			}
 			c->received += l;
 			if (add_fragment(c->cache, c->from, rb->data, l) == 1) c->tries = 0;
 			info->chunk_remaining -= l;
@@ -458,7 +479,7 @@ int get_header(struct read_buffer *rb)
 
 void http_got_header(struct connection *c, struct read_buffer *rb)
 {
-	int cf;
+	off_t cf;
 	int state = c->state != S_PROC ? S_GETH : S_PROC;
 	unsigned char *head;
 	unsigned char *cookie, *ch;
@@ -501,13 +522,9 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 		abort_connection(c);
 		return;
 	}
-	if (!(head = mem_alloc(a + 1))) {
-		setcstate(c, S_OUT_OF_MEM);
-		abort_connection(c);
-		return;
-	}
+	head = mem_alloc(a + 1);
 	memcpy(head, rb->data, a); head[a] = 0;
-	if (check_http_server_bugs(host, c->info, head)) {
+	if (check_http_server_bugs(host, c->info, head) && is_connection_restartable(c)) {
 		mem_free(head);
 		setcstate(c, S_RESTART);
 		retry_connection(c);
@@ -539,15 +556,18 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 	}
 	if (h == 204) {
 		mem_free(head);
-		setcstate(c, S_OK);
+		setcstate(c, S_HTTP_204);
 		http_end_request(c, 0);
 		return;
 	}
-	if (!c->cache && get_cache_entry(c->url, &c->cache)) {
-		mem_free(head);
-		setcstate(c, S_OUT_OF_MEM);
-		abort_connection(c);
-		return;
+	if (!c->cache) {
+		if (get_cache_entry(c->url, &c->cache)) {
+			mem_free(head);
+			setcstate(c, S_OUT_OF_MEM);
+			abort_connection(c);
+			return;
+		}
+		c->cache->refcount--;
 	}
 	e = c->cache;
 	if (e->head) mem_free(e->head);
@@ -565,21 +585,13 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 	}
 #endif
 	if (e->redirect) mem_free(e->redirect), e->redirect = NULL;
-	if (h == 301 || h == 302 || h == 303) {
+	if (h == 301 || h == 302 || h == 303 || h == 307) {
 		if ((d = parse_http_header(e->head, "Location", NULL))) {
 			if (e->redirect) mem_free(e->redirect);
 			e->redirect = d;
 			e->redirect_get = h == 303;
 		}
 	}
- 	/*if (h == 401) if ((d = parse_http_header(e->head, "WWW-Authenticate", NULL))) {
- 		if (!strncasecmp(d, "Basic", 5)) {
- 			add_auth_entry(host, get_param(d, "realm"));
- 			add_questions_entry(do_auth_dialog);
- 			setcstate(c, S_QUESTIONS);
-  		}
- 		mem_free(d);
-  	}*/
 	kill_buffer_data(rb, a);
 	info->close = 0;
 	info->length = -1;
@@ -594,11 +606,21 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 		if (strlen(d) > 6) {
 			d[5] = 0;
 			if (!(strcasecmp(d, "bytes")) && d[6] >= '0' && d[6] <= '9') {
-				int f = strtol(d + 6, NULL, 10);
-				if (f >= 0) c->from = f;
+#if defined(HAVE_STRTOLL)
+				long long f = strtoll(d + 6, NULL, 10);
+#elif defined(HAVE_STRTOQ)
+				longlong f = strtoq(d + 6, NULL, 10);
+#else
+				long f = strtol(d + 6, NULL, 10);
+				if (f == MAXLONG) f = -1;
+#endif
+				if (f >= 0 && (off_t)f >= 0 && (off_t)f == f) c->from = f;
 			}
 		}
 		mem_free(d);
+	} else if (h == 206) {
+/* Hmm ... some servers send 206 partial but don't send Content-Range */
+		c->from = cf;
 	}
 	if (cf && !c->from && !c->unrestartable) c->unrestartable = 1;
 	if (c->from > cf || c->from < 0) {
@@ -608,17 +630,27 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 	}
 	if ((d = parse_http_header(e->head, "Content-Length", NULL))) {
 		unsigned char *ep;
-		int l = strtol(d, (char **)&ep, 10);
-		if (!*ep && l >= 0) {
+#if defined(HAVE_STRTOLL)
+		long long l = strtoll(d, (char **)(void *)&ep, 10);
+#elif defined(HAVE_STRTOQ)
+		longlong l = strtoq(d, (char **)(void *)&ep, 10);
+#else
+		long l = strtol(d, (char **)(void *)&ep, 10);
+		if (l == MAXLONG) l = -1;
+#endif
+		if (!*ep && l >= 0 && (off_t)l >= 0 && (off_t)l == l) {
 			if (!info->close || version >= 11) info->length = l;
-			c->est_length = c->from + l;
+			if (c->from + l >= 0) c->est_length = c->from + l;
 		}
 		mem_free(d);
 	}
 	if ((d = parse_http_header(e->head, "Accept-Ranges", NULL))) {
 		if (!strcasecmp(d, "none") && !c->unrestartable) c->unrestartable = 1;
 		mem_free(d);
-	} else if (!c->unrestartable && !c->from) c->unrestartable = 1;
+	} else {
+		if (!c->unrestartable && !c->from) c->unrestartable = 1;
+	}
+	if (info->bl_flags & BL_NO_RANGE && !c->unrestartable) c->unrestartable = 1;
 	if ((d = parse_http_header(e->head, "Transfer-Encoding", NULL))) {
 		if (!strcasecmp(d, "chunked")) {
 			info->length = -2;
