@@ -4,7 +4,6 @@
 #define IN_BUF_SIZE	16
 
 #define USE_TWIN_MOUSE	1
-#define USE_X_MOUSE	2
 #define TW_BUTT_LEFT	1
 #define TW_BUTT_MIDDLE	2
 #define TW_BUTT_RIGHT	4
@@ -55,6 +54,14 @@ void write_ev_queue(struct itrm *itrm)
 	if (!itrm->eqlen) set_handlers(itrm->sock_out, get_handler(itrm->sock_out, H_READ), NULL, get_handler(itrm->sock_out, H_ERROR), get_handler(itrm->sock_out, H_DATA));
 }
 
+void queue_event(struct itrm *itrm, unsigned char *data, int len);
+
+void mouse_queue_event(struct itrm *itrm, unsigned char *data, int len)
+{
+	if (itrm->blocked) return;
+	queue_event(itrm, data, len);
+}
+
 void queue_event(struct itrm *itrm, unsigned char *data, int len)
 {
 	int w = 0;
@@ -65,12 +72,8 @@ void queue_event(struct itrm *itrm, unsigned char *data, int len)
 		return;
 	}
 	if (w < len) {
-		unsigned char *c;
-		if (!(c = mem_realloc(itrm->ev_queue, itrm->eqlen + len - w))) {
-			free_trm(itrm);
-			return;
-		}
-		itrm->ev_queue = c;
+		if ((unsigned)itrm->eqlen + (unsigned)(len - w) > MAXINT) overalloc();
+		itrm->ev_queue = mem_realloc(itrm->ev_queue, itrm->eqlen + len - w);
 		memcpy(itrm->ev_queue + itrm->eqlen, data + w, len - w);
 		itrm->eqlen += len - w;
 		set_handlers(itrm->sock_out, get_handler(itrm->sock_out, H_READ), (void (*)(void *))write_ev_queue, (void (*)(void *))free_trm, itrm);
@@ -88,33 +91,37 @@ unsigned char *init_seq = "\033[?1000h\033[?47h\0337";
 unsigned char *term_seq = "\033[2J\033[?1000l\033[?47l\0338\b \b";
 */
 
-unsigned char *init_seq = "\033)0\0337";
-unsigned char *init_seq_x_mouse = "\033[?1000h";
-unsigned char *init_seq_tw_mouse = "\033[?9h";
-unsigned char *term_seq = "\033[2J\0338\r \b";
-unsigned char *term_seq_x_mouse = "\033[?1000l";
-unsigned char *term_seq_tw_mouse = "\033[?9l";
+unsigned char init_seq[] = "\033)0\0337";
+unsigned char init_seq_x_mouse[] = "\033[?1000h";
+unsigned char init_seq_tw_mouse[] = "\033[?9h";
+unsigned char term_seq[] = "\033[2J\0338\r \b";
+unsigned char term_seq_x_mouse[] = "\033[?1000l\015       \015";
+unsigned char term_seq_tw_mouse[] = "\033[?9l";
 
 /*unsigned char *term_seq = "\033[2J\033[?1000l\0338\b \b";*/
 
-void send_init_sequence(int h,int flags)
+void send_init_sequence(int h, int flags)
 {
+	want_draw();
 	hard_write(h, init_seq, strlen(init_seq));
 	if (flags & USE_TWIN_MOUSE) {
 		hard_write(h, init_seq_tw_mouse, strlen(init_seq_tw_mouse));
-	} else if (flags & USE_X_MOUSE) {
+	} else {
 		hard_write(h, init_seq_x_mouse, strlen(init_seq_x_mouse));
 	}
+	done_draw();
 }
 
 void send_term_sequence(int h,int flags)
 {
+	want_draw();
 	hard_write(h, term_seq, strlen(term_seq));
 	if (flags & USE_TWIN_MOUSE) {
 		hard_write(h, term_seq_tw_mouse, strlen(term_seq_tw_mouse));
-	} else if (flags & USE_X_MOUSE) {
+	} else {
 		hard_write(h, term_seq_x_mouse, strlen(term_seq_x_mouse));
 	}
+	done_draw();
 }
 
 void resize_terminal()
@@ -124,14 +131,63 @@ void resize_terminal()
 	if (get_terminal_size(ditrm->std_out, &x, &y)) return;
 	ev.x = x;
 	ev.y = y;
-	queue_event(ditrm, (char *)&ev, sizeof(struct event));
+	queue_event(ditrm, (unsigned char *)&ev, sizeof(struct event));
+}
+
+int ttcgetattr(int fd, struct termios *t)
+{
+	int r;
+#ifdef SIGTTOU
+	interruptible_signal(SIGTTOU, 1);
+#endif
+#ifdef SIGTTIN
+	interruptible_signal(SIGTTIN, 1);
+#endif
+	r = tcgetattr(fd, t);
+#ifdef SIGTTOU
+	interruptible_signal(SIGTTOU, 0);
+#endif
+#ifdef SIGTTIN
+	interruptible_signal(SIGTTIN, 0);
+#endif
+	return r;
+}
+
+int ttcsetattr(int fd, int a, struct termios *t)
+{
+	int r;
+#ifdef SIGTTOU
+	interruptible_signal(SIGTTOU, 1);
+#endif
+#ifdef SIGTTIN
+	interruptible_signal(SIGTTIN, 1);
+#endif
+	r = tcsetattr(fd, a, t);
+#ifdef SIGTTOU
+	interruptible_signal(SIGTTOU, 0);
+#endif
+#ifdef SIGTTIN
+	interruptible_signal(SIGTTIN, 0);
+#endif
+	return r;
 }
 
 int setraw(int fd, struct termios *p)
 {
 	struct termios t;
 	memset(&t, 0, sizeof(struct termios));
-	if (tcgetattr(fd, &t)) return -1;
+	if (ttcgetattr(fd, &t)) {
+		/*fprintf(stderr, "getattr result %s\n", strerror(errno));*/
+		/* If the terminal was destroyed (the user logged off),
+		 * we fake success here so that we can destroy the terminal
+		 * later.
+		 *
+		 * Linux returns EIO
+		 * FreeBSD returns ENXIO
+		 */
+		if (errno == EIO || errno == ENXIO) return 0;
+		return -1;
+	}
 	if (p) memcpy(p, &t, sizeof(struct termios));
 	os_cfmakeraw(&t);
 	t.c_lflag |= ISIG;
@@ -139,7 +195,10 @@ int setraw(int fd, struct termios *p)
 	t.c_lflag |= TOSTOP;
 #endif
 	t.c_oflag |= OPOST;
-	if (tcsetattr(fd, TCSANOW, &t)) return -1;
+	if (ttcsetattr(fd, TCSANOW, &t)) {
+		/*fprintf(stderr, "setattr result %s\n", strerror(errno));*/
+		return -1;
+	}
 	return 0;
 }
 
@@ -154,7 +213,7 @@ void handle_trm(int std_in, int std_out, int sock_in, int sock_out, int ctl_in,
 		error("ERROR: could not get terminal size");
 		return;
 	}
-	if (!(itrm = mem_alloc(sizeof(struct itrm)))) return;
+	itrm = mem_alloc(sizeof(struct itrm));
 	ditrm = itrm;
 	itrm->std_in = std_in;
 	itrm->std_out = std_out;
@@ -172,48 +231,40 @@ void handle_trm(int std_in, int std_out, int sock_in, int sock_out, int ctl_in,
 	ev.x = x;
 	ev.y = y;
 	handle_terminal_resize(ctl_in, resize_terminal);
-	queue_event(itrm, (char *)&ev, sizeof(struct event));
+	queue_event(itrm, (unsigned char *)&ev, sizeof(struct event));
 	xwin = is_xterm() * ENV_XWIN + can_twterm() * ENV_TWIN + (!!getenv("STY")) * ENV_SCREEN + get_system_env();
 	itrm->flags = 0;
 	if (!(ts = getenv("TERM"))) ts = "";
 	if ((xwin & ENV_TWIN) && !strcmp(ts,"linux")) itrm->flags |= USE_TWIN_MOUSE;
-	if ((xwin & ENV_XWIN)) itrm->flags |= USE_X_MOUSE;
 	if (strlen(ts) >= MAX_TERM_LEN) queue_event(itrm, ts, MAX_TERM_LEN);
 	else {
 		unsigned char *mm;
 		int ll = MAX_TERM_LEN - strlen(ts);
 		queue_event(itrm, ts, strlen(ts));
-		if (!(mm = mem_alloc(ll))) {
-			free_trm(itrm);
-			return;
-		}
+		mm = mem_alloc(ll);
 		memset(mm, 0, ll);
 		queue_event(itrm, mm, ll);
 		mem_free(mm);
 	}
-	if (!(ts = get_cwd()) && !(ts = stracpy(""))) goto neni_pamet;
+	if (!(ts = get_cwd())) ts = stracpy("");
 	if (strlen(ts) >= MAX_CWD_LEN) queue_event(itrm, ts, MAX_CWD_LEN);
 	else {
 		unsigned char *mm;
 		int ll = MAX_CWD_LEN - strlen(ts);
 		queue_event(itrm, ts, strlen(ts));
-		if (!(mm = mem_alloc(ll))) {
-			free_trm(itrm);
-			return;
-		}
+		mm = mem_alloc(ll);
 		memset(mm, 0, ll);
 		queue_event(itrm, mm, ll);
 		mem_free(mm);
 	}
 	mem_free(ts);
-	neni_pamet:
-	queue_event(itrm, (char *)&xwin, sizeof(int));
-	queue_event(itrm, (char *)&init_len, sizeof(int));
-	queue_event(itrm, (char *)init_string, init_len);
-	itrm->mouse_h = handle_mouse(0, (void (*)(void *, unsigned char *, int))queue_event, itrm);
+	queue_event(itrm, (unsigned char *)&xwin, sizeof(int));
+	queue_event(itrm, (unsigned char *)&init_len, sizeof(int));
+	queue_event(itrm, (unsigned char *)init_string, init_len);
+	itrm->mouse_h = handle_mouse(0, (void (*)(void *, unsigned char *, int))mouse_queue_event, itrm);
 	itrm->orig_title = get_window_title();
 	set_window_title("Links");
-	send_init_sequence(std_out,itrm->flags);
+	send_init_sequence(std_out, itrm->flags);
 }
 
 void unblock_itrm_x(void *h)
@@ -221,23 +272,20 @@ void unblock_itrm_x(void *h)
 	close_handle(h);
 	if (!ditrm) return;
 	unblock_itrm(0);
-	resize_terminal();
 }
 
 int unblock_itrm(int fd)
 {
 	struct itrm *itrm = ditrm;
 	if (!itrm) return -1;
-	/*if (ditrm->sock_out != fd) {
-		internal("unblock_itrm: bad fd: %d", fd);
-		return -1;
-	}*/
 	if (itrm->ctl_in >= 0 && setraw(itrm->ctl_in, NULL)) return -1;
+	if (itrm->blocked != fd + 1) return -2;
 	itrm->blocked = 0;
-	send_init_sequence(itrm->std_out,itrm->flags);
+	send_init_sequence(itrm->std_out, itrm->flags);
 	set_handlers(itrm->std_in, (void (*)(void *))in_kbd, NULL, (void (*)(void *))free_trm, itrm);
 	handle_terminal_resize(itrm->ctl_in, resize_terminal);
 	unblock_stdin();
+	resize_terminal();
 	return 0;
 }
 
@@ -245,15 +293,12 @@ void block_itrm(int fd)
 {
 	struct itrm *itrm = ditrm;
 	if (!itrm) return;
-	/*if (ditrm->sock_out != fd) {
-		internal("block_itrm: bad fd: %d", fd);
-		return;
-	}*/
-	itrm->blocked = 1;
+	if (itrm->blocked) return;
+	itrm->blocked = fd + 1;
 	block_stdin();
 	unhandle_terminal_resize(itrm->ctl_in);
 	send_term_sequence(itrm->std_out,itrm->flags);
-	tcsetattr(itrm->ctl_in, TCSANOW, &itrm->t);
+	ttcsetattr(itrm->ctl_in, TCSANOW, &itrm->t);
 	set_handlers(itrm->std_in, NULL, NULL, (void (*)(void *))free_trm, itrm);
 }
 
@@ -264,7 +309,7 @@ void free_trm(struct itrm *itrm)
 	if (itrm->orig_title) mem_free(itrm->orig_title), itrm->orig_title = NULL;
 	unhandle_terminal_resize(itrm->ctl_in);
 	send_term_sequence(itrm->std_out,itrm->flags);
-	tcsetattr(itrm->ctl_in, TCSANOW, &itrm->t);
+	ttcsetattr(itrm->ctl_in, TCSANOW, &itrm->t);
 	if (itrm->mouse_h) unhandle_mouse(itrm->mouse_h);
 	set_handlers(itrm->std_in, NULL, NULL, NULL, NULL);
 	set_handlers(itrm->sock_in, NULL, NULL, NULL, NULL);
@@ -276,6 +321,11 @@ void free_trm(struct itrm *itrm)
 	if (itrm == ditrm) ditrm = NULL;
 }
 
+void fatal_tty_exit(void)
+{
+	if (ditrm) ttcsetattr(ditrm->ctl_in, TCSANOW, &ditrm->t);
+}
+
 void resize_terminal_x(unsigned char *text)
 {
 	int x, y;
@@ -302,20 +352,19 @@ void dispatch_special(unsigned char *text)
 
 unsigned char buf[OUT_BUF_SIZE];
 
-/*#define RD ({ char cc; if (p < c) cc = buf[p++]; else if ((dl = hard_read(itrm->sock_in, &cc, 1)) <= 0) {debug("%d %d", dl, errno);goto fr;} cc; })*/
-/*udefine RD ({ char cc; if (p < c) cc = buf[p++]; else if ((hard_read(itrm->sock_in, &cc, 1)) <= 0) goto fr; cc; })*/
 #define RD(xx) { unsigned char cc; if (p < c) cc = buf[p++]; else if ((hard_read(itrm->sock_in, &cc, 1)) <= 0) goto fr; xx = cc; }
 
 void in_sock(struct itrm *itrm)
 {
 	unsigned char *path, *delete;
 	int pl, dl;
-	char ch;
+	unsigned char ch;
 	int fg;
 	int c, i, p;
 	if ((c = read(itrm->sock_in, buf, OUT_BUF_SIZE)) <= 0) {
 		fr:
 		free_trm(itrm);
+		terminate = 1;
 		return;
 	}
 	qwerty:
@@ -360,20 +409,19 @@ void in_sock(struct itrm *itrm)
 			if (*delete) unlink(delete);
 			goto to_je_ale_hnus;
 		}
-		if (!(param = mem_alloc(strlen(path) + strlen(delete) + 3))) goto to_je_ale_hnus;
+		param = mem_alloc(strlen(path) + strlen(delete) + 3);
 		param[0] = fg;
 		strcpy(param + 1, path);
 		strcpy(param + 1 + strlen(path) + 1, delete);
-		if (fg == 1) block_itrm(itrm->ctl_in);
+		if (fg == 1) block_itrm(0);
 		if ((blockh = start_thread((void (*)(void *, int))exec_thread, param, strlen(path) + strlen(delete) + 3)) == -1) {
-			if (fg == 1) unblock_itrm(itrm->ctl_in);
+			if (fg == 1) unblock_itrm(0);
 			mem_free(param);
 			goto to_je_ale_hnus;
 		}
 		mem_free(param);
 		if (fg == 1) {
 			set_handlers(blockh, (void (*)(void *))unblock_itrm_x, NULL, (void (*)(void *))unblock_itrm_x, (void *)blockh);
-			/*block_itrm(itrm->ctl_in);*/
 		} else {
 			set_handlers(blockh, close_handle, NULL, close_handle, (void *)blockh);
 		}
@@ -388,24 +436,7 @@ void in_sock(struct itrm *itrm)
 
 int process_queue(struct itrm *);
 
-void kbd_timeout(struct itrm *itrm)
-{
-	struct event ev = {EV_KBD, KBD_ESC, 0, 0};
-	itrm->tm = -1;
-	if (can_read(itrm->std_in)) {
-		in_kbd(itrm);
-		return;
-	}
-	if (!itrm->qlen) {
-		internal("timeout on empty queue");
-		return;
-	}
-	queue_event(itrm, (char *)&ev, sizeof(struct event));
-	if (--itrm->qlen) memmove(itrm->kqueue, itrm->kqueue+1, itrm->qlen);
-	while (process_queue(itrm)) ;
-}
-
-int get_esc_code(char *str, int len, char *code, int *num, int *el)
+int get_esc_code(unsigned char *str, int len, unsigned char *code, int *num, int *el)
 {
 	int pos;
 	*num = 0;
@@ -420,95 +451,190 @@ int get_esc_code(char *str, int len, char *code, int *num, int *el)
 	return -1;
 }
 
+void kbd_timeout(struct itrm *itrm)
+{
+	struct event ev = {EV_KBD, KBD_ESC, 0, 0};
+	unsigned char code;
+	int num;
+	int len;
+	itrm->tm = -1;
+	if (can_read(itrm->std_in)) {
+		in_kbd(itrm);
+		return;
+	}
+	if (!itrm->qlen) {
+		internal("timeout on empty queue");
+		return;
+	}
+	queue_event(itrm, (unsigned char *)&ev, sizeof(struct event));
+	if (get_esc_code(itrm->kqueue, itrm->qlen, &code, &num, &len)) len = 1;
+	itrm->qlen -= len;
+	memmove(itrm->kqueue, itrm->kqueue + len, itrm->qlen);
+	while (process_queue(itrm)) ;
+}
+
 struct key {
 	int x, y;
 };
 
 struct key os2xtd[256] = {
 /* 0 */
-0,0,0,0,' ',KBD_CTRL,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,KBD_BS,KBD_ALT,0,0,
+{0,0},{0,0},{' ',KBD_CTRL},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{KBD_BS,KBD_ALT},{0,0},
 /* 16 */
-'Q',KBD_ALT,'W',KBD_ALT,'E',KBD_ALT,'R',KBD_ALT,'T',KBD_ALT,'Y',KBD_ALT,'U',KBD_ALT,'I',KBD_ALT,
+{'Q',KBD_ALT},{'W',KBD_ALT},{'E',KBD_ALT},{'R',KBD_ALT},{'T',KBD_ALT},{'Y',KBD_ALT},{'U',KBD_ALT},{'I',KBD_ALT},
 /* 24 */
-'O',KBD_ALT,'P',KBD_ALT,'[',KBD_ALT,']',KBD_ALT,KBD_ENTER,KBD_ALT,0,0,'A',KBD_ALT,'S',KBD_ALT,
+{'O',KBD_ALT},{'P',KBD_ALT},{'[',KBD_ALT},{']',KBD_ALT},{KBD_ENTER,KBD_ALT},{0,0},{'A',KBD_ALT},{'S',KBD_ALT},
 /* 32 */
-'D',KBD_ALT,'F',KBD_ALT,'G',KBD_ALT,'H',KBD_ALT,'J',KBD_ALT,'K',KBD_ALT,'L',KBD_ALT,';',KBD_ALT,
+{'D',KBD_ALT},{'F',KBD_ALT},{'G',KBD_ALT},{'H',KBD_ALT},{'J',KBD_ALT},{'K',KBD_ALT},{'L',KBD_ALT},{';',KBD_ALT},
 /* 40 */
-'\'',KBD_ALT,'`',KBD_ALT,0,0,'\\',KBD_ALT,'Z',KBD_ALT,'X',KBD_ALT,'C',KBD_ALT,'V',KBD_ALT,
+{'\'',KBD_ALT},{'`',KBD_ALT},{0,0},{'\\',KBD_ALT},{'Z',KBD_ALT},{'X',KBD_ALT},{'C',KBD_ALT},{'V',KBD_ALT},
 /* 48 */
-'B',KBD_ALT,'N',KBD_ALT,'M',KBD_ALT,',',KBD_ALT,'.',KBD_ALT,'/',KBD_ALT,0,0,'*',KBD_ALT,
+{'B',KBD_ALT},{'N',KBD_ALT},{'M',KBD_ALT},{',',KBD_ALT},{'.',KBD_ALT},{'/',KBD_ALT},{0,0},{'*',KBD_ALT},
 /* 56 */
-0,0,' ',KBD_ALT,0,0,KBD_F1,0,KBD_F2,0,KBD_F3,0,KBD_F4,0,KBD_F5,0,
+{0,0},{' ',KBD_ALT},{0,0},{KBD_F1,0},{KBD_F2,0},{KBD_F3,0},{KBD_F4,0},{KBD_F5,0},
 /* 64 */
-KBD_F6,0,KBD_F7,0,KBD_F8,0,KBD_F9,0,KBD_F10,0,0,0,0,0,KBD_HOME,0,
+{KBD_F6,0},{KBD_F7,0},{KBD_F8,0},{KBD_F9,0},{KBD_F10,0},{0,0},{0,0},{KBD_HOME,0},
 /* 72 */
-KBD_UP,0,KBD_PAGE_UP,0,'-',KBD_ALT,KBD_LEFT,0,'5',0,KBD_RIGHT,0,'+',KBD_ALT,KBD_END,0,
+{KBD_UP,0},{KBD_PAGE_UP,0},{'-',KBD_ALT},{KBD_LEFT,0},{'5',0},{KBD_RIGHT,0},{'+',KBD_ALT},{KBD_END,0},
 /* 80 */
-KBD_DOWN,0,KBD_PAGE_DOWN,0,KBD_INS,0,KBD_DEL,0,0,0,0,0,0,0,0,0,
+{KBD_DOWN,0},{KBD_PAGE_DOWN,0},{KBD_INS,0},{KBD_DEL,0},{0,0},{0,0},{0,0},{0,0},
 /* 88 */
-0,0,0,0,0,0,0,0,0,0,0,0,KBD_F1,KBD_CTRL,KBD_F2,KBD_CTRL,
+{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{KBD_F1,KBD_CTRL},{KBD_F2,KBD_CTRL},
 /* 96 */
-KBD_F3,KBD_CTRL,KBD_F4,KBD_CTRL,KBD_F5,KBD_CTRL,KBD_F6,KBD_CTRL,KBD_F7,KBD_CTRL,KBD_F8,KBD_CTRL,KBD_F9,KBD_CTRL,KBD_F10,KBD_CTRL,
+{KBD_F3,KBD_CTRL},{KBD_F4,KBD_CTRL},{KBD_F5,KBD_CTRL},{KBD_F6,KBD_CTRL},{KBD_F7,KBD_CTRL},{KBD_F8,KBD_CTRL},{KBD_F9,KBD_CTRL},{KBD_F10,KBD_CTRL},
 /* 104 */
-KBD_F1,KBD_ALT,KBD_F2,KBD_ALT,KBD_F3,KBD_ALT,KBD_F4,KBD_ALT,KBD_F5,KBD_ALT,KBD_F6,KBD_ALT,KBD_F7,KBD_ALT,KBD_F8,KBD_ALT,
+{KBD_F1,KBD_ALT},{KBD_F2,KBD_ALT},{KBD_F3,KBD_ALT},{KBD_F4,KBD_ALT},{KBD_F5,KBD_ALT},{KBD_F6,KBD_ALT},{KBD_F7,KBD_ALT},{KBD_F8,KBD_ALT},
 /* 112 */
-KBD_F9,KBD_ALT,KBD_F10,KBD_ALT,0,0,KBD_LEFT,KBD_CTRL,KBD_RIGHT,KBD_CTRL,KBD_END,KBD_CTRL,KBD_PAGE_DOWN,KBD_CTRL,KBD_HOME,KBD_CTRL,
+{KBD_F9,KBD_ALT},{KBD_F10,KBD_ALT},{0,0},{KBD_LEFT,KBD_CTRL},{KBD_RIGHT,KBD_CTRL},{KBD_END,KBD_CTRL},{KBD_PAGE_DOWN,KBD_CTRL},{KBD_HOME,KBD_CTRL},
 /* 120 */
-'1',KBD_ALT,'2',KBD_ALT,'3',KBD_ALT,'4',KBD_ALT,'5',KBD_ALT,'6',KBD_ALT,'7',KBD_ALT,'8',KBD_ALT,
+{'1',KBD_ALT},{'2',KBD_ALT},{'3',KBD_ALT},{'4',KBD_ALT},{'5',KBD_ALT},{'6',KBD_ALT},{'7',KBD_ALT},{'8',KBD_ALT},
 /* 128 */
-'9',KBD_ALT,'0',KBD_ALT,'-',KBD_ALT,'=',KBD_ALT,KBD_PAGE_UP,KBD_CTRL,KBD_F11,0,KBD_F12,0,0,0,
+{'9',KBD_ALT},{'0',KBD_ALT},{'-',KBD_ALT},{'=',KBD_ALT},{KBD_PAGE_UP,KBD_CTRL},{KBD_F11,0},{KBD_F12,0},{0,0},
 /* 136 */
-0,0,KBD_F11,KBD_CTRL,KBD_F12,KBD_CTRL,KBD_F11,KBD_ALT,KBD_F12,KBD_ALT,KBD_UP,KBD_CTRL,'-',KBD_CTRL,'5',KBD_CTRL,
+{0,0},{KBD_F11,KBD_CTRL},{KBD_F12,KBD_CTRL},{KBD_F11,KBD_ALT},{KBD_F12,KBD_ALT},{KBD_UP,KBD_CTRL},{'-',KBD_CTRL},{'5',KBD_CTRL},
 /* 144 */
-'+',KBD_CTRL,KBD_DOWN,KBD_CTRL,KBD_INS,KBD_CTRL,KBD_DEL,KBD_CTRL,KBD_TAB,KBD_CTRL,0,0,0,0,KBD_HOME,KBD_ALT,
+{'+',KBD_CTRL},{KBD_DOWN,KBD_CTRL},{KBD_INS,KBD_CTRL},{KBD_DEL,KBD_CTRL},{KBD_TAB,KBD_CTRL},{0,0},{0,0},{KBD_HOME,KBD_ALT},
 /* 152 */
-KBD_UP,KBD_ALT,KBD_PAGE_UP,KBD_ALT,0,0,KBD_LEFT,KBD_ALT,0,0,KBD_RIGHT,KBD_ALT,0,0,KBD_END,KBD_ALT,
+{KBD_UP,KBD_ALT},{KBD_PAGE_UP,KBD_ALT},{0,0},{KBD_LEFT,KBD_ALT},{0,0},{KBD_RIGHT,KBD_ALT},{0,0},{KBD_END,KBD_ALT},
 /* 160 */
-KBD_DOWN,KBD_ALT,KBD_PAGE_DOWN,KBD_ALT,KBD_INS,KBD_ALT,KBD_DEL,KBD_ALT,0,0,KBD_TAB,KBD_ALT,KBD_ENTER,KBD_ALT,0,0,
+{KBD_DOWN,KBD_ALT},{KBD_PAGE_DOWN,KBD_ALT},{KBD_INS,KBD_ALT},{KBD_DEL,KBD_ALT},{0,0},{KBD_TAB,KBD_ALT},{KBD_ENTER,KBD_ALT},{0,0},
 /* 168 */
-0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
 /* 176 */
-0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
 /* 192 */
-0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
 /* 208 */
-0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
 /* 224 */
-0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
 /* 240 */
-0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
+{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},{0,0},
 /* 256 */
 };
 
 int xterm_button = -1;
 
+int is_interix(void)
+{
+#ifdef INTERIX
+	return 1;
+#else
+	unsigned char *term = getenv("TERM");
+	return term && !strncmp(term, "interix", 7);
+#endif
+}
+
+int is_ibm(void)
+{
+	unsigned char *term = getenv("TERM");
+	return term && !strncmp(term, "ibm", 3);
+}
+
 int process_queue(struct itrm *itrm)
 {
-	struct event ev = {EV_KBD, -1, 0, 0};
+	struct event ev = { EV_KBD, -1, 0, 0 };
 	int el = 0;
 	if (!itrm->qlen) goto end;
+	/*{
+		int i;
+		fprintf(stderr, "queue:");
+		for (i = 0; i < itrm->qlen; i++) {
+			fprintf(stderr, " %d", itrm->kqueue[i]);
+			if (itrm->kqueue[i] >= 32) fprintf(stderr, "[%c]", itrm->kqueue[i]);
+		}
+		fprintf(stderr, ".\n");
+	}*/
 	if (itrm->kqueue[0] == '\033') {
 		if (itrm->qlen < 2) goto ret;
 		if (itrm->kqueue[1] == '[' || itrm->kqueue[1] == 'O') {
-			char c;
+			unsigned char c = 0;
 			int v;
 			if (itrm->qlen >= 4 && itrm->kqueue[2] == '[') {
 				if (itrm->kqueue[3] < 'A' || itrm->kqueue[3] > 'L') goto ret;
-				ev.x = KBD_F1 - (itrm->kqueue[3] - 'A');
+				ev.x = KBD_F1 + (itrm->kqueue[3] - 'A');
 				el = 4;
 			} else if (get_esc_code(itrm->kqueue, itrm->qlen, &c, &v, &el)) goto ret;
 			else switch (c) {
+				case 'L':
+				case '@': ev.x = KBD_INS; break;
 				case 'A': ev.x = KBD_UP; break;
 				case 'B': ev.x = KBD_DOWN; break;
 				case 'C': ev.x = KBD_RIGHT; break;
 				case 'D': ev.x = KBD_LEFT; break;
 				case 'F':
+				case 'K':
 				case 'e': ev.x = KBD_END; break;
 				case 'H':
 				case 0: ev.x = KBD_HOME; break;
+				case 'V':
 				case 'I': ev.x = KBD_PAGE_UP; break;
+				case 'U':
+					if (is_interix()) {
+						ev.x = KBD_END;
+						break;
+					}
 				case 'G': ev.x = KBD_PAGE_DOWN; break;
+				case 'P':
+					if (is_ibm()) {
+						ev.x = KBD_DEL;
+						break;
+					}
+					ev.x = KBD_F1; break;
+				case 'Q': ev.x = KBD_F2; break;
+				case 'S':
+					if (is_interix()) {
+						ev.x = KBD_PAGE_UP;
+						break;
+					}
+					ev.x = KBD_F4; break;
+				case 'T':
+					if (is_interix()) {
+						ev.x = KBD_PAGE_DOWN;
+						break;
+					}
+					ev.x = KBD_F5; break;
+				case 'W': ev.x = KBD_F8; break;
+				case 'X': ev.x = KBD_F9; break;
+				case 'Y': ev.x = KBD_F11; break;
 
+				case 'q': switch (v) {
+					case 139: ev.x = KBD_INS; break;
+					case 146: ev.x = KBD_END; break;
+					case 150: ev.x = KBD_PAGE_UP; break;
+					case 154: ev.x = KBD_PAGE_DOWN; break;
+					default: if (v >= 1 && v <= 48) {
+							int fn = (v - 1) % 12;
+							int mod = (v - 1) / 12;
+							ev.x = KBD_F1 + fn;
+							if (mod == 1)
+								ev.y |= KBD_SHIFT;
+							if (mod == 2)
+								ev.y |= KBD_CTRL;
+							if (mod == 3)
+								ev.y |= KBD_ALT;
+						} break;
+					} break;
 				case 'z': switch (v) {
 					case 247: ev.x = KBD_INS; break;
 					case 214: ev.x = KBD_HOME; break;
@@ -535,7 +661,8 @@ int process_queue(struct itrm *itrm)
 					case 24: ev.x = KBD_F12; break;
 					} break;
 				case 'R':
-						  resize_terminal (); break ;
+					resize_terminal();
+					break ;
 				case 'M': if (itrm->qlen - el < 3) goto ret;
 					if (v == 5) {
 						if (xterm_button == -1) xterm_button = 0; /* */
@@ -572,15 +699,48 @@ int process_queue(struct itrm *itrm)
 			}
 		} else {
 			el = 2;
-			if (itrm->kqueue[1] >= ' ') {
-				ev.x = itrm->kqueue[1];
-				ev.y = KBD_ALT;
-				goto l2;
-			}
 			if (itrm->kqueue[1] == '\033') {
 				if (itrm->qlen >= 3 && (itrm->kqueue[2] == '[' || itrm->kqueue[2] == 'O')) el = 1;
 				ev.x = KBD_ESC;
 				goto l2;
+			} else if (itrm->kqueue[1] == 127) {
+				ev.x = KBD_DEL;
+				ev.y = 0;
+				goto l2;
+			} else if (itrm->kqueue[1] == 'F' && is_interix()) {
+				int num;
+				if (itrm->qlen < 3) goto ret;
+				if (itrm->kqueue[2] >= '1' && itrm->kqueue[2] <= '9') {
+					num = itrm->kqueue[2] - '1';
+				} else if (itrm->kqueue[2] >= 'A' && itrm->kqueue[2] <= 'Z') {
+					num = itrm->kqueue[2] - 'A' + 9;
+				} else if (itrm->kqueue[2] >= 'a' && itrm->kqueue[2] <= 'k') {
+					num = itrm->kqueue[2] - 'a' + 35;
+				} else if (itrm->kqueue[2] >= 'm' && itrm->kqueue[2] <= 'z') {
+					num = itrm->kqueue[2] - 'a' + 34;
+				} else goto do_alt;
+				if (num < 12) {
+					ev.x = KBD_F1 + num;
+				} else if (num < 24) {
+					ev.x = KBD_F1 + (num - 12);
+					ev.y |= KBD_SHIFT;
+				} else if (num < 36) {
+					ev.x = KBD_F1 + (num - 24);
+					ev.y |= KBD_ALT;
+				} else if (num < 48) {
+					ev.x = KBD_F1 + (num - 36);
+					ev.y |= KBD_CTRL;
+				} else {
+					ev.x = KBD_F1 + (num - 48);
+					ev.y |= KBD_SHIFT | KBD_CTRL;
+				}
+				el = 3;
+				goto l1;
+			} else {
+do_alt:
+				ev.x = itrm->kqueue[1];
+				ev.y |= KBD_ALT;
+				goto l2;
 			}
 		}
 		goto l1;
@@ -602,12 +762,15 @@ int process_queue(struct itrm *itrm)
 	if (ev.x == 6) ev.x = KBD_PAGE_DOWN;*/
 	if (ev.x == 8) ev.x = KBD_BS;
 	if (ev.x == 9) ev.x = KBD_TAB;
-	if (ev.x == 10) ev.x = KBD_ENTER, ev.y = KBD_CTRL;
+	if (ev.x == 10) ev.x = KBD_ENTER/*, ev.y = KBD_CTRL*/;
 	if (ev.x == 13) ev.x = KBD_ENTER;
-	if (ev.x == 127) ev.x = KBD_BS;
-	if (ev.x < ' ') {
+	if (ev.x == 127) {
+		if (is_interix()) ev.x = KBD_DEL;
+		else ev.x = KBD_BS;
+	}
+	if (ev.x >= 0 && ev.x < ' ') {
 		ev.x += 'A' - 1;
-		ev.y = KBD_CTRL;
+		ev.y |= KBD_CTRL;
 	}
 	l1:
 	if (itrm->qlen < el) {
@@ -615,14 +778,14 @@ int process_queue(struct itrm *itrm)
 		itrm->qlen = el;
 	}
 	if (ev.x != -1) {
-		queue_event(itrm, (char *)&ev, sizeof(struct event));
+		queue_event(itrm, (unsigned char *)&ev, sizeof(struct event));
 		memmove(itrm->kqueue, itrm->kqueue + el, itrm->qlen -= el);
 	} else {
 		/*printf("%d %d\n", itrm->qlen, el);fflush(stdout);*/
 		memmove(itrm->kqueue, itrm->kqueue + el, itrm->qlen -= el);
 	}
 	end:
-	if (itrm->qlen < IN_BUF_SIZE) set_handlers(itrm->std_in, (void (*)(void *))in_kbd, NULL, (void (*)(void *))free_trm, itrm);
+	if (itrm->qlen < IN_BUF_SIZE && !itrm->blocked) set_handlers(itrm->std_in, (void (*)(void *))in_kbd, NULL, (void (*)(void *))free_trm, itrm);
 	return el;
 	ret:
 	itrm->tm = install_timer(ESC_TIMEOUT, (void (*)(void *))kbd_timeout, itrm);
@@ -640,7 +803,9 @@ void in_kbd(struct itrm *itrm)
 		return;
 	}
 	if ((r = read(itrm->std_in, itrm->kqueue + itrm->qlen, IN_BUF_SIZE - itrm->qlen)) <= 0) {
-		free_trm(itrm);
+		struct event ev = { EV_ABORT, 0, 0, 0 };
+		set_handlers(itrm->std_in, NULL, NULL, (void (*)(void *))free_trm, itrm);
+		queue_event(itrm, (unsigned char *)&ev, sizeof(struct event));
 		return;
 	}
 	if ((itrm->qlen += r) > IN_BUF_SIZE) {
