@@ -39,29 +39,36 @@ void shutdown_trans()
 		}
 }
 
+static inline int is_direct_text(unsigned char *text)
+{
+/* Do not compare to dummyarray directly - thwart some misoptimizations */
+	unsigned char * volatile dm = dummyarray;
+	return !(text >= dm && text < dm + T__N_TEXTS);
+}
+
 unsigned char *get_text_translation(unsigned char *text, struct terminal *term)
 {
 	unsigned char **current_tra;
 	struct conv_table *conv_table;
 	unsigned char *trn;
-	if (text < dummyarray || text > dummyarray + T__N_TEXTS) return text;
-	if ((current_tra = translation_array[current_language][term->spec->charset])) {
+	int ch = term ? term->spec->charset : 0;
+	if (is_direct_text(text)) return text;
+	if ((current_tra = translation_array[current_language][ch])) {
 		unsigned char *tt;
 		if ((trn = current_tra[text - dummyarray])) return trn;
 		tr:
 		if (!(tt = translations[current_language].t[text - dummyarray].name)) {
 			trn = stracpy(translation_english[text - dummyarray].name);
 		} else {
-			conv_table = get_translation_table(current_lang_charset, term->spec->charset);
+			conv_table = get_translation_table(current_lang_charset, ch);
 			trn = convert_string(conv_table, tt, strlen(tt));
 		}
 		current_tra[text - dummyarray] = trn;
 	} else {
-		if (current_lang_charset && term->spec->charset != current_lang_charset) {
-			if ((current_tra = translation_array[current_language][term->spec->charset] = mem_alloc(sizeof (unsigned char **) * T__N_TEXTS))) {
-				memset(current_tra, 0, sizeof (unsigned char **) * T__N_TEXTS);
-				goto tr;
-			}
+		if (current_lang_charset && ch != current_lang_charset) {
+			current_tra = translation_array[current_language][ch] = mem_alloc(sizeof (unsigned char **) * T__N_TEXTS);
+			memset(current_tra, 0, sizeof (unsigned char **) * T__N_TEXTS);
+			goto tr;
 		}
 		if (!(trn = translations[current_language].t[text - dummyarray].name)) {
 			trn = translations[current_language].t[text - dummyarray].name = translation_english[text - dummyarray].name;	/* modifying translation structure */
@@ -72,7 +79,7 @@ unsigned char *get_text_translation(unsigned char *text, struct terminal *term)
 
 unsigned char *get_english_translation(unsigned char *text)
 {
-	if (text < dummyarray || text > dummyarray + T__N_TEXTS) return text;
+	if (is_direct_text(text)) return text;
 	return translation_english[text - dummyarray].name;
 }
 
