@@ -2,6 +2,18 @@
 #define __EXTENSIONS__
 #endif
 
+#ifndef _LARGEFILE_SOURCE
+#define _LARGEFILE_SOURCE	1
+#endif
+
+#ifndef _ALL_SOURCE
+#define _ALL_SOURCE             1
+#endif
+
+#ifndef _FILE_OFFSET_BITS
+#define _FILE_OFFSET_BITS	64
+#endif
+
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
@@ -23,6 +35,14 @@
 #endif
 #include <sys/types.h>
 
+#ifndef __USE_XOPEN
+#define U_X
+#define __USE_XOPEN
+#endif
+#ifndef _XOPEN_SOURCE
+#define X_S
+#define _XOPEN_SOURCE	5	/* The 5 is a kludge to get a strptime() prototype in NetBSD */
+#endif
 #ifdef TIME_WITH_SYS_TIME
 #ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
@@ -37,6 +57,13 @@
 #include <time.h>
 #endif
 #endif
+#ifdef X_S
+#undef _XOPEN_SOURCE
+#endif
+#ifdef U_X
+#undef __USE_XOPEN
+#endif
+
 
 #include <sys/stat.h>
 #ifdef HAVE_FCNTL_H
@@ -61,12 +88,21 @@
 #ifdef HAVE_SYS_CYGWIN_H
 #include <sys/cygwin.h>
 #endif
+#ifdef HAVE_INTERIX_INTERIX_H
+#include <interix/interix.h>
+#endif
 #ifdef HAVE_IO_H
 #include <io.h>
 #endif
 #ifdef HAVE_SYS_UTSNAME_H
 #include <sys/utsname.h>
 #endif
+#ifdef HAVE_PWD_H
+#include <pwd.h>
+#endif
+#ifdef HAVE_GRP_H
+#include <grp.h>
+#endif
 /*
 #ifdef HAVE_WAIT_H
 #include <wait.h>
@@ -86,47 +122,117 @@
 #ifdef HAVE_NETINET_IP_H
 #include <netinet/ip.h>
 #endif
+#ifdef HAVE_ARPA_INET_H
+#include <arpa/inet.h>
+#endif
 #include <utime.h>
 
+#include <termios.h>
+
 #ifdef HAVE_LONG_LONG
 #define longlong long long
 #else
-#define longlong long
+#define longlong double
 #endif
 
-#include <termios.h>
+#if defined(__INTERIX) && defined(HAVE_STRTOQ)
+extern quad_t
+#if defined(__cdecl) || defined(_MSC_VER)
+__cdecl
+#endif
+strtoq(const char *, char **, int);
+#endif
+
+#if defined(__hpux) && defined(__LP64__)
+#undef HAVE_SOCKLEN_T
+#endif
+
+#ifndef HAVE_SOCKLEN_T
+#define socklen_t int
+#endif
+
+#ifndef PF_INET
+#define PF_INET AF_INET
+#endif
+#ifndef PF_UNIX
+#define PF_UNIX AF_UNIX
+#endif
 
 #ifdef HAVE_SSL
 #include <openssl/ssl.h>
+#include <openssl/rand.h>
 #endif
 
 #include "os_depx.h"
 
 #include "setup.h"
 
-#define LINKS_COPYRIGHT "(C) 1999 - 2003 Mikulas Patocka"
-#define LINKS_COPYRIGHT_8859_1 "(C) 1999 - 2003 Mikuls Patocka"
-#define LINKS_COPYRIGHT_8859_2 "(C) 1999 - 2003 Mikul Patoka"
+#define LINKS_COPYRIGHT "(C) 1999 - 2011 Mikulas Patocka"
+#define LINKS_COPYRIGHT_8859_1 "(C) 1999 - 2011 Mikul\341s Patocka"
+#define LINKS_COPYRIGHT_8859_2 "(C) 1999 - 2011 Mikul\341\271 Pato\350ka"
 
+#ifdef HAVE_POINTER_COMPARISON_BUG
+#define DUMMY ((void *)1L)
+#else
 #define DUMMY ((void *)-1L)
+#endif
 
-/*
-#if defined(HAVE_GETPID) && defined(HAVE_SETPGID) && defined(HAVE_GETPGID)
-#define USE_PGRP
-#ifndef HAVE_SETPGRP
-static inline int setpgrp()
-{
-	return getpgid(0);
-}
+#define RET_OK		0
+#define RET_ERROR	1
+#define RET_SIGNAL	2
+#define RET_SYNTAX	3
+#define RET_FATAL	4
+
+#ifndef HAVE_MEMMOVE
+void *memmove(void *, const void *, size_t);
 #endif
-#ifndef HAVE_GETPGRP
-static inline int getpgrp()
-{
-	return setpgid(0,0);
-}
+#ifndef HAVE_RAISE
+int raise(int);
 #endif
+#ifndef HAVE_STRTOUL
+unsigned long strtoul(const char *, char **, int);
+#endif
+#ifndef HAVE_STRERROR
+char *strerror(int);
+#endif
+#ifndef HAVE_GETTIMEOFDAY
+struct timeval {
+	long tv_sec;
+	long tv_usec;
+};
+struct timezone {
+	int tz_minuteswest;
+	int tz_dsttime;
+};
+int gettimeofday(struct timeval *tv, struct timezone *tz);
+#endif
+#ifndef HAVE_STRLEN
+size_t strlen(const char *s);
+#endif
+#ifndef HAVE_STRCPY
+char *strcpy(char *dst, const char *src);
+#endif
+#ifndef HAVE_STRCHR
+char *strchr(const char *s, int c);
+#endif
+#ifndef HAVE_STRRCHR
+char *strrchr(const char *s, int c);
+#endif
+#ifndef HAVE_STRCMP
+int strcmp(const char *s1, const char *s2);
+#endif
+#ifndef HAVE_STRNCMP
+int strncmp(const char *s1, const char *s2, size_t n);
+#endif
+#ifndef HAVE_STRCSPN
+size_t strcspn(const char *s, const char *reject);
+#endif
+#ifndef HAVE_STRSTR
+char *strstr(const char *haystack, const char *needle);
+#endif
+#ifndef HAVE_TEMPNAM
+char *tempnam(const char *dir, const char *pfx);
 #endif
-*/
 
 #define option option_dirty_workaround_for_name_clash_with_include_on_cygwin
 #define table table_dirty_workaround_for_name_clash_with_libraries_on_macos
@@ -142,7 +248,7 @@ void int_error(unsigned char *, ...);
 extern int errline;
 extern unsigned char *errfile;
 
-#define internal errfile = __FILE__, errline = __LINE__, int_error
+#define internal while (1) errfile = __FILE__, errline = __LINE__, int_error
 #define debug errfile = __FILE__, errline = __LINE__, debug_msg
 
 #ifdef SPECIAL_MALLOC
@@ -165,6 +271,8 @@ void *sp_realloc(void *, size_t);
 
 /* inline */
 
+void fatal_tty_exit(void);
+
 #ifdef LEAK_DEBUG
 
 extern long mem_amount;
@@ -172,23 +280,32 @@ extern long last_mem_amount;
 
 #ifndef LEAK_DEBUG_LIST
 struct alloc_header {
-	int size;
+	size_t size;
 };
 #else
 struct alloc_header {
 	struct alloc_header *next;
 	struct alloc_header *prev;
-	int size;
+	size_t size;
 	int line;
 	unsigned char *file;
 	unsigned char *comment;
 };
 #endif
 
-#define L_D_S ((sizeof(struct alloc_header) + 7) & ~7)
+#define L_D_S ((sizeof(struct alloc_header) + 15) & ~15)
 
 #endif
 
+#define overalloc()							\
+do {									\
+	error("ERROR: attempting to allocate too large block at %s:%d", __FILE__, __LINE__);\
+	fatal_tty_exit();						\
+	exit(RET_FATAL);						\
+} while (1)	/* while (1) is not a type --- it's here to allow the
+	compiler that doesn't know that exit doesn't return to do better
+	optimizations */
+
 #ifdef LEAK_DEBUG
 
 void *debug_mem_alloc(unsigned char *, int, size_t);
@@ -206,8 +323,11 @@ static inline void *mem_alloc(size_t size)
 {
 	void *p;
 	if (!size) return DUMMY;
+	if (size > MAXINT) overalloc();
 	if (!(p = malloc(size))) {
-		error("ERROR: out of memory (malloc returned NULL)\n");
+		error("ERROR: out of memory (malloc returned NULL)");
+		fatal_tty_exit();
+		exit(RET_FATAL);
 		return NULL;
 	}
 	return p;
@@ -234,8 +354,11 @@ static inline void *mem_realloc(void *p, size_t size)
 		mem_free(p);
 		return DUMMY;
 	}
+	if (size > MAXINT) overalloc();
 	if (!(p = realloc(p, size))) {
-		error("ERROR: out of memory (realloc returned NULL)\n");
+		error("ERROR: out of memory (realloc returned NULL)");
+		fatal_tty_exit();
+		exit(RET_FATAL);
 		return NULL;
 	}
 	return p;
@@ -250,7 +373,7 @@ static inline void set_mem_comment(void *p, unsigned char *c, int l) {}
 
 static inline unsigned char upcase(unsigned char a)
 {
-	if (a>='a' && a<='z') a -= 0x20;
+	if (a >= 'a' && a <= 'z') a -= 0x20;
 	return a;
 }
 
@@ -270,13 +393,12 @@ static inline int cmpbeg(unsigned char *str, unsigned char *b)
 
 #if !(defined(LEAK_DEBUG) && defined(LEAK_DEBUG_LIST))
 
-static inline unsigned char *memacpy(const unsigned char *src, int len)
+static inline unsigned char *memacpy(const unsigned char *src, size_t len)
 {
 	unsigned char *m;
-	if ((m = mem_alloc(len + 1))) {
-		memcpy(m, src, len);
-		m[len] = 0;
-	}
+	m = mem_alloc(len + 1);
+	memcpy(m, src, len);
+	m[len] = 0;
 	return m;
 }
 
@@ -287,13 +409,12 @@ static inline unsigned char *stracpy(const unsigned char *src)
 
 #else
 
-static inline unsigned char *debug_memacpy(unsigned char *f, int l, unsigned char *src, int len)
+static inline unsigned char *debug_memacpy(unsigned char *f, int l, unsigned char *src, size_t len)
 {
 	unsigned char *m;
-	if ((m = debug_mem_alloc(f, l, len + 1))) {
-		memcpy(m, src, len);
-		m[len] = 0;
-	}
+	m = debug_mem_alloc(f, l, len + 1);
+	memcpy(m, src, len);
+	m[len] = 0;
 	return m;
 }
 
@@ -308,16 +429,16 @@ static inline unsigned char *debug_stracpy(unsigned char *f, int l, unsigned cha
 
 #endif
 
-static inline int snprint(unsigned char *s, int n, unsigned num)
+static inline int snprint(unsigned char *s, int n, off_t num)
 {
-	int q = 1;
+	off_t q = 1;
 	while (q <= num / 10) q *= 10;
 	while (n-- > 1 && q) *(s++) = num / q + '0', num %= q, q /= 10;
 	*s = 0;
 	return !!q;
 }
 
-static inline int snzprint(unsigned char *s, int n, int num)
+static inline int snzprint(unsigned char *s, int n, off_t num)
 {
 	if (n > 1 && num < 0) *(s++) = '-', num = -num, n--;
 	return snprint(s, n, num);
@@ -326,51 +447,59 @@ static inline int snzprint(unsigned char *s, int n, int num)
 static inline void add_to_strn(unsigned char **s, unsigned char *a)
 {
 	unsigned char *p;
-	if (!(p = mem_realloc(*s, strlen(*s) + strlen(a) + 1))) return;
+	size_t l1 = strlen(*s), l2 = strlen(a);
+	if (((l1 | l2) | (l1 + l2 + 1)) > MAXINT) overalloc();
+	p = mem_realloc(*s, l1 + l2 + 1);
 	strcat(p, a);
 	*s = p;
 }
 
-#define ALLOC_GR	0x100		/* must be power of 2 */
+#define ALLOC_GR	0x040		/* must be power of 2 */
 
 #define init_str() init_str_x(__FILE__, __LINE__)
 
 static inline unsigned char *init_str_x(unsigned char *file, int line)
 {
-	unsigned char *p;
-	if ((p = debug_mem_alloc(file, line, ALLOC_GR))) *p = 0;
-	return p;
+	unsigned char *s = debug_mem_alloc(file, line, ALLOC_GR);
+	*s = 0;
+	return s;
 }
 
 static inline void add_to_str(unsigned char **s, int *l, unsigned char *a)
 {
 	unsigned char *p;
-	int ll = strlen(a);
-	if ((*l & ~(ALLOC_GR - 1)) != ((*l + ll) & ~(ALLOC_GR - 1)) &&
-	   (!(p = mem_realloc(*s, (*l + ll + ALLOC_GR) & ~(ALLOC_GR - 1))) ||
-	   !(*s = p))) return;
-	strcpy(*s + *l, a); *l += ll;
+	size_t ll = strlen(a);
+	if (ll > MAXINT || *l + ll > MAXINT) overalloc();
+	p = *s;
+	if (((size_t)*l & ~(ALLOC_GR - 1)) != ((*l + ll) & ~(ALLOC_GR - 1))) {
+		p = *s = mem_realloc(*s, (*l + ll + ALLOC_GR) & ~(ALLOC_GR - 1));
+	}
+	strcpy(p + *l, a); *l += ll;
 }
 
 static inline void add_bytes_to_str(unsigned char **s, int *l, unsigned char *a, int ll)
 {
 	unsigned char *p;
-	if ((*l & ~(ALLOC_GR - 1)) != ((*l + ll) & ~(ALLOC_GR - 1)) &&
-	   (!(p = mem_realloc(*s, (*l + ll + ALLOC_GR) & ~(ALLOC_GR - 1))) ||
-	   !(*s = p))) return;
-	memcpy(*s + *l, a, ll); (*s)[*l += ll] = 0;
+	if (*l + ll < 0) overalloc();
+	p = *s;
+	if ((*l & ~(ALLOC_GR - 1)) != ((*l + ll) & ~(ALLOC_GR - 1))) {
+		p = *s = mem_realloc(*s, (*l + ll + ALLOC_GR) & ~(ALLOC_GR - 1));
+	}
+	memcpy(p + *l, a, ll); p[*l += ll] = 0;
 }
 
 static inline void add_chr_to_str(unsigned char **s, int *l, unsigned char a)
 {
 	unsigned char *p;
-	if ((*l & (ALLOC_GR - 1)) == ALLOC_GR - 1 &&
-	   (!(p = mem_realloc(*s, (*l + 1 + ALLOC_GR) & ~(ALLOC_GR - 1))) ||
-	   !(*s = p))) return;
-	*(*s + *l) = a; *(*s + ++(*l)) = 0;
+	if (*l + 1 < 0) overalloc();
+	p = *s;
+	if ((*l & (ALLOC_GR - 1)) == ALLOC_GR - 1) {
+		p = *s = mem_realloc(*s, (*l + 1 + ALLOC_GR) & ~(ALLOC_GR - 1));
+	}
+	*(p + *l) = a; *(p + ++(*l)) = 0;
 }
 
-static inline void add_num_to_str(unsigned char **s, int *l, int n)
+static inline void add_num_to_str(unsigned char **s, int *l, off_t n)
 {
 	unsigned char a[64];
 	/*sprintf(a, "%d", n);*/
@@ -389,7 +518,9 @@ static inline void add_knum_to_str(unsigned char **s, int *l, int n)
 
 static inline long strtolx(unsigned char *c, unsigned char **end)
 {
-	long l = strtol(c, (char **)end, 10);
+	long l;
+	if (c[0] == '0' && upcase(c[1]) == 'X' && c[2]) l = strtol((char *)c + 2, (char **)end, 16);
+	else l = strtol((char *)c, (char **)end, 10);
 	if (!*end) return l;
 	if (upcase(**end) == 'K') {
 		(*end)++;
@@ -406,31 +537,15 @@ static inline long strtolx(unsigned char *c, unsigned char **end)
 	return l;
 }
 
-static inline unsigned char *copy_string(unsigned char **dst, unsigned char *src)
-{
-	if ((*dst = src) && (*dst = mem_alloc(strlen(src) + 1))) strcpy(*dst, src);
-	return *dst;
-}
-
 /* Copies at most dst_size chars into dst. Ensures null termination of dst. */
-static inline unsigned char *safe_strncpy(unsigned char *dst, const unsigned char *src, size_t dst_size) {
-/*	size_t to_copy;
-
+static inline unsigned char *safe_strncpy(unsigned char *dst, const unsigned char *src, size_t dst_size)
+{
+	size_t to_copy;
+	if (!dst_size) return dst;
 	to_copy = strlen(src);
-
-	/ Ensure that the url size is not greater than str_size / 
-	if (dst_size < to_copy)
-		to_copy = dst_size - 1;
-
-	strncpy(dst, src, to_copy);
-
-	/ Ensure null termination /
-	dst[to_copy] = '\0';
-	
-	return dst;*/
-
-	strncpy(dst, src, dst_size);
-	if (strlen(src) >= dst_size) dst[dst_size - 1] = 0;
+	if (to_copy >= dst_size) to_copy = dst_size - 1;
+	memcpy(dst, src, to_copy);
+	memset(dst + to_copy, 0, dst_size - to_copy);
 	return dst;
 }
 
@@ -450,11 +565,12 @@ struct xlist_head {
 #define del_from_list(x) {do_not_optimize_here(x); ((struct list_head *)(x)->next)->prev=(x)->prev; ((struct list_head *)(x)->prev)->next=(x)->next; do_not_optimize_here(x);}
 /*#define add_to_list(l,x) {(x)->next=(l).next; (x)->prev=(typeof(x))&(l); (l).next=(x); if ((l).prev==&(l)) (l).prev=(x);}*/
 
-#ifdef HAVE_TYPEOF
+#if defined(HAVE_TYPEOF) && !(defined(__GNUC__) && __GNUC__ >= 4)
+/* GCC 4 emits warnings about this and I haven't found a way to stop it */
 #define add_at_pos(p,x) do {do_not_optimize_here(p); (x)->next=(p)->next; (x)->prev=(p); (p)->next=(x); (x)->next->prev=(x);do_not_optimize_here(p);} while(0)
-#define add_to_list(l,x) add_at_pos((typeof(x))&(l),(x))
-#define foreach(e,l) for ((e)=(l).next; (e)!=(typeof(e))&(l); (e)=(e)->next)
-#define foreachback(e,l) for ((e)=(l).prev; (e)!=(typeof(e))&(l); (e)=(e)->prev)
+#define add_to_list(l,x) add_at_pos((typeof(x))(void *)&(l),(x))
+#define foreach(e,l) for ((e)=(l).next; (e)!=(typeof(e))(void *)&(l); (e)=(e)->next)
+#define foreachback(e,l) for ((e)=(l).prev; (e)!=(typeof(e))(void *)&(l); (e)=(e)->prev)
 #else
 #define add_at_pos(p,x) do {do_not_optimize_here(p); (x)->next=(p)->next; (x)->prev=(void *)(p); (p)->next=(x); (x)->next->prev=(x); do_not_optimize_here(p); } while(0)
 #define add_to_list(l,x) add_at_pos(&(l),(x))
@@ -472,9 +588,9 @@ static inline int isA(unsigned char c)
 	        c == '_' || c == '-';
 }
 
-static inline int casecmp(unsigned char *c1, unsigned char *c2, int len)
+static inline int casecmp(unsigned char *c1, unsigned char *c2, size_t len)
 {
-	int i;
+	size_t i;
 	for (i = 0; i < len; i++) if (upcase(c1[i]) != upcase(c2[i])) return 1;
 	return 0;
 }
@@ -541,9 +657,16 @@ void set_clipboard_text(char *);
 void set_window_title(unsigned char *);
 unsigned char *get_window_title();
 int is_safe_in_shell(unsigned char);
+unsigned char *escape_path(unsigned char *);
 void check_shell_security(unsigned char **);
+int check_shell_url(unsigned char *);
 void block_stdin();
 void unblock_stdin();
+void init_os(void);
+void init_os_terminal(void);
+void get_path_to_exe(void);
+unsigned char *os_conv_to_external_path(unsigned char *, unsigned char *);
+unsigned char *os_fixup_external_program(unsigned char *);
 int exe(char *);
 int resize_window(int, int);
 int can_resize_window(int);
@@ -551,10 +674,11 @@ int can_open_os_shell(int);
 struct open_in_new *get_open_in_new(int);
 void set_highpri();
 #ifdef HAVE_OPEN_PREALLOC
-int open_prealloc(char *, int, int, int);
-void prealloc_truncate(int, int);
+int can_prealloc(char *);
+int open_prealloc(char *, int, int, off_t);
+void prealloc_truncate(int, off_t);
 #else
-static inline void prealloc_truncate(int x, int y) {}
+static inline void prealloc_truncate(int x, off_t y) {}
 #endif
 void os_cfmakeraw(struct termios *t);
 
@@ -565,6 +689,7 @@ void os_cfmakeraw(struct termios *t);
 #endif
 
 typedef long ttime;
+typedef unsigned long uttime;
 typedef unsigned tcount;
 
 extern int terminate;
@@ -585,6 +710,7 @@ ttime get_time();
 void *get_handler(int, int);
 void set_handlers(int, void (*)(void *), void (*)(void *), void (*)(void *), void *);
 void install_signal_handler(int, void (*)(void *), void *, int);
+void interruptible_signal(int sig, int in);
 void set_sigcld();
 
 /* dns.c */
@@ -606,11 +732,11 @@ struct cache_entry {
 	unsigned char *head;
 	unsigned char *redirect;
 	int redirect_get;
-	int length;
+	off_t length;
 	int incomplete;
 	int tgc;
 	unsigned char *last_modified;
-	int data_size;
+	off_t data_size;
 	struct list_head frag;
 	tcount count;
 	int refcount;
@@ -622,20 +748,22 @@ struct cache_entry {
 struct fragment {
 	struct fragment *next;
 	struct fragment *prev;
-	int offset;
-	int length;
-	int real_length;
+	off_t offset;
+	off_t length;
+	off_t real_length;
 	unsigned char data[1];
 };
 
+extern int page_size;
+
+void init_cache();
 long cache_info(int);
 int find_in_cache(unsigned char *, struct cache_entry **);
 int get_cache_entry(unsigned char *, struct cache_entry **);
-int get_cache_data(struct cache_entry *e, unsigned char **, int *);
-int add_fragment(struct cache_entry *, int, unsigned char *, int);
+int add_fragment(struct cache_entry *, off_t, unsigned char *, off_t);
 void defrag_entry(struct cache_entry *);
-void truncate_entry(struct cache_entry *, int, int);
-void free_entry_to(struct cache_entry *, int);
+void truncate_entry(struct cache_entry *, off_t, int);
+void free_entry_to(struct cache_entry *, off_t);
 void delete_entry_content(struct cache_entry *);
 void delete_cache_entry(struct cache_entry *);
 void garbage_collection(int);
@@ -653,8 +781,8 @@ void garbage_collection(int);
 
 struct remaining_info {
 	int valid;
-	int size, loaded, last_loaded, cur_loaded;
-	int pos;
+	off_t size, loaded, last_loaded, cur_loaded;
+	off_t pos;
 	ttime elapsed;
 	ttime last_time;
 	ttime dis_b;
@@ -670,7 +798,7 @@ struct connection {
 	int running;
 	int state;
 	int prev_error;
-	int from;
+	off_t from;
 	int pri[N_PRI];
 	int no_cache;
 	int sock1;
@@ -681,10 +809,11 @@ struct connection {
 	struct list_head statuss;
 	void *info;
 	void *buffer;
+	struct conn_info *newconn;
 	void (*conn_func)(void *);
 	struct cache_entry *cache;
-	int received;
-	int est_length;
+	off_t received;
+	off_t est_length;
 	int unrestartable;
 	struct remaining_info prg;
 	int timer;
@@ -695,6 +824,21 @@ struct connection {
 #endif
 };
 
+extern struct list_head queue;
+
+struct k_conn {
+	struct k_conn *next;
+	struct k_conn *prev;
+	void (*protocol)(struct connection *);
+	unsigned char *host;
+	int port;
+	int conn;
+	ttime timeout;
+	ttime add_time;
+};
+
+extern struct list_head keepalive_connections;
+
 static inline int getpri(struct connection *c)
 {
 	int i;
@@ -718,39 +862,46 @@ static inline int getpri(struct connection *c)
 #define S_GETH		6
 #define S_PROC		7
 #define S_TRANS		8
-#define S_QUESTIONS	9
 
-#define S_WAIT_REDIR		-999
-#define S_OK			-1000
-#define S_INTERRUPTED		-1001
-#define S_EXCEPT		-1002
-#define S_INTERNAL		-1003
-#define S_OUT_OF_MEM		-1004
-#define S_NO_DNS		-1005
-#define S_CANT_WRITE		-1006
-#define S_CANT_READ		-1007
-#define S_MODIFIED		-1008
-#define S_BAD_URL		-1009
-#define S_TIMEOUT		-1010
-#define S_RESTART		-1011
-#define S_STATE			-1012
+#define S_OK			(-2000000000)
+#define S_INTERRUPTED		(-2000000001)
+#define S_EXCEPT		(-2000000002)
+#define S_INTERNAL		(-2000000003)
+#define S_OUT_OF_MEM		(-2000000004)
+#define S_NO_DNS		(-2000000005)
+#define S_CANT_WRITE		(-2000000006)
+#define S_CANT_READ		(-2000000007)
+#define S_MODIFIED		(-2000000008)
+#define S_BAD_URL		(-2000000009)
+#define S_TIMEOUT		(-2000000010)
+#define S_RESTART		(-2000000011)
+#define S_STATE			(-2000000012)
+#define S_LARGE_FILE		(-2000000014)
 
-#define S_HTTP_ERROR		-1100
-#define S_HTTP_100		-1101
-#define S_HTTP_204		-1102
+#define S_HTTP_ERROR		(-2000000100)
+#define S_HTTP_100		(-2000000101)
+#define S_HTTP_204		(-2000000102)
 
-#define S_FILE_TYPE		-1200
-#define S_FILE_ERROR		-1201
+#define S_FILE_TYPE		(-2000000200)
+#define S_FILE_ERROR		(-2000000201)
 
-#define S_FTP_ERROR		-1300
-#define S_FTP_UNAVAIL		-1301
-#define S_FTP_LOGIN		-1302
-#define S_FTP_PORT		-1303
-#define S_FTP_NO_FILE		-1304
-#define S_FTP_FILE_ERROR	-1305
+#define S_FTP_ERROR		(-2000000300)
+#define S_FTP_UNAVAIL		(-2000000301)
+#define S_FTP_LOGIN		(-2000000302)
+#define S_FTP_PORT		(-2000000303)
+#define S_FTP_NO_FILE		(-2000000304)
+#define S_FTP_FILE_ERROR	(-2000000305)
 
-#define S_SSL_ERROR		-1400
-#define S_NO_SSL		-1401
+#define S_SSL_ERROR		(-2000000400)
+#define S_NO_SSL		(-2000000401)
+
+#define S_NO_SMB_CLIENT		(-2000000500)
+
+#define S_WAIT_REDIR		(-2000000600)
+
+#define S_UNKNOWN_ERROR		(-2000000700)
+
+#define S_MAX			(-2000000900)
 
 extern struct s_msg_dsc {
 	int n;
@@ -770,30 +921,20 @@ struct status {
 	struct remaining_info *prg;
 };
 
-struct http_auth_basic {
-	struct http_auth_basic *next;
-	struct http_auth_basic *prev;
-	int valid;
-	unsigned char *url;
-	int url_len;
-	unsigned char *realm;
-	unsigned char *uid;
-	unsigned char *passwd;
-};
-
-void check_queue();
+void check_queue(void *dummy);
 long connect_info(int);
 void send_connection_info(struct connection *c);
 void setcstate(struct connection *c, int);
 int get_keepalive_socket(struct connection *c);
 void add_keepalive_socket(struct connection *c, ttime);
 void run_connection(struct connection *c);
+int is_connection_restartable(struct connection *c);
 void retry_connection(struct connection *c);
 void abort_connection(struct connection *c);
 void end_connection(struct connection *c);
 int load_url(unsigned char *, struct status *, int, int);
 void change_connection(struct status *, struct status *, int);
-void detach_connection(struct status *, int);
+void detach_connection(struct status *, off_t);
 void abort_all_connections();
 void abort_background_connections();
 int is_entry_used(struct cache_entry *);
@@ -803,15 +944,11 @@ void add_blacklist_entry(unsigned char *, int);
 void del_blacklist_entry(unsigned char *, int);
 int get_blacklist_flags(unsigned char *);
 void free_blacklist();
-void update_noproxy();
-void free_noproxy();
-unsigned char *find_auth(unsigned char *);
-void add_auth_entry(unsigned char*, unsigned char *);
-void del_auth_entry(struct http_auth_basic *);
-void free_auth();
 
 #define BL_HTTP10		1
-#define BL_NO_CHARSET		2
+#define BL_NO_ACCEPT_LANGUAGE	2
+#define BL_NO_CHARSET		4
+#define BL_NO_RANGE		8
 
 /* url.c */
 
@@ -838,6 +975,7 @@ unsigned char *join_urls(unsigned char *, unsigned char *);
 unsigned char *translate_url(unsigned char *, unsigned char *);
 unsigned char *extract_position(unsigned char *);
 void get_filename_from_url(unsigned char *, unsigned char **, int *);
+void add_conv_str(unsigned char **s, int *l, unsigned char *b, int ll, int encode_special);
 
 /* connect.c */
 
@@ -891,8 +1029,10 @@ void finger_func(struct connection *);
 
 /* ftp.c */
 
-extern int fast_ftp;
-extern int passive_ftp;
+#if defined(IP_TOS) && defined(IPTOS_THROUGHPUT)
+#define HAVE_IPTOS
+#endif
+
 void ftp_func(struct connection *);
 
 /* smb.c */
@@ -904,6 +1044,7 @@ void smb_func(struct connection *);
 void mailto_func(struct session *, unsigned char *);
 void telnet_func(struct session *, unsigned char *);
 void tn3270_func(struct session *, unsigned char *);
+void mms_func(struct session *, unsigned char *);
 
 /* kbd.c */
 
@@ -957,6 +1098,13 @@ void dispatch_special(unsigned char *);
 void kbd_ctrl_c();
 int is_blocked();
 
+extern unsigned char init_seq[];
+extern unsigned char init_seq_x_mouse[];
+extern unsigned char init_seq_tw_mouse[];
+extern unsigned char term_seq[];
+extern unsigned char term_seq_x_mouse[];
+extern unsigned char term_seq_tw_mouse[];
+
 /* terminal.c */
 
 typedef unsigned short chr;
@@ -994,6 +1142,7 @@ struct window {
 #define ENV_BE		8
 #define ENV_TWIN	16
 #define ENV_WIN32	32
+#define ENV_INTERIX	64
 
 struct terminal {
 	struct terminal *next;
@@ -1110,7 +1259,7 @@ int n_languages();
 unsigned char *language_name(int);
 
 #define _(_x_, _y_) get_text_translation(_x_, _y_)
-#define TEXT(x) (dummyarray + x)
+#define TEXT_(x) (dummyarray + x) /* TEXT causes name clash on windows */
 
 /* af_unix.c */
 
@@ -1119,15 +1268,11 @@ void af_unix_close();
 
 /* main.c */
 
-#define RET_OK		0
-#define RET_ERROR	1
-#define RET_SIGNAL	2
-#define RET_SYNTAX	3
-#define RET_FATAL	4
-
 extern int retval;
 
 extern unsigned char *path_to_exe;
+extern unsigned char **g_argv;
+extern int g_argc;
 
 void unhandle_terminal_signals(struct terminal *term);
 int attach_terminal(int, int, int, void *, int);
@@ -1171,6 +1316,7 @@ extern struct list_head extensions;
 extern struct list_head mailto_prog;
 extern struct list_head telnet_prog;
 extern struct list_head tn3270_prog;
+extern struct list_head mms_prog;
 
 unsigned char *get_content_type(unsigned char *, unsigned char *);
 struct assoc *get_type_assoc(struct terminal *term, unsigned char *);
@@ -1187,6 +1333,8 @@ void menu_add_ext(struct terminal *, void *, void *);
 void menu_del_ext(struct terminal *, void *, void *);
 void menu_list_ext(struct terminal *, void *, void *);
 
+int is_html_type(unsigned char *ct);
+
 /* session.c */
 
 struct link_def {
@@ -1323,6 +1471,7 @@ struct document_options {
 	struct rgb default_link;
 	struct rgb default_vlink;
 	unsigned char *framename;
+	int real_cp;    /* codepage of document. Does not really belong here. Must not be compared. Used only in get_attr_val */
 };
 
 static inline void ds2do(struct document_setup *ds, struct document_options *doo)
@@ -1414,9 +1563,13 @@ struct f_data_c {
 struct view_state {
 	int view_pos;
 	int view_posx;
+	int orig_view_pos;
+	int orig_view_posx;
 	int current_link;
+	int orig_link;
 	int plain;
 	unsigned char *goto_position;
+	unsigned char *goto_position_end;
 	struct form_state *form_info;
 	int form_info_len;
 	unsigned char url[1];
@@ -1458,7 +1611,7 @@ struct download {
 	unsigned char *url;
 	struct status stat;
 	unsigned char *file;
-	int last_pos;
+	off_t last_pos;
 	int handle;
 	int redirect_cnt;
 	unsigned char *prog;
@@ -1522,9 +1675,10 @@ extern struct list_head sessions;
 time_t parse_http_date(const char *);
 unsigned char *encode_url(unsigned char *);
 unsigned char *decode_url(unsigned char *);
-unsigned char *subst_file(unsigned char *, unsigned char *);
+unsigned char *subst_file(unsigned char *, unsigned char *, int);
 int are_there_downloads();
 void free_strerror_buf();
+int get_error_from_errno(int errn);
 unsigned char *get_err_msg(int);
 void print_screen_status(struct session *);
 void print_error_dialog(struct session *, struct status *, unsigned char *);
@@ -1558,8 +1712,6 @@ unsigned char *get_current_title(struct session *, unsigned char *, size_t);
 
 unsigned char *get_current_link_url(struct session *, unsigned char *, size_t);
 
-void add_questions_entry(void *);
-
 /* bfu.c */
 
 struct memory_list {
@@ -1710,7 +1862,7 @@ int cancel_dialog(struct dialog_data *, struct dialog_item_data *);
 void msg_box(struct terminal *, struct memory_list *, unsigned char *, int, /*unsigned char *, void *, int,*/ ...);
 void input_field_fn(struct dialog_data *);
 void input_field(struct terminal *, struct memory_list *, unsigned char *, unsigned char *, unsigned char *, unsigned char *, void *, struct history *, int, unsigned char *, int, int, int (*)(struct dialog_data *, struct dialog_item_data *), void (*)(void *, unsigned char *), void (*)(void *));
-void add_to_history(struct history *, unsigned char *);
+void add_to_history(struct history *, unsigned char *, int);
 
 void box_sel_move(struct dialog_item_data *, int ); 
 void show_dlg_item_box(struct dialog_data *, struct dialog_item_data *);
@@ -1718,6 +1870,8 @@ void box_sel_set_visible(struct dialog_item_data *, int );
 
 /* menu.c */
 
+extern struct history goto_url_history;
+
 void activate_bfu_technology(struct session *, int);
 void dialog_goto_url(struct session *ses, char *url);
 void dialog_save_url(struct session *ses);
@@ -1727,6 +1881,7 @@ void search_dlg(struct session *, struct f_data_c *, int);
 void search_back_dlg(struct session *, struct f_data_c *, int);
 void exit_prog(struct terminal *, void *, struct session *);
 void do_auth_dialog(struct session *);
+
 /* charsets.c */
 
 #include "codepage.h"
@@ -1739,7 +1894,9 @@ struct conv_table {
 	} u;
 };
 
+int cp2u(unsigned char ch, int from);
 struct conv_table *get_translation_table(int, int);
+int get_entity_number(unsigned char *st, int l);
 unsigned char *get_entity_string(unsigned char *, int, int);
 unsigned char *convert_string(struct conv_table *, unsigned char *, int);
 int get_cp_index(unsigned char *n);
@@ -1748,6 +1905,9 @@ unsigned char *get_cp_mime_name(int);
 int is_cp_special(int);
 void free_conv_table();
 
+unsigned char charset_upcase(unsigned char, int);
+void charset_upcase_string(unsigned char **, int);
+
 /* view.c */
 
 int can_open_in_new(struct terminal *);
@@ -1866,24 +2026,25 @@ struct html_element {
 	struct frameset_desc *frameset;
 };
 
+extern int get_attr_val_nl;
+
 extern struct list_head html_stack;
 extern int line_breax;
 
 extern unsigned char *startf;
 extern unsigned char *eofff;
 
-#define format (((struct html_element *)html_stack.next)->attr)
+#define format_ (((struct html_element *)html_stack.next)->attr)
 #define par_format (((struct html_element *)html_stack.next)->parattr)
 #define html_top (*(struct html_element *)html_stack.next)
 
 extern void *ff;
-extern void (*put_chars_f)(void *, unsigned char *, int);
+extern int (*put_chars_f)(void *, unsigned char *, int);
 extern void (*line_break_f)(void *);
-extern void (*init_f)(void *);
 extern void *(*special_f)(void *, int, ...);
 
 void ln_break(int, void (*)(void *), void *);
-void put_chrs(unsigned char *, int, void (*)(void *, unsigned char *, int), void *);
+void put_chrs(unsigned char *, int, int (*)(void *, unsigned char *, int), void *);
 
 extern int table_level;
 extern int empty_format;
@@ -1903,7 +2064,7 @@ int get_bgcolor(unsigned char *, struct rgb *);
 void html_stack_dup();
 void kill_html_stack_item(struct html_element *);
 unsigned char *skip_comment(unsigned char *, unsigned char *);
-void parse_html(unsigned char *, unsigned char *, void (*)(void *, unsigned char *, int), void (*)(void *), void (*)(void *), void *(*)(void *, int, ...), void *, unsigned char *);
+void parse_html(unsigned char *, unsigned char *, int (*)(void *, unsigned char *, int), void (*)(void *), void *(*)(void *, int, ...), void *, unsigned char *);
 int get_image_map(unsigned char *, unsigned char *, unsigned char *, unsigned char *a, struct menu_item **, struct memory_list **, unsigned char *, unsigned char *, int, int, int);
 void scan_http_equiv(unsigned char *, unsigned char *, unsigned char **, int *, unsigned char **);
 
@@ -1954,8 +2115,8 @@ extern struct document_options *d_opt;
 extern int last_link_to_move;
 extern int margin;
 
-int xxpand_line(struct part *, int, int);
-int xxpand_lines(struct part *, int);
+void xxpand_line(struct part *, int, int);
+void xxpand_lines(struct part *, int);
 void xset_hchar(struct part *, int, int, unsigned);
 void xset_hchars(struct part *, int, int, int, unsigned);
 void align_line(struct part *, int);
@@ -1995,14 +2156,16 @@ struct option {
 
 unsigned char *parse_options(int, unsigned char *[]);
 void init_home();
+unsigned char *read_config_file(unsigned char *name);
+int write_to_config_file(unsigned char *, unsigned char *);
 unsigned char *get_token(unsigned char **line);
 void load_config();
 void write_config(struct terminal *);
 void write_html_config(struct terminal *);
 void end_config();
 
-int load_url_history();
-int save_url_history();
+void load_url_history();
+void save_url_history();
 
 extern int anonymous;
 
@@ -2027,13 +2190,14 @@ extern int max_connections;
 extern int max_connections_to_host;
 extern int max_tries;
 extern int screen_width;
+extern int dump_codepage;
 extern int receive_timeout;
 extern int unrestartable_receive_timeout;
 
 extern struct document_setup dds;
 
 extern int max_format_cache_entries;
-extern long memory_cache_size;
+extern int memory_cache_size;
 
 extern struct rgb default_fg;
 extern struct rgb default_bg;
@@ -2042,7 +2206,6 @@ extern struct rgb default_vlink;
 
 extern unsigned char http_proxy[];
 extern unsigned char ftp_proxy[];
-extern unsigned char no_proxy_for[];
 extern unsigned char download_dir[];
 
 struct http_bugs {
@@ -2055,7 +2218,14 @@ struct http_bugs {
 
 extern struct http_bugs http_bugs;
 
-extern unsigned char default_anon_pass[];
+struct ftp_options {
+	unsigned char anon_pass[MAX_STR_LEN];
+	int fast_ftp;
+	int passive_ftp;
+	int set_tos;
+};
+
+extern struct ftp_options ftp_options;
 
 /* bookmarks.c */
 
@@ -2077,17 +2247,12 @@ struct dlg_data_item_data_box {
 	int list_len;	/* Number of items in the list */
 };
 
-/* Which fields to free when zapping a box_item. Bitwise or these. */
-enum box_item_free {NOTHING = 0, TEXT = 1 , DATA = 2};
 /* An item in a box */
 struct box_item {
 	struct box_item *next;
 	struct box_item *prev;
 	unsigned char *text;	/* Text to display */
-	void (*on_hilight)(struct terminal *, struct dlg_data_item_data_box *, struct box_item *);	/* Run when this item is hilighted */
-	int (*on_selected)(struct terminal *, struct dlg_data_item_data_box *, struct box_item *);	/* Run when the user selects on this item. Returns pointer to the box_item that should be selected after execution*/
 	void *data;	/* data */
-	enum box_item_free free_i;
 };
 
 
