@@ -31,33 +31,60 @@ void sig_ign(void *x)
 {
 }
 
+int fg_poll_timer = -1;
+void poll_fg(void *t);
+
 void sig_tstp(struct terminal *t)
 {
-#ifdef SIGSTOP
-	int pid = getpid();
-	block_itrm(0);
-#if defined (SIGCONT) && defined(SIGTTOU)
-	if (!fork()) {
-		sleep(1);
-		kill(pid, SIGCONT);
-		exit(0);
+#if defined(SIGSTOP) && !defined(NO_CTRL_Z)
+#if defined(SIGCONT) && defined(SIGTTOU) && defined(HAVE_GETPID)
+	pid_t pid = getpid();
+	pid_t newpid;
+#endif
+	block_itrm(1);
+#if defined(SIGCONT) && defined(SIGTTOU)
+	if (!(newpid = fork())) {
+		while (1) {
+			sleep(1);
+			kill(pid, SIGCONT);
+		}
 	}
 #endif
 	raise(SIGSTOP);
+#if defined(SIGCONT) && defined(SIGTTOU)
+	if (newpid != -1) kill(newpid, SIGKILL);
 #endif
+#endif
+	if (fg_poll_timer != -1) kill_timer(fg_poll_timer);
+	fg_poll_timer = install_timer(FG_POLL_TIME, poll_fg, t);
+}
+
+void poll_fg(void *t)
+{
+	int r;
+	fg_poll_timer = -1;
+	r = unblock_itrm(1);
+	if (r == -1) {
+		fg_poll_timer = install_timer(FG_POLL_TIME, poll_fg, t);
+	}
+	if (r == -2) {
+		/* This will unblock externally spawned viewer, if it exists */
+#ifdef SIGCONT
+		kill(0, SIGCONT);
+#endif
+	}
 }
 
 void sig_cont(struct terminal *t)
 {
-	if (!unblock_itrm(0)) /*redraw_terminal_cls(t)*/resize_terminal();
-	/*else register_bottom_half(raise, SIGSTOP);*/
+	unblock_itrm(1);
 }
 
 void handle_basic_signals(struct terminal *term)
 {
 	install_signal_handler(SIGHUP, (void (*)(void *))sig_intr, term, 0);
 	install_signal_handler(SIGINT, (void (*)(void *))sig_ctrl_c, term, 0);
-	install_signal_handler(SIGTERM, (void (*)(void *))sig_terminate, term, 0);
+	/*install_signal_handler(SIGTERM, (void (*)(void *))sig_terminate, term, 0);*/
 #ifdef SIGTSTP
 	install_signal_handler(SIGTSTP, (void (*)(void *))sig_tstp, term, 0);
 #endif
@@ -72,24 +99,6 @@ void handle_basic_signals(struct terminal *term)
 #endif
 }
 
-/*void handle_slave_signals(struct terminal *term)
-{
-	install_signal_handler(SIGHUP, (void (*)(void *))sig_terminate, term, 0);
-	install_signal_handler(SIGINT, (void (*)(void *))sig_terminate, term, 0);
-	install_signal_handler(SIGTERM, (void (*)(void *))sig_terminate, term, 0);
-#ifdef SIGTSTP
-	install_signal_handler(SIGTSTP, (void (*)(void *))sig_tstp, term, 0);
-#endif
-#ifdef SIGTTIN
-	install_signal_handler(SIGTTIN, (void (*)(void *))sig_tstp, term, 0);
-#endif
-#ifdef SIGTTOU
-	install_signal_handler(SIGTTOU, (void (*)(void *))sig_ign, term, 0);
-#endif
-#ifdef SIGCONT
-	install_signal_handler(SIGCONT, (void (*)(void *))sig_cont, term, 0);
-#endif
-}*/
 
 void unhandle_terminal_signals(struct terminal *term)
 {
@@ -107,13 +116,14 @@ void unhandle_terminal_signals(struct terminal *term)
 #ifdef SIGCONT
 	install_signal_handler(SIGCONT, NULL, NULL, 0);
 #endif
+	if (fg_poll_timer != -1) kill_timer(fg_poll_timer), fg_poll_timer = -1;
 }
 
 void unhandle_basic_signals(struct terminal *term)
 {
 	install_signal_handler(SIGHUP, NULL, NULL, 0);
 	install_signal_handler(SIGINT, NULL, NULL, 0);
-	install_signal_handler(SIGTERM, NULL, NULL, 0);
+	/*install_signal_handler(SIGTERM, NULL, NULL, 0);*/
 #ifdef SIGTSTP
 	install_signal_handler(SIGTSTP, NULL, NULL, 0);
 #endif
@@ -126,6 +136,7 @@ void unhandle_basic_signals(struct terminal *term)
 #ifdef SIGCONT
 	install_signal_handler(SIGCONT, NULL, NULL, 0);
 #endif
+	if (fg_poll_timer != -1) kill_timer(fg_poll_timer), fg_poll_timer = -1;
 }
 
 int terminal_pipe[2];
@@ -147,7 +158,7 @@ int attach_terminal(int in, int out, int ctl, void *info, int len)
 }
 
 struct status dump_stat;
-int dump_pos;
+off_t dump_pos;
 
 int dump_red_count = 0;
 
@@ -191,8 +202,7 @@ void end_dump(struct status *stat, void *p)
 		struct document_options o;
 		struct view_state *vs;
 		struct f_data_c fd;
-		if (!(vs = mem_alloc(sizeof(struct view_state) + strlen(stat->ce->url) + 1)))
-			goto terminate;
+		vs = mem_alloc(sizeof(struct view_state) + strlen(stat->ce->url) + 1);
 		memset(&o, 0, sizeof(struct document_options));
 		memset(vs, 0, sizeof(struct view_state));
 		memset(&fd, 0, sizeof(struct f_data_c));
@@ -201,7 +211,7 @@ void end_dump(struct status *stat, void *p)
 		o.xw = screen_width;
 		o.yw = 25;
 		o.col = 0;
-		o.cp = 0;
+		o.cp = dump_codepage == -1 ? 0 : dump_codepage;
 		ds2do(&dds, &o);
 		o.plain = 0;
 		o.frames = 0;
@@ -228,8 +238,8 @@ void end_dump(struct status *stat, void *p)
 	terminate = 1;
 }
 
-int ac;
-unsigned char **av;
+int g_argc;
+unsigned char **g_argv;
 
 unsigned char *path_to_exe;
 
@@ -244,18 +254,29 @@ void init()
 	init_trans();
 	set_sigcld();
 	init_home();
+	init_cache();
 	init_keymaps();
 
 /* OS/2 has some stupid bug and the pipe must be created before socket :-/ */
 	if (c_pipe(terminal_pipe)) {
 		error("ERROR: can't create pipe for internal communication");
+		retval = RET_FATAL;
 		goto ttt;
 	}
-	if (!(u = parse_options(ac - 1, av + 1))) goto ttt;
+	if (!(u = parse_options(g_argc - 1, g_argv + 1))) {
+		retval = RET_SYNTAX;
+		goto ttt;
+	}
+	if (!dmp) {
+		init_os_terminal();
+	}
 	if (!no_connect && (uh = bind_to_af_unix()) != -1) {
 		close(terminal_pipe[0]);
 		close(terminal_pipe[1]);
-		if (!(info = create_session_info(base_session, u, &len))) goto ttt;
+		if (!(info = create_session_info(base_session, u, &len))) {
+			retval = RET_FATAL;
+			goto ttt;
+		}
 		handle_trm(get_input_handle(), get_output_handle(), uh, uh, get_ctl_handle(), info, len);
 		handle_basic_signals(NULL);	/* OK, this is race condition, but it must be so; GPM installs it's own buggy TSTP handler */
 		mem_free(info);
@@ -267,11 +288,10 @@ void init()
 	read_bookmarks();
 	load_url_history();
 	init_cookies();
-	u = parse_options(ac - 1, av + 1);
+	u = parse_options(g_argc - 1, g_argv + 1);
 	if (!u) {
 		ttt:
 		terminate = 1;
-		retval = RET_SYNTAX;
 		return;
 	}
 	if (!dmp) {
@@ -285,12 +305,16 @@ void init()
 		close(terminal_pipe[1]);
 		if (!*u) {
 			fprintf(stderr, "URL expected after %s\n.", dmp == D_DUMP ? "-dump" : "-source");
+			retval = RET_SYNTAX;
 			goto ttt;
 		}
 		dump_stat.end = end_dump;
 		dump_pos = 0;
 		if (!(uu = translate_url(u, wd = get_cwd()))) uu = stracpy(u);
-		if (load_url(uu, &dump_stat, PRI_MAIN, 0)) goto ttt;
+		if (load_url(uu, &dump_stat, PRI_MAIN, 0)) {
+			retval = RET_FATAL;
+			goto ttt;
+		}
 		mem_free(uu);
 		if (wd) mem_free(wd);
 	}
@@ -315,7 +339,6 @@ void terminate_all_subsystems()
 	free_history_lists();
 	free_term_specs();
 	free_types();
-	free_auth();
 	if (init_b) finalize_bookmarks();
 	free_keymaps();
 	free_conv_table();
@@ -326,13 +349,17 @@ void terminate_all_subsystems()
 	free_strerror_buf();
 	shutdown_trans();
 	terminate_osdep();
+	if (fg_poll_timer != -1) kill_timer(fg_poll_timer), fg_poll_timer = -1;
 }
 
 int main(int argc, char *argv[])
 {
-	path_to_exe = argv[0];
-	ac = argc;
-	av = (unsigned char **)argv;
+	g_argc = argc;
+	g_argv = (unsigned char **)argv;
+
+	init_os();
+
+	get_path_to_exe();
 
 	select_loop(init);
 	terminate_all_subsystems();
