@@ -2,7 +2,15 @@
 #ifdef INT_MAX
 #define MAXINT INT_MAX
 #else
-#define MAXINT 0x7fffffff
+#define MAXINT ((int)((unsigned int)-1 >> 1))
+#endif
+#endif
+
+#ifndef MAXLONG
+#ifdef LONG_MAX
+#define MAXLONG LONG_MAX
+#else
+#define MAXLONG ((long)((unsigned long)-1L >> 1))
 #endif
 #endif
 
@@ -10,18 +18,13 @@
 #define SA_RESTART	0
 #endif
 
-/*#ifdef sparc
-#define htons(x) (x)
-#endif*/
-
-#ifndef HAVE_CFMAKERAW
-void cfmakeraw(struct termios *t);
-#endif
-
 #ifdef __EMX__
 #define strcasecmp stricmp
+#define strncasecmp strnicmp
 #define read _read
 #define write _write
+#define getcwd _getcwd2
+#define chdir _chdir2
 #endif
 
 #ifdef BEOS
@@ -44,14 +47,30 @@ void cfmakeraw(struct termios *t);
 #ifndef SO_ERROR
 #define SO_ERROR	10001
 #endif
-#ifdef errno
-#undef errno
-#endif
-#define errno 1
 #endif
 
 #if defined(O_SIZE) && defined(__EMX__)
 #define HAVE_OPEN_PREALLOC
 #endif
 
+#if defined(__WATCOMC__) && defined(_WCRTLINK)
+#define LIBC_CALLBACK	_WCRTLINK
+#else
+#define LIBC_CALLBACK
+#endif
+
+#if defined(__WATCOMC__) && defined(__LINUX__)
+#define SIGNAL_HANDLER __declspec(__cdecl)
+#else
+#define SIGNAL_HANDLER
+#endif
+
+#if defined(HAVE_HERROR) && defined(__GNUC__) && defined(__hpux)
+#undef HAVE_HERROR
+#endif
+
+#ifdef HAVE_MAXINT_CONVERSION_BUG
+#undef MAXINT
+#define MAXINT 0x7FFFF000
+#endif
 
