@@ -15,17 +15,6 @@ struct h_conn {
 
 struct list_head h_conns = {&h_conns, &h_conns};
 
-struct k_conn {
-	struct k_conn *next;
-	struct k_conn *prev;
-	void (*protocol)(struct connection *);
-	unsigned char *host;
-	int port;
-	int conn;
-	ttime timeout;
-	ttime add_time;
-};
-
 struct list_head keepalive_connections = {&keepalive_connections, &keepalive_connections};
 
 long connect_info(int type)
@@ -80,6 +69,7 @@ void stat_timer(struct connection *c)
 {
 	ttime a;
 	struct remaining_info *r = &c->prg;
+	if (getpri(c) == PRI_CANCEL && (c->est_length > (longlong)memory_cache_size * MAX_CACHED_OBJECT || c->from > (longlong)memory_cache_size * MAX_CACHED_OBJECT)) register_bottom_half(check_queue, NULL);
 	r->loaded = c->received;
 	if ((r->size = c->est_length) < (r->pos = c->from) && r->size != -1)
 		r->size = c->from;
@@ -191,6 +181,10 @@ void free_connection_data(struct connection *c)
 		mem_free(c->buffer);
 		c->buffer = NULL;
 	}
+	if (c->newconn) {
+		mem_free(c->newconn);
+		c->newconn = NULL;
+	}
 	if (c->info) {
 		mem_free(c->info);
 		c->info = NULL;
@@ -200,13 +194,15 @@ void free_connection_data(struct connection *c)
 		internal("active connections underflow");
 		active_connections = 0;
 	}
-	if (c->state != S_WAIT) if ((h = is_host_on_list(c))) {
-		if (!--h->conn) {
-			del_from_list(h);
-			mem_free(h->host);
-			mem_free(h);
-		}
-	} else internal("suspending connection that is not on the list (state %d)", c->state);
+	if (c->state != S_WAIT) {
+		if ((h = is_host_on_list(c))) {
+			if (!--h->conn) {
+				del_from_list(h);
+				mem_free(h->host);
+				mem_free(h);
+			}
+		} else internal("suspending connection that is not on the list (state %d)", c->state);
+	}
 }
 
 void send_connection_info(struct connection *c)
@@ -242,7 +238,7 @@ void add_keepalive_socket(struct connection *c, ttime timeout)
 		internal("keepalive connection not connected");
 		goto del;
 	}
-	if (!(k = mem_alloc(sizeof(struct k_conn)))) goto close;
+	k = mem_alloc(sizeof(struct k_conn));
 	if ((k->port = get_port(c->url)) == -1 || !(k->protocol = get_protocol_handle(c->url)) || !(k->host = get_host_and_pass(c->url))) {
 		mem_free(k);
 		del_connection(c);
@@ -257,14 +253,14 @@ void add_keepalive_socket(struct connection *c, ttime timeout)
 #ifdef DEBUG
 	check_queue_bugs();
 #endif
-	register_bottom_half((void (*)(void *))check_queue, NULL);
+	register_bottom_half(check_queue, NULL);
 	return;
 	close:
 	close(c->sock1);
 #ifdef DEBUG
 	check_queue_bugs();
 #endif
-	register_bottom_half((void (*)(void *))check_queue, NULL);
+	register_bottom_half(check_queue, NULL);
 }
 
 void del_keepalive_socket(struct k_conn *kc)
@@ -403,17 +399,23 @@ void run_connection(struct connection *c)
 	func(c);
 }
 
+int is_connection_restartable(struct connection *c)
+{
+	return !(c->unrestartable >= 2 || (c->tries + 1 >= (max_tries ? max_tries : 1000)));
+}
+
 void retry_connection(struct connection *c)
 {
 	interrupt_connection(c);
-	if (c->unrestartable >= 2 || (++c->tries >= (max_tries ? max_tries : 1000))) {
+	if (!is_connection_restartable(c)) {
 		/*send_connection_info(c);*/
 		del_connection(c);
 #ifdef DEBUG
 		check_queue_bugs();
 #endif
-		register_bottom_half((void (*)(void *))check_queue, NULL);
+		register_bottom_half(check_queue, NULL);
 	} else {
+		c->tries++;
 		c->prev_error = c->state;
 		run_connection(c);
 	}
@@ -427,21 +429,22 @@ void abort_connection(struct connection *c)
 #ifdef DEBUG
 	check_queue_bugs();
 #endif
-	register_bottom_half((void (*)(void *))check_queue, NULL);
+	register_bottom_half(check_queue, NULL);
 }
 
 int try_connection(struct connection *c)
 {
 	struct h_conn *hc = NULL;
 	if ((hc = is_host_on_list(c))) {
-		if (hc->conn >= max_connections_to_host)
+		if (hc->conn >= max_connections_to_host) {
 			if (try_to_suspend_connection(c, hc->host)) return 0;
 			else return -1;
+		}
 	}
-	if (active_connections >= max_connections)
+	if (active_connections >= max_connections) {
 		if (try_to_suspend_connection(c, NULL)) return 0;
 		else return -1;
-	connect_ok:
+	}
 	run_connection(c);
 	return 1;
 }
@@ -479,7 +482,7 @@ void check_queue_bugs()
 }
 #endif
 
-void check_queue()
+void check_queue(void *dummy)
 {
 	struct connection *c;
 	again:
@@ -512,6 +515,10 @@ void check_queue()
 			setcstate(c, S_INTERRUPTED);
 			del_connection(c);
 			goto again2;
+		} else if (c->est_length > (longlong)memory_cache_size * MAX_CACHED_OBJECT || c->from > (longlong)memory_cache_size * MAX_CACHED_OBJECT) {
+			setcstate(c, S_INTERRUPTED);
+			abort_connection(c);
+			goto again2;
 		}
 	}
 #ifdef DEBUG
@@ -521,12 +528,12 @@ void check_queue()
 
 unsigned char *get_proxy(unsigned char *url)
 {
-	int l = strlen(url);
+	size_t l = strlen(url);
 	unsigned char *proxy = NULL;
 	unsigned char *u;
 	if (*http_proxy && l >= 7 && !casecmp(url, "http://", 7)) proxy = http_proxy;
 	if (*ftp_proxy && l >= 6 && !casecmp(url, "ftp://", 6)) proxy = ftp_proxy;
-	if (!(u = mem_alloc(l + 1 + (proxy ? strlen(proxy) + 9 : 0)))) return NULL;
+	u = mem_alloc(l + 1 + (proxy ? strlen(proxy) + 9 : 0));
 	if (proxy) strcpy(u, "proxy://"), strcat(u, proxy), strcat(u, "/");
 	else *u = 0;
 	strcat(u, url);
@@ -553,12 +560,25 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 	}
 #endif
 	if (stat) stat->state = S_OUT_OF_MEM, stat->prev_error = 0;
-	if (no_cache <= NC_CACHE && !find_in_cache(url, &e) && !e->incomplete) {
+	if (no_cache <= NC_CACHE && !find_in_cache(url, &e)) {
+		if (e->incomplete) {
+			e->refcount--;
+			goto skip_cache;
+		}
 		if (stat) {
 			stat->ce = e;
 			stat->state = S_OK;
 			if (stat->end) stat->end(stat, stat->data);
 		}
+		e->refcount--;
+		return 0;
+	}
+	skip_cache:
+	if (!casecmp(url, "proxy://", 8)) {
+		if (stat) {
+			stat->state = S_BAD_URL;
+			if (stat->end) stat->end(stat, stat->data);
+		}
 		return 0;
 	}
 	if (!(u = get_proxy(url))) {
@@ -571,7 +591,7 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 			del_from_list(c);
 			c->pri[pri]++;
 			add_to_queue(c);
-			register_bottom_half((void (*)(void *))check_queue, NULL);
+			register_bottom_half(check_queue, NULL);
 		} else c->pri[pri]++;
 		if (stat) {
 			stat->prg = &c->prg;
@@ -585,11 +605,7 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 #endif
 		return 0;
 	}
-	if (!(c = mem_alloc(sizeof(struct connection)))) {
-		stat->end(stat, stat->data);
-		mem_free(u);
-		return -1;
-	}
+	c = mem_alloc(sizeof(struct connection));
 	memset(c, 0, sizeof(struct connection));
 	c->count = connection_count++;
 	c->url = u;
@@ -603,6 +619,7 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 	c->dnsquery = NULL;
 	c->info = NULL;
 	c->buffer = NULL;
+	c->newconn = NULL;
 	c->cache = NULL;
 	c->tries = 0;
 	init_list(c->statuss);
@@ -621,7 +638,7 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 #ifdef DEBUG
 	check_queue_bugs();
 #endif
-	register_bottom_half((void (*)(void *))check_queue, NULL);
+	register_bottom_half(check_queue, NULL);
 	return 0;
 }
 
@@ -671,26 +688,26 @@ void change_connection(struct status *oldstat, struct status *newstat, int newpr
 #ifdef DEBUG
 	check_queue_bugs();
 #endif
-	register_bottom_half((void (*)(void *))check_queue, NULL);
+	register_bottom_half(check_queue, NULL);
 }
 
-void detach_connection(struct status *stat, int pos)
+void detach_connection(struct status *stat, off_t pos)
 {
 	struct connection *c;
-	int i, l;
+	int i;
+	off_t l;
 	if (stat->state < 0) return;
 	c = stat->c;
 	if (c->detached) goto detach_done;
 	if (!c->cache) return;
 	if (c->est_length == -1) l = c->from;
 	else l = c->est_length;
-	if (l < memory_cache_size * MAX_CACHED_OBJECT) return;
+	if (l < (longlong)memory_cache_size * MAX_CACHED_OBJECT) return;
 	l = 0;
 	for (i = 0; i < PRI_CANCEL; i++) l += c->pri[i];
 	if (!l) internal("detaching free connection");
 	delete_unused_format_cache_entries();
 	if (l != 1 || c->cache->refcount) return;
-	/*debug("detached");*/
 	c->cache->url[0] = 0;
 	c->detached = 1;
 	detach_done:
@@ -736,12 +753,14 @@ void abort_background_connections()
 	while (1) {
 		int j;
 		struct connection *c = (void *)&queue;
-		for (j = 0; j <= i; j++) if ((c = c->next) == (void *)&queue) return;
+		for (j = 0; j <= i; j++) if ((c = c->next) == (void *)&queue) goto brk;
 		if (getpri(c) >= PRI_CANCEL) {
 			setcstate(c, S_INTERRUPTED);
 			abort_connection(c);
 		} else i++;
 	}
+	brk:
+	abort_all_keepalive_connections();
 }
 
 int is_entry_used(struct cache_entry *e)
@@ -767,7 +786,7 @@ void add_blacklist_entry(unsigned char *host, int flags)
 		b->flags |= flags;
 		return;
 	}
-	if (!(b = mem_alloc(sizeof(struct blacklist_entry) + strlen(host) + 1))) return;
+	b = mem_alloc(sizeof(struct blacklist_entry) + strlen(host) + 1);
 	b->flags = flags;
 	strcpy(b->host, host);
 	add_to_list(blacklist, b);
@@ -798,150 +817,51 @@ void free_blacklist()
 	free_list(blacklist);
 }
 
-struct list_head http_auth_basic_list = { &http_auth_basic_list, &http_auth_basic_list };
-
-struct http_auth_basic *find_auth_entry(unsigned char *url)
-{
-	struct http_auth_basic *a;
-	foreach(a, http_auth_basic_list) if (!strncmp(a->url, url, a->url_len)) return a;
-	return NULL;
-}
-
-void add_auth_entry(unsigned char* url, unsigned char *realm)
-{
-	struct http_auth_basic *a;
-	if ((a = find_auth_entry(url))) {
-		a->valid = 0;
-		if (a->realm) mem_free(a->realm);
-		a->realm = realm;
-		del_from_list(a);
-		add_to_list(http_auth_basic_list, a);
-		return;
-	}
-	if (!(a = mem_alloc(sizeof(struct http_auth_basic)))) return;
-	memset(a, 0, sizeof(struct http_auth_basic));
-	while (url[a->url_len] && !end_of_dir(url[a->url_len])) a->url_len++;
-	while (a->url_len && url[a->url_len - 1] != '/') a->url_len--;
-	if (a->url_len < 8) goto err;
-	if (!(a->url = mem_alloc(a->url_len + 1))) {
-		err:
-		mem_free(a);
-		return;
-	}
-	memcpy(a->url, url, a->url_len);
-	a->url[a->url_len] = 0;
-	a->realm = realm;
-	add_to_list(http_auth_basic_list, a);
-	return;
-}
-
-unsigned char *base64_encode(unsigned char *);
-
-unsigned char *find_auth(unsigned char *url)
-{
-	struct http_auth_basic *a = NULL;
-	unsigned char *uid, *b64;
-	if ((a = find_auth_entry(url))) {
-		int n;
-		if (!(uid = mem_alloc(strlen(a->uid) + strlen(a->passwd) + 2))) return NULL;
-		strcpy(uid, a->uid);
-		uid[n = strlen(a->uid)] = ':';
-		strcpy(&uid[n + 1], a->passwd);
-		b64 = base64_encode(uid);
-		mem_free(uid);
-		return b64;
-	}
-	return NULL;
-}
-
-void del_auth_entry(struct http_auth_basic *a)
-{
-		if (a->url) mem_free(a->url);
-		if (a->realm) mem_free(a->realm);
-		if (a->uid) mem_free(a->uid);
-		if (a->passwd) mem_free(a->passwd);
-		del_from_list(a);
-		mem_free(a);
-}
-
-void free_auth()
-{
-	while (!list_empty(http_auth_basic_list)) del_auth_entry(http_auth_basic_list.next);
-}
-
-unsigned char base64_chars[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
-
-unsigned char *base64_encode(unsigned char *in)
-{
-	unsigned char *out, *outstr;
-	int inlen = strlen(in);
-	if (!(outstr = out = mem_alloc(((inlen / 3) + 1) * 4 + 1 ))) return NULL;
-	while (inlen >= 3) {
-		*out++ = base64_chars[ (int)(*in >> 2) ];
-		*out++ = base64_chars[ (int)((*in << 4 | *(in + 1) >> 4) & 63) ];
-		*out++ = base64_chars[ (int)((*(in + 1) << 2 | *(in + 2) >> 6) & 63) ];
-		*out++ = base64_chars[ (int)(*(in + 2) & 63) ];
-		inlen -= 3; in += 3;
-	}
-	if (inlen == 1) {
-		*out++ = base64_chars[ (int)(*in >> 2) ];
-		*out++ = base64_chars[ (int)(*in << 4 & 63) ];
-		*out++ = '=';
-		*out++ = '=';
-	}
-	if (inlen == 2) {
-		*out++ = base64_chars[ (int)(*in >> 2) ];
-		*out++ = base64_chars[ (int)((*in << 4 | *(in + 1) >> 4) & 63) ];
-		*out++ = base64_chars[ (int)((*(in + 1) << 2) & 63) ];
-		*out++ = '=';
-	}
-	*out = 0;
-	return outstr;
-}
-
 struct s_msg_dsc msg_dsc[] = {
-	S_WAIT,			TEXT(T_WAITING_IN_QUEUE),
-	S_DNS,			TEXT(T_LOOKING_UP_HOST),
-	S_CONN,			TEXT(T_MAKING_CONNECTION),
-	S_SSL_NEG,		TEXT(T_SSL_NEGOTIATION),
-	S_SENT,			TEXT(T_REQUEST_SENT),
-	S_LOGIN,		TEXT(T_LOGGING_IN),
-	S_GETH,			TEXT(T_GETTING_HEADERS),
-	S_PROC,			TEXT(T_SERVER_IS_PROCESSING_REQUEST),
-	S_TRANS,		TEXT(T_TRANSFERRING),
+	{ S_WAIT,		TEXT_(T_WAITING_IN_QUEUE) },
+	{ S_DNS,		TEXT_(T_LOOKING_UP_HOST) },
+	{ S_CONN,		TEXT_(T_MAKING_CONNECTION) },
+	{ S_SSL_NEG,		TEXT_(T_SSL_NEGOTIATION) },
+	{ S_SENT,		TEXT_(T_REQUEST_SENT) },
+	{ S_LOGIN,		TEXT_(T_LOGGING_IN) },
+	{ S_GETH,		TEXT_(T_GETTING_HEADERS) },
+	{ S_PROC,		TEXT_(T_SERVER_IS_PROCESSING_REQUEST) },
+	{ S_TRANS,		TEXT_(T_TRANSFERRING) },
 
-	S_WAIT_REDIR,		TEXT(T_WAITING_FOR_REDIRECT_CONFIRMATION),
-	S_OK,			TEXT(T_OK),
-	S_INTERRUPTED,		TEXT(T_INTERRUPTED),
-	S_EXCEPT,		TEXT(T_SOCKET_EXCEPTION),
-	S_INTERNAL,		TEXT(T_INTERNAL_ERROR),
-	S_OUT_OF_MEM,		TEXT(T_OUT_OF_MEMORY),
-	S_NO_DNS,		TEXT(T_HOST_NOT_FOUND),
-	S_CANT_WRITE,		TEXT(T_ERROR_WRITING_TO_SOCKET),
-	S_CANT_READ,		TEXT(T_ERROR_READING_FROM_SOCKET),
-	S_MODIFIED,		TEXT(T_DATA_MODIFIED),
-	S_BAD_URL,		TEXT(T_BAD_URL_SYNTAX),
-	S_TIMEOUT,		TEXT(T_RECEIVE_TIMEOUT),
-	S_RESTART,		TEXT(T_REQUEST_MUST_BE_RESTARTED),
-	S_STATE,		TEXT(T_CANT_GET_SOCKET_STATE),
+	{ S_OK,			TEXT_(T_OK) },
+	{ S_INTERRUPTED,	TEXT_(T_INTERRUPTED) },
+	{ S_EXCEPT,		TEXT_(T_SOCKET_EXCEPTION) },
+	{ S_INTERNAL,		TEXT_(T_INTERNAL_ERROR) },
+	{ S_OUT_OF_MEM,		TEXT_(T_OUT_OF_MEMORY) },
+	{ S_NO_DNS,		TEXT_(T_HOST_NOT_FOUND) },
+	{ S_CANT_WRITE,		TEXT_(T_ERROR_WRITING_TO_SOCKET) },
+	{ S_CANT_READ,		TEXT_(T_ERROR_READING_FROM_SOCKET) },
+	{ S_MODIFIED,		TEXT_(T_DATA_MODIFIED) },
+	{ S_BAD_URL,		TEXT_(T_BAD_URL_SYNTAX) },
+	{ S_TIMEOUT,		TEXT_(T_RECEIVE_TIMEOUT) },
+	{ S_RESTART,		TEXT_(T_REQUEST_MUST_BE_RESTARTED) },
+	{ S_STATE,		TEXT_(T_CANT_GET_SOCKET_STATE) },
+	{ S_LARGE_FILE,		TEXT_(T_TOO_LARGE_FILE)},
 
-	S_HTTP_ERROR,		TEXT(T_BAD_HTTP_RESPONSE),
-	S_HTTP_100,		TEXT(T_HTTP_100),
-	S_HTTP_204,		TEXT(T_NO_CONTENT),
+	{ S_HTTP_ERROR,		TEXT_(T_BAD_HTTP_RESPONSE) },
+	{ S_HTTP_100,		TEXT_(T_HTTP_100) },
+	{ S_HTTP_204,		TEXT_(T_NO_CONTENT) },
 
-	S_FILE_TYPE,		TEXT(T_UNKNOWN_FILE_TYPE),
-	S_FILE_ERROR,		TEXT(T_ERROR_OPENING_FILE),
+	{ S_FILE_TYPE,		TEXT_(T_UNKNOWN_FILE_TYPE) },
+	{ S_FILE_ERROR,		TEXT_(T_ERROR_OPENING_FILE) },
 
-	S_FTP_ERROR,		TEXT(T_BAD_FTP_RESPONSE),
-	S_FTP_UNAVAIL,		TEXT(T_FTP_SERVICE_UNAVAILABLE),
-	S_FTP_LOGIN,		TEXT(T_BAD_FTP_LOGIN),
-	S_FTP_PORT,		TEXT(T_FTP_PORT_COMMAND_FAILED),
-	S_FTP_NO_FILE,		TEXT(T_FILE_NOT_FOUND),
-	S_FTP_FILE_ERROR,	TEXT(T_FTP_FILE_ERROR),
-#ifdef HAVE_SSL
-	S_SSL_ERROR,		TEXT(T_SSL_ERROR),
-#else
-	S_NO_SSL,		TEXT(T_NO_SSL),
-#endif
-	0,		NULL
+	{ S_FTP_ERROR,		TEXT_(T_BAD_FTP_RESPONSE) },
+	{ S_FTP_UNAVAIL,	TEXT_(T_FTP_SERVICE_UNAVAILABLE) },
+	{ S_FTP_LOGIN,		TEXT_(T_BAD_FTP_LOGIN) },
+	{ S_FTP_PORT,		TEXT_(T_FTP_PORT_COMMAND_FAILED) },
+	{ S_FTP_NO_FILE,	TEXT_(T_FILE_NOT_FOUND) },
+	{ S_FTP_FILE_ERROR,	TEXT_(T_FTP_FILE_ERROR) },
+
+	{ S_SSL_ERROR,		TEXT_(T_SSL_ERROR) },
+	{ S_NO_SSL,		TEXT_(T_NO_SSL) },
+
+	{ S_NO_SMB_CLIENT,	TEXT_(T_NO_SMB_CLIENT) },
+
+	{ S_WAIT_REDIR,		TEXT_(T_WAITING_FOR_REDIRECT_CONFIRMATION) },
+	{ 0,			NULL },
 };
