@@ -121,7 +121,11 @@ int install_timer(ttime t, void (*func)(void *), void *data)
 	tm->interval = t;
 	tm->func = func;
 	tm->data = data;
-	tm->id = timer_id++;
+	new_id:
+	tm->id = timer_id;
+	timer_id++;
+	if (timer_id == MAXINT) timer_id = 0;
+	foreach(tt, timers) if (tt->id == tm->id) goto new_id;
 	foreach(tt, timers) if (tt->interval >= t) break;
 	add_at_pos(tt->prev, tm);
 	return tm->id;
@@ -144,8 +148,12 @@ void kill_timer(int id)
 
 void *get_handler(int fd, int tp)
 {
-	if (fd < 0 || fd >= FD_SETSIZE) {
-		internal("get_handler: handle %d >= FD_SETSIZE %d", fd, FD_SETSIZE);
+	if (fd < 0)
+		internal("get_handler: handle %d", fd);
+	if (fd >= (int)FD_SETSIZE) {
+		error("too big handle %d");
+		fatal_tty_exit();
+		exit(RET_FATAL);
 		return NULL;
 	}
 	switch (tp) {
@@ -160,8 +168,12 @@ void *get_handler(int fd, int tp)
 
 void set_handlers(int fd, void (*read_func)(void *), void (*write_func)(void *), void (*error_func)(void *), void *data)
 {
-	if (fd < 0 || fd >= FD_SETSIZE) {
-		internal("set_handlers: handle %d >= FD_SETSIZE %d", fd, FD_SETSIZE);
+	if (fd < 0)
+		internal("set_handlers: handle %d", fd);
+	if (fd >= (int)FD_SETSIZE) {
+		error("too big handle %d");
+		fatal_tty_exit();
+		exit(RET_FATAL);
 		return;
 	}
 	threads[fd].read_func = read_func;
@@ -205,37 +217,49 @@ struct signal_handler {
 int signal_mask[NUM_SIGNALS];
 struct signal_handler signal_handlers[NUM_SIGNALS];
 
-int critical_section = 0;
+int signal_pipe[2];
 
-void check_for_select_race();
-
-void got_signal(int sig)
+void signal_break(void *data)
 {
+	char c;
+	while (can_read(signal_pipe[0])) read(signal_pipe[0], &c, 1);
+}
+
+SIGNAL_HANDLER void got_signal(int sig)
+{
+	int sv_errno = errno;
 	if (sig >= NUM_SIGNALS || sig < 0) {
-		error("ERROR: bad signal number: %d", sig);
-		return;
+		/*error("ERROR: bad signal number: %d", sig);*/
+		goto ret;
 	}
-	if (!signal_handlers[sig].fn) return;
+	if (!signal_handlers[sig].fn) goto ret;
 	if (signal_handlers[sig].critical) {
 		signal_handlers[sig].fn(signal_handlers[sig].data);
-		return;
+		goto ret;
 	}
 	signal_mask[sig] = 1;
-	check_for_select_race();
+	ret:
+	if (can_write(signal_pipe[1])) write(signal_pipe[1], "x", 1);
+	errno = sv_errno;
 }
 
+struct sigaction sa_zero;
+
 void install_signal_handler(int sig, void (*fn)(void *), void *data, int critical)
 {
-	struct sigaction sa;
+	struct sigaction sa = sa_zero;
 	if (sig >= NUM_SIGNALS || sig < 0) {
 		internal("bad signal number: %d", sig);
 		return;
 	}
-	memset(&sa, 0, sizeof sa);
 	if (!fn) sa.sa_handler = SIG_IGN;
-	else sa.sa_handler = got_signal;
+	else sa.sa_handler = (void *)got_signal;
+#ifdef HAVE_SIGFILLSET
 	sigfillset(&sa.sa_mask);
-	/*sa.sa_flags = SA_RESTART;*/
+#else
+	memset(&sa.sa_mask, -1, sizeof sa.sa_mask);
+#endif
+	sa.sa_flags = SA_RESTART;
 	if (!fn) sigaction(sig, &sa, NULL);
 	signal_handlers[sig].fn = fn;
 	signal_handlers[sig].data = data;
@@ -243,33 +267,22 @@ void install_signal_handler(int sig, void (*fn)(void *), void *data, int critica
 	if (fn) sigaction(sig, &sa, NULL);
 }
 
-int pending_alarm = 0;
-
-void alarm_handler(void *x)
+void interruptible_signal(int sig, int in)
 {
-	pending_alarm = 0;
-	check_for_select_race();
-}
-
-void check_for_select_race()
-{
-	if (critical_section) {
-#ifdef SIGALRM
-		install_signal_handler(SIGALRM, alarm_handler, NULL, 1);
-#endif
-		pending_alarm = 1;
-#ifdef HAVE_ALARM
-		/*alarm(1);*/
-#endif
+	struct sigaction sa = sa_zero;
+	if (sig >= NUM_SIGNALS || sig < 0) {
+		internal("bad signal number: %d", sig);
+		return;
 	}
-}
-
-void uninstall_alarm()
-{
-	pending_alarm = 0;
-#ifdef HAVE_ALARM
-	alarm(0);
+	if (!signal_handlers[sig].fn) return;
+	sa.sa_handler = (void *)got_signal;
+#ifdef HAVE_SIGFILLSET
+	sigfillset(&sa.sa_mask);
+#else
+	memset(&sa.sa_mask, -1, sizeof sa.sa_mask);
 #endif
+	if (!in) sa.sa_flags = SA_RESTART;
+	sigaction(sig, &sa, NULL);
 }
 
 int check_signals()
@@ -290,7 +303,7 @@ void sigchld(void *p)
 #ifndef WNOHANG
 	wait(NULL);
 #else
-	while ((int)waitpid(-1, NULL, WNOHANG) > 0) ;
+	while (waitpid(-1, NULL, WNOHANG) > 0) ;
 #endif
 }
 
@@ -303,6 +316,9 @@ int terminate = 0;
 
 void select_loop(void (*init)())
 {
+	struct stat st;
+	if (stat(".", &st) && getenv("HOME")) chdir(getenv("HOME"));
+	memset(&sa_zero, 0, sizeof sa_zero);
 	memset(signal_mask, 0, sizeof signal_mask);
 	memset(signal_handlers, 0, sizeof signal_handlers);
 	FD_ZERO(&w_read);
@@ -311,6 +327,14 @@ void select_loop(void (*init)())
 	w_max = 0;
 	last_time = get_time();
 	signal(SIGPIPE, SIG_IGN);
+	if (c_pipe(signal_pipe)) {
+		error("ERROR: can't create pipe for signal handling");
+		fatal_tty_exit();
+		exit(RET_FATAL);
+	}
+	fcntl(signal_pipe[0], F_SETFL, O_NONBLOCK);
+	fcntl(signal_pipe[1], F_SETFL, O_NONBLOCK);
+	set_handlers(signal_pipe[0], signal_break, NULL, NULL, NULL);
 	init();
 	CHK_BH;
 	while (!terminate) {
@@ -320,12 +344,24 @@ void select_loop(void (*init)())
 		check_signals();
 		check_timers();
 		redraw_all_terminals();
+#ifdef OS_BAD_SIGNALS
+	/* Cygwin has buggy signals that sometimes don't interrupt select.
+	   So don't wait indefinitely in it. */
+		tv.tv_sec = 1;
+		tv.tv_usec = 0;
+		tm = &tv;
+#endif
 		if (!list_empty(timers)) {
-			ttime tt = ((struct timer *)&timers)->next->interval + 1;
+			ttime tt = ((struct timer *)timers.next)->interval + 1;
 			if (tt < 0) tt = 0;
-			tv.tv_sec = tt / 1000;
-			tv.tv_usec = (tt % 1000) * 1000;
-			tm = &tv;
+#ifdef OS_BAD_SIGNALS
+			if (tt < 1000)
+#endif
+			{
+				tv.tv_sec = tt / 1000;
+				tv.tv_usec = (tt % 1000) * 1000;
+				tm = &tv;
+			}
 		}
 		memcpy(&x_read, &w_read, sizeof(fd_set));
 		memcpy(&x_write, &w_write, sizeof(fd_set));
@@ -333,9 +369,7 @@ void select_loop(void (*init)())
 		/*rep_sel:*/
 		if (terminate) break;
 		if (!w_max && list_empty(timers)) break;
-		critical_section = 1;
 		if (check_signals()) {
-			critical_section = 0;
 			continue;
 		}
 			/*{
@@ -349,13 +383,13 @@ void select_loop(void (*init)())
 				fflush(stdout);
 			}*/
 		if ((n = select(w_max, &x_read, &x_write, &x_error, tm)) < 0) {
-			critical_section = 0;
-			uninstall_alarm();
-			if (errno != EINTR) error("ERROR: select failed: %d", errno);
+			if (errno != EINTR) {
+				error("ERROR: select failed: %s", strerror(errno));
+				fatal_tty_exit();
+				exit(RET_FATAL);
+			}
 			continue;
 		}
-		critical_section = 0;
-		uninstall_alarm();
 		check_signals();
 		/*printf("sel: %d\n", n);*/
 		check_timers();
