@@ -4,9 +4,11 @@
 #define VERSION_STRING			VERSION
 #endif
 
+#if 0
 #define DEBUG
 #define LEAK_DEBUG
 #define LEAK_DEBUG_LIST
+#endif
 
 #define LINKS_MANUAL_URL		"http://links.sourceforge.net/docs/manual-0.90-en/"
 
@@ -14,6 +16,8 @@
 #define LINKS_PORT			23456
 #define MAX_BIND_TRIES			3
 
+#define FG_POLL_TIME			500
+
 #define DNS_TIMEOUT			3600000UL
 
 #define HTTP_KEEPALIVE_TIMEOUT		60000
@@ -22,11 +26,14 @@
 #define KEEPALIVE_CHECK_TIME		20000
 
 #define MAX_REDIRECTS			10
+#define MAX_CACHED_REDIRECTS		5
+
+#define DOWNLOAD_NAME_TRIES		10000
 
 #define MEMORY_CACHE_GC_PERCENT		9/10
 #define MAX_CACHED_OBJECT		1/4
 
-#define MAX_HISTORY_ITEMS		256
+#define MAX_HISTORY_ITEMS		4096
 #define MENU_HOTKEY_SPACE		2
 
 #define COL(x)				((x)*0x100)
@@ -68,7 +75,7 @@
 #define COLOR_STATUS			COL(070)
 #define COLOR_STATUS_BG			COL(007)
 
-#define ESC_TIMEOUT			100
+#define ESC_TIMEOUT			200
 
 #define DISPLAY_TIME_MIN		200
 #define DISPLAY_TIME			20
@@ -82,6 +89,8 @@
 #define HTML_FRAME_CHAR_HEIGHT		16
 #define HTML_TABLE_2ND_PASS
 #define HTML_DEFAULT_INPUT_SIZE		20
+#define HTML_DEFAULT_TEXTAREA_WIDTH	40
+#define HTML_DEFAULT_TEXTAREA_HEIGHT	7
 
 #define MAX_INPUT_URL_LEN		4096
 
@@ -92,3 +101,6 @@
 #define RESOURCE_INFO_REFRESH		100
 
 #define DOWN_DLG_MIN			20
+
+#define MAX_UID_LEN			20
+#define MAX_PASSWD_LEN			20
