@@ -38,11 +38,12 @@ unsigned char *get_cwd()
 	int bufsize = 128;
 	unsigned char *buf;
 	while (1) {
-		if (!(buf = mem_alloc(bufsize))) return NULL;
+		buf = mem_alloc(bufsize);
 		if (getcwd(buf, bufsize)) return buf;
 		mem_free(buf);
 		if (errno == EINTR) continue;
 		if (errno != ERANGE) return NULL;
+		if ((unsigned)bufsize > MAXINT - 128) overalloc();
 		bufsize += 128;
 	}
 	return NULL;
@@ -58,17 +59,17 @@ struct list_head terminals = {&terminals, &terminals};
 void alloc_term_screen(struct terminal *term, int x, int y)
 {
 	unsigned *s, *t;
-	if ((s = mem_realloc(term->screen, x * y * sizeof(unsigned)))) {
-		if ((t = mem_realloc(term->last_screen, x * y * sizeof(unsigned)))) {
-			memset(t, -1, x * y * sizeof(unsigned));
-			term->x = x;
-			term->y = y;
-			term->last_screen = t;
-			memset(s, 0, x * y * sizeof(unsigned));
-			term->screen = s;
-			term->dirty = 1;
-		}
-	}
+	if (x && (unsigned)x * (unsigned)y / (unsigned)x != (unsigned)y) overalloc();
+	if ((unsigned)x * (unsigned)y > MAXINT / sizeof(unsigned)) overalloc();
+	s = mem_realloc(term->screen, x * y * sizeof(unsigned));
+	t = mem_realloc(term->last_screen, x * y * sizeof(unsigned));
+	memset(t, -1, x * y * sizeof(unsigned));
+	term->x = x;
+	term->y = y;
+	term->last_screen = t;
+	memset(s, 0, x * y * sizeof(unsigned));
+	term->screen = s;
+	term->dirty = 1;
 }
 
 void in_term(struct terminal *);
@@ -164,10 +165,7 @@ void add_window_at_pos(struct terminal *term, void (*handler)(struct window *, s
 	struct window *win;
 	ev.x = term->x;
 	ev.y = term->y;
-	if (!(win = mem_alloc(sizeof (struct window)))) {
-		mem_free(data);
-		return;
-	}
+	win = mem_alloc(sizeof (struct window));
 	win->handler = handler;
 	win->data = data;
 	win->term = term;
@@ -178,7 +176,7 @@ void add_window_at_pos(struct terminal *term, void (*handler)(struct window *, s
 
 void add_window(struct terminal *term, void (*handler)(struct window *, struct event *, int), void *data)
 {
-	add_window_at_pos(term, handler, data, (struct window *)&term->windows);
+	add_window_at_pos(term, handler, data, (struct window *)(void *)&term->windows);
 }
 
 void delete_window(struct window *win)
@@ -238,7 +236,7 @@ void empty_window_handler(struct window *win, struct event *ev, int fwd)
 	void (*fn)(void *) = ewd->fn;
 	void *data = ewd->data;
 	if (ewd->b) return;
-	switch (ev->ev) {
+	switch ((int)ev->ev) {
 		case EV_INIT:
 		case EV_RESIZE:
 		case EV_REDRAW:
@@ -259,7 +257,7 @@ void empty_window_handler(struct window *win, struct event *ev, int fwd)
 void add_empty_window(struct terminal *term, void (*fn)(void *), void *data)
 {
 	struct ewd *ewd;
-	if (!(ewd = mem_alloc(sizeof(struct ewd)))) return;
+	ewd = mem_alloc(sizeof(struct ewd));
 	ewd->fn = fn;
 	ewd->data = data;
 	ewd->b = 0;
@@ -273,7 +271,7 @@ void free_term_specs()
 
 struct list_head term_specs = {&term_specs, &term_specs};
 
-struct term_spec dumb_term = { NULL, NULL, "", 0, 1, 0, 0, 0 };
+struct term_spec dumb_term = { NULL, NULL, "", 0, 1, 0, 0, 0, 0 };
 
 struct term_spec *get_term_spec(unsigned char *term)
 {
@@ -286,7 +284,7 @@ struct term_spec *new_term_spec(unsigned char *term)
 {
 	struct term_spec *t;
 	foreach(t, term_specs) if (!strcasecmp(t->term, term)) return t;
-	if (!(t = mem_alloc(sizeof(struct term_spec)))) return NULL;
+	t = mem_alloc(sizeof(struct term_spec));
 	memcpy(t, &dumb_term, sizeof(struct term_spec));
 	if (strlen(term) < MAX_TERM_LEN) strcpy(t->term, term);
 	else memcpy(t->term, term, MAX_TERM_LEN - 1), t->term[MAX_TERM_LEN - 1] = 0;
@@ -305,10 +303,7 @@ struct terminal *init_term(int fdin, int fdout, void (*root_window)(struct windo
 {
 	struct terminal *term;
 	struct window *win;
-	if (!(term = mem_alloc(sizeof (struct terminal)))) {
-		check_if_no_terminal();
-		return NULL;
-	}
+	term = mem_alloc(sizeof (struct terminal));
 	memset(term, 0, sizeof(struct terminal));
 	term->fdin = fdin;
 	term->fdout = fdout;
@@ -330,11 +325,7 @@ struct terminal *init_term(int fdin, int fdout, void (*root_window)(struct windo
 	term->input_queue = DUMMY;
 	term->qlen = 0;
 	init_list(term->windows);
-	if (!(win = mem_alloc(sizeof (struct window)))) {
-		mem_free(term);
-		check_if_no_terminal();
-		return NULL;
-	}
+	win = mem_alloc(sizeof (struct window));
 	win->handler = root_window;
 	win->data = NULL;
 	win->term = term;
@@ -350,10 +341,8 @@ void in_term(struct terminal *term)
 	struct event *ev;
 	int r;
 	unsigned char *iq;
-	if (!(iq = mem_realloc(term->input_queue, term->qlen + ALLOC_GR))) {
-		destroy_terminal(term);
-		return;
-	}
+	if ((unsigned)term->qlen + ALLOC_GR > MAXINT) overalloc();
+	iq = mem_realloc(term->input_queue, term->qlen + ALLOC_GR);
 	term->input_queue = iq;
 	if ((r = read(term->fdin, iq + term->qlen, ALLOC_GR)) <= 0) {
 		if (r == -1 && errno != ECONNRESET) error("ERROR: error %d on terminal: could not read event", errno);
@@ -362,7 +351,7 @@ void in_term(struct terminal *term)
 	}
 	term->qlen += r;
 	test_queue:
-	if (term->qlen < sizeof(struct event)) return;
+	if ((size_t)term->qlen < sizeof(struct event)) return;
 	ev = (struct event *)iq;
 	r = sizeof(struct event);
 	if (ev->ev != EV_INIT && ev->ev != EV_RESIZE && ev->ev != EV_REDRAW && ev->ev != EV_KBD && ev->ev != EV_MOUSE && ev->ev != EV_ABORT) {
@@ -371,9 +360,9 @@ void in_term(struct terminal *term)
 	}
 	if (ev->ev == EV_INIT) {
 		int init_len;
-		if (term->qlen < sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * sizeof(int)) return;
+		if ((size_t)term->qlen < sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * sizeof(int)) return;
 		init_len = *(int *)(iq + sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + sizeof(int));
-		if (term->qlen < sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * sizeof(int) + init_len) return;
+		if ((size_t)term->qlen < sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * sizeof(int) + init_len) return;
 		memcpy(term->term, iq + sizeof(struct event), MAX_TERM_LEN);
 		term->term[MAX_TERM_LEN - 1] = 0;
 		memcpy(term->cwd, iq + sizeof(struct event) + MAX_TERM_LEN, MAX_CWD_LEN);
@@ -404,8 +393,8 @@ void in_term(struct terminal *term)
 			ev->y = term->y;
 			goto send_redraw;
 		}
-		else if (ev->ev == EV_KBD && ev->x == KBD_CTRL_C) ((struct window *)&term->windows)->prev->handler(term->windows.prev, ev, 0);
-		else ((struct window *)&term->windows)->next->handler(term->windows.next, ev, 0);
+		else if (ev->ev == EV_KBD && ev->x == KBD_CTRL_C) ((struct window *)term->windows.prev)->handler(term->windows.prev, ev, 0);
+		else ((struct window *)term->windows.next)->handler(term->windows.next, ev, 0);
 	}
 	if (ev->ev == EV_ABORT) {
 		destroy_terminal(term);
@@ -418,7 +407,7 @@ void in_term(struct terminal *term)
 	goto test_queue;
 }
 
-inline int getcompcode(int c)
+static inline int getcompcode(int c)
 {
 	return (c<<1 | (c&4)>>2) & 7;
 }
@@ -457,7 +446,7 @@ unsigned char frame_restrict[48] = {
 	unsigned char A = ch >> 8 & 0x7f;						\
 	if (s->mode == TERM_LINUX) {							\
 		if (s->m11_hack) {							\
-			if (ch >> 15 != mode) {						\
+			if ((int)(ch >> 15) != mode) {					\
 				if (!(mode = ch >> 15)) add_to_str(&a, &l, "\033[10m");	\
 				else add_to_str(&a, &l, "\033[11m");			\
 			}								\
@@ -466,7 +455,7 @@ unsigned char frame_restrict[48] = {
 			if (frame_restrict[c - 176]) c = frame_restrict[c - 176];	\
 		}									\
 	} else if (s->mode == TERM_VT100) {						\
-		if (ch >> 15 != mode) {							\
+		if ((int)(ch >> 15) != mode) {						\
 			if (!(mode = ch >> 15)) add_to_str(&a, &l, "\x0f");		\
 			else add_to_str(&a, &l, "\x0e");				\
 		}									\
@@ -506,22 +495,23 @@ void redraw_all_terminals()
 void redraw_screen(struct terminal *term)
 {
 	int x, y, p = 0;
-	int cx = -1, cy = -1;
+	int cx = term->lcx, cy = term->lcy;
 	unsigned char *a;
 	int attrib = -1;
 	int mode = -1;
 	int l = 0;
 	struct term_spec *s;
 	if (!term->dirty || (term->master && is_blocked())) return;
-	if (!(a = init_str())) return;
+	a = init_str();
 	s = term->spec;
 	for (y = 0; y < term->y; y++)
 		for (x = 0; x < term->x; x++, p++) {
 			if (y == term->y - 1 && x == term->x - 1) break;
 			if (term->screen[p] == term->last_screen[p]) continue;
-			if ((term->screen[p] & 0x3800) == (term->last_screen[p] & 0x3800) && ((term->screen[p] & 0xff) == 0 || (term->screen[p] & 0xff) == 1 || (term->screen[p] & 0xff) == ' ') && ((term->last_screen[p] & 0xff) == 0 || (term->last_screen[p] & 0xff) == 1 || (term->last_screen[p] & 0xff) == ' ')) continue;
+			if ((term->screen[p] & 0x3800) == (term->last_screen[p] & 0x3800) && ((term->screen[p] & 0xff) == 0 || (term->screen[p] & 0xff) == 1 || (term->screen[p] & 0xff) == ' ') && ((term->last_screen[p] & 0xff) == 0 || (term->last_screen[p] & 0xff) == 1 || (term->last_screen[p] & 0xff) == ' ') && (x != term->cx || y != term->cy)) continue;
+			term->last_screen[p] = term->screen[p];
 			if (cx == x && cy == y) goto pc;/*PRINT_CHAR(p)*/
-			else if (cy == y && x - cx < 10) {
+			else if (cy == y && x - cx < 10 && x - cx > 0) {
 				int i;
 				for (i = x - cx; i >= 0; i--) PRINT_CHAR(p - i);
 			} else {
@@ -541,7 +531,9 @@ void redraw_screen(struct terminal *term)
 		if (s->mode == TERM_LINUX && s->m11_hack) add_to_str(&a, &l, "\033[10m");
 		if (s->mode == TERM_VT100) add_to_str(&a, &l, "\x0f");
 	}
-	if (l || term->cx != term->lcx || term->cy != term->lcy) {
+	term->lcx = cx;
+	term->lcy = cy;
+	if (term->cx != term->lcx || term->cy != term->lcy) {
 		term->lcx = term->cx;
 		term->lcy = term->cy;
 		add_to_str(&a, &l, "\033[");
@@ -554,7 +546,6 @@ void redraw_screen(struct terminal *term)
 	hard_write(term->fdout, a, l);
 	if (l && term->master) done_draw();
 	mem_free(a);
-	memcpy(term->last_screen, term->screen, term->x * term->y * sizeof(int));
 	term->dirty = 0;
 }
 
@@ -580,7 +571,7 @@ void destroy_terminal(struct terminal *term)
 		free_all_itrms();
 #ifndef NO_FORK_ON_EXIT
 		if (!list_empty(terminals)) {
-			if (fork()) exit(0);
+			if (fork() > 0) _exit(0);
 		}
 #endif
 	}
@@ -728,7 +719,7 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 				unlink(delete);
 				return;
 			}
-			if (!(param = mem_alloc(strlen(path) + strlen(delete) + 3))) return;
+			param = mem_alloc(strlen(path) + strlen(delete) + 3);
 			param[0] = fg;
 			strcpy(param + 1, path);
 			strcpy(param + 1 + strlen(path) + 1, delete);
@@ -750,14 +741,13 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 		}
 	} else {
 		unsigned char *data;
-		if ((data = mem_alloc(strlen(path) + strlen(delete) + 4))) {
-			data[0] = 0;
-			data[1] = fg;
-			strcpy(data + 2, path);
-			strcpy(data + 3 + strlen(path), delete);
-			hard_write(term->fdout, data, strlen(path) + strlen(delete) + 4);
-			mem_free(data);
-		}
+		data = mem_alloc(strlen(path) + strlen(delete) + 4);
+		data[0] = 0;
+		data[1] = fg;
+		strcpy(data + 2, path);
+		strcpy(data + 3 + strlen(path), delete);
+		hard_write(term->fdout, data, strlen(path) + strlen(delete) + 4);
+		mem_free(data);
 		/*char x = 0;
 		hard_write(term->fdout, &x, 1);
 		x = fg;
@@ -770,7 +760,7 @@ void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char
 void do_terminal_function(struct terminal *term, unsigned char code, unsigned char *data)
 {
 	unsigned char *x_data;
-	if (!(x_data = mem_alloc(strlen(data) + 2))) return;
+	x_data = mem_alloc(strlen(data) + 2);
 	x_data[0] = code;
 	strcpy(x_data + 1, data);
 	exec_on_terminal(term, NULL, x_data, 0);
@@ -779,8 +769,27 @@ void do_terminal_function(struct terminal *term, unsigned char code, unsigned ch
 
 void set_terminal_title(struct terminal *term, unsigned char *title)
 {
-	if (term->title && !strcmp(title, term->title)) return;
+	int i;
+	for (i = 0; i < 10000; i++) if (!title[i]) goto s;
+	title[10000] = 0;
+	s:
+	if (strchr(title, 1)) {
+		unsigned char *a, *b;
+		for (a = title, b = title; *a; a++) if (*a != 1) *b++ = *a;
+		*b = 0;
+	}
+	if (term->title && !strcmp(title, term->title)) goto ret;
 	if (term->title) mem_free(term->title);
 	term->title = stracpy(title);
+#ifdef SET_WINDOW_TITLE_UTF_8
+	{
+		struct conv_table *table;
+		mem_free(title);
+		table = get_translation_table(term->spec->charset, get_cp_index("utf-8"));
+		title = convert_string(table, term->title, strlen(term->title));
+	}
+#endif
 	do_terminal_function(term, TERM_FN_TITLE, title);
+	ret:
+	mem_free(title);
 }
