@@ -3,6 +3,7 @@
 struct list_head mailto_prog = { &mailto_prog, &mailto_prog };
 struct list_head telnet_prog = { &telnet_prog, &telnet_prog };
 struct list_head tn3270_prog = { &tn3270_prog, &tn3270_prog };
+struct list_head mms_prog = { &mms_prog, &mms_prog };
 
 struct list_head assoc = { &assoc, &assoc };
 
@@ -105,20 +106,22 @@ void free_types()
 	free_list(telnet_prog);
 	foreach(p, tn3270_prog) mem_free(p->prog);
 	free_list(tn3270_prog);
+	foreach(p, mms_prog) mem_free(p->prog);
+	free_list(mms_prog);
 }
 
 unsigned char *ct_msg[] = {
-	TEXT(T_LABEL),
-	TEXT(T_CONTENT_TYPES),
-	TEXT(T_PROGRAM__IS_REPLACED_WITH_FILE_NAME),
+	TEXT_(T_LABEL),
+	TEXT_(T_CONTENT_TYPES),
+	TEXT_(T_PROGRAM__IS_REPLACED_WITH_FILE_NAME),
 #ifdef ASSOC_BLOCK
-	TEXT(T_BLOCK_TERMINAL_WHILE_PROGRAM_RUNNING),
+	TEXT_(T_BLOCK_TERMINAL_WHILE_PROGRAM_RUNNING),
 #endif
 #ifdef ASSOC_CONS_XWIN
-	TEXT(T_RUN_ON_TERMINAL),
-	TEXT(T_RUN_IN_XWINDOW),
+	TEXT_(T_RUN_ON_TERMINAL),
+	TEXT_(T_RUN_IN_XWINDOW),
 #endif
-	TEXT(T_ASK_BEFORE_OPENING),
+	TEXT_(T_ASK_BEFORE_OPENING),
 };
 
 void add_ct_fn(struct dialog_data *dlg)
@@ -192,8 +195,13 @@ void update_assoc(struct assoc *new)
 		}
 		return;
 	}
+	foreach(repl, assoc) if (!strcmp(repl->label, new->label) && !strcmp(repl->ct, new->ct) && !strcmp(repl->prog, new->prog) && repl->block == new->block && repl->cons == new->cons && repl->xwin == new->xwin && repl->ask == new->ask && repl->system == new->system) {
+		del_from_list(repl);
+		add_to_list(assoc, repl);
+		return;
+	}
 	new->cnt = get_assoc_cnt();
-	if (!(repl = mem_alloc(sizeof(struct assoc)))) return;
+	repl = mem_alloc(sizeof(struct assoc));
 	add_to_list(assoc, repl);
 	replace:
 	repl->label = stracpy(new->label);
@@ -209,7 +217,7 @@ void update_assoc(struct assoc *new)
 
 void really_del_ct(void *fcp)
 {
-	int fc = (int)fcp;
+	tcount fc = (int)fcp;
 	struct assoc *del;
 	foreach(del, assoc) if (del->cnt == fc) goto ok;
 	return;
@@ -221,7 +229,7 @@ void menu_del_ct(struct terminal *term, void *fcp, void *xxx2)
 {
 	unsigned char *str;
 	int l;
-	int fc = (int)fcp;
+	tcount fc = (int)fcp;
 	struct assoc *del;
 	foreach(del, assoc) if (del->cnt == fc) goto ok;
 	return;
@@ -230,13 +238,13 @@ void menu_del_ct(struct terminal *term, void *fcp, void *xxx2)
 	add_to_str(&str, &l, del->ct);
 	add_to_str(&str, &l, " -> ");
 	add_to_str(&str, &l, del->prog);
-	msg_box(term, getml(str, NULL), TEXT(T_DELETE_ASSOCIATION), AL_CENTER | AL_EXTD_TEXT, TEXT(T_DELETE_ASSOCIATION), ": ", str, "?", NULL, fcp, 2, TEXT(T_YES), really_del_ct, B_ENTER, TEXT(T_NO), NULL, B_ESC);
+	msg_box(term, getml(str, NULL), TEXT_(T_DELETE_ASSOCIATION), AL_CENTER | AL_EXTD_TEXT, TEXT_(T_DELETE_ASSOCIATION), ": ", str, "?", NULL, fcp, 2, TEXT_(T_YES), really_del_ct, B_ENTER, TEXT_(T_NO), NULL, B_ESC);
 }
 
 void menu_add_ct(struct terminal *term, void *fcp, void *xxx2)
 {
 	int p;
-	int fc = (int)fcp;
+	tcount fc = (int)fcp;
 	struct assoc *new, *from;
 	unsigned char *label;
 	unsigned char *ct;
@@ -248,16 +256,16 @@ void menu_add_ct(struct terminal *term, void *fcp, void *xxx2)
 	}
 	from = NULL;
 	ok:
-	if (!(d = mem_alloc(sizeof(struct dialog) + 10 * sizeof(struct dialog_item) + sizeof(struct assoc) + 3 * MAX_STR_LEN))) return;
+	d = mem_alloc(sizeof(struct dialog) + 10 * sizeof(struct dialog_item) + sizeof(struct assoc) + 3 * MAX_STR_LEN);
 	memset(d, 0, sizeof(struct dialog) + 10 * sizeof(struct dialog_item) + sizeof(struct assoc) + 3 * MAX_STR_LEN);
 	new = (struct assoc *)&d->items[10];
 	new->label = label = (unsigned char *)(new + 1);
 	new->ct = ct = label + MAX_STR_LEN;
 	new->prog = prog = ct + MAX_STR_LEN;
 	if (from) {
-		strncpy(label, from->label, MAX_STR_LEN - 1);
-		strncpy(ct, from->ct, MAX_STR_LEN - 1);
-		strncpy(prog, from->prog, MAX_STR_LEN - 1);
+		safe_strncpy(label, from->label, MAX_STR_LEN);
+		safe_strncpy(ct, from->ct, MAX_STR_LEN);
+		safe_strncpy(prog, from->prog, MAX_STR_LEN);
 		new->block = from->block;
 		new->cons = from->cons;
 		new->xwin = from->xwin;
@@ -269,7 +277,7 @@ void menu_add_ct(struct terminal *term, void *fcp, void *xxx2)
 		new->ask = 1;
 		new->system = SYSTEM_ID;
 	}
-	d->title = TEXT(T_ASSOCIATION);
+	d->title = TEXT_(T_ASSOCIATION);
 	d->fn = add_ct_fn;
 	d->refresh = (void (*)(void *))update_assoc;
 	d->refresh_data = new;
@@ -305,18 +313,18 @@ void menu_add_ct(struct terminal *term, void *fcp, void *xxx2)
 	d->items[p].type = D_BUTTON;
 	d->items[p].gid = B_ENTER;
 	d->items[p].fn = ok_dialog;
-	d->items[p++].text = TEXT(T_OK);
+	d->items[p++].text = TEXT_(T_OK);
 	d->items[p].type = D_BUTTON;
 	d->items[p].gid = B_ESC;
-	d->items[p].text = TEXT(T_CANCEL);
+	d->items[p].text = TEXT_(T_CANCEL);
 	d->items[p++].fn = cancel_dialog;
 	d->items[p++].type = D_END;
 	do_dialog(term, d, getml(d, NULL));
 }
 
 struct menu_item mi_no_assoc[] = {
-	TEXT(T_NO_ASSOCIATIONS), "", M_BAR, NULL, NULL, 0, 0,
-	NULL, NULL, 0, NULL, NULL, 0, 0
+	{ TEXT_(T_NO_ASSOCIATIONS), "", M_BAR, NULL, NULL, 0, 0 },
+	{ NULL, NULL, 0, NULL, NULL, 0, 0 },
 };
 
 void menu_list_assoc(struct terminal *term, void *fn, void *xxx)
@@ -333,8 +341,8 @@ void menu_list_assoc(struct terminal *term, void *fn, void *xxx)
 }
 
 unsigned char *ext_msg[] = {
-	TEXT(T_EXTENSION_S),
-	TEXT(T_CONTENT_TYPE),
+	TEXT_(T_EXTENSION_S),
+	TEXT_(T_CONTENT_TYPE),
 };
 
 void add_ext_fn(struct dialog_data *dlg)
@@ -387,8 +395,13 @@ void update_ext(struct extension *new)
 		}
 		return;
 	}
+	foreach(repl, extensions) if (!strcmp(repl->ext, new->ext) && !strcmp(repl->ct, new->ct)) {
+		del_from_list(repl);
+		add_to_list(extensions, repl);
+		return;
+	}
 	new->cnt = get_assoc_cnt();
-	if (!(repl = mem_alloc(sizeof(struct extension)))) return;
+	repl = mem_alloc(sizeof(struct extension));
 	add_to_list(extensions, repl);
 	replace:
 	repl->ext = stracpy(new->ext);
@@ -398,7 +411,7 @@ void update_ext(struct extension *new)
 
 void really_del_ext(void *fcp)
 {
-	int fc = (int)fcp;
+	tcount fc = (int)fcp;
 	struct extension *del;
 	foreach(del, extensions) if (del->cnt == fc) goto ok;
 	return;
@@ -410,7 +423,7 @@ void menu_del_ext(struct terminal *term, void *fcp, void *xxx2)
 {
 	unsigned char *str;
 	int l;
-	int fc = (int)fcp;
+	tcount fc = (int)fcp;
 	struct extension *del;
 	foreach(del, extensions) if (del->cnt == fc) goto ok;
 	return;
@@ -419,12 +432,12 @@ void menu_del_ext(struct terminal *term, void *fcp, void *xxx2)
 	add_to_str(&str, &l, del->ext);
 	add_to_str(&str, &l, " -> ");
 	add_to_str(&str, &l, del->ct);
-	msg_box(term, getml(str, NULL), TEXT(T_DELETE_EXTENSION), AL_CENTER | AL_EXTD_TEXT, TEXT(T_DELETE_EXTENSION), " ", str, "?", NULL, fcp, 2, TEXT(T_YES), really_del_ext, B_ENTER, TEXT(T_NO), NULL, B_ESC);
+	msg_box(term, getml(str, NULL), TEXT_(T_DELETE_EXTENSION), AL_CENTER | AL_EXTD_TEXT, TEXT_(T_DELETE_EXTENSION), " ", str, "?", NULL, fcp, 2, TEXT_(T_YES), really_del_ext, B_ENTER, TEXT_(T_NO), NULL, B_ESC);
 }
 
 void menu_add_ext(struct terminal *term, void *fcp, void *xxx2)
 {
-	int fc = (int)fcp;
+	tcount fc = (int)fcp;
 	struct extension *new, *from;
 	unsigned char *ext;
 	unsigned char *ct;
@@ -435,17 +448,17 @@ void menu_add_ext(struct terminal *term, void *fcp, void *xxx2)
 	}
 	from = NULL;
 	ok:
-	if (!(d = mem_alloc(sizeof(struct dialog) + 5 * sizeof(struct dialog_item) + sizeof(struct extension) + 2 * MAX_STR_LEN))) return;
+	d = mem_alloc(sizeof(struct dialog) + 5 * sizeof(struct dialog_item) + sizeof(struct extension) + 2 * MAX_STR_LEN);
 	memset(d, 0, sizeof(struct dialog) + 5 * sizeof(struct dialog_item) + sizeof(struct extension) + 2 * MAX_STR_LEN);
 	new = (struct extension *)&d->items[5];
 	new->ext = ext = (unsigned char *)(new + 1);
 	new->ct = ct = ext + MAX_STR_LEN;
 	if (from) {
-		strncpy(ext, from->ext, MAX_STR_LEN - 1);
-		strncpy(ct, from->ct, MAX_STR_LEN - 1);
+		safe_strncpy(ext, from->ext, MAX_STR_LEN);
+		safe_strncpy(ct, from->ct, MAX_STR_LEN);
 		new->cnt = from->cnt;
 	}
-	d->title = TEXT(T_EXTENSION);
+	d->title = TEXT_(T_EXTENSION);
 	d->fn = add_ext_fn;
 	d->refresh = (void (*)(void *))update_ext;
 	d->refresh_data = new;
@@ -460,18 +473,18 @@ void menu_add_ext(struct terminal *term, void *fcp, void *xxx2)
 	d->items[2].type = D_BUTTON;
 	d->items[2].gid = B_ENTER;
 	d->items[2].fn = ok_dialog;
-	d->items[2].text = TEXT(T_OK);
+	d->items[2].text = TEXT_(T_OK);
 	d->items[3].type = D_BUTTON;
 	d->items[3].gid = B_ESC;
-	d->items[3].text = TEXT(T_CANCEL);
+	d->items[3].text = TEXT_(T_CANCEL);
 	d->items[3].fn = cancel_dialog;
 	d->items[4].type = D_END;
 	do_dialog(term, d, getml(d, NULL));
 }
 
 struct menu_item mi_no_ext[] = {
-	TEXT(T_NO_EXTENSIONS), "", M_BAR, NULL, NULL, 0, 0,
-	NULL, NULL, 0, NULL, NULL, 0, 0
+	{ TEXT_(T_NO_EXTENSIONS), "", M_BAR, NULL, NULL, 0, 0 },
+	{ NULL, NULL, 0, NULL, NULL, 0, 0 },
 };
 
 void menu_list_ext(struct terminal *term, void *fn, void *xxx)
@@ -494,14 +507,12 @@ void update_prog(struct list_head *l, unsigned char *p, int s)
 		mem_free(repl->prog);
 		goto ss;
 	}
-	if (!(repl = mem_alloc(sizeof(struct protocol_program)))) return;
+	repl = mem_alloc(sizeof(struct protocol_program));
 	add_to_list(*l, repl);
 	repl->system = s;
 	ss:
-	if ((repl->prog = mem_alloc(MAX_STR_LEN))) {
-		strncpy(repl->prog, p, MAX_STR_LEN);
-		repl->prog[MAX_STR_LEN - 1] = 0;
-	}
+	repl->prog = mem_alloc(MAX_STR_LEN);
+	safe_strncpy(repl->prog, p, MAX_STR_LEN);
 }
 
 unsigned char *get_prog(struct list_head *l)
@@ -512,3 +523,8 @@ unsigned char *get_prog(struct list_head *l)
 	foreach(repl, *l) if (repl->system == SYSTEM_ID) return repl->prog;
 	return NULL;
 }
+
+int is_html_type(unsigned char *ct)
+{
+	return !strcasecmp(ct, "text/html") || !strcasecmp(ct, "text/x-server-parsed-html") || !casecmp(ct, "application/xhtml", strlen("application/xhtml"));
+}
