@@ -16,18 +16,21 @@ struct {
 		{"proxy", 3128, proxy_func, NULL, 0, 1, 1, 1},
 		{"ftp", 21, ftp_func, NULL, 0, 1, 1, 0},
 		{"finger", 79, finger_func, NULL, 0, 1, 1, 0},
+#ifndef DISABLE_SMB
 		{"smb", 139, smb_func, NULL, 0, 1, 1, 0},
+#endif
 		{"mailto", 0, NULL, mailto_func, 0, 0, 0, 0},
 		{"telnet", 0, NULL, telnet_func, 0, 0, 0, 0},
 		{"tn3270", 0, NULL, tn3270_func, 0, 0, 0, 0},
-		{NULL, 0, NULL}
+		{"mms", 0, NULL, mms_func, 1, 0, 1, 0},
+		{NULL, 0, NULL, NULL, 0, 0, 0, 0}
 };
 
 int check_protocol(unsigned char *p, int l)
 {
 	int i;
 	for (i = 0; protocols[i].prot; i++)
-		if (!casecmp(protocols[i].prot, p, l)) {
+		if (!casecmp(protocols[i].prot, p, l) && (int)strlen(protocols[i].prot) == l) {
 			return i;
 		}
 	return -1;
@@ -77,10 +80,13 @@ int parse_url(unsigned char *url, int *prlen, unsigned char **user, int *uslen,
 		return 0;
 	}
 	p += 3;
-	q = p + strcspn(p, "@/");
+	q = p + strcspn(p, "@/?");
 	if (!*q && protocols[a].need_slash_after_host) return -1;
 	if (*q == '@') {
-		unsigned char *pp = strchr(p, ':');
+		unsigned char *pp;
+		while (strcspn(q + 1, "@") < strcspn(q + 1, "/?"))
+			q = q + 1 + strcspn(q + 1, "@");
+		pp = strchr(p, ':');
 		if (!pp || pp > q) {
 			if (user) *user = p;
 			if (uslen) *uslen = q - p;
@@ -92,7 +98,7 @@ int parse_url(unsigned char *url, int *prlen, unsigned char **user, int *uslen,
 		}
 		p = q + 1;
 	} 
-	q = p + strcspn(p, ":/");
+	q = p + strcspn(p, ":/?");
 	if (!*q && protocols[a].need_slash_after_host) return -1;
 	if (host) *host = p;
 	if (holen) *holen = q - p;
@@ -105,7 +111,7 @@ int parse_url(unsigned char *url, int *prlen, unsigned char **user, int *uslen,
 		for (cc = 0; cc < pp - q - 1; cc++) if (q[cc+1] < '0' || q[cc+1] > '9') return -1;
 		q = pp;
 	}
-	if (*q) q++;
+	if (*q && *q != '?') q++;
 	p = q;
 	p_c[0] = POST_CHAR;
 	p_c[1] = 0;
@@ -169,15 +175,17 @@ int get_port(unsigned char *url)
 {
 	unsigned char *h;
 	int hl;
-	int n = -1;
+	long n = -1;
 	if (parse_url(url, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &h, &hl, NULL, NULL, NULL)) return -1;
 	if (h) {
 		n = strtol(h, NULL, 10);
-		if (n) return n;
+		if (n && n < MAXINT) return n;
 	}
 	if ((h = get_protocol_name(url))) {
-		get_prot_info(h, &n, NULL, NULL, NULL);
+		int nn = -1;	/* against warning */
+		get_prot_info(h, &nn, NULL, NULL, NULL);
 		mem_free(h);
+		n = nn;
 	}
 	return n;
 }
@@ -213,35 +221,6 @@ unsigned char *get_url_data(unsigned char *url)
 	return d;
 }
 
-/*void translate_directories(unsigned char *url)
-{
-	unsigned char *p;
-	unsigned char *dd = get_url_data(url);
-	unsigned char *d = dd;
-	if (!d || d == url || *--d != '/') return;
-	r:
-	p = d + strcspn(d, "/;?#");
-	if (p[0] != '/') return;
-	if (p[1] == '.' && p[2] == '/') {
-		memmove(p, p + 2, strlen(p + 2) + 1);
-		d = p;
-		goto r;
-	}
-	if (p[1] == '.' && p[2] == '.' && p[3] == '/') {
-		unsigned char *e;
-		for (e = p - 1; e >= dd; e--) if (*e == '/') {
-			memmove(e, p + 3, strlen(p + 3) + 1);
-			d = e;
-			goto r;
-		}
-		memmove(dd, p + 3, strlen(p + 3) + 1);
-		d = dd;
-		goto r;
-	}
-	d = p + 1;
-	goto r;
-}*/
-
 #define dsep(x) (lo ? dir_sep(x) : (x) == '/')
 
 void translate_directories(unsigned char *url)
@@ -265,23 +244,13 @@ void translate_directories(unsigned char *url)
 		goto r;
 	}
 	if (dsep(s[0]) && s[1] == '.' && s[2] == '.' && (dsep(s[3]) || !s[3])) {
-		unsigned char *d1 = d;
 		while (d > dd) {
 			d--;
-			if (dsep(*d)) {
-				if (d + 3 == d1 && d[1] == '.' && d[2] == '.') {
-					d = d1;
-					goto p;
-				}
-				goto b;
-			}
+			if (dsep(*d)) goto b;
 		}
-		/*d = d1;
-		goto p;*/
 		b:
-		*d++ = *s;
-		if (s[3]) s += 4;
-		else s += 3;
+		if (!s[3]) *d++ = *s;
+		s += 3;
 		goto r;
 	}
 	p:
@@ -297,7 +266,10 @@ void insert_wd(unsigned char **up, unsigned char *cwd)
 #ifdef DOS_FS
 	if (upcase(url[7]) >= 'A' && upcase(url[7]) <= 'Z' && url[8] == ':' && dir_sep(url[9])) return;
 #endif
-	if (!(url = mem_alloc(strlen(*up) + strlen(cwd) + 2))) return;
+#ifdef SPAD
+	if (_is_absolute(url + 7) != _ABS_NO) return;
+#endif
+	url = mem_alloc(strlen(*up) + strlen(cwd) + 2);
 	memcpy(url, *up, 7);
 	strcpy(url + 7, cwd);
 	if (!dir_sep(cwd[strlen(cwd) - 1])) strcat(url, "/");
@@ -311,23 +283,44 @@ unsigned char *join_urls(unsigned char *base, unsigned char *rel)
 	unsigned char *p, *n, *pp;
 	int l;
 	int lo = !casecmp(base, "file://", 7);
-	if (rel[0] == '#') {
-		if (!(n = stracpy(base))) return NULL;
+	if (rel[0] == '#' || !rel[0]) {
+		n = stracpy(base);
 		for (p = n; *p && *p != POST_CHAR && *p != '#'; p++) ;
 		*p = 0;
 		add_to_strn(&n, rel);
 		translate_directories(n);
 		return n;
 	}
+	if (rel[0] == '?' || rel[0] == '&') {
+		unsigned char rj[3];
+		unsigned char *d = get_url_data(base);
+		if (!d) goto bad_base;
+		rj[0] = rel[0];
+		rj[1] = POST_CHAR;
+		rj[2] = 0;
+		d += strcspn(d, rj);
+		n = memacpy(base, d - base);
+		add_to_strn(&n, rel);
+		translate_directories(n);
+		return n;
+	}
 	if (rel[0] == '/' && rel[1] == '/') {
 		unsigned char *s, *n;
 		if (!(s = strstr(base, "//"))) {
-			internal("bad base url: %s", base);
-			return NULL;
+			if (!(s = strchr(base, ':'))) {
+				bad_base:
+				internal("bad base url: %s", base);
+				return NULL;
+			}
+			s++;
 		}
 		n = memacpy(base, s - base);
 		add_to_strn(&n, rel);
-		return n;
+		if (!parse_url(n, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) {
+			translate_directories(n);
+			return n;
+		}
+		mem_free(n);
 	}
 	if (!casecmp("proxy://", rel, 8)) goto prx;
 	if (!parse_url(rel, &l, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) {
@@ -335,26 +328,26 @@ unsigned char *join_urls(unsigned char *base, unsigned char *rel)
 		translate_directories(n);
 		return n;
 	}
-	if ((n = stracpy(rel))) {
-		while (n[0] && n[strlen(n) - 1] <= ' ') n[strlen(n) - 1] = 0;
-		add_to_strn(&n, "/");
-		if (!parse_url(n, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) {
-			translate_directories(n);
-			return n;
-		}
-		mem_free(n);
+	n = stracpy(rel);
+	while (n[0] && n[strlen(n) - 1] <= ' ') n[strlen(n) - 1] = 0;
+	add_to_strn(&n, "/");
+	if (!parse_url(n, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) {
+		translate_directories(n);
+		return n;
 	}
+	mem_free(n);
 	prx:
 	if (parse_url(base, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &p, NULL, NULL) || !p) {
-		internal("bad base url");
-		return NULL;
+		goto bad_base;
 	}
 	if (!dsep(*p)) p--;
-	if (!dsep(rel[0])) for (pp = p; *pp; pp++) {
+	if (end_of_dir(rel[0])) for (; *p; p++) {
+		if (end_of_dir(*p)) break;
+	} else if (!dsep(rel[0])) for (pp = p; *pp; pp++) {
 		if (end_of_dir(*pp)) break;
 		if (dsep(*pp)) p = pp + 1;
 	}
-	if (!(n = mem_alloc(p - base + strlen(rel) + 1))) return NULL;
+	n = mem_alloc(p - base + strlen(rel) + 1);
 	memcpy(n, base, p - base);
 	strcpy(n + (p - base), rel);
 	translate_directories(n);
@@ -363,18 +356,21 @@ unsigned char *join_urls(unsigned char *base, unsigned char *rel)
 
 unsigned char *translate_url(unsigned char *url, unsigned char *cwd)
 {
-	static int r = 0;
 	unsigned char *ch;
 	unsigned char *nu, *da;
-	while(*url == ' ') url++;
-	if (!casecmp("proxy://", url, 8)) goto prx;
+	unsigned char *prefix;
+	int sl;
+	while (*url == ' ') url++;
+	if (!casecmp("proxy://", url, 8)) return NULL;
 	if (!parse_url(url, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &da, NULL, NULL)) {
 		nu = stracpy(url);
 		insert_wd(&nu, cwd);
 		translate_directories(nu);
 		return nu;
 	}
-	if (strstr(url, "//") && (nu = stracpy(url))) {
+	if (strchr(url, POST_CHAR)) return NULL;
+	if (strstr(url, "://")) {
+		nu = stracpy(url);
 		add_to_strn(&nu, "/");
 		if (!parse_url(nu, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) {
 			insert_wd(&nu, cwd);
@@ -383,24 +379,27 @@ unsigned char *translate_url(unsigned char *url, unsigned char *cwd)
 		}
 		mem_free(nu);
 	}
-	prx:
 	ch = url + strcspn(url, ".:/@");
+	prefix = "file://";
+	sl = 0;
 	if (*ch != ':' || *(url + strcspn(url, "/@")) == '@') {
-		unsigned char *prefix = "file://";
-		int sl = 0;
 		if (*url != '.' && *ch == '.') {
 			unsigned char *f, *e;
 			int i;
 			for (e = ch + 1; *(f = e + strcspn(e, ".:/")) == '.'; e = f + 1) ;
 			for (i = 0; i < f - e; i++) if (e[i] >= '0' && e[i] <= '9') goto http;
-			if (f - e == 2) http: prefix = "http://", sl = 1;
-			else {
-				unsigned char *tld[] = { "com", "edu", "net", "org", "gov", "mil", "int", "arpa", "aero", "biz", "coop", "info", "museum", "name", "pro", NULL };
-				for (i = 0; tld[i]; i++) if (f - e == strlen(tld[i]) && !casecmp(tld[i], e, f - e)) goto http;
+			if (f - e == 2) {
+				http:
+				prefix = "http://", sl = 1;
+			} else {
+				unsigned char *tld[] = { "com", "edu", "net", "org", "gov", "mil", "int", "arpa", "aero", "biz", "coop", "info", "museum", "name", "pro", "cat", "jobs", "mobi", "travel", "tel", NULL };
+				for (i = 0; tld[i]; i++) if ((size_t)(f - e) == strlen(tld[i]) && !casecmp(tld[i], e, f - e)) goto http;
 			}
 		}
 		if (*ch == '@' || *ch == ':' || !cmpbeg(url, "ftp.")) prefix = "ftp://", sl = 1;
-		if (!(nu = stracpy(prefix))) return NULL;
+		goto set_prefix;
+		set_prefix:
+		nu = stracpy(prefix);
 		add_to_strn(&nu, url);
 		if (sl && !strchr(url, '/')) add_to_strn(&nu, "/");
 		if (parse_url(nu, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) mem_free(nu), nu = NULL;
@@ -410,7 +409,13 @@ unsigned char *translate_url(unsigned char *url, unsigned char *cwd)
 		}
 		return nu;
 	}
-	if (!(nu = memacpy(url, ch - url + 1))) return NULL;
+#ifdef DOS_FS
+	if (ch == url + 1) goto set_prefix;
+#endif
+#ifdef SPAD
+	if (_is_local(url)) goto set_prefix;
+#endif
+	nu = memacpy(url, ch - url + 1);
 	add_to_strn(&nu, "//");
 	add_to_strn(&nu, ch + 1);
 	if (!parse_url(nu, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) {
@@ -431,10 +436,10 @@ unsigned char *translate_url(unsigned char *url, unsigned char *cwd)
 unsigned char *extract_position(unsigned char *url)
 {
 	unsigned char *u, *uu, *r;
+	if ((u = get_url_data(url))) url = u;
 	if (!(u = strchr(url, POST_CHAR))) u = url + strlen(url);
-	uu = u;
-	while (--uu >= url && *uu != '#') ;
-	if (uu < url || !(r = mem_alloc(u - uu))) return NULL;
+	if (!(uu = memchr(url, '#', u - url))) return NULL;
+	r = mem_alloc(u - uu);
 	memcpy(r, uu + 1, u - uu - 1);
 	r[u - uu - 1] = 0;
 	memmove(uu, u, strlen(u) + 1);
@@ -453,3 +458,35 @@ void get_filename_from_url(unsigned char *url, unsigned char **s, int *l)
 	}
 	*l = url - *s;
 }
+
+#define accept_char(x)	((x) != '"' && (x) != '\'' && (x) != '&' && (x) != '<' && (x) != '>')
+#define special_char(x)	((x) == '%' || (x) == '#')
+
+void add_conv_str(unsigned char **s, int *l, unsigned char *b, int ll, int encode_special)
+{
+	for (; ll; ll--, b++) {
+		if ((unsigned char)*b < ' ') continue;
+		if (special_char(*b) && encode_special == 1) {
+			unsigned char h[4];
+			sprintf(h, "%%%02X", (unsigned)*b & 0xff);
+			add_to_str(s, l, h);
+		} else if (*b == '%' && encode_special <= -1 && ll > 2 && ((b[1] >= '0' && b[1] <= '9') || (b[1] >= 'A' && b[1] <= 'F') || (b[1] >= 'a' && b[1] <= 'f'))) {
+			unsigned char h = 0;
+			int i;
+			for (i = 1; i < 3; i++) {
+				if (b[i] >= '0' && b[i] <= '9') h = h * 16 + b[i] - '0';
+				if (b[i] >= 'A' && b[i] <= 'F') h = h * 16 + b[i] - 'A' + 10;
+				if (b[i] >= 'a' && b[i] <= 'f') h = h * 16 + b[i] - 'a' + 10;
+			}
+			if (h >= ' ') add_chr_to_str(s, l, h);
+			ll -= 2;
+			b += 2;
+		} else if (accept_char(*b) || encode_special == -2) {
+			add_chr_to_str(s, l, *b);
+		} else {
+			add_to_str(s, l, "&#");
+			add_num_to_str(s, l, (int)*b);
+			add_chr_to_str(s, l, ';');
+		}
+	}
+}
